/*=============================================================================
        File: barray2d.hh
     Purpose:       
    Revision: $Id: barray2d.hh,v 1.1.1.1 1998/04/06 21:02:31 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1998 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/

#ifndef _Matrix_barray2d_h_
#define _Matrix_barray2d_h_


#include <fstream.h>
#include <iomanip.h>
#include "specialType.hh"

// Predefining every friend functions 
// This is required by latest ISO C++ draft

template <class T> class Basic2DArray ;
template <class T> istream& operator>>(istream& is, Basic2DArray<T>& ary);
template <class T> ostream& operator<<(ostream& os, const Basic2DArray<T>& ary);


/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: Basic2DArray --- A basic templated array class in two dimensions
  Description: This is a basis array class for two dimensionnal storage and 
               retrieval.

   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
template<class T> class Basic2DArray
{
public:
  const int& rows;
  const int& cols;
  Basic2DArray() ;
  Basic2DArray(const int r, const int c) ;
  Basic2DArray(const Basic2DArray<T>& f2);
  Basic2DArray(T* p, const int r, const int c) ;
  
  virtual ~Basic2DArray();
  
  Basic2DArray<T>& operator=(const Basic2DArray<T>& f2);
  
  virtual void resize(const int nr, const int nc);  // destructive resizing
  void resize(const Basic2DArray<T>& A) { resize(A.rows,A.cols) ; }
  virtual void resizeKeep(const int nr, const int nc) ; // non-destructive resizzing

  virtual void reset(const T val = 0.0);
  T operator=(const T val) { reset(val); return val; }

  virtual T* operator[](const int i) { return vm[i]; } // returns a pointer to the start of a row $i$
  virtual T* operator[](const int i) const { return vm[i];} // returns a pointer to the start of a row $i$ 
  
  T& operator()(const int i,const int j)       { return elem(i,j); } // calls elem(i,j)
  T  operator()(const int i,const int j) const { return elem(i,j); } // calls elem(i,j)

  void io_elem_width(int w) { width = w ; }
  void io_by_rows()    { by_columns = 0; }  // row at a time ASCII I/O
  void io_by_columns() { by_columns = 1; }  // column at a time ASCII I/O

#ifdef ANSI_ISO_CPP
  friend istream& operator>> <>(istream& is, Basic2DArray<T>& ary);
  friend ostream& operator<< <>(ostream& os, const Basic2DArray<T>& ary);
#else  
  friend istream& operator>>(istream& is, Basic2DArray<T>& ary);
  friend ostream& operator<<(ostream& os, const Basic2DArray<T>& ary);
#endif

#ifdef DEBUG_MATRIX
  virtual T& elem(const int,const int);  // returns an error message if the index is out of range
  virtual T  elem(const int,const int) const;   // returns an error message if the index is out of range
#else
#ifdef COLUMN_ORDER
  virtual T& elem(const int i,const int j) 
    { return vm[j][i] ; }  // no error message are generated if the index are out of range
  virtual T  elem(const int i,const int j) const
    { return vm[j][i] ; }  // no error message are generated if the index are out of range
#else
  virtual T& elem(const int i,const int j) 
    { return vm[i][j] ; }  // no error message are generated if the index are out of range
  virtual T  elem(const int i,const int j) const
    { return vm[i][j] ; }  // no error message are generated if the index are out of range
#endif
#endif
  
protected:
  int by_columns,width;
  int rz, cz;           // the number of rows and columns
  T *m;       // the memory is allocated here
  T **vm ;    // for referencing, vm[i][j] is faster than m[i+cols*j]
  int created ;

  void init(const int r = 1, const int c = 1);

};


typedef Basic2DArray<int> Array2D_INT ;            
typedef Basic2DArray<char> Array2D_BYTE ;          
typedef Basic2DArray<double> Array2D_DOUBLE ;      
typedef Basic2DArray<Complex> Array2D_COMPLEX ;    
typedef Basic2DArray<unsigned char> Array2D_UBYTE ;
typedef Basic2DArray<Point3D> Array2D_Point3D ;
typedef Basic2DArray<CtrlPnt4D> Array2D_CtrlPnt4D ;
typedef Basic2DArray<Coordinate> Array2D_Coordinate ;


#if defined(__sgi ) && !defined(__GNUG__)

// The SGI compilers needs these definitions in the header
// no one knows why, but at least now it works.

void
Basic2DArray<CtrlPnt4D>::init(const int r,const int c)
{
  if ( r <= 0  || c <= 0 )
    {
      PlError error("Basic2DArray<T>::init");
      error << "bad Basic2DArray<T> size " << r << ", " << c << endl ;
      error.fatal() ;
    }
	
  rz = r;	cz = c;

  m = new CtrlPnt4D [rz*cz];
#ifdef COLUMN_ORDER
  vm = new CtrlPnt4D* [cz] ;
#else
  vm = new CtrlPnt4D* [rz] ;
#endif

  // we want to have all the data in a consecutive memory segment
  // but it has to be in column order instead of row order
  // i.e.
  //   m1 m5 m9  m13
  //   m2 m6 m10 m14
  //   m3 m7 m11 m15
  //   m4 m8 m12 m16
  // This is done so that OpenGL can be used effectively
  // since the data portion will now be in column order.
  // Note that the order of the elements themselves are
  // still defined by the COLUMN_ORDER macro

  float* dn ;
  dn = new float[rz*cz*4] ;

  int i ;
#ifdef COLUMN_ORDER
  for(i=0;i<rz*cz;++i){
    if(m[i].created && m[i].data)
      delete [] m[i].data ;
    m[i].created = 0 ;
    m[i].data = &dn[i*4] ;
  }
#else
  for(i=0;i<rz;++i){
    for(int j=0;j<cz;++j){
      if(m[i*cz+j].created && m[i*cz+j].data)
	delete [] m[i*cz+j].data ;
      m[i*cz+j].created = 0 ;
      m[i*cz+j].data = &dn[(rz*j+i)*4] ;
    }
  }
#endif
  
  CtrlPnt4D *p1 ;
  int sze = rz*cz ;
  p1 = m-1 ;

  if(sze>0)
    m[0].created = 1 ;

  for(i = 0; i < sze; i++)
      *(++p1) = (CtrlPnt4D)0;
#ifdef COLUMN_ORDER
  for(i=0;i<cz;++i)
    vm[i] = &m[i*rz] ;
#else
  for(i=0;i<rz;++i)
    vm[i] = &m[i*cz] ;
#endif
}

void 
Basic2DArray<CtrlPnt4D>::resizeKeep(const int nr,const int nc)
{
  if(nr==rz && nc==cz){
    return ;
  }

  CtrlPnt4D *mn ;
  CtrlPnt4D *p,*pn ;

  mn = new CtrlPnt4D[nr*nc] ;
  
  // we want to have all the data in a consecutive memory segment
  // but it has to be in column order instead of row order
  // i.e.
  //   m1 m5 m9  m13
  //   m2 m6 m10 m14
  //   m3 m7 m11 m15
  //   m4 m8 m12 m16
  // This is done so that OpenGL can be used effectively
  // since the data is column ordered.
  // Note that the elements themselves are still ordered
  // according to COLUMN_ORDER

  float* dn ;
  dn = new float[nr*nc*4] ;

  int i,j ;
#ifdef COLUMN_ORDER
  for(i=0;i<nr*nc;++i){
    if(mn[i].created && mn[i].data)
      delete [] mn[i].data ;
    mn[i].created = 0 ; 
    mn[i].data = &dn[i*4] ;
  }
  for(j=0;j<minimum(nc,cz);j++){
    p = &m[j*rz] -1;
    pn = &mn[j*nr] -1 ;
    for(i=0;i<minimum(nr,rz);i++){
      *(++pn) = *(++p) ;
    }
  }

  for(j=cz;j<nc;j++)
    pn = &mn[j*nr+rz]-1 ;
    for(i=rz;i<nr;i++){
      *(pn++) = (CtrlPnt4D)0 ;
  }

#else
  for(i=0;i<nr;++i){
    for(j=0;j<nc;++j){
      if(mn[i*nc+j].created && mn[i*nc+j].data)
	delete [] mn[i*nc+j].data ;
      mn[i*nc+j].created = 0 ; 
      mn[i*nc+j].data = &dn[(nr*j+i)*4] ;
    }
  }
  for(i=0;i<minimum(nr,rz);i++){
    p = &m[i*cz] -1;
    pn = &mn[i*nc] -1 ;
    for(j=0;j<minimum(nc,cz);j++){
      *(++pn) = *(++p) ;
    }
  }

  for(i=rz;i<nr;i++){
    pn = &mn[i*nc+cz]-1 ;
    for(j=cz;j<nc;j++)
      *(pn++) = (CtrlPnt4D)0 ;
  }

#endif

  rz = nr ;
  cz = nc ;

  if(m && created){
    delete []m ;
  }
  m = mn ;
  if((nr*nc)>0)
    m[0].created = 1 ;
  if(vm)
    delete []vm ;
#ifdef COLUMN_ORDER
  vm = new CtrlPnt4D* [cz] ;
  for(i=0;i<cz;++i)
    vm[i] = &m[i*rz] ;
#else
  vm = new CtrlPnt4D* [rz] ;
  for(i=0;i<rz;++i)
    vm[i] = &m[i*cz] ;
#endif
}

#endif



#ifdef CPP_SRC_INCLUDE
#include "../src/barray2d.cc"
#endif


#endif

