/*=============================================================================
        File: matrixMat.hh
     Purpose:       
    Revision: $Id: matrixMat.hh,v 1.1.1.1 1998/04/06 21:02:31 lavoie Exp $
  Created by: Philippe Lavoie          (22 Oct, 1997)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _matrix_matrixMat_h_
#define _matrix_matrixMat_h_

#include "matrix.hh"

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: LUMatrix --- a class for the LU decomposition of a matrix
  Description: This class was adapted from a class in the Kalman library
               written by Skip Carter (\verb|skip@taygeta.oc.nps.navy.mil|)
	       available from 
	       \verb|ftp:://usc.edu/pub/C-numanal/kalman.tar.gz|

	       It performs the LU decomposition of a matrix. It can be
	       used to find it's inverse and to solve linear systems.

	       This class can {\em only} be used with floating point values
	       (float or double).

   Created by: Skip Carter
  Modified by: Philippe Lavoie (22 Oct. 1997)
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
template <class T>
class LUMatrix : public PlMatrix<T> 
{
public:
  LUMatrix(int r, int c) : PlMatrix<T>(r,c),pivot(pivot_) { pivot_.resize(r) ; }
  LUMatrix()  : PlMatrix<T>(),pivot(pivot_)    {; }
  LUMatrix(const LUMatrix<T>& lu): PlMatrix<T>(lu),pivot(pivot_) { pivot_ = lu.pivot_;}
  LUMatrix(const PlMatrix<T>& a): PlMatrix<T>(a.rows,a.cols),pivot(pivot_) { decompose(a) ; }
  
  void resize(const int r, const int c) { PlMatrix<T>::resize(r,c) ; pivot_.resize(r) ; }
  LUMatrix& operator=(const LUMatrix<T>&);
  LUMatrix& decompose(const PlMatrix<T> &a);
  T determinant() ;

  PlMatrix<T> inverse();
  void inverseIn(PlMatrix<T>&);
  const PlVector<int>& pivot ;

private:
  PlVector<int> pivot_ ;
  
protected:
  int errval ;
  int sign ;
};


/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: SVDMatrix --- a matrix for the SVD decomposition
  Description: This matrix was adapted from the Numerical Math Class Library
               developed by Oleg Kiselyov and available from
	       \verb|http://www.lh.com/oleg/ftp/packages.html|.	       
	       
	       This class can {\em only} be used with floating point values
	       (float or double).

	       The following comments are from Oleg Kiselyov.

	       Singular Value Decomposition of a rectangular matrix
			    $ A = U  Sig  V'$
	       where matrices $U$ and $V$ are orthogonal and $Sig$ is a 
	       diagonal matrix.

	       The singular value decomposition is performed by constructing 
	       an SVD  object from an $M\times N$ matrix $A$ with $M \ge N$
	       (that is, at least as many rows as columns). Note, in case 
	       $M > N,$ matrix $Sig$ has to be a $M \times N$ diagonal
	       matrix. However, it has only $N$ diagonal elements, which 
	       we store in a vector sig.

	       {\bf Algorithm:}	Bidiagonalization with Householder 
	       reflections followed by a modification of a QR-algorithm. 
	       For more details, see G.E. Forsythe, M.A. Malcolm, C.B. Moler 
	       Computer methods for mathematical computations, Prentice-Hall, 
	       1977.  However, in the present implementation, matrices $U$ 
	       and $V$ are computed right away rather than delayed until 
	       after all Householder reflections.

	       This code is based for the most part on a Algol68 code I 
	       (Oleg Kiselyov) wrote ca. 1987.
	  
	       Look at the source code for more information about the 
	       algorithm.

   Created by: Oleg Kiselyov
  Modified by: Philippe Lavoie (22 Oct. 1997)
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
template <class T>
class SVDMatrix 
{
public:
  SVDMatrix():U(U_), V(V_), sig(sig_) { ; }
  SVDMatrix(const PlMatrix<T>& A);
  
  const PlMatrix<T>& U ;
  const PlMatrix<T>& V ;
  const PlVector<T>& sig ;

  int decompose(const PlMatrix<T>& A) ;
  void minMax(T& min_sig, T& max_sig) const;
  double q_cond_number(void) const;	

  void cut_off(const double min_sig);
  void inverseIn(PlMatrix<T>& inv, double tau=0) ;
  PlMatrix<T> inverse(double tau=0) ;
  int solve(const PlMatrix<T>& B, PlMatrix<T>& X, double tau=0) ;

protected:
  int M,N;	// Dimensions of the problem ($M \gt N$)
  PlMatrix<T> U_;		// $M \times M$ orthogonal matrix $U$
  PlMatrix<T> V_;		// $N \times N$ orthogonal matrix $V$
  PlVector<T> sig_;	// Vector of $N$ unordered singular values
  
  // Internal procedures used in SVD
  double left_householder(PlMatrix<T>& A, const int i);
  double right_householder(PlMatrix<T>& A, const int i);
  double bidiagonalize(PlVector<T>& super_diag, const PlMatrix<T>& _A);

  void rotate(PlMatrix<T>& U, const int i, const int j,
	      const double cos_ph, const double sin_ph);
  void rip_through(PlVector<T>& super_diag, const int k, const int l, const double eps);
  int get_submatrix_to_work_on(PlVector<T>& super_diag, const int k, const double eps);
  void diagonalize(PlVector<T>& super_diag, const double eps);

};

template <class T> 
void solve(const PlMatrix<T>&A, const PlMatrix<T>&B, PlMatrix<T>&C) ;

template <class T>
PlMatrix<T> inverse(const PlMatrix<T>& A) ;

#ifdef CPP_SRC_INCLUDE
#include "../src/matrixMat.cc"
#endif


#endif
