/*============================================================================
        File: specialType.hh
     Purpose: Defines types used by my programs. This file should not exist
              but due to a bug in the sparc compiler it has to be included 
              in the matrix library.
    Revision: $Id: specialType.hh,v 1.5 1998/06/03 15:11:21 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _Matrix_specialType_h_
#define _Matrix_specialType_h_

#include "internals.hh"

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: Vector2 --- A vector of points in 2 dimensions
  Description: A basic vector class for points in 2 dimensions.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class Vector2 {
 public:
  double x, y ;
  Vector2(double X=0, double Y=0) : x(X),y(Y) {}
  Vector2(const Vector2 &v) : x(v.x), y(v.y) {}

  Vector2& operator=(const Vector2& v) { x=v.x ; y=v.y ; return *this ;} ;
  Vector2& operator+=(const Vector2& v) {x+=v.x ; y+= v.y ; return *this;} ;
  Vector2& operator-=(const Vector2& v) {x-=v.x ; y-= v.y ; return *this;} ;
  Vector2& operator*=(double v) {x*=v ; y*= v ; return *this;} ;
  Vector2& operator/=(double v) {x/=v ; y/= v ; return *this;} ;
  friend int operator==(const Vector2&,double) ;
  friend int operator==(const Vector2& a,const Vector2& b){return a.x==b.x && a.y==b.y ;}
  friend int operator!=(const Vector2& a,const Vector2& b){return a.x!=b.x || a.y!=b.y ;}
  
  friend Vector2 operator+(const Vector2&,const Vector2&) ;
  friend Vector2 operator-(const Vector2&,const Vector2&) ;
  friend double operator*(const Vector2& a,const Vector2& b){return dot(a,b) ;}
  friend Vector2 operator*(const Vector2& a,const double b) {return b*a;}
  friend Vector2 operator*(const double a,const Vector2& b) ;
  friend Vector2 operator/(const Vector2&,const double) ;

  friend double angle(const Vector2&,const Vector2&) ;
  friend double abs(const Vector2& a) {return sqrt(abs2(a));}
  friend double abs2(const Vector2& a) {return a.x*a.x+a.y*a.y;}
  friend double dot(const Vector2&, const Vector2&) ;
  friend ostream& operator<<(ostream& os, const Vector2& v2);
  friend istream& operator>>(istream& is, Vector2& v2);

};

typedef Vector2 Point2 ;

inline int operator<(const Vector2& a, const Vector2& b) { 
  return abs(a)<abs(b) ? 1 : 0 ;}
inline int operator>(const Vector2& a, const Vector2& b) { 
  return abs(a)>abs(b) ? 1 : 0 ;}
inline int operator<=(const Vector2& a, const Vector2& b) { 
  return abs(a)<=abs(b) ? 1 : 0 ;}
inline int operator>=(const Vector2& a, const Vector2& b) { 
  return abs(a)>=abs(b) ? 1 : 0 ;}

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: Color --- A class for a RGB color 
  Description: Defines a class containing the RGB component of a color pixel.
               Each component is an unsigned char valid for the range 
	       $[0..255]$.

   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class Color {
 public:
  unsigned char r,g,b ;
  Color(const unsigned char R=0, const unsigned char G=0, const unsigned char B=0) : r(R),g(G),b(B) {} 
  Color& operator+=(const Color& a) { r+=a.r ; g+=a.g ; b+=a.b; return *this ;} // the += operator
  Color& operator-=(const Color& a) { r+=a.r ; g+=a.g ; b+=a.b; return *this;} // the -= operator
  Color& operator*=(double a) { r = (unsigned char)(a*double(r)) ; g = (unsigned char)(a*double(g)) ; b = (unsigned char)(a*double(b)) ; return *this ; }
  Color& operator/=(double a) { r = (unsigned char)(double(r)/a) ; g = (unsigned char)(double(g)/a) ; b = (unsigned char)(double(b)/a) ; return *this ; }
  Color& operator=(const Color& a) { r=a.r ; g=a.g; b=a.b ; return *this ; } // the assignment operator
  friend Color operator*(const double d, const Color& a) ; // multiplies a color by a double 
  friend Color operator*(const Color& a, const Color& b) ; // multiplies a color by another color
  friend Color operator+(const Color& a, const Color& b) ; // Addition of two colors


  friend ostream& operator<<(ostream& os, const Color& point);
  friend istream& operator>>(istream& os, Color& point);
};


inline int operator==(const Color& a, const Color& b) { return (b.r==a.r)&&(b.g==a.g)&&(b.b==a.b) ; } // the equality operator
inline int operator!=(const Color& a, const Color& b) { return !((b.r==a.r)||(b.g==a.g)||(b.b==a.b));} // the inequality operator


inline int operator<(const Color& a, const Color& b){ 
  return (a.r<b.r && a.g<b.g && a.b<b.b) ; } // the smaller than operator
inline int operator>(const Color& a, const Color& b){ 
  return (a.r>b.r && a.g>b.g && a.b>b.b) ; } // the greater than operator
inline int operator<=(const Color& a, const Color& b){ 
  return (a.r<=b.r && a.g<=b.g && a.b<=b.b) ; } // the smaller or equal operator
inline int operator>=(const Color& a, const Color& b){ 
  return (a.r>=b.r && a.g>=b.g && a.b>=b.b) ; } // the greater or equal operator

extern Color whiteColor ;
extern Color redColor ;
extern Color blueColor ;
extern Color greenColor ;
extern Color yellowColor ;
extern Color cyanColor ;
extern Color magentaColor ;
extern Color gray80Color ;
extern Color gray50Color ;
extern Color blackColor ;


inline Color operator*(const double d, const Color& a) {
  Color result ;
  result.r = (unsigned char) (d*a.r) ;
  result.g = (unsigned char) (d*a.g) ;
  result.b = (unsigned char) (d*a.b) ;
  return result ;
}

inline Color operator*(const Color& a, const Color& b) {
  Color result;
  result.r = a.r*b.r ;
  result.g = a.g*b.g ;
  result.b = a.b*b.b ;
  return result ;
}

inline Color operator+(const Color& a, const Color& b) {
  Color result(a);
  result.r += b.r ;
  result.g += b.g ;
  result.b += b.b ;
  return result ;
}

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: ColorF --- A class for a RGB color defined with float variables
  Description: Defines a class containing the RGB component of a color pixel. 
               Each component is defined with a float valid in the range 
	       $[0..1]$.

   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class ColorF {
 public:
  float r,g,b ;

  ColorF(float R=0.0, float G=0.0, float B=0.0) : r(R),g(G),b(B) {} 

  ColorF& operator=(const ColorF& a) { r=a.r ; g=a.g; b=a.b ; return *this ; }
};


/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: Point3D --- A point in 3D space
  Description: Defines a point in 3D space $(x,y,z)$ along with basic 
               mathematical operators.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class Point3D {
public:
  float data[3] ;
  float& x ;
  float& y ;
  float& z ;
  Point3D() : x(data[0]), y(data[1]), z(data[2]) { x = y = z = 0 ;}
  Point3D(float a) : x(data[0]), y(data[1]), z(data[2]) { x = y = z = a ;}
  Point3D(float X, float Y, float Z) : x(data[0]), y(data[1]), z(data[2]) {x =X ; y=Y ; z=Z ;}
  Point3D(const Point3D& a) : x(data[0]), y(data[1]), z(data[2]) { x = a.x ; y = a.y ; z = a.z ; }

  
  Point3D& operator=(const Point3D& v) { x=v.x ; y=v.y ; z=v.z ;  return *this ;} ;
  Point3D& operator+=(const Point3D& v) {x+=v.x ; y+= v.y ; z+=v.z ;  return *this;} ;
  Point3D& operator-=(const Point3D& v) {x-=v.x ; y-= v.y ; z-=v.z ;  return *this;} ;
  Point3D& operator*=(float v) {x*=v ; y*= v ; z*= v;  return *this;} ;
  Point3D& operator/=(float v) {x/=v ; y/= v ; z/= v ; return *this;} ;
  friend int operator==(const Point3D&,float) ;
  friend int operator==(const Point3D& a,const Point3D& b);
  friend int operator!=(const Point3D& a,const Point3D& b);
  
  friend Point3D operator+(const Point3D&,const Point3D&) ;
  friend Point3D operator-(const Point3D&,const Point3D&) ;
  friend float operator*(const Point3D& a,const Point3D& b){return dot(a,b) ;}
  friend Point3D operator*(const Point3D& a,const float b) {return b*a;}
  friend Point3D operator*(const float a,const Point3D& b) ;
  friend Point3D operator/(const Point3D&,const float) ;

  friend float angle(const Point3D&,const Point3D&) ;
  friend float abs(const Point3D& a) {return (float)sqrt(abs2(a));}
  friend float abs2(const Point3D& a) {return a.x*a.x+a.y*a.y+a.z*a.z;}
  friend float dot(const Point3D&, const Point3D&) ;
  friend Point3D crossProduct(const Point3D& a, const Point3D& b) ;
  Point3D unitLength() const;
 

  friend ostream& operator<<(ostream& os, const Point3D& point);
  friend istream& operator>>(istream& os, Point3D& point);
};

inline int operator<(const Point3D& a, const Point3D& b) { 
  return a.x<b.x || a.y<b.y || a.z<b.z ;}
inline int operator>(const Point3D& a, const Point3D& b) { 
  return a.x>b.x || a.y>b.y || a.z>b.z ;}
inline int operator<=(const Point3D& a, const Point3D& b) { 
  return a.x<=b.x || a.y<=b.y || a.z<=b.z ;}
inline int operator>=(const Point3D& a, const Point3D& b) { 
  return a.x>=b.x || a.y>=b.y || a.z>=b.z ;}


/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: CtrlPnt4D --- A class for point in the homogeneous 4D space
  Description: When using B-splines, homogeneous points are used. They are 
               defined with 4D points as $(x,y,z,w)$. In 3D space, a 4D point 
	       is equivalent to $(\frac{x}{w},\frac{y}{w},\frac{z}{w})$.

	       This class also defines basic mathematical operators for 4D 
	       homogeneous points.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class CtrlPnt4D {
public:
  float *data ; // this is actually a data[4] array having the x,y,z,w  info
  int created; // usefull to change the data pointer

  CtrlPnt4D(): data(new float[4]), created(1) { x() = y() = z() = 0 ; w() = 0.0 ; }
  CtrlPnt4D(float a): data(new float[4]), created(1) { x() = y() = z() = a ; w() = a ; }
  CtrlPnt4D(float X, float Y, float Z, float W) : data(new float[4]), created(1) { x() = X ; y() = Y ; z() = Z ; w()=W ; }
  CtrlPnt4D(const CtrlPnt4D& a): data(new float[4]), created(1) { x() = a.x() ; y() = a.y() ; z() = a.z() ; w() = a.w() ; }
  CtrlPnt4D(const Point3D& a): data(new float[4]), created(1) {  x() = a.x ; y() = a.y ; z() = a.z ; w() = 1.0 ; }
  ~CtrlPnt4D() { if(created) if(data) delete []data ; }

  inline float& x() { return data[0] ;}
  inline float x() const { return data[0] ; }
  inline float& y() { return data[1] ;}
  inline float y() const { return data[1] ;}
  inline float& z() { return data[2] ;}
  inline float z() const { return data[2] ;}
  inline float& w() { return data[3] ;}
  inline float w() const { return data[3] ;}
  

  CtrlPnt4D& operator=(const CtrlPnt4D& v) ;
  CtrlPnt4D& operator=(const Point3D& v) {x()=v.x ; y()=v.y ; z()=v.z ; w()=1.0 ; return *this;}
  CtrlPnt4D& operator=(const float v) { x() = y() = z() = w() = v ; return *this ;} ;
  CtrlPnt4D& operator+=(const CtrlPnt4D& v) ;
  CtrlPnt4D& operator-=(const CtrlPnt4D& v) ; 
  CtrlPnt4D& operator*=(float v)  ;
  CtrlPnt4D& operator/=(float v)  ;
  friend int operator==(const CtrlPnt4D&,float) ;
  friend int operator==(const CtrlPnt4D& a,const CtrlPnt4D& b);
  friend int operator!=(const CtrlPnt4D& a,const CtrlPnt4D& b);
  
  friend CtrlPnt4D operator+(const CtrlPnt4D&,const CtrlPnt4D&) ;
  friend CtrlPnt4D operator-(const CtrlPnt4D&,const CtrlPnt4D&) ;
  friend float operator*(const CtrlPnt4D& a,const CtrlPnt4D& b){return dot(a,b) ;}
  friend CtrlPnt4D operator*(const CtrlPnt4D& a,const float b) ;
  friend CtrlPnt4D operator*(const float a,const CtrlPnt4D& b) ;
  friend CtrlPnt4D operator/(const CtrlPnt4D&,const float) ;

  friend float angle(const CtrlPnt4D&,const CtrlPnt4D&) ;
  friend float abs(const CtrlPnt4D& a) {return sqrt(abs2(a));}
  friend float abs2(const CtrlPnt4D& a) ;
  friend float dot(const CtrlPnt4D&, const CtrlPnt4D&) ;

  friend Point3D project(const CtrlPnt4D& ) ;
  friend float distance3D(const CtrlPnt4D& a, const CtrlPnt4D& b) { return abs(project(a)-project(b)) ;}

  friend ostream& operator<<(ostream& os, const CtrlPnt4D& point);
  friend istream& operator>>(istream& os, CtrlPnt4D& point);

  void move(const Point3D& m) ;

};

inline int operator<(const CtrlPnt4D& a, const CtrlPnt4D& b) { 
  return abs(a)<abs(b) ? 1 : 0 ;}
inline int operator>(const CtrlPnt4D& a, const CtrlPnt4D& b) { 
  return abs(a)>abs(b) ? 1 : 0 ;}
inline int operator<=(const CtrlPnt4D& a, const CtrlPnt4D& b) { 
  return abs(a)<=abs(b) ? 1 : 0 ;}
inline int operator>=(const CtrlPnt4D& a, const CtrlPnt4D& b) { 
  return abs(a)>=abs(b) ? 1 : 0 ;}

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: Coordinate --- A class for a matrix coordinate point
  Description: 
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class Coordinate {
public:
  int i,j ;
  Coordinate() 
    { i=0;j=0;} 
  Coordinate(int I, int J) 
    { i=I ; j=J ; }
  Coordinate(int a) 
    { i=j=a; }

  const Coordinate& operator=(const Coordinate& c)
    {i=c.i; j=c.j; return *this ;}
  friend int operator==(const Coordinate& a, const Coordinate& b)
    { return ( a.i == b.i && a.j == b.j ); }

  friend float distance2(const Coordinate& a){ return float(a.i*a.i) + float(a.j*a.j) ; }
  friend float distance(const Coordinate& a) { return sqrt(distance(a)) ; }

  friend Coordinate operator-(const Coordinate& a, const Coordinate& b){ Coordinate m ; m.i = a.i-b.i ; m.j = a.j-b.j ; return m ; }
  friend Coordinate operator+(const Coordinate& a, const Coordinate& b){ Coordinate m ; m.i = a.i+b.i ; m.j = a.j+b.j ; return m ; }

  friend ostream& operator<<(ostream& os, const Coordinate& point);
  friend istream& operator>>(istream& os, Coordinate& point);
};

inline int operator<(const Coordinate& a, const Coordinate& b){ 
  return (a.i<b.i && a.j<b.j ) ; } // the smaller than operator
inline int operator>(const Coordinate& a, const Coordinate& b){ 
  return (a.i>b.i && a.j>b.j ) ; } // the greater than operator
inline int operator<=(const Coordinate& a, const Coordinate& b){ 
  return (a.i<=b.i && a.j<=b.j ) ; } // the smaller or equal operator
inline int operator>=(const Coordinate& a, const Coordinate& b){ 
  return (a.i>=b.i && a.j>=b.j ) ; } // the greater or equal operator

// Some often used inline functions

template <class T>
inline T absolute(T a){
  return a<0.0 ? -a : a ;
}

template <class T>
inline T minimum(T a, T b){
  return a<b ? a : b ;
}

template <class T>
inline T maximum(T a, T b){
    return a>b ? a : b ;
}

// Use minimumRef or maximumRef if you want to pass by reference
// You should only use this for special types and not for
// the base types
template <class T>
inline T minimumRef(const T &a, const T &b){
  return a<b ? a : b ;
}

template <class T>
inline T maximumRef(const T &a, const T &b){
    return a>b ? a : b ;
}

// definition for Complex, CtrlPnt4D, Point3D, Vector2 and Coordinate
// follows

inline Complex minimum(Complex a, Complex b){
  double r,i ;
  r = minimum(real(a),real(b)) ;
  i = minimum(imag(a),imag(b)) ;
  return Complex(r,i) ;
}

inline CtrlPnt4D minimum(CtrlPnt4D a, CtrlPnt4D b){
  CtrlPnt4D m ;
  m.x() = minimum(a.x(),b.x()) ;
  m.y() = minimum(a.y(),b.y()) ;
  m.z() = minimum(a.z(),b.z()) ;
  m.w() = minimum(a.w(),b.w()) ;
  return m ; 
}

inline Point3D minimum(Point3D a, Point3D  b){
  Point3D m ;
  m.x = minimum(a.x,b.x) ;
  m.y = minimum(a.y,b.y) ;
  m.z = minimum(a.z,b.z) ;
  return m ; 
}

inline Vector2 minimum(Vector2 a, Vector2 b){
  Vector2 m ;
  m.x = minimum(a.x,b.x) ;
  m.y = minimum(a.y,b.y) ;
  return m ;
}

inline Coordinate minimum(Coordinate a, Coordinate b){
  Coordinate m ;
  m.i = minimum(a.i,b.i) ;
  m.j = minimum(a.j,b.j) ;
  return m ;
}

inline Complex maximum(Complex a, Complex b){
  double r,i ;
  r = maximum(real(a),real(b)) ;
  i = maximum(imag(a),imag(b)) ;
  return Complex(r,i) ;
}

inline CtrlPnt4D maximum(CtrlPnt4D a, CtrlPnt4D b){
  CtrlPnt4D m ;
  m.x() = maximum(a.x(),b.x()) ;
  m.y() = maximum(a.y(),b.y()) ;
  m.z() = maximum(a.z(),b.z()) ;
  m.w() = maximum(a.w(),b.w()) ;
  return m ; 
}

inline Point3D maximum(Point3D a,Point3D  b){
  Point3D m ;
  m.x = maximum(a.x,b.x) ;
  m.y = maximum(a.y,b.y) ;
  m.z = maximum(a.z,b.z) ;
  return m ; 
}

inline Vector2 maximum(Vector2 a, Vector2 b){
  Vector2 m ;
  m.x = maximum(a.x,b.x) ;
  m.y = maximum(a.y,b.y) ;
  return m ;
}

inline Coordinate maximum(Coordinate a, Coordinate b){
  Coordinate m ;
  m.i = maximum(a.i,b.i) ;
  m.j = maximum(a.j,b.j) ;
  return m ;
}

inline Complex minimumRef(const Complex &a, const Complex &b){
  double r,i ;
  r = minimum(real(a),real(b)) ;
  i = minimum(imag(a),imag(b)) ;
  return Complex(r,i) ;
}

inline CtrlPnt4D minimumRef(const CtrlPnt4D &a, const CtrlPnt4D &b){
  CtrlPnt4D m ;
  m.x() = minimum(a.x(),b.x()) ;
  m.y() = minimum(a.y(),b.y()) ;
  m.z() = minimum(a.z(),b.z()) ;
  m.w() = minimum(a.w(),b.w()) ;
  return m ; 
}

inline Point3D minimumRef(const Point3D &a,const Point3D  &b){
  Point3D m ;
  m.x = minimum(a.x,b.x) ;
  m.y = minimum(a.y,b.y) ;
  m.z = minimum(a.z,b.z) ;
  return m ; 
}

inline Vector2 minimumRef(const Vector2 &a, const Vector2 &b){
  Vector2 m ;
  m.x = minimum(a.x,b.x) ;
  m.y = minimum(a.y,b.y) ;
  return m ;
}

inline Coordinate minimumRef(const Coordinate &a, const Coordinate &b){
  Coordinate m ;
  m.i = minimum(a.i,b.i) ;
  m.j = minimum(a.j,b.j) ;
  return m ;
}

inline Complex maximumRef(const Complex &a, const Complex &b){
  double r,i ;
  r = maximum(real(a),real(b)) ;
  i = maximum(imag(a),imag(b)) ;
  return Complex(r,i) ;
}

inline CtrlPnt4D maximumRef(const CtrlPnt4D &a, const CtrlPnt4D &b){
  CtrlPnt4D m ;
  m.x() = maximum(a.x(),b.x()) ;
  m.y() = maximum(a.y(),b.y()) ;
  m.z() = maximum(a.z(),b.z()) ;
  m.w() = maximum(a.w(),b.w()) ;
  return m ; 
}

inline Point3D maximumRef(const Point3D &a,const Point3D  &b){
  Point3D m ;
  m.x = maximum(a.x,b.x) ;
  m.y = maximum(a.y,b.y) ;
  m.z = maximum(a.z,b.z) ;
  return m ; 
}

inline Vector2 maximumRef(const Vector2 &a, const Vector2 &b){
  Vector2 m ;
  m.x = maximum(a.x,b.x) ;
  m.y = maximum(a.y,b.y) ;
  return m ;
}

inline Coordinate maximumRef(const Coordinate &a, const Coordinate &b){
  Coordinate m ;
  m.i = maximum(a.i,b.i) ;
  m.j = maximum(a.j,b.j) ;
  return m ;
}

inline Point3D crossProduct(const Point3D& a, const Point3D& b){
  Point3D prod ;
  prod.x = a.y*b.z - a.z*b.y ;
  prod.y = a.z*b.x - a.x*b.z ;
  prod.z = a.x*b.y - a.y*b.x ;
  return prod ;
}

inline Point3D Point3D::unitLength() const{
  float d = abs(*this) ;
  Point3D u ;
  u.x = x/d ;
  u.y = y/d ;
  u.z = z/d ;
  return u ;
}


// These functions are needed for qSort (in vector.hh and matrix.hh)

int compareInt(const void* a, const void* b) ;
int compareFloat(const void* a, const void* b) ;
int compareDouble(const void* a, const void* b) ;


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator*= --- multiplies itself with a float
  Description: Multiplies itself with a float
        Input: v  --> the value to multiply itself with
       Output: the control point multiplied by $v$.
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
inline CtrlPnt4D& CtrlPnt4D::operator*=(float v){
  data[0] *= v ; 
  data[1] *= v ;
  data[2] *= v ;
  data[3] *= v ; 
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* --- multiplying a point with a float
  Description: Multiplying a point with a float
        Input: b --> the point to multiply
	       a --> the floating point value to multiply with
       Output: $ab$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
inline CtrlPnt4D operator*(const CtrlPnt4D& b, const float a) {
  CtrlPnt4D mul(b) ;
  mul *= a ; 
  return mul ;
};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* --- multiplying a point with a float
  Description: Multiplying a point with a float
        Input: a --> the floating point value to multiply with
               b --> the point to multiply
       Output: $ab$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
inline CtrlPnt4D operator*(const float a,const CtrlPnt4D& b) {
  CtrlPnt4D mul(b) ;
  mul *= a ;
  return mul ;
};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: boundTo --- bounds a variable to a region
  Description: Bounds a value to a region, the variable $a$ is not changed
               if it's already in the region. Otherwise it is set to one 
	       of the boundary value.
        Input: a <-> the variable to bound
   	     low --> the low bound value
	    high --> the higest bound value
       Output: The variable $a$ is now in the range $[low,high]$.
 Restrictions: The variable low {\em must} be smaller then high.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> inline void boundTo(T& a, T low, T high){
  if(a<=low) a = low ;
  if(a>=high) a = high ;
}



#ifdef CPP_SRC_INCLUDE
#include "../src/specialType.cc"
#endif



#endif 
