/*=============================================================================
        File: error.cc
     Purpose: Implementation of general error reporting class
    Revision: $Id: error.cc,v 1.2 1998/05/01 19:53:38 lavoie Exp $
  Created by:    Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/

#include <stream.h>
#include <string.h>
#include "internals.hh"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlError --- constructor 
  Description:
        Input: title --> the title of the routine where the error occurs
       Output: 
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlError::PlError(const char *title) :  ostrstream(), prog(0)
{
  int l = strlen( title ) + 1;

  prog = new char[l];
  strcpy(prog, title);
  init (&__my_sb);
  clear() ;

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: report --- reports an error message
  Description: The usual call to this function is made through the 
               $\ll$ operator, {\em i.e.} 
	       \begin{verbatim}
	         PlError error("routine name") ;
		 error << "An error occured " ;
		 error.report() ;
	       \end{verbatim}
        Input: msg --> the message to report
       Output: 
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlError::report(const char *msg)
{
  if ( msg == 0 )		// message must have been sent via <<
    cerr << str() ;
  else
    cerr << msg;

  cerr << '\n';
#ifdef DEBUG_MATRIX
  cerr << "\n\nThe program is now in an infinte loop. Press CTRL-c to exit.\n" ;
#endif
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: warning --- reports a warning error message
  Description: The usual call to this function is made through the 
               $\ll$ operator, {\em i.e.} 
	       \begin{verbatim}
	         PlError error("routine name") ;
		 error << "An error occured " ;
		 error.warning() ;
	       \end{verbatim}
        Input: msg --> the message to report
       Output: 
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlError::warning(const char *msg)
{
  cerr << "\nRoutine: " << prog << "\nWarning: ";

  report( msg );

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: fatal --- reports a fatal error
  Description: Reports a fatal error. If the \verb.DEBUG_MATRIX. flag has been
               set then the routine starts an infinte loop. An exit(1) is
	       executed otherwise.

	       The usual call to this function is made through the 
               $\ll$ operator, {\em i.e.} 
	       \begin{verbatim}
	         PlError error("routine name") ;
		 error << "A fatal error occured " ;
		 error.fatal() ;
	       \end{verbatim}

        Input: msg --> a message to report
       Output: 
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlError::fatal(const char *msg)
{
  cerr << "\nRoutine: " << prog << "\nFatal error: ";
  
  report( msg );
#ifdef DEBUG_MATRIX
  while(1){ ; }
#else
  exit(1);
#endif
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: memory --- reports a memory allocation error
  Description: Reports a memory allocation error. If the \verb.DEBUG_MATRIX.
               flag has been set then the routine starts an infinte loop. 
	       An exit(1) is executed otherwise.
	       
	       The usual call to this function is made through the 
               $\ll$ operator, {\em i.e.} 
	       \begin{verbatim}
	         PlError error("variable name") ;
		 float* var = new float[bigSize] ;
		 error.memory(var) ;
	       \end{verbatim}

        Input: p --> the pointer to test
       Output: 
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlError::memory(const void *p)
{
  if ( p == 0)
    {
      cerr << "\nRoutine: " << prog << " Memory allocation error\n";
#ifdef DEBUG_MATRIX
      while (1) { ; }
#else
      exit(1);
#endif
    }  
}
