/*=============================================================================
        File: image.cc
     Purpose:
    Revision: $Id: image.cc,v 1.2 1998/05/08 20:13:25 lavoie Exp $
  Created by:    Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/

#include "image.hh"
#include <stdio.h>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawLine --- draws a line from point 1 to point 2
  Description: Draws a line without aliasing from point $(i_1,j_1)$ to 
               $(i_2,j_2)$ with the value {\tt color}.
        Input: i1 --> the row of point 1
               j1 --> the column of point 1
	       i2 --> the row of point 2
	       j2 --> the column of point 2
            color --> the line is drawn with this value
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
void MatrixImage<T>::drawLine(int i1, int j1, int i2, int j2, T color){
  int i,j ;
  double mx,b ;
  if(i1<0 || i2 <0 || j1<0 || j2<0 || i1>rows || 
     i2>rows || j1>=cols || j2>=cols ){
    PlError error("MatrixImage<T>::drawLine") ;
    error << "Error in drawing line\n Invalid index ("<< i1 << ", " << j1 << ") to ( " << i2 << ", " << j2 << ") \n" ;
    error.warning() ;
    return ;
  }

  // check if line is vertical
  if(j1==j2){
    for(i=minimum(i1,i2);i<=maximum(i1,i2);i++)
     operator()(i,j1) = color ;
    return ;
  }
  mx = (double)(i1-i2)/(double)(j1-j2) ;
  b = (double)i1 - mx*j1 ;
  if(absolute(i1-i2)>absolute(j1-j2)){ // draw vertically
    if(i1>i2){
      for(i=i1;i>=i2;i--){
	j = int(((double)i-b)/mx) ;
	operator()(i,j) = color ;
      }
    }
    else{
      for(i=i1;i<=i2;i++){
	j = (int)((i-b)/mx) ;
	operator()(i,j) = color ;
      }
    }
  }
  else{
    if(j1>j2){
      for(j=j1;j>=j2;j--){
	i = (int)(mx*j+b) ;
	operator()(i,j) = color ;
      }
    }
    else{
      for(j=j1;j<=j2;j++){
	i = (int)(mx*j+b) ;
	operator()(i,j) = color ;
      }
    }
  }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawPoint --- draws a point of radius $r$
  Description: Draws a point of radius $r$ at location $(i,j)$ with the 
               value {\tt color}.
        Input: i --> the row of the center point to draw
               j --> the column of the center point to draw
               r --> the radius of the point
           color --> the value to draw the point with
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
void MatrixImage<T>::drawPoint(int i, int j, double r , T color){
  for(int y=i-int(ceil(r)) ; y<i+int(ceil(r)) ; y++)
    for(int x = j-int(ceil(r)) ; x<j+int(ceil(r)) ; x++){
      if(y>=0 && y<rows && x>=0 && x<cols){
	if( ((y-i)*(y-i)+(x-j)*(x-j))<= r*r)
	  operator()(y,x) = color ;
      }
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: store --- copies an image to a Matrix
  Description: copies the image to a matrix of the same type.
        Input: a <-- the matrix to store the image to
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
void MatrixImage<T>::store(PlMatrix<T>& a){
  if(a.rows != rows || a.cols != cols) {
    a.resize(rows,cols) ;
  }
  T *aptr, *bptr ;
  int size,i ;
  aptr = &a(0,0)-1 ;
  bptr = m-1 ;
  size = cols*rows ;
  for(i=0;i<size;i++)
    *(++aptr) = *(++bptr) ;  
}

#ifdef WITH_IMAGE_MAGICK

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: IM_ImageT --- default constructor
  Description:
        Input:
       Output:
 Restrictions: The \verb.IM_ImageT. class must be of type {\tt Color} or type 
               {\tt unsigned char}.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
IM_ImageT<T>::IM_ImageT():MatrixImage<T>(){
  autoSave = 0 ;
  file_name = 0 ;
  image = 0 ;
  GetImageInfo(&image_info) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: IM_ImageT --- constructor with a filename and a save flag
  Description: The image in {\tt filename} will be read. If the save flag is 
               set to 1, then when the destructor is called the image is 
	       saved to {\tt filename}.

	       If en error occurs, it is not reported. The use of the read 
	       and write member functions is recommended instead of this 
	       constructor.
        Input: filename --> the file to read an image from
                   save --> a flag indicating if autosave should occur, defaults to 0
       Output:
 Restrictions: The \verb.IM_ImageT. class must be of type {\tt Color} or type 
               {\tt unsigned char}. The image stored in {\tt filename} must 
	       be readable from the Image Magick library.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
IM_ImageT<T>::IM_ImageT(const char* filename, int save):MatrixImage<T>(){
  autoSave = save ;
  file_name = new char[1024] ; // 1024 characters for the name of a file, should be enough
  (void)strcpy(file_name,filename) ;
  GetImageInfo(&image_info) ;
  read(filename) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  IM_ImageT --- constructor specifying the size of the image
  Description:
        Input: r --> the number of rows for the image
             c --> the number of columns for the image
       Output:
 Restrictions: The \verb.IM_ImageT. class must be of type {\tt Color} or type 
               {\tt unsigned char}.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
IM_ImageT<T>::IM_ImageT(const int r, const int c):MatrixImage<T>(r,c){
  autoSave = 0 ;
  file_name = 0 ;
  image = 0 ;
  GetImageInfo(&image_info) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: read ---  reads from a file an image
  Description: Reads an image file. The filename should be in the format 
               ``imagename.type'' or ``type:imagename''.
        Input: filename --> the name of the file to read
       Output:
 Restrictions: The \verb.IM_ImageT. class must be of type {\tt Color} or type 
               {\tt unsigned char}. The image stored in {\tt filename} must 
	       be readable from the Image Magick library.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
int IM_ImageT<T>::read(const char* filename) {
  PlError error("IM_ImageT<T>::read") ;
  error << "An image of this type is NOT supported!\n" ;
  error.fatal() ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: write --- writes to a file an image
  Description: Writes to a file an image. The filename should be in the 
               format "imagename.type" or "type:imagename".
        Input: filename --> the file to write to. 
       Output:
 Restrictions: The \verb.IM_ImageT. class must be of type {\tt Color} or type 
               {\tt unsigned char}. {\tt filename} must specify a valid type 
	       for the Image Magick library.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
int IM_ImageT<T>::write(const char* filename) {
  PlError error("IM_ImageT<T>::write") ;
  error << "An image of this type is NOT supported!\n" ;
  error.fatal() ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: setMatrix --- sets the matrix internal data from the image data
  Description: The matrix data is set to the one from the image data.
               This step is necessary since the ImageMagick data format 
	       can't be made compatible with the matrix data format. The data 
	       between the two internal formats must be copied each time 
	       because of this.
        Input: filename --> the file to write to. 
       Output:
 Restrictions: The image data {\em must} be initialized properly.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
void IM_ImageT<T>::setMatrix(){
  PlError error("IM_ImageT<T>::setImage") ;
  error << "An image of this type is NOT supported!\n" ;
  error.fatal() ;
}

void IM_ImageT<Color>::setMatrix(){
#ifdef COLUMN_ORDER
  for(int i=0;i<rows;++i)
    for(int j=0;j<cols;++j){
      vm[j][i].r = image->pixels[i*cols+j].red ;
      vm[j][i].g = image->pixels[i*cols+j].green ;
      vm[j][i].b = image->pixels[i*cols+j].blue ;
    }
#else
  int size=rows*cols ;
  for(int i=0;i<size;++i){
    m[i].r = image->pixels[i].red ;
    m[i].g = image->pixels[i].green ;
    m[i].b = image->pixels[i].blue;
  }
#endif
}

void IM_ImageT<unsigned char>::setMatrix(){
#ifdef COLUMN_ORDER
  for(int i=0;i<rows;++i)
    for(int j=0;j<cols;++j)
      vm[j][i] = image->pixels[i*cols+j].red ;
#else
  int size=rows*cols ;
  for(int i=0;i<size;++i){
    m[i] = image->pixels[i].red ;
  }
#endif
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: setImage --- sets the image data from the matrix data.
  Description: Sets the image data from the matrix data. This step is necessary
               since the ImageMagick data format can't be made compatible with
	       the matrix data format. The data between the two internal 
	       formats must be copied each time because of this.
        Input: filename --> the file to write to. 
       Output:
 Restrictions: image must be initialized.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
void IM_ImageT<T>::setImage(){
  PlError error("IM_ImageT<T>::setImage") ;
  error << "An image of this type is NOT supported!\n" ;
  error.fatal() ;
}

void IM_ImageT<Color>::setImage(){
  if(image->rows != (unsigned int)rows || image->columns != (unsigned int)cols){
    image->rows = rows ;
    image->columns = cols ;
    image->packets = rows*cols ;
    if(image->pixels)
      delete []image->pixels ;
    image->pixels = (RunlengthPacket*)malloc(sizeof(RunlengthPacket)*image->packets) ;
  }

#ifdef COLUMN_ORDER
  for(int i=0;i<rows;++i)
    for(int j=0;j<cols;++j){
      image->pixels[j*rows+i].red = vm[j][i].r ;
      image->pixels[j*rows+i].green = vm[j][i].g ; 
      image->pixels[j*rows+i].blue =  vm[j][i].b ; 
      image->pixels[j*rows+i].index =  0;
      image->pixels[j*rows+i].length = 0 ;
  }
#else
  int size=rows*cols ;
    for(int i=0;i<size;++i){
    image->pixels[i].red = m[i].r ;
    image->pixels[i].green = m[i].g ; 
    image->pixels[i].blue =  m[i].b ; 
    image->pixels[i].index =  0;
    image->pixels[i].length = 0 ;
  }
#endif
}

void IM_ImageT<unsigned char>::setImage(){
  if(image->rows != (unsigned int)rows || image->columns != (unsigned int)cols){
    image->rows = rows ;
    image->columns = cols ;
    image->packets = rows*cols ;
    if(image->pixels)
      delete []image->pixels ;
    image->pixels = (RunlengthPacket*)malloc(sizeof(RunlengthPacket)*image->packets) ;
  }

#ifdef COLUMN_ORDER
  for(int i=0;i<rows;++i)
    for(int j=0;j<cols;++j){
      image->pixels[j*rows+i].red = vm[j][i] ;
      image->pixels[j*rows+i].green = vm[j][i] ; 
      image->pixels[j*rows+i].blue =  vm[j][i] ; 
      image->pixels[j*rows+i].index =  vm[j][i] ;
      image->pixels[j*rows+i].length = 0 ;
  }
#else
  int size=rows*cols ;
  for(int i=0;i<size;++i){
    image->pixels[i].red = m[i] ;
    image->pixels[i].green = m[i] ;
    image->pixels[i].blue = m[i] ;
    image->pixels[i].index = m[i] ; 
    image->pixels[i].length = 0 ;
  }
#endif
}



int IM_ImageT<Color>::read(const char* filename) {
  (void)strcpy(image_info.filename,filename) ;
  if(image)
    DestroyImage(image) ;
  image=ReadImage(&image_info) ;
  if(!image)
    return 0 ;
  UncompressImage(image) ;
  resize(image->rows,image->columns) ;

  setMatrix() ;

  return 1 ;
}

int IM_ImageT<unsigned char>::read(const char* filename) {
  (void)strcpy(image_info.filename,filename) ;
  image_info.monochrome = 1 ;
  if(image)
    DestroyImage(image) ;
  image=ReadImage(&image_info) ;
  if(!image)
    return 0 ;
  UncompressImage(image) ;
  resize(image->rows,image->columns) ;

  setMatrix() ;

  return 1 ;
}

int IM_ImageT<Color>::write(const char* filename) {

  if(image)
    DestroyImage(image) ;

  image=AllocateImage(&image_info) ;

  if(!image)
    return 0 ;
  
  (void)strcpy(image->filename,filename) ;

  image->rows = rows ;
  image->columns = cols ;
  image->packets=rows*cols ;
  image->pixels = (RunlengthPacket*)malloc(sizeof(RunlengthPacket)*image->packets) ;

  setImage() ;

  return WriteImage(&image_info,image) ;
}

int IM_ImageT<unsigned char>::write(const char* filename) {
  if(image)
    DestroyImage(image) ;

  image=AllocateImage(&image_info) ;

  if(!image)
    return 0 ;
  
  (void)strcpy(image->filename,filename) ;

  image->rows = rows ;
  image->columns = cols ;
  image->packets=rows*cols ;
  image->pixels = (RunlengthPacket*)malloc(sizeof(RunlengthPacket)*image->packets) ;

  setImage() ;

  image_info.monochrome = 1 ;  
  return WriteImage(&image_info,image) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: ~IM_ImateT --- destructor
  Description: If the autosave flag has been set, it saves the image to a file.
        Input:
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
IM_ImageT<T>::~IM_ImageT(){
  if(autoSave && file_name)
    write(file_name) ;
  if(image)
    DestroyImage(image) ;

  DestroyImageInfo(&image_info) ;
  if(file_name){
    delete []file_name ;
  }
}

#if defined(WITH_IMAGE_MAGICK) && defined(__GNUG__)
template class IM_ImageT<unsigned char> ;
template class IM_ImageT<Color> ;
#endif

#endif

#if defined(__GNUG__) && !defined(CPP_SRC_INCLUDE)

// instantiation for g++ to be used with -fno-implicit-templates

template class MatrixImage<int> ;
template class MatrixImage<float> ;
template class MatrixImage<double> ;
template class MatrixImage<char> ;
template class MatrixImage<unsigned char> ;
template class MatrixImage<Color> ;



#else

void dumbSparc_Image(void){
  MatrixImage<unsigned char> a ;
  MatrixImage<int> b ;
  MatrixImage<double> c ;
  MatrixImage<char> d ;
  MatrixImage<Color> aa ;
  MatrixImage<Color> g ;
#ifdef WITH_IMAGE_MAGICK
  IM_ImageT<unsigned char> A;
  IM_ImageT<Color> B;
#endif
}

#endif
