/*=============================================================================
        File: matrix.cc
     Purpose:
    Revision: $Id: matrix.cc,v 1.3 1998/06/03 14:22:24 lavoie Exp $
  Created by:    Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1998 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/


#include "internals.hh"
#include <fstream.h>
#include <string.h>


#include "matrix.hh"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator= --- assignment operator
  Description: Assignment operator
        Input: a --> the matrix to copy
       Output:
 Restrictions: the matrix a must have compatible dimensions
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T>& PlMatrix<T>::operator=(const PlMatrix<T> &a){
  int i;
  
  if ( this == &a )
    return *this;
  
  if ( a.rows != rows || a.cols != cols ){
    resize(a.rows,a.cols) ;
  }
  
  int sze = rows*cols ;
  T *ptr, *aptr ;
  ptr = m-1 ;
  aptr = a.m-1 ;
  
  for (i = 0; i < sze; ++i)
    *(++ptr) = *(++aptr) ;
  
  by_columns = a.by_columns;
  
  return *this;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: submatrix --- sets the submatrix $(s_r,s_c)$ to $a$.
  Description: Sets the submatrix at $(s_r,s_c)$ to $a$. The matrix can be 
               viewed as
	       \[ \left[\begin{array}{ccc}
	         sub_{0,0} & \vdots & sub_{0,m} \\
	         \cdots & \cdots & \cdots \\
	         sub_{n,0} & \vdots & sub_{n,m} \end{array} \right] \]
	       where each sub matrix is of the size of $a$. This function 
	       replaces $sub_{s_r,s_c}$ with $a$.

        Input: sr --> the row of the submatrix
               sc --> the column of the submatrix
                a --> the submatrix to copy from
       Output:
 Restrictions: Since the size of $a$ defines the size of the submatrices, 
               this size must be such that a submatrix located at $(s_r,s_c)$ 
	       exists.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
void PlMatrix<T>::submatrix(int sr, int sc, PlMatrix<T> &a)
{
  int rwz,coz,i,j;
  
  if ( rows % a.rows != 0 || cols % a.cols != 0 || rows < a.rows || cols < a.cols )
    {
      PlError error("PlMatrix<T>::submatrix");
      error << "PlMatrix and submatrix incommensurate" ;
      error.fatal() ;
    }
  
  if ( sr >= rows/a.rows || sr < 0 || sc >= cols/a.cols || sc < 0 )
    {
      PlError error("PlMatrix<T>::submatrix");
      error << "Submatrix location out of bounds.\nrowblock " << sr << ", " << rows/a.rows << " colblock " << sc << ", " << a.cols << endl ;
      error.fatal() ;
    }
  rwz = sr*a.rows;
  coz = sc*a.cols;
  
#ifdef COLUMN_ORDER
  for ( i = 0; i < a.rows; ++i )
    for(j=0;j<a.cols;++j)
      elem(i+rwz,j+coz) = a(i,j) ;
#else
  T *ptr, *aptr ;
  aptr = a.m - 1;
  for ( i = 0; i < a.rows; ++i )
    {
      ptr = &m[(i+rwz)*cols+coz]-1 ;
      for ( j = 0; j < a.cols; ++j)
	*(++ptr) = *(++aptr) ;
    }  
#endif
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: as --- copies a matrix into this matrix starting at 
                      index $(rw,cl)$.
  Description: The matrix $a$ is copied into this one starting at $(rw,cl)$.
        Input: rw --> the row to insert the matrix at
               cl --> the column to insert the matrix at
                a --> the matrix to insert
       Output:
 Restrictions: The matrix $a$ must fit inside the matrix starting from 
               $(rw,cl)$.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> void
PlMatrix<T>::as(int rw, int cl, PlMatrix<T>& a)
{
  // Assign matrix a to this matrix at (i,j)
  int i, j;
  
  if ( (rw + a.rows) > rows || ( cl + a.cols) > cols) {
    PlError error("PlMatrix<T>::as") ;
    error << "PlMatrix A will not fit in this PlMatrix at " << rw << ", " << cl << endl ;
    error.fatal() ;
  }

#ifdef COLUMN_ORDER
  for(i=0;i<a.rows;++i)
    for(j=0;j<a.cols;++j)
      elem(i+rw,j+cl) = a(i,j) ;
#else
  T *pptr,*aptr ;
  aptr = a.m-1 ;
  for ( i = 0; i<a.rows; ++i) {
    pptr = &m[(i+rw)*cols+cl]-1 ;
    for ( j = 0; j < a.cols; ++j)
      *(++pptr) = *(++aptr);
  }
#endif
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: get --- returns the matrix of size $(n_r,n_c)$ starting at 
                       $rw,cl$.
  Description: This generates a matrix of size $(n_r,n_c)$ with its first 
               point located at $rw,cl$.
        Input: rw --> the index of the row 
               cl --> the index of the column
	       nr --> the number of rows of the matrix to generate
               nc --> the number of coluns of the matrix to generate
       Output: the matrix of size $n_r,n_c$ starting at index $(rw,cl)$.
 Restrictions: The matrix to return must fit inside the original matrix.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T> 
PlMatrix<T>::get(int rw, int cl, int nr, int nc) const
{
  PlMatrix<T> getmat(nr,nc) ;
  if ( (rw+nr) > rows || (cl+nc) > cols) {
    PlError error("PlMatrix<T>::get") ;
    error << "PlMatrix of size "<< nr << ", " << nc << " not available at " << rw << ", " << cl << endl ;
    error.fatal() ;
  }
  
  int i, j;

#ifdef COLUMN_ORDER
  for(i=0;i<nr;++i)
    for(j=0;j<nc;++j)
      getmat(i,j) = elem(i+rw,j+cl) ;
#else
  T *pptr,*aptr ;
  aptr = getmat.m-1;
  for (i = 0; i < nr; ++i) {
    pptr = &m[(i+rw)*cols+cl]-1 ;
    for ( j = 0; j < nc; ++j)
      *(++aptr) = *(++pptr) ;
  }
#endif
  return getmat ;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: norm --- Finds the first norm of the matrix
  Description: Finds the first norm of the matrix
        Input:
       Output: the norm of the matrix
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> 
double PlMatrix<T>::norm(void){
  int i,j ;
  double sum, maxsum;
  int init=0 ;
  T *pptr ;
  pptr = m-1 ;
  maxsum = 0 ; // Silence the warning message
  for(i=0;i<rows;++i){
    sum = 0 ;
    for ( j = 0; j < cols; ++j) 
      sum += *(++pptr) ;
    if(init)
      maxsum = (maxsum>sum) ? maxsum : sum;
    else{
      maxsum = sum ;
      init = 1;
    }
  }
  return maxsum;
}

double
PlMatrix<Complex>::norm(void){
  int i,j ;
  double sumR, sumI, maxsum;
  int init=0 ;
  Complex *ptr ;
  ptr = m-1 ;
  maxsum = -1 ; // shuts up the warning messages
  for(i=0;i<rows;++i){
    sumR = 0.0 ;
    sumI = 0.0 ;
    for ( j = 0; j < cols; ++j) {
      sumR += real(*ptr)*real(*ptr) ;
      sumI += imag(*ptr)*imag(*ptr) ;
    }
    if(init)
      maxsum = (maxsum>(sumR+sumI)) ? maxsum : (sumR+sumI);
    else{
      maxsum = (sumR+sumI) ;
      init = 1;
    }
    ++ptr ;
  }
  return sqrt(maxsum);
}

double
PlMatrix<Color>::norm(void) {
  PlError error("PlMatrix<Color>::norm()") ;
  error << "ERROR: you can't get the norm of a Color matrix\n" ;
  error.fatal() ;
  return 0 ;
}

double
PlMatrix<Point3D>::norm(void) {
  int i,j ;
  double sumX, sumY, sumZ, maxsum;
  int init=0 ;
  Point3D *ptr ;
  ptr = m-1 ;
  maxsum = -1 ; // shuts up the warning messages 
  for(i=0;i<rows;++i){
    sumX = 0.0 ;
    sumY = 0.0 ;
    sumZ = 0.0 ;
    for ( j = 0; j < cols; ++j) {
      sumX += (*ptr).x * (*ptr).x ;
      sumY += (*ptr).y * (*ptr).y ;
      sumZ += (*ptr).z * (*ptr).z ;
    }
    if(init)
      maxsum = (maxsum>(sumX+sumY+sumZ)) ? maxsum : (sumX+sumY+sumZ);
    else{
      maxsum = (sumX+sumY+sumZ) ;
      init = 1;
    }
    ++ptr ;
  }
  return sqrt(maxsum);
}

double
PlMatrix<CtrlPnt4D>::norm(void) {
  int i,j ;
  double sumX, sumY, sumZ, sumW, maxsum;
  int init=0 ;
  CtrlPnt4D *ptr ;
  ptr = m-1 ;
  maxsum = -1 ; // shuts up the warning messages 
  for(i=0;i<rows;++i){
    sumX = 0.0 ;
    sumY = 0.0 ;
    sumZ = 0.0 ;
    sumW = 0.0 ;
    for ( j = 0; j < cols; ++j) {
      sumX += (*ptr).x() * (*ptr).x() ;
      sumY += (*ptr).y() * (*ptr).y() ;
      sumZ += (*ptr).z() * (*ptr).z() ;
      sumW += (*ptr).w() * (*ptr).w() ;
    }
    if(init)
      maxsum = (maxsum>(sumX+sumY+sumZ+sumW)) ? maxsum : (sumX+sumY+sumZ+sumW);
    else{
      maxsum = (sumX+sumY+sumZ+sumW) ;
      init = 1;
    }
    ++ptr ;
  }
  return sqrt(maxsum);
}

double
PlMatrix<Vector2>::norm(void){
  int i,j ;
  double sumX, sumY, maxsum;
  int init=0 ;
  Vector2 *ptr ;
  ptr = m-1 ;
  maxsum = -1 ; // shuts up the warning messages 
  for(i=0;i<rows;++i){
    sumX = 0.0 ;
    sumY = 0.0 ;
    for ( j = 0; j < cols; ++j) {
      sumX += (*ptr).x * (*ptr).x ;
      sumY += (*ptr).y * (*ptr).y ;
    }
    if(init)
      maxsum = (maxsum>(sumX+sumY)) ? maxsum : (sumX+sumY);
    else{
      maxsum = (sumX+sumY) ;
      init = 1;
    }
    ++ptr ;
  }
  return sqrt(maxsum);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: diag --- sets the diagonal of the matrix to $a$
  Description: Sets the diagonal points of the matrix to $a$. The diagonal 
               points are $(0,0),(1,1),(2,2)$,etc.
        Input: a --> the value to set the diagonal to
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
void PlMatrix<T>::diag(const T a)
{
  int i, iend;
  
  iend = rows;
  if ( iend > cols )
    iend = cols;
  
  for (i = 0; i < iend; ++i)
    elem(i,i) = a;

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: getDiag --- returns the diagonal of the matrix
  Description: Returns a vector with the component $[i]$ being set to the 
               component $(i,i)$ of the matrix.
        Input: 
       Output: the vector representing the diagonal of the matrix.
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlVector<T> PlMatrix<T>::getDiag(){
  int i, iend;
  PlVector<T> vec(minimum(rows,cols)) ;
  iend = minimum(rows,cols);
  for (i = 0; i < iend; ++i)
      vec[i] = elem(i,i);
  return vec ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator+= --- increase every elements
  Description: Increase everey elements by a double
        Input: a --> the value to increase the elements by
       Output: a reference to itself
 Restrictions: 
   Created by: Philippe Lavoie (1 June 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T>&
PlMatrix<T>::operator+=(double a)
{
  T *p1 ;
  p1 = m-1 ;
  const int size = rows*cols ;
  for(int i=0;i<size; ++i)
    *(++p1) += a ;  
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator-= --- decrease every elements
  Description: Decrease everey elements by a double
        Input: a --> the value to decrease the elements by
       Output: a reference to itself
 Restrictions: 
   Created by: Philippe Lavoie (1 June 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T>&
PlMatrix<T>::operator-=(double a)
{
  T *p1 ;
  p1 = m-1 ;
  const int size = rows*cols ;
  for(int i=0;i<size; ++i)
    *(++p1) -= a ;  
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator*= --- multiply every elements
  Description: Multiply everey elements by a double
        Input: a --> the value to mutiply the elements with
       Output: a reference to itself
 Restrictions: 
   Created by: Philippe Lavoie (1 June 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T>&
PlMatrix<T>::operator*=(double a)
{
  T *p1 ;
  p1 = m-1 ;
  const int size = rows*cols ;
  for(int i=0;i<size; ++i)
    *(++p1) *= a ;  
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator/= --- divide every elements
  Description: Divide everey elements by a double
        Input: a --> the value to divide the elements with
       Output: a reference to itself
 Restrictions: 
   Created by: Philippe Lavoie (1 June 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T>&
PlMatrix<T>::operator/=(double a)
{
  T *p1 ;
  p1 = m-1 ;
  const int size = rows*cols ;
  for(int i=0;i<size; ++i)
    *(++p1) /= a ;  
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator+= --- adds a matrix to itself
  Description: Adds a matrix to itself
        Input: a --> the matrix to increment itself with
       Output: a reference to itself
 Restrictions: the matrix a must have a size compatible with the matrix
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T>&
PlMatrix<T>::operator+=(const PlMatrix<T> &a)
{
  if ( a.rows != rows || a.cols != cols )
    {
      PlError error("PlMatrix<T>::operator+=") ;
      if ( rows != a.rows )
	error << "Matrices are of diferent size, a.rows = " << rows << " and b.rows = " << a.rows << endl ;
      if ( cols != a.cols)
	error << "Matrices are of diferent size, a.cols = " << cols << " and b.cols = " << a.cols << endl ;
      error.fatal() ;
    }

  int i, sze ;
  T *aptr,*sptr ;
  aptr = a.m - 1 ;
  sptr = m - 1 ;
  sze = rows*cols ;
  for (i = 0; i < sze; ++i){
      *(++sptr) += *(++aptr) ;
  }
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator+ --- the addition operator
  Description: The addition operator
        Input: a --> the first matrix to add
               b --> the second matrix to add
       Output: a matrix equal to $a+b$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T> operator+(const PlMatrix<T> &a,const PlMatrix<T> &b)
{
  PlMatrix<T> sum(a) ;
  sum += b ;
  return sum;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator-= --- self substraction
  Description: This will substract the matrix a from the matrix. The result is
               thus $matrix = matrix - a$.
        Input: a --> the matrix to substract
       Output: a reference to itself 
 Restrictions: The matrix $a$ must be compatible with this matrix
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T> PlMatrix<T>&
PlMatrix<T>::operator-=(const PlMatrix<T> &a)
{
  if ( a.rows != rows || a.cols != cols )
    {
      PlError error("PlMatrix<T>::operator-=") ;
      if ( rows != a.rows )
	error << "Matrices are of diferent size, a.rows = " << rows << " and b.rows = " << a.rows << endl ;
      if ( cols != a.cols)
	error << "Matrices are of diferent size, a.cols = " << cols << " and b.cols = " << a.cols << endl ;
      error.fatal() ;
    }

  int i, size;
  T *aptr,*sptr ;
  aptr = a.m - 1 ;
  sptr = m - 1 ;
  size = rows*cols ;
  for (i = 0; i < size; ++i){
      *(++sptr) -= *(++aptr) ;
  }
  return *this ;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator- --- the substraction operator
  Description: The substraction operator
        Input: a --> the matrix to substract from 
	       b --> the matrix to substract
       Output: a matrix equal to $a-b$
 Restrictions: the matrix must be compatible
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T> operator-(const PlMatrix<T> &a,const PlMatrix<T> &b)
{
  PlMatrix<T> diff(a) ;
  diff -= b ;
  return diff;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* --- the multiplication operator
  Description: Multiplies a with b.
        Input: a --> a matrix
               b --> the matrix to multiply with
       Output: A matrix equal to $a b$
 Restrictions: The matrix must be compatible for the multiplication: a.cols == b.rows
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T> operator*(const PlMatrix<T> &a,const PlMatrix<T> &b)
{
  if ( a.cols != b.rows )
    {
      PlError error("PlMatrix<T> operator*(PlMatrix<T>&,PlMatrix<T>&)");
      error << "PlMatrix<T> a * PlMatrix<T> b incommensurate, a.cols = " << a.cols << ", b.rows = " << b.rows << endl ;
      error.fatal() ;
    }

  int i, j, k, row=a.rows, col=b.cols,size = a.cols;
  PlMatrix<T> prod(row,col);
  T zero = (T)0;
  
#ifdef COLUMN_ORDER
  for(i=0;i<row;++i)
    for(j=0;j<size;++j)
      if(a(i,j) != zero){
	for(k=0;k<col; ++k)
	  prod(i,k) += a(i,j)* b(j,k) ;
      }
#else
  T *pptr,*aptr,*bptr ;
  aptr = a.m ;
  for (i = 0; i < row; ++i)
    for (j = 0; j < size; ++j){
      if ( *aptr != zero )
	{
	  pptr = prod[i]-1;
	  bptr = b[j]-1;
	  for (k = 0; k < col; ++k){
	    *(++pptr) += *aptr * *(++bptr);
	  }
	}
      ++aptr;
    }
#endif
  return prod;
  
}

PlMatrix<Vector2> operator*(const PlMatrix<Vector2> &a,const PlMatrix<Vector2> &b)
{
  PlError error("PlMatrix<Vector2> operator*(PlMatrix<Vector2>,PlMatrix<Vector2>) ");
  error << "These types of matrices can't be multiplied together and yield a PlMatrix<Vector2>\n";
  error.fatal();
   
  PlMatrix<Vector2> prod(a);
  b(0,0) ; // shuts up the warning messages
  return prod;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* --- multiplication of a double and a matrix
  Description: Multiplies a double with a matrix
        Input: d --> the double value to scale the matrix with
               a --> the matrix to scale with the double value $d$
       Output: A matrix equal to $d A$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T>  operator*(const double d,
			       const PlMatrix<T> &a)
{
  int i, size=a.rows*a.cols ;
  PlMatrix<T> b(a.rows,a.cols);
  
  T *bptr,*aptr ;
  bptr = b.m - 1 ;
  aptr = a.m - 1 ;
  for (i = 0; i < size; ++i)
    *(++bptr) = (T)(d * (*(++aptr))) ;
  return b;
  
}

PlMatrix<Complex>  operator*(const double d,
			   const PlMatrix<Complex> &a)
{
  int i, size=a.rows*a.cols ;
  PlMatrix<Complex> b(a.rows,a.cols);
  
  Complex *bptr,*aptr ;
  bptr = b.m - 1 ;
  aptr = a.m - 1 ;
  for (i = 0; i < size; ++i)
    *(++bptr) = (Complex)(d * (*(++aptr))) ;
  return b;
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* --- multiplies a matrix with a complex value
  Description: Multiplies a matrix with a complex value
        Input: d --> a complex value
               a --> a matrix
       Output: returns a matrix equal to $d A$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T>  operator*(const Complex &d,
			       const PlMatrix<T> &a)
{
  int i, size=a.rows*a.cols ;
  PlMatrix<T> b(a.rows,a.cols);
  
  T *bptr,*aptr ;
  bptr = b.m - 1 ;
  aptr = a.m - 1 ;
  for (i = 0; i < size; ++i)
    *(++bptr) = (T)real(d) * *(++aptr) ;
  return b;
}

PlMatrix<Complex>  operator*(const Complex &d,
			       const PlMatrix<Complex> &a)
{
  int i, size=a.rows*a.cols ;
  PlMatrix<Complex> b(a.rows,a.cols);
  
  Complex *bptr,*aptr ;
  bptr = b.m - 1 ;
  aptr = a.m - 1 ;
  for (i = 0; i < size; ++i)
    *(++bptr) = d * *(++aptr) ;
  return b;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* --- multiplies a matrix with a vector
  Description: Multiplies a matrix with a vector
        Input: a --> the matrix
               x --> the vector
       Output: returns a vector representing $a x$
 Restrictions: The matrix and the vector must be of compatible sizes
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlVector<T> operator*(const PlMatrix<T> &a, const PlVector<T> &x)
{
  if ( a.cols != x.size() )
    {
      PlError error("PlMatrix<T> operator*(PlMatrix<T>& a,PlVector<T>& b)");
      error << "a * b incommensurate, a.cols = " << a.cols << ", b.rows = " << x.size() << endl ;
      error.fatal() ;
    }
  
  int i, k, row = a.rows, size = a.cols;
  PlVector<T> prod(row);
  T zero = (T)0;
  
  T *pptr,*aptr,*xptr ;
  aptr = a.m - 1 ;
  pptr = &prod[0] ;
  for (i = 0; i < row; ++i){
    xptr = x.memory()-1 ;
    for (k = 0, *pptr = zero; k < size; ++k)
      *pptr += *(++aptr) * *(++xptr) ;
    ++pptr ;
  }
  
  return prod;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator== --- the equality operator
  Description: Every elements are compared with each others. If one of them
               in matrix $a$ is not equal to the one in matrix $b$, then
	       the result is negative.

        Input: a --> the first matrix to compare 
               b --> the second matrix to compare
       Output: 1 if equal, 0 otherwise
 Restrictions: The matrices must be of compatible sizes
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
int operator==(const PlMatrix<T> &a,const PlMatrix<T> &b)
{
  if ( a.rows != b.rows || a.cols != b.cols )
    {
      PlError error("operator==(const PlMatrix<T>&,const PlMatrix<T>&)");
      if ( b.rows != a.rows )
	error << "Matrices are of diferent size, a.rows = " << a.rows << " and b.rows = " << b.rows << endl ;
      if ( b.cols != a.cols)
	error << "Matrices are of diferent size, a.cols = " << a.cols << " and b.cols = " << b.cols << endl ;
      error.fatal() ;
    }
  
  int i, j, row = a.rows, col = a.cols;
  int l = 1;
  
  for (i = 0; i < row; ++i)
    for (j = 0; j < col; ++j)
      l = l && ( a.elem(i,j) == b.elem(i,j) );
  
  return l;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: comm --- computes $a b - b a$ 
  Description: Computes $a b - b a$ 
        Input: a --> the a matrix
               b --> the b matrix
       Output:
 Restrictions: The a and b matrix must be compatible for the comm operation.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T> comm(const PlMatrix<T> &a,const PlMatrix<T> &b)
{
  PlMatrix<T> r = a * b - b * a;
  
  return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: trace --- The sum of all diagonal elements
  Description: The sum of all diagonal elements
        Input: a --> the matrix to trace
       Output: a value representing the sum of all diagonal elements
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
T PlMatrix<T>::trace() const
{
  int size = rows;
  T sum = (T)0;
  
  if ( size > cols )
    size = cols;
  
  for (int d = 0; d < size; ++d)
    sum += elem(d,d) ;
  
  return sum;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: herm --- computes the hermitian of the matrix
  Description: This functions returns a matrix for which every elements $(i,j)$
               correspond to the element $(j,i)$ of the original matrix.
        Input: a --> the a matrix 
       Output: A matrix corresponding to the hermitian of $a$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T> PlMatrix<T>::herm() const
{
  int i, j, r = cols, c = rows;
  PlMatrix<T> adj(r,c);
  
  for (i = 0; i < r; ++i)
    for (j = 0; j < c; ++j)
      adj.elem(i,j) = elem(j,i) ;

  return adj;

}

PlMatrix<Complex> PlMatrix<Complex>::herm() const
{
  int i, j, r = cols, c = rows;
  PlMatrix<Complex> adj(r,c);
  
  for (i = 0; i < r; ++i)
    for (j = 0; j < c; ++j)
      adj.elem(i,j) = conj(elem(j,i)) ;

  return adj;

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: transpose --- returns the transpose of the matrix
  Description: Returns the transpose of the matrix
        Input: a --> the matrix to transpose
       Output: the transpose of $a$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template<class T>
PlMatrix<T> PlMatrix<T>::transpose() const
{					
  // same as hermitian for real PlMatrix<T>
  int i, j, r = cols, c = rows;
  PlMatrix<T> adj(r,c);
  
  for (i = 0; i < r; ++i)
    for (j = 0; j < c; ++j)
      adj.elem(i,j) = elem(j,i) ;
  
  
  return adj; 
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: read --- read a matrix file
  Description: Reads a matrix file. The format of a file is
{\tt rows cols data...}, where rows and cols are int and data is a vector of the matrix type.
        Input: filename --> the name of the file to read
       Output: 1 if reading the file was successfull, 0 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
int PlMatrix<T>::read(char* filename) {
  ifstream fin(filename) ;
  if(!fin) {
    resize(1,1) ;
    return 0 ;
  }
  int r,c ;
  char *type ;
  type = new char[6] ;
  if(!fin.read(type,sizeof(char)*6)) return 0 ;
  r = strncmp(type,"matrix",6) ;
  if(r) return 0 ;
  if(!fin.read((char*)&r,sizeof(int))) return 0 ;
  if(!fin.read((char*)&c,sizeof(int))) return 0 ;
  resize(r,c) ;
  if(!fin.read((char*)m,sizeof(T)*r*c)) return 0 ;

  delete []type ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: read --- read a raw file containing a matrix of size $(r,c)$
  Description: Reads a file containing raw data of a matrix of size $(r,c)$.
        Input: filename --> the name of the file to read
                      r --> the number of rows 
                      c --> the number of columns
       Output: 1 if reading the file was successfull, 0 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
int PlMatrix<T>::read(char* filename,int r, int c) {
  ifstream fin(filename) ;
  if(!fin) {
    resize(1,1) ;
    return 0 ;
  }
  resize(r,c) ;
  if(!fin.read((char*)m,sizeof(T)*r*c)) return 0 ;

  return 1 ;
}

int PlMatrix<Color>::read(char* filename,int r, int c) {
  ifstream fin(filename) ;
  if(!fin) {
    resize(1,1) ;
    return 0 ;
  }
  resize(r,c) ;
  PlMatrix<unsigned char> T(r,c) ;
  unsigned char* p = T[0] ;
  if(!fin.read((char*)p,sizeof(unsigned char)*r*c)) return 0 ;
  int i,j ;
  for(i=0;i<r ;++i)
    for(j=0;j<c ;++j)
      elem(i,j).r = T(i,j) ;
  if(!fin.read((char*)p,sizeof(unsigned char)*r*c)) return 0 ;
  for(i=0;i<r ;i++)
    for(j=0;j<c ;++j)
      elem(i,j).g = T(i,j) ;
  if(!fin.read((char*)p,sizeof(unsigned char)*r*c)) return 0 ;
  for(i=0;i<r ;i++)
    for(j=0;j<c ;++j)
      elem(i,j).b = T(i,j) ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: write --- write a matrix into a PlMatrix file
  Description: Writes a matrix file. The format of the file is
               {\tt rows cols data...}, where rows and cols are int and data 
	       is a vector of the matrix type.
        Input: filename --> the name of the file to write to
       Output:
 Restrictions: 1 if reading the file was successfull, 0 otherwise
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
int PlMatrix<T>::write(char* filename) {
  ofstream fout(filename) ;
  if(!fout)
    return 0 ;
  int r,c ;
  r = rows ; c = cols ;
  if(!fout.write((char*)&"matrix",sizeof(char)*6)) return 0 ;
  if(!fout.write((char*)&r,sizeof(int))) return 0 ;
  if(!fout.write((char*)&c,sizeof(int))) return 0 ;
  if(!fout.write((char*)m,sizeof(T)*r*c)) return 0 ;
  return 1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writeRaw --- write the raw data to a file
  Description: Writes the raw data to a file. The size information is 
               {\em not} written to the file.
        Input: filename --> the name of the file to write to
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
int PlMatrix<T>::writeRaw(char* filename) {
  ofstream fout(filename) ;
  if(!fout)
    return 0 ;
  if(!fout.write((char*)m,sizeof(T)*rows*cols)) return 0 ;
  return 1;
}


template <class T>
void PlMatrix<T>::qSort(){
  PlError error("PlMatrix<T>::qSort()");
  error << "qSort is not defined for that type.\nPlease defined it in your .C file!";
  error.fatal() ;
}

void PlMatrix<int>::qSort(){
  qsort((char*)m,rows*cols,sizeof(int),compareInt) ;
}

void PlMatrix<float>::qSort(){
  qsort((char*)m,rows*cols,sizeof(float),compareFloat) ;
}

void PlMatrix<double>::qSort(){
  qsort((char*)m,rows*cols,sizeof(float),compareDouble) ;
}



#if defined(__GNUG__) && !defined(CPP_SRC_INCLUDE)
// Template instantation for g++ to be used with -fno-implicit-templates

template class PlMatrix<int> ;

template PlMatrix<int> operator+(const PlMatrix<int>&,const PlMatrix<int>&);
template PlMatrix<int> operator-(const PlMatrix<int>&,const PlMatrix<int>&);
template PlMatrix<int> operator*(const PlMatrix<int>&,const PlMatrix<int>&);
template PlMatrix<int> operator*(const double,const PlMatrix<int>&);
template PlMatrix<int> operator*(const Complex&,const PlMatrix<int>&);
template PlVector<int> operator*(const PlMatrix<int>&,PlVector<int>&);
template int operator==(const PlMatrix<int>&,const PlMatrix<int>&);
template int operator!=(const PlMatrix<int>&,const PlMatrix<int>&);
template PlMatrix<int> comm(const PlMatrix<int>&,const PlMatrix<int>&);
template PlMatrix<int> herm(const PlMatrix<int>&);
template PlMatrix<int> transpose(const PlMatrix<int>&);
template int trace(const PlMatrix<int>& );	

// Byte instantation

template class PlMatrix<char> ;

template PlMatrix<char> operator+(const PlMatrix<char>&,const PlMatrix<char>&);
template PlMatrix<char> operator-(const PlMatrix<char>&,const PlMatrix<char>&);
template PlMatrix<char> operator*(const PlMatrix<char>&,const PlMatrix<char>&);
template PlMatrix<char> operator*(const double,const PlMatrix<char>&);
template PlMatrix<char> operator*(const Complex&,const PlMatrix<char>&);
template PlVector<char> operator*(const PlMatrix<char>&,PlVector<char>&);
template int operator!=(const PlMatrix<char>&,const PlMatrix<char>&);
template int operator==(const PlMatrix<char>&,const PlMatrix<char>&);
template PlMatrix<char> comm(const PlMatrix<char>&,const PlMatrix<char>&);
template PlMatrix<char> herm(const PlMatrix<char>&);
template PlMatrix<char> transpose(const PlMatrix<char>&);
template char trace(const PlMatrix<char>& );	

// Float instantiation

template class PlMatrix<float> ;

template PlMatrix<float> operator+(const PlMatrix<float>&,const PlMatrix<float>&);
template PlMatrix<float> operator-(const PlMatrix<float>&,const PlMatrix<float>&);
template PlMatrix<float> operator*(const PlMatrix<float>&,const PlMatrix<float>&);
template PlMatrix<float> operator*(const double,const PlMatrix<float>&);
template PlMatrix<float> operator*(const Complex&,const PlMatrix<float>&);
template PlVector<float> operator*(const PlMatrix<float>&,PlVector<float>&);
template float operator==(const PlMatrix<float>&,const PlMatrix<float>&);
template float operator!=(const PlMatrix<float>&,const PlMatrix<float>&);
template PlMatrix<float> comm(const PlMatrix<float>&,const PlMatrix<float>&);
template PlMatrix<float> herm(const PlMatrix<float>&);
template PlMatrix<float> transpose(const PlMatrix<float>&);
template float trace(const PlMatrix<float>& );	

// Double instantation

template class PlMatrix<double> ;

template PlMatrix<double> operator+(const PlMatrix<double>&,const PlMatrix<double>&);
template PlMatrix<double> operator-(const PlMatrix<double>&,const PlMatrix<double>&);
template PlMatrix<double> operator*(const PlMatrix<double>&,const PlMatrix<double>&);
template PlMatrix<double> operator*(const double,const PlMatrix<double>&);
template PlMatrix<double> operator*(const Complex&,const PlMatrix<double>&);
template PlVector<double> operator*(const PlMatrix<double>&,PlVector<double>&);
template int operator==(const PlMatrix<double>&,const PlMatrix<double>&);
template int operator!=(const PlMatrix<double>&,const PlMatrix<double>&);
template PlMatrix<double> comm(const PlMatrix<double>&,const PlMatrix<double>&);
template PlMatrix<double> herm(const PlMatrix<double>&);
template PlMatrix<double> transpose(const PlMatrix<double>&);
template double trace(const PlMatrix<double>& );	

// Complex instantation

template class PlMatrix<Complex> ;

template PlMatrix<Complex> operator+(const PlMatrix<Complex>&,const PlMatrix<Complex>&);
template PlMatrix<Complex> operator-(const PlMatrix<Complex>&,const PlMatrix<Complex>&);
template PlMatrix<Complex> operator*(const PlMatrix<Complex>&,const PlMatrix<Complex>&);
#ifndef GNU_GCC_2_8
template PlMatrix<Complex> operator*(const double,const PlMatrix<Complex>&);
template PlMatrix<Complex> operator*(const Complex&,const PlMatrix<Complex>&);
#endif
template PlVector<Complex> operator*(const PlMatrix<Complex>&,PlVector<Complex>&);
template int operator==(const PlMatrix<Complex>&,const PlMatrix<Complex>&);
template int operator!=(const PlMatrix<Complex>&,const PlMatrix<Complex>&);
template PlMatrix<Complex> comm(const PlMatrix<Complex>&,const PlMatrix<Complex>&);
template PlMatrix<Complex> herm(const PlMatrix<Complex>&);
template PlMatrix<Complex> transpose(const PlMatrix<Complex>&);
template Complex trace(const PlMatrix<Complex>& );	

// UBYTE instantation

template class PlMatrix<unsigned char> ;

template PlMatrix<unsigned char> operator+(const PlMatrix<unsigned char>&,const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> operator-(const PlMatrix<unsigned char>&,const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> operator*(const PlMatrix<unsigned char>&,const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> operator*(const double,const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> operator*(const Complex&,const PlMatrix<unsigned char>&);
template PlVector<unsigned char> operator*(const PlMatrix<unsigned char>&,PlVector<unsigned char>&);
template int operator==(const PlMatrix<unsigned char>&,const PlMatrix<unsigned char>&);
template int operator!=(const PlMatrix<unsigned char>&,const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> comm(const PlMatrix<unsigned char>&,const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> herm(const PlMatrix<unsigned char>&);
template PlMatrix<unsigned char> transpose(const PlMatrix<unsigned char>&);
template unsigned char trace(const PlMatrix<unsigned char>& );	


// Vector2 instantiation

template class PlMatrix<Vector2> ;

template PlMatrix<Vector2> operator+(const PlMatrix<Vector2>&,const PlMatrix<Vector2>&);
template PlMatrix<Vector2> operator-(const PlMatrix<Vector2>&,const PlMatrix<Vector2>&);
#ifndef GNU_GCC_2_8
template PlMatrix<Vector2> operator*(const PlMatrix<Vector2>&,const PlMatrix<Vector2>&);
template PlMatrix<Vector2> operator*(const double,const PlMatrix<Vector2>&);
#endif
template int operator==(const PlMatrix<Vector2>&,const PlMatrix<Vector2>&);
template int operator!=(const PlMatrix<Vector2>&,const PlMatrix<Vector2>&);
template PlMatrix<Vector2> comm(const PlMatrix<Vector2>&,const PlMatrix<Vector2>&);
template PlMatrix<Vector2> herm(const PlMatrix<Vector2>&);
template PlMatrix<Vector2> transpose(const PlMatrix<Vector2>&);
template Vector2 trace(const PlMatrix<Vector2>& );	

// Point3D instantiation

template class PlMatrix<Point3D> ;

template PlMatrix<Point3D> operator+(const PlMatrix<Point3D>&,const PlMatrix<Point3D>&);
template PlMatrix<Point3D> operator-(const PlMatrix<Point3D>&,const PlMatrix<Point3D>&);
template PlMatrix<Point3D> operator*(const PlMatrix<Point3D>&,const PlMatrix<Point3D>&);
template PlMatrix<Point3D> operator*(const double,const PlMatrix<Point3D>&);
template int operator==(const PlMatrix<Point3D>&,const PlMatrix<Point3D>&);
template int operator!=(const PlMatrix<Point3D>&,const PlMatrix<Point3D>&);
template PlMatrix<Point3D> comm(const PlMatrix<Point3D>&,const PlMatrix<Point3D>&);
template PlMatrix<Point3D> herm(const PlMatrix<Point3D>&);
template PlMatrix<Point3D> transpose(const PlMatrix<Point3D>&);
template Point3D trace(const PlMatrix<Point3D>& );	

// Color instantiation

template class PlMatrix<Color> ;

template PlMatrix<Color> operator+(const PlMatrix<Color>&,const PlMatrix<Color>&);
template PlMatrix<Color> operator-(const PlMatrix<Color>&,const PlMatrix<Color>&);
template PlMatrix<Color> operator*(const PlMatrix<Color>&,const PlMatrix<Color>&);
template PlMatrix<Color> operator*(const double,const PlMatrix<Color>&);
template int operator==(const PlMatrix<Color>&,const PlMatrix<Color>&);
template int operator!=(const PlMatrix<Color>&,const PlMatrix<Color>&);
template PlMatrix<Color> comm(const PlMatrix<Color>&,const PlMatrix<Color>&);
template PlMatrix<Color> herm(const PlMatrix<Color>&);
template PlMatrix<Color> transpose(const PlMatrix<Color>&);
template Color trace(const PlMatrix<Color>& );	

// CtrlPnt4D instantiation

template class PlMatrix<CtrlPnt4D> ;

template PlMatrix<CtrlPnt4D> operator+(const PlMatrix<CtrlPnt4D>&,const PlMatrix<CtrlPnt4D>&);
template PlMatrix<CtrlPnt4D> operator-(const PlMatrix<CtrlPnt4D>&,const PlMatrix<CtrlPnt4D>&);
template PlMatrix<CtrlPnt4D> operator*(const PlMatrix<CtrlPnt4D>&,const PlMatrix<CtrlPnt4D>&);
template PlMatrix<CtrlPnt4D> operator*(const double,const PlMatrix<CtrlPnt4D>&);
template int operator==(const PlMatrix<CtrlPnt4D>&,const PlMatrix<CtrlPnt4D>&);
template int operator!=(const PlMatrix<CtrlPnt4D>&,const PlMatrix<CtrlPnt4D>&);
template PlMatrix<CtrlPnt4D> comm(const PlMatrix<CtrlPnt4D>&,const PlMatrix<CtrlPnt4D>&);
template PlMatrix<CtrlPnt4D> herm(const PlMatrix<CtrlPnt4D>&);
template PlMatrix<CtrlPnt4D> transpose(const PlMatrix<CtrlPnt4D>&);
template CtrlPnt4D trace(const PlMatrix<CtrlPnt4D>& );	

#else

void dumbSPARC_PlMatrix(){
  PlMatrix<int> a ;
  PlMatrix<char> b ;
  PlMatrix<unsigned char> c ;
  PlMatrix<double> d ;
  PlVector<double> dv ;
  PlMatrix<Vector2> e ;
  PlMatrix<Complex> f ;
  PlMatrix<CtrlPnt4D> g ;
  PlMatrix<Point3D> h ;
  PlMatrix<float> i ;
  PlMatrix<Color> j ;
  double dd=2.0 ;

  a = a + a*a - transpose(a) ;
  b = b + b*b - transpose(b) ; 
  c = c + c*c - transpose(c) ;
  d = d + d*d - transpose(d) + dd*d;  
  dv = d*dv ;
  e = e + e*e - transpose(e) ;
  f = f + f*f - transpose(f) ;
  g = g + g*g - transpose(g) ;
  h = h + h*h - h ;
  i = i + i*i - transpose(i) ;
  j = j ;

  a != a ;
  b != b ;
  c != c ;
  d != d ;
  e != e ;
  f != f ;
  g != g ;
  h != h ;
  i != i ;
  j != j ;
}

#endif
