/*=============================================================================
        File: mlapack.cc
     Purpose:
    Revision: $Id: mlapack.cc,v 1.1.1.1 1998/04/06 21:02:32 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
	  This library is free software; you can redistribute it and/or
	  modify it under the terms of the GNU Library General Public
	  License as published by the Free Software Foundation; either
	  version 2 of the License, or (at your option) any later version.
 
	  This library is distributed in the hope that it will be useful,
	  but WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  Library General Public License for more details.
 
	  You should have received a copy of the GNU Library General Public
	  License along with this library; if not, write to the Free
	  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#include "mlapack.hh"

#ifdef WITH_LAPACK
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: svd --- computes the singular value decomposition (SVD)
  Description: svd computes the singular value decomposition of a
               general matrix, $A=USV^T$, where U and V are orthogonal/unitary
               and S is diagonal, real, and non-negative. Here, $V^T$ means
               transpose(V).

               For $m \times n$ matrix A, U is $m \times m$, S is 
	       $m \times n$, and V is $n \times n$ where
               the columns of U and V contain the left and right singular
               vectors respectively and the diagonal elements of S are the
               singular values in descending sorted order.

               svd uses the DGESVD routine from LAPACK to obtain the
               decomposition.

        Input: A --> input matrix object
               U <-- matrix object to receive the U matrix
	       S <-- matrix object to receive the S matrix
	       V <-- matrix object to receive the V matrix
       Output: \begin{description}
               \item[0:] successful exit. 
               \item[\ensuremath{<} 0:]  if return = -i, the i-th argument had an illegal value.
               \item[\ensuremath{>} 0:]  if DBDSQR did not converge, INFO specifies how many
                       superdiagonals of an intermediate bidiagonal form B
                       did not converge to zero. See info on DGESVD for more
                       information.
	       \end{description}
 Restrictions: You should read the LAPACK documentation to find out more
               about the error messages
   Created by: Philippe Lavoie (19 December 1996)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int 
LAPACK::svd(const Matrix_DOUBLE &A, 
	    Matrix_DOUBLE &U, 
	    Matrix_DOUBLE &S, 
	    Matrix_DOUBLE &V) {
  double *work,*svec ;
  long m,n ;
  long work_len,s_len,info ;
  char jobu = 'A' ;
  char jobvt = 'A' ;
  m = A.rows ;
  n = A.cols ;
  U.resize(m,m) ;
  S.resize(m,n) ;
  V.resize(n,n) ;

  work_len = 2*(2*minimum(m,n)+maximum(m,n)) ;
  work = new double[work_len] ;
  s_len = minimum(m,n) ;
  svec = new double[s_len] ;

#ifdef COLUMN_ORDER
  // Call the FORTRAN routine
  DGESVD(&jobu,&jobvt,&m,&n,A.m,&m,
	 svec,U.m,&m,V.m,&n,work,&work_len,&info);
#else
  Matrix_DOUBLE Ut,Vt ;
  Matrix_DOUBLE At ;
  Ut.resize(m,m) ;
  Vt.resize(n,n) ;
  At.resize(n,m) ;
  At = transpose(A) ;
  // Call the FORTRAN routine
  DGESVD(&jobu,&jobvt,&m,&n,At.m,&m,
	 svec,Ut.m,&m,Vt.m,&n,work,&work_len,&info);
  U = transpose(Ut) ;
  V = transpose(Vt) ;
#endif
  S.reset(0) ;
  for(int i=0;i<minimum(m,n);i++)
    S(i,i) = svec[i] ;
  last_error = info ;
  delete []svec ;
  delete []work ;
  return info ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: lse --- compute least-squares solution to a set of linear 
                       equations
  Description: this routine employs the LAPACK DGELS driver to obtain the 
               least-squares
               solution to a set of linear equations $A X = B$ where
               $B$ is made up of column vectors, each of which is assumed
               to be an independent right-hand-side, and $X$ is made up
               of column vectors, each of which is an independent solution
               to the corresponding $X$ column vector.

               For an $m \times n$ matrix A, if $m \geq n$ then the solution 
	       is obtained using the QR factorization; 
	       if $m < n$, then the solution is obtained using the LQ 
	       factorization.

	       If an error occurs, an error message will be given.
	       You should read the LAPACK documentation to find out more
               about the error messages.

        Input: A --> $A$ matrix
               B --> matrix of column vectors for the right hand side
	       X <-- matrix of column vectors for the solution 
       Output: 1 on success, 0 on failure
 Restrictions: The matrices must be of compatible sizes.
   Created by: Philippe Lavoie (19 December 1996)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int
LAPACK::lse(const Matrix_DOUBLE &A, const Matrix_DOUBLE &B, Matrix_DOUBLE &X){
  double *work;
  char trans;
  long lwork,info ;
  long m,n,nrhs,ldb ;

  trans = 'N' ;
  m = A.rows ;
  n = A.cols ;
  ldb = maximum(m,n) ;
  nrhs = B.cols ;
  if(m==0 || n==0){
    PlError error("LAPACK::lse");
    error << "A has size " << m << ", " << n << endl ;
    error.fatal() ;
  }
  if(B.rows != m){
    PlError error("LAPACK::mlse");
    error << "B doesn't have the proper dimenstion (" << B.rows << " instead of " << m << ")\n" ;
    error.fatal() ;
  }
  lwork = minimum(m,n)+maximum(maximum(m,n),(long)B.cols)*8;
  work = new double[lwork] ;

#ifdef COLUMN_ORDER
  DGELS(&trans,&m,&n,&nrhs,A.m,&m,B.m,&m,work,&lwork,&info) ;
  X = B.get(0,0,A.cols,B.cols) ;
#else
  Matrix_DOUBLE At,Bt ;
  At.resize(n,m) ;
  At = transpose(A) ;
  Bt.resize(B.cols,B.rows) ;
  Bt = transpose(B) ;
  DGELS(&trans,&m,&n,&nrhs,At.m,&m,Bt.m,&m,work,&lwork,&info) ;
  X = transpose(Bt).get(0,0,A.cols,B.cols) ;
#endif
  delete []work ;

  last_error = info ;
  if(info<0 && print_error){
    PlError error("LAPACK::lse") ;
    error << "DGELS returned the error message " << info << endl ;
    error.warning() ;
    return 0 ;
  }
  return 1 ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: inverse --- computes the inverse of a matrix
  Description: Computes the inverse of a matrix using the DGETRI function
               from the LAPACK package

	       If $last\underline{ }error$ is set to 0, the inverse has been 
	       computed without any problems. If \verb.last_error.
	       is set to a value 
	       $i$ greater than 0, then the value of $A(i,i)$ is exactly 0,  
	       the matrix is singular and the inverse can not be computed. 
	       If $last\underline{ }error$ is smaller than 0 then the routine 
	       has a bug. If you see this error occuring, please inform the 
	       author.

        Input: A --> input matrix object
       Output: $A^{-1}$, the inverse of matrix A.
 Restrictions: The matrix must be squared.
   Created by: Philippe Lavoie (19 December 1996)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Matrix_DOUBLE 
LAPACK::inverse(const Matrix_DOUBLE& A) {
  Matrix_DOUBLE Inverse ; 
  long *pivot ;
  double *work ;
  long pi_len,work_len ;
  long info ;

  if(A.rows != A.cols){
    PlError error("Matrix<double> LAPACK::invm(Matrix<double>&)");
    error << "The matrix is not squared!\n" ;
    error.fatal() ;
  }

  pi_len = A.rows ;
  work_len = pi_len*8 ;
  pivot = new long[pi_len] ;
  work = new double[work_len] ;
  Inverse.resize(A.rows,A.cols) ;

#ifdef COLUMN_ORDER
  long Acols, Arows ;
  Acols = A.cols ;
  Arows = A.rows ;
  Inverse = A ;
  DGETRF(&Acols,&Arows,Inverse.m,&Acols,pivot,&info) ;
#else
  Matrix_DOUBLE At ;
  At.resize(A.cols,A.rows) ;
  At = transpose(A) ;
  
  // Compute the LU factorization
  long Atcols, Atrows ;
  Atcols = At.cols ;
  Atrows = At.rows ;
  DGETRF(&Atcols,&Atrows,At.m,&Atcols,pivot,&info) ;
#endif
  last_error = info ;
  if (info != 0 && print_error)
    {
      PlError error("LAPACK::invm");
      error << "LAPACK {dz}getrf() returned with INFO=" << info ; 
      error << "\n\twhere INFO < 0 -> for INFO = -i, the i'th argument had an illegal value and\n\tINFO > 0 -> for INFO = i, U(i,i) is exactly zero. Matrix is singular\n\tand the inverse cannot be computed.\n" ;
      error.warning() ;
      delete []work ;
      delete []pivot ;
      return Inverse;
    }
  
  // obtain the inverse from LU factorization
#ifdef COLUMN_ORDER
  DGETRI(&Acols,Inverse.m,&Acols,pivot,work,&work_len,&info);
#else
  DGETRI(&Atcols,At.m,&Atcols,pivot,work,&work_len,&info);
#endif
  last_error = info ;
  if (info != 0 && print_error)
    {
      PlError error("LAPACK::invm");
      error << "LAPACK {dz}getri() returned with INFO=" << info ; 
      error << "\n\twhere INFO < 0 -> for INFO = -i, the i'th argument had an illegal value and\n\tINFO > 0 -> for INFO = i, U(i,i) is exactly zero. Matrix is singular\n\tand the inverse cannot be computed.\n" ;
      error.warning() ;
      delete []work ;
      delete []pivot ;
      return Inverse ;
    }

#ifdef COLUMN_ORDER
  // Inverse is already properly setup
#else
  Inverse = transpose(At) ;
#endif

  delete []work ;
  delete []pivot ;
  return Inverse;
}

LAPACK LAPack(0);

#endif // WITH_LAPACK
