/*=============================================================================
        File: hnurbsS.hh
     Purpose:       
    Revision: $Id: hnurbsS.hh,v 2.6 1998/06/03 15:43:41 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1998 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _nurbs_hnurbs_h_
#define _nurbs_hnurbs_h_

#include "nurbsS.hh"
#include MatrixLib(vector.hh)

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlHNurbsSurface --- a hierarchichal NURBS surface class
  Description: This class can represent and manipulate a hierarchical NURBS
               surface. The surface is composed of points in 4D. It
	       can have any degree and have any number of control points.

	       This does not correspond to the HBsplines given by
	       Forsey. However I hope that it will be usefull for
	       interactive manipulations of NURBS surfaces.

	       Other aspects of my implementation are different. They
	       will be documented when the class is fully functionnal.

	       Presently there is only a limited set of functions available
	       for this class. 

   Created by: Philippe Lavoie (28 September 1997)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlHNurbsSurface: public PlNurbsSurface {
public:
  PlHNurbsSurface() ; 
  PlHNurbsSurface(const PlNurbsSurface& S) ; 
  PlHNurbsSurface(const PlHNurbsSurface& S) ;
  PlHNurbsSurface(PlHNurbsSurface* base) ;
  PlHNurbsSurface(PlHNurbsSurface* base, const PlVector<float>& xU, const PlVector<float>& xV) ;
  PlHNurbsSurface(PlHNurbsSurface* base, const PlHNurbsSurface &surf) ;
  virtual ~PlHNurbsSurface() ;

  PlHNurbsSurface* baseLevel() const { return baseLevel_;}
  PlHNurbsSurface* nextLevel() const { return nextLevel_;}
  PlHNurbsSurface* firstLevel() const { return firstLevel_;}
  PlHNurbsSurface* lastLevel() const { return lastLevel_; }

  void splitUV(int nu, int nv, PlVector<float> &nU, PlVector<float> &nV) ;
  void splitUV(int nu, int su, int nv, int sv, PlVector<float> &nU, PlVector<float> &nV) ;
  
  virtual PlHNurbsSurface* addLevel(int nsplit) ;
  virtual PlHNurbsSurface* addLevel() ;

  virtual void copy(const PlHNurbsSurface& nS) ;

  int modifies(float u, float v) ;

  
  CtrlPnt4D operator()(float u, float v) const { return pointAt(u,v,-1) ; }
  CtrlPnt4D pointAt(float u, float v, int lod) const ;

  void deriveAt(float u, float v, int, Matrix_CtrlPnt4D&, int lod=-1) const;
  void deriveAt(float u, float v, int, Matrix_Point3D&, int lod=-1) const;

  int movePointOffset(float u, float v, const Point3D& delta) ; 


  int initBase(int force=0) ;
  virtual void updateSurface(int i0=-1, int j0=-1) ;
  virtual void updateLevels(int updateLevel=-1) ;

  int isoCurveU(float u, PlNurbsCurve& c,int lod=-1) const ;
  int isoCurveV(float v, PlNurbsCurve& c,int lod=-1) const ;

  // I/O functions
  int read(const char* filename);
  int write(const char* filename) const;
  virtual int read(ifstream &fin) ;
  int write(ofstream &fout) const ;

  Matrix_CtrlPnt4D offset ;
  Vector_FLOAT rU,rV ;

  const int& level;

  int maxLevel() const ;

  int modifiedN() const { return updateN ; }

  void refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  void refineKnotU(const Vector_FLOAT& X) ;
  void refineKnotV(const Vector_FLOAT& X) ;


  void axis(int i, int j, Point3D& xaxis, Point3D& yaxis, Point3D& zaxis)const;

protected:
  PlNurbsSurface baseSurf ;
  PlHNurbsSurface *baseLevel_,*nextLevel_,*firstLevel_,*lastLevel_ ;

  Matrix_Point3D ivec,jvec,kvec ;

  int level_ ;
  int updateN,baseUpdateN ;
  int update_ ; // indicates that one of the offset point was modified

  float uS_,uE_,vS_,vE_ ;
  float uD,vD ; 
  
};

inline void 
PlHNurbsSurface::axis(int i, int j, Point3D& xaxis, Point3D& yaxis, Point3D& zaxis) const 
{ 
  if(baseLevel_) { 
    xaxis = ivec(i,j) ; 
    yaxis = jvec(i,j) ; 
    zaxis = kvec(i,j) ; 
  } 
  else{
    xaxis = Point3D(1,0,0) ; 
    yaxis = Point3D(0,1,0) ; 
    zaxis = Point3D(0,0,1) ; 
  }
}

#endif // _nurbs_hnurbs_h_
