/*=============================================================================
        File: hnurbsS_sp.hh
     Purpose:       
    Revision: $Id: hnurbsS_sp.hh,v 2.5 1998/06/03 18:23:55 lavoie Exp $
  Created by: Philippe Lavoie          (14 May, 1998)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _nurbs_hnurbsS_sp_h_
#define _nurbs_hnurbsS_sp_h_

#include "hnurbsS.hh"

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlHNurbsSurfaceSP --- A NURBS surface with surface point
  Description: A Nurbs surface with surface point manipulators. This allows 
               someone to modify the point on a surface for which a 
	       control point has maximal influence over it. This might 
	       provide a more intuitive method to modify a surface. 

   Created by: Philippe Lavoie (14 May, 1998)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlHNurbsSurfaceSP : public PlHNurbsSurface{
public:
  PlHNurbsSurfaceSP() ; 
  PlHNurbsSurfaceSP(const PlNurbsSurface& S) ; 
  PlHNurbsSurfaceSP(const PlHNurbsSurface& S) ;
  PlHNurbsSurfaceSP(const PlHNurbsSurfaceSP& S) ;
  PlHNurbsSurfaceSP(PlHNurbsSurface* base) ;
  PlHNurbsSurfaceSP(PlHNurbsSurface* base, const PlVector<float>& xU, const PlVector<float>& xV) ;
  PlHNurbsSurfaceSP(PlHNurbsSurface* base, const PlHNurbsSurface &surf) ;

  virtual void resizeKeep(int Pu, int Pv, int DegU, int DegV) ;
  
  virtual void refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  virtual void refineKnotU(const Vector_FLOAT& X);
  virtual void refineKnotV(const Vector_FLOAT& X);
  
  virtual void mergeKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  virtual void mergeKnotU(const Vector_FLOAT& X);
  virtual void mergeKnotV(const Vector_FLOAT& X);

  virtual void updateSurface(int i0=-1, int j0=-1) ;  
  virtual void updateLevels(int updateLevel=-1) ;

  virtual PlHNurbsSurfaceSP* addLevel(int nsplit, int s=1) ;
  virtual PlHNurbsSurfaceSP* addLevel() ;

  virtual void copy(const PlHNurbsSurface& nS) ;

  virtual int read(ifstream &fin) ;
  

  void modSurfCPby(int i, int j, const CtrlPnt4D& a) ;

  float maxAtUV(int i, int j) const { return maxAtU_[i]*maxAtV_[j] ; }
  float maxAtU(int i) const { return maxAtU_[i] ; }
  float maxAtV(int i) const { return maxAtV_[i] ; }

  CtrlPnt4D surfP(int i,int j) const 
    { return point4D(maxAtU_[i],maxAtV_[j]); }

  CtrlPnt4D surfP(int i,int j, int lod) const
    { return pointAt(maxAtU_[i],maxAtV_[j],lod) ; }

  void updateMaxUV() 
    { updateMaxU() ; updateMaxV() ; }
  void updateMaxU() ;
  void updateMaxV() ;

  int okMax() { return (maxU.n<=1)?0:1 ; }

protected:

  Vector_FLOAT maxU,maxV ;
  Vector_FLOAT maxAtU_,maxAtV_ ;
};


inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP() : PlHNurbsSurface() { 
  ;
}

inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP(const PlNurbsSurface& nS)  : PlHNurbsSurface(nS) { 
  updateMaxUV() ; 
}

inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP(PlHNurbsSurface* base): PlHNurbsSurface(base) {
  updateMaxUV() ; 
}

inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP(PlHNurbsSurface* base, const PlVector<float>& xU, const PlVector<float>& xV) : PlHNurbsSurface(base,xU,xV) {
  updateMaxUV() ;
}

inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP(const PlHNurbsSurface& S) : PlHNurbsSurface() {
  copy(S) ;
}
inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP(const PlHNurbsSurfaceSP& S) : PlHNurbsSurface() {
  copy(S) ;
}

inline PlHNurbsSurfaceSP::PlHNurbsSurfaceSP(PlHNurbsSurface* base, const PlHNurbsSurface &surf):
  PlHNurbsSurface(base)
{
  copy(surf) ;
  updateMaxUV() ;
}


inline void 
PlHNurbsSurfaceSP::resizeKeep(int Pu, int Pv, int DegU, int DegV) { 
  PlHNurbsSurface::resizeKeep(Pu,Pv,DegU,DegV) ; 
  updateMaxUV() ; 
}

inline void 
PlHNurbsSurfaceSP::refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) {
  PlHNurbsSurface::refineKnots(nU,nV) ;
  updateMaxUV() ;
}
  
inline void 
PlHNurbsSurfaceSP::refineKnotU(const Vector_FLOAT& X){
  PlHNurbsSurface::refineKnotU(X) ;
  updateMaxU() ;
}

inline void 
PlHNurbsSurfaceSP::refineKnotV(const Vector_FLOAT& X){
  PlHNurbsSurface::refineKnotV(X) ;
  updateMaxV() ;
}

  
inline void 
PlHNurbsSurfaceSP::mergeKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) {
  PlHNurbsSurface::mergeKnots(nU,nV) ;
  updateMaxUV() ;
}


inline void 
PlHNurbsSurfaceSP::mergeKnotU(const Vector_FLOAT& X){
  PlHNurbsSurface::mergeKnotU(X) ;
  updateMaxU() ;
}


inline void 
PlHNurbsSurfaceSP::mergeKnotV(const Vector_FLOAT& X){
  PlHNurbsSurface::mergeKnotV(X) ;
  updateMaxV() ;
}




#endif
