/*=============================================================================
        File: nurbs.hh
     Purpose:       
    Revision: $Id: nurbs.hh,v 2.6 1998/05/25 21:08:12 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _nurbs_nurbs_h_
#define _nurbs_nurbs_h_

class PlNurbsCurveArray ;

#include "curve.hh"
#include "matrixRT.hh"
#include MatrixLib(list.hh)

// see in file curve.hh for to see if INCLUDE_MATRIX_HEADER is defined

#include MatrixLib(vector.hh)
#include MatrixLib(matrix.hh)
#include MatrixLib(error.hh)
#include MatrixLib(image.hh)
#include MatrixLib(matrixMat.hh)

// forward decleration of PlNurbsSurface
class PlNurbsSurface ;


#undef INCLUDE_CFSQP
#ifdef INCLUDE_CFSQP
#include <cfsqpusr.h>
#endif

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlNurbsCurve --- A NURBS curve class
  Description: This class is used to represent and manipulate NURBS curve. 
               The curves are composed of points in 4D. They can have any 
	       degree and have any number of control points.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlNurbsCurve : public PlParaCurve{
public:
  PlNurbsCurve() ;
  PlNurbsCurve(const PlNurbsCurve& nurb) ;
  PlNurbsCurve(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int degree=3) ;
  PlNurbsCurve(const Vector_Point3D& P1, const Vector_FLOAT &W, const Vector_FLOAT &U1, int degree=3) ;
  virtual ~PlNurbsCurve() { } // empty destructor

  // Reference to internal data
  const int& degree ; // a reference to the degree of the curve
  const Vector_CtrlPnt4D& CtrlP ; // a reference to the vector of control points
  const Vector_FLOAT& Knot ; // a reference to the vector of knots
  
  // basic functions
  void resize(int n, int Deg) ;
  virtual void reset(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int degree) ;
  virtual PlNurbsCurve& operator=(const PlNurbsCurve&) ;

  // Evaluattion functions
  virtual CtrlPnt4D operator()(float u) const;
  friend CtrlPnt4D C(float u, const PlNurbsCurve& nurb) {return nurb(u) ; } // a function interface to operator()
  friend Point3D Cp(float u, const PlNurbsCurve& nurb) {return project(nurb(u)) ;} // returns the curvePoint in 3D

  
  // derivative functions
  void deriveAt(float u, int, Vector_CtrlPnt4D&) const;
  void deriveAt(float u, int, Vector_Point3D&) const;
  Point3D derive3D(float u, int d) const;
  CtrlPnt4D derive(float u, int d) const;
  Point3D normal(float u, const Point3D& v) const ;
  

  // Basis functions
  float basisFun(float u, int i, int p=-1) const ;
  void basisFuns(float u, int span, Vector_FLOAT& N) const ;
  void dersBasisFuns(int n,float u, int span, Matrix_FLOAT& N) const;

  // Knot functions
  float minKnot() const { return U[0] ; }
  float maxKnot() const { return U[U.n-1] ; }
  int findSpan(float u) const ;
  void findMultSpan(float u, int& r, int& s) const;
  int findMult(int r) const;
  int findKnot(float u) const;
  float getRemovalBnd(int r, int s) const;

  virtual void removeKnot(int r, int s, int num);
  virtual void removeKnotsBound(const Vector_FLOAT& ub, Vector_FLOAT& ek, float E) ;
  virtual void knotInsertion(float u, int r,PlNurbsCurve& nc);
  virtual void refineKnotVector(const Vector_FLOAT& X);
  virtual void mergeKnotVector(const Vector_FLOAT &Um);
  
  // Curve fitting functions
  int leastSquares(const Vector_Point3D& Q, int degC, int n ) ;
  int leastSquares(const Vector_Point3D& Q, int degC, int n, const Vector_FLOAT& ub);
  int leastSquares(const Vector_CtrlPnt4D& Q, int degC, int n, const Vector_FLOAT& ub);
  int leastSquares(const Vector_Point3D& Q, int degC, int n, const Vector_FLOAT& ub, const Vector_FLOAT& knot);
  int leastSquares(const Vector_CtrlPnt4D& Q, int degC, int n, const Vector_FLOAT& ub, const Vector_FLOAT& knot);
  void globalApproxErrBnd(Vector_Point3D& Q, int degree, float E) ;
  void globalApproxErrBnd(Vector_Point3D& Q, Vector_FLOAT& ub, int degree, float E) ;
  void globalApproxErrBnd2(Vector_Point3D& Q, int degree, float E) ;
  void globalApproxErrBnd3(Vector_Point3D& Q, int degree, float E) ;
  void globalApproxErrBnd3(Vector_Point3D& Q, const Vector_FLOAT &ub, int degree, float E) ;

  void globalInterp(const Vector_Point3D& Q, int d);
  void globalInterp(const Vector_Point3D& Q, const Vector_FLOAT& ub, int d);
  void globalInterp(const Vector_CtrlPnt4D& Q, int d);
  void globalInterp(const Vector_CtrlPnt4D& Q, const Vector_FLOAT& U, int d);
  void globalInterp(const Vector_CtrlPnt4D& Q, const Vector_FLOAT& ub, const Vector_FLOAT& U, int d);

  void globalInterpD(const Vector_Point3D& Q, const Vector_Point3D& D, int d, int unitD, float a=1.0);
  
  void projectTo(const Point3D& p, float guess, float& u, Point3D& r, float e1=0.001, float e2=0.001,int maxTry=100) const;

  // Generate type of curve
  void makeCircle(const Point3D& O, const Point3D& X, const Point3D& Y, float r, double as, double ae);
  void makeCircle(const Point3D& O, float r, double as, double ae);
  virtual void degreeElevate(int t);
  friend void generateCompatibleCurves(PlNurbsCurveArray &ca);
  void decompose(PlNurbsCurveArray& c) const ;
  
  int splitAt(float u, PlNurbsCurve& cl, PlNurbsCurve& cu) const;
  int mergeOf(const PlNurbsCurve& cl, const PlNurbsCurve &cu) ;

  // Modifies the NURBS curve
  void transform(const MatrixRT<float>& A) ;
  void modCP(int i,const CtrlPnt4D& a) { P[i] = a ; } // To manipulate the value of the control point $P[i]$
  void modCPby(int i,const CtrlPnt4D& a) { P[i] += a ; } // To manipulate the value of the control point $P[i]$
  virtual void modKnot(const Vector_FLOAT& knot) { if(knot.n-deg-1==P.n) U = knot ; }  // to change the values of the knot vector only if the size is compatible with P.n
  
  int movePoint(float u, const Point3D& delta) ;
  int movePoint(float u, const Vector_Point3D& delta) ;
  int movePoint(const Vector_FLOAT& ur, const Vector_Point3D& D);
  int movePoint(const Vector_FLOAT& ur, const Vector_Point3D& D, const Vector_INT& Dr, const Vector_INT& Dk) ;
  int movePoint(const Vector_FLOAT& ur, const Vector_Point3D& D, const Vector_INT& Dr, const Vector_INT& Dk, const Vector_INT& fixCP);

  // I/O functions
  int read(const char*) ;
  int write(const char*) const ;
  virtual int read(ifstream &fin) ;
  int write(ofstream &fout) const ;
  int writePS(const char*,int cp=0,float magFact=0.0,float dash=5.0) const ;
  int writePSp(const char*,const Vector_Point3D&,const Vector_Point3D&, int cp=0,float magFact=0.0,float dash=5.0) const ;
  
  int writeVRML(const char* filename,float radius,int K, const Color& color,int Nu,int Nv, float u_s, float u_e) const ;
  int writeVRML(const char* filename,float radius=1,int K=5, const Color& color=whiteColor,int Nu=20,int Nv=20) const { return writeVRML(filename,radius,K,color,Nu,Nv,U[0],U[U.n-1]) ; } // writes the curve to a VRML file
  void drawImg(Image_UBYTE& Img,unsigned char color=255,float step=0.01) ;
  void drawImg(Image_Color& Img,const Color& color,float step=0.01) ;
  void drawAaImg(Image_Color& Img, const Color& color, int precision=3,int alpha=1) ;
  void drawAaImg(Image_Color& Img, const Color& color, const PlNurbsCurve& profile, int precision=3,int alpha=1) ;
  PlNurbsSurface drawAaImg(Image_Color& Img, const Color& color, const PlNurbsCurve& profile, const PlNurbsCurve &scaling, int precision=3,int alpha=1) ;

  BasicList<Point3D> tesselate(float tolerance, BasicList<float> *uk) const ;

protected:
  Vector_CtrlPnt4D P; // the vector of control points
  Vector_FLOAT U ;  // the knot vector
  int deg ;  // the degree of the NURBS curve

};


float chordLengthParam(const Vector_Point3D& Q, Vector_FLOAT &ub);
void binomialCoef(Matrix_FLOAT& Bin) ;
Vector_FLOAT knotUnion(const Vector_FLOAT& Ua, const Vector_FLOAT& Ub);
float nurbsBasisFun(float u, int i, int p, const Vector_FLOAT& U) ; 
void nurbsBasisFuns(float u, int span, int deg, const Vector_FLOAT& U, Vector_FLOAT& N);
void nurbsDersBasisFuns(int n, float u, int span, int deg, const Vector_FLOAT& U, Matrix_FLOAT& ders) ;
int intersectLine(const Point3D& p1, const Point3D& t1, const Point3D& p2, const Point3D& t2, Point3D& p);
void knotAveraging(const Vector_FLOAT& uk, int deg, Vector_FLOAT& U) ;
void averagingKnots(const Vector_FLOAT& U, int deg, Vector_FLOAT& uk);
int findSpan(float u, const Vector_FLOAT& U, int deg);

int maxInfluence(int i, const Vector_FLOAT& U, int p, float &u);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: getCoordinates --- finds the matrix coordinate i,j of a Point3D
  Description: Finds the closest coordinate of a matrix(i,j) to the $x,y$ 
               coordinates of a 3D point.
	      
        Input: p --> the 3D point
               i <-- the coordinate for the row (corresponds to $y$)
	       j <-- the coordinate for the column (corresponds to $x$)
	       rows --> the number of rows in the matrix
	       cols --> the number of columns in the matrix
       Output: 1 if the coordinates are in the matrix range, 0 otherwise
 Restrictions: 
   Created by: Philippe Lavoie (25 July 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
inline int getCoordinates(const Point3D& p, int& i, int& j, int rows, int cols){
  i = int(rint(p.y)) ;
  j = int(rint(p.x)) ;
  if(i>=rows) return 0 ;
  if(j>=cols) return 0 ;
  if(i<0) return 0 ;
  if(j<0) return 0 ;
  return 1 ;
}


/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlNurbsCurveArray --- an array of PlNurbsCurve
  Description: This class represents an array of PlNurbsCurve.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlNurbsCurveArray {
public:
  const int& n ;
  PlNurbsCurveArray(PlNurbsCurve* Ca, int size) ;
  PlNurbsCurveArray():n(sze) { C = 0 ; sze = 0 ; rsize = 0 ;}
  virtual ~PlNurbsCurveArray(){ if(C){ for(int i=0;i<rsize;i++) delete C[i];  delete []C ; }}
  
  virtual PlNurbsCurve& operator[](int i) { return *(C[i]) ; }
  virtual PlNurbsCurve operator[](int i) const { return *(C[i]) ; }
  
  virtual void resize(int s) ;
  void init(PlNurbsCurve* Ca, int size) ;
  int read(const char *filename);
  int write(const char *filename);
protected:
  int sze ; // the number of NURBS curves in the array
  int rsize ; // the number of space allocated for the arra
  PlNurbsCurve** C ; // An array of pointers to NURBS curves
};



#endif 

