/*=============================================================================
        File: nurbsS.H
     Purpose:       
    Revision: $Id: nurbsS.hh,v 2.4 1998/05/25 21:07:49 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/

#ifndef _nurbs_nurbsS_h_
#define _nurbs_nurbsS_h_


#include "surface.hh"
#include "nurbs.hh"

enum Direction { u_direction=1, v_direction=2, both_direction=3} ;

class PlNurbsSurfaceArray ; // forward decleration

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlNurbsSurface --- a class to represent a NURBS surface
  Description: The NURBS surface is composed of points in 4D. It can have 
               any degree in both the $u$ and the $v$ direction.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlNurbsSurface : public PlParaSurface {
public:
  PlNurbsSurface() ;
  PlNurbsSurface(const PlNurbsSurface& nS) ;
  PlNurbsSurface(int DegU, int DegV, const Vector_FLOAT& Uk, const Vector_FLOAT& Vk, const Matrix_CtrlPnt4D& Cp) ;
  PlNurbsSurface(int DegU, int DegV, Vector_FLOAT& Uk, Vector_FLOAT& Vk, Matrix_Point3D& Cp, Matrix_FLOAT& W) ;
  virtual ~PlNurbsSurface() {;}
  
public:  
  // Reference to internal data
  const Vector_FLOAT& KnotU ;
  const Vector_FLOAT& KnotV ;
  const Matrix_CtrlPnt4D& CtrlP ;
  const int& degreeU ;
  const int& degreeV ;

  // Basic operators
  virtual PlNurbsSurface& operator=(const PlNurbsSurface&) ;
  void resize(int Pu, int Pv, int DegU, int DegV) ;
  virtual void resizeKeep(int Pu, int Pv, int DegU, int DegV) ;
  int ok() ;
  
  // Basis functions
  virtual CtrlPnt4D operator()(float u, float v) const ;

  void basisFuns(float u, float v, int spanU, int spanV, Vector_FLOAT& Nu, Vector_FLOAT& Nv) const ;
  void basisFunsU(float u, int span, Vector_FLOAT& N) const ;
  void basisFunsV(float u, int span, Vector_FLOAT& N) const ;
  

  // Derivative functions
  void deriveAt(float u, float v, int d, Matrix_Point3D& skl) const ;
  void deriveAt(float u, float v, int d, Matrix_CtrlPnt4D& skl) const ;
  Point3D normal(float u, float v) const ;

  
  // Surface fitting functions

  void globalInterp(const Matrix_Point3D& Q, int pU, int pV);
  void globalInterp(const Matrix_CtrlPnt4D& Q, int pU, int pV);
  void leastSquares(const Matrix_Point3D& Q, int pU, int pV, int nU, int nV) ;

  friend void gordonSurface(PlNurbsCurveArray& lU, PlNurbsCurveArray& lV, const Matrix_Point3D& intersections, PlNurbsSurface& gS);
  friend int surfMeshParams(const Matrix_Point3D& Q, Vector_FLOAT& uk, Vector_FLOAT& vl);
  friend void globalSurfInterpXY(const Matrix_Point3D& Q, int pU, int pV, PlNurbsSurface& S);
  friend void globalSurfInterpXY(const Matrix_Point3D& Q, int pU, int pV, PlNurbsSurface& S, const Vector_FLOAT& uk, const Vector_FLOAT& vk);
  friend void globalSurfApprox(const Matrix_Point3D& Q, int pU, int pV, PlNurbsSurface& S, double error=0);



  // Surface generation function
  int skinV(PlNurbsCurveArray& ca, int degV);
  int skinU(PlNurbsCurveArray& ca, int degU);
  void sweep(const PlNurbsCurve& T, const PlNurbsCurve& C, const PlNurbsCurve& Sv, int K,int useAy=0, int invAz=0) ;
  void sweep(const PlNurbsCurve& T, const PlNurbsCurve& C, int K,int useAy=0, int invAz=0) ;
  void makeFromRevolution(const PlNurbsCurve& profile, const Point3D& S, const Point3D& T, double theta=2*M_PI) ;
  void makeSphere(const Point3D& O, float r) ; 


  void degreeElevate(int tU, int tV) ;
  virtual void degreeElevateU(int tU) ;
  virtual void degreeElevateV(int tV) ;

  
  int decompose(PlNurbsSurfaceArray& Sa) const ;

  // Knot functions
  void findSpan(float u, float v, int& spanU, int& spanV) const ;
  int findSpanU(float u) const ;
  int findSpanV(float v) const ;  
  
  int findMultU(int r) const ;
  int findMultV(int r) const ;

  virtual void refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  virtual void refineKnotU(const Vector_FLOAT& X);
  virtual void refineKnotV(const Vector_FLOAT& X);
  
  virtual void mergeKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  virtual void mergeKnotU(const Vector_FLOAT& X);
  virtual void mergeKnotV(const Vector_FLOAT& X);
  

  // special functions
  void isoCurveU(float u, PlNurbsCurve& c) const ;
  void isoCurveV(float v, PlNurbsCurve& c) const ;

  // I/O functions
  int read(const char* filename);
  int write(const char* filename) const;
  virtual int read(ifstream &fin) ;
  int write(ofstream &fout) const ;
  int writeVRML(const char* filename,const Color& color,int Nu,int Nv, float u_s, float u_e, float v_s, float v_e) const { return PlParaSurface::writeVRML(filename,color,Nu,Nv,u_s,u_e,v_s,v_e);}
  int writeVRML(const char* filename,float tolerance, const Color& color) const ; 
  int writeVRML(const char* filename,const Color& color=whiteColor,int Nu=20,int Nv=20) const { return PlParaSurface::writeVRML(filename,color,Nu,Nv,U[0],U[U.n-1],V[0],V[V.n-1]); } 

  int writePOVRAY(ostream& povray, int patch_type=1, double flatness=0.01, int num_u_steps=8, int num_v_steps=8) const ;
  int writePOVRAY(float, ostream& povray, const Color& color=Color(250,250,250),int smooth=0 , float ambient=0.2, float diffuse=0.6) const ;
  int writePOVRAY(const char *filename, const Color& color=Color(250,250,250), const Point3D& view=Point3D(0,1,0), const Point3D& up=Point3D(0,0,1), int patch_type=1, double flatness=0.01, int num_u_steps=8, int num_v_steps=8) const ;
  int writePOVRAY(float tolerance, const char *filename, const Color& color=Color(250,250,250), const Point3D& view=Point3D(0,1,0), const Point3D& up=Point3D(0,0,1), int smooth=0, float ambient=0.2, float diffuse=0.6) const ;

  int writeRIB(ostream& rib) const ;
  int writeRIB(const char* filename, const Color& color=Color(255,255,0), const Point3D& view=Point3D(0,1,0)) const ;

  void tesselate(float tolerance, BasicList<Point3D> &points, BasicList<int> &connect, BasicList<Point3D> *normal=0) const ;

  // Modifies the shape of the surface
  void transform(const MatrixRT<float>& A) ;
  void modCP(int i, int j, const CtrlPnt4D& p) { P(i,j) = p ; } // modifies a control point
  void modCPby(int i, int j, const CtrlPnt4D& p) { P(i,j) += p ; } // modifies a control point
  
  float& modU(int i) { return U[i] ; }  
  float modU(int i) const { return U[i]; }
  float& modV(int i) { return V[i] ; }
  float modV(int i) const { return V[i]; }
  
  void modKnotU(const Vector_FLOAT& uKnot) { if(P.rows+degU+1==uKnot.n) U=uKnot ; } // modifies the U knot vector if uKnot is of a proper size
  void modKnotV(const Vector_FLOAT& vKnot) { if(P.cols+degV+1==vKnot.n) V=vKnot ; } // modifies the U knot vector if uKnot is of a proper size
  
  int movePoint(float u, float v, const Point3D& delta);
  int movePoint(const Vector_FLOAT& ur, const Vector_FLOAT& vr, const Vector_Point3D& D, const Vector_INT& Du, const Vector_INT& Dv) ;
  int movePoint(const Vector_FLOAT& ur, const Vector_FLOAT& vr, const Vector_Point3D& D, const Vector_INT& Du, const Vector_INT& Dv, const Vector_INT& Dk, const Vector_INT& Dl) ;
  int movePoint(const Vector_FLOAT& ur, const Vector_FLOAT& vr, const Vector_Point3D& D, const Vector_INT& Du, const Vector_INT& Dv, const Vector_INT& Dk, const Vector_INT& Dl, const BasicArray<Coordinate>& fixCP) ;

protected:
  Vector_FLOAT U ;
  Vector_FLOAT V ;
  Matrix_CtrlPnt4D P ;
  int degU ;
  int degV ;

public:
  PlNurbsSurface& transpose(void) ;
};

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlNurbsSurfaceArray --- an array of PlNurbsSurface
  Description: This class represents an array of PlNurbsSurface.
   Created by: Philippe Lavoie (4 Oct. 1996)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlNurbsSurfaceArray {
public:
  const int& n ;
  PlNurbsSurfaceArray(PlNurbsSurface* Sa, int size) ;
  PlNurbsSurfaceArray():n(sze) { S = 0 ; sze = 0 ; rsize = 0 ;}
  virtual ~PlNurbsSurfaceArray(){ if(S){ for(int i=0;i<rsize;i++) delete S[i];  delete []S ; }}
  
  virtual PlNurbsSurface& operator[](int i) { return *(S[i]) ; }
  virtual PlNurbsSurface operator[](int i) const { return *(S[i]) ; }

  virtual void resize(int s) ;
  void init(PlNurbsSurface* Sa, int size) ;

  PlNurbsSurfaceArray& operator=(const PlNurbsSurfaceArray& Sa) ;

protected:
  int sze ; // the number of NURBS curves in the array
  int rsize ; // the number of space allocated for the arra
  PlNurbsSurface** S ; // An array of pointers to NURBS curves
};

// In case someone wants the old behavior for degree elevation
inline void degreeElevate(const PlNurbsSurface& S, int tU, int tV, PlNurbsSurface& nS) { nS = S ; nS.degreeElevate(tU,tV) ; }
inline void degreeElevateU(const PlNurbsSurface& S, int tU, PlNurbsSurface& nS) { nS = S ; nS.degreeElevateU(tU) ; }
inline void degreeElevateV(const PlNurbsSurface& S, int tV, PlNurbsSurface& nS) { nS = S ; nS.degreeElevateV(tV) ; }


#endif 
