/*=============================================================================
        File: nurbsS_sp.hh
     Purpose:       
    Revision: $Id: nurbsS_sp.hh,v 2.6 1998/05/29 18:29:13 lavoie Exp $
  Created by: Philippe Lavoie          (8 May, 1998)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _nurbs_nurbsS_sp_h_
#define _nurbs_nurbsS_sp_h_

#include "nurbsS.hh"

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlNurbsSurfaceSP --- A NURBS surface with surface point
  Description: A Nurbs surface with surface point manipulators. This allows 
               someone to modify the point on a surface for which a 
	       control point has maximal influence over it. This might 
	       provide a more intuitive method to modify a surface. 

   Created by: Philippe Lavoie (8 May, 1998)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlNurbsSurfaceSP : public PlNurbsSurface{
public:
  PlNurbsSurfaceSP() ;
  PlNurbsSurfaceSP(const PlNurbsSurface& nS)  ;
  PlNurbsSurfaceSP(const PlNurbsSurfaceSP& nS)  ;
  PlNurbsSurfaceSP(int DegU, int DegV, const Vector_FLOAT& Uk, const Vector_FLOAT& Vk, const Matrix_CtrlPnt4D& Cp)  ;
  PlNurbsSurfaceSP(int DegU, int DegV, Vector_FLOAT& Uk, Vector_FLOAT& Vk, Matrix_Point3D& Cp, Matrix_FLOAT& W)  ;


  virtual PlNurbsSurface& operator=(const PlNurbsSurface& a) ;
  virtual PlNurbsSurfaceSP& operator=(const PlNurbsSurfaceSP& a) ;
  
  virtual void resizeKeep(int Pu, int Pv, int DegU, int DegV) ;
  

  virtual void refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  virtual void refineKnotU(const Vector_FLOAT& X);
  virtual void refineKnotV(const Vector_FLOAT& X);
  
  virtual void mergeKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) ;
  virtual void mergeKnotU(const Vector_FLOAT& X);
  virtual void mergeKnotV(const Vector_FLOAT& X);

  virtual int read(ifstream &fin);

  
  virtual void degreeElevateU(int tU) ;
  virtual void degreeElevateV(int tV) ;

  PlNurbsSurfaceSP generateParallel(float d) const ; 


  void modSurfCPby(int i, int j, const CtrlPnt4D& a) 
    { P(i,j) +=  a / (maxU[i]*maxV[j]) ;  }
  void modSurfCP(int i, int j, const CtrlPnt4D& a) 
    { modSurfCPby(i,j,a-surfP(i,j)) ;  }

  float maxAtUV(int i, int j) const { return maxAtU_[i]*maxAtV_[j] ; }
  float maxAtU(int i) const { return maxAtU_[i] ; }
  float maxAtV(int i) const { return maxAtV_[i] ; }

  CtrlPnt4D surfP(int i,int j) const 
    { return point4D(maxAtU_[i],maxAtV_[j]); }

  void updateMaxUV() 
    { updateMaxU() ; updateMaxV() ; }
  void updateMaxU() ;
  void updateMaxV() ;

  int okMax() { return (maxU.n<=1)?0:1 ; }

protected:

  Vector_FLOAT maxU,maxV ;
  Vector_FLOAT maxAtU_,maxAtV_ ;
};


inline PlNurbsSurfaceSP::PlNurbsSurfaceSP() : PlNurbsSurface() { 
  ;
}

inline PlNurbsSurfaceSP::PlNurbsSurfaceSP(const PlNurbsSurface& nS)  : PlNurbsSurface(nS) { 
  updateMaxUV() ; 
}

inline PlNurbsSurfaceSP::PlNurbsSurfaceSP(const PlNurbsSurfaceSP& nS)  : PlNurbsSurface(nS) { 
  maxU = nS.maxU ; 
  maxV = nS.maxV ; 
  maxAtU_ = nS.maxAtU_ ; 
  maxAtV_ = nS.maxAtV_ ; 
}

inline PlNurbsSurfaceSP::PlNurbsSurfaceSP(int DegU, int DegV, const Vector_FLOAT& Uk, const Vector_FLOAT& Vk, const Matrix_CtrlPnt4D& Cp)  : PlNurbsSurface(DegU,DegV,Uk,Vk,Cp) { 
  updateMaxUV() ; 
}

inline PlNurbsSurfaceSP::PlNurbsSurfaceSP(int DegU, int DegV, Vector_FLOAT& Uk, Vector_FLOAT& Vk, Matrix_Point3D& Cp, Matrix_FLOAT& W)  : PlNurbsSurface(DegU,DegV,Uk,Vk,Cp,W) { 
  updateMaxUV() ; 
}

inline PlNurbsSurface& 
PlNurbsSurfaceSP::operator=(const PlNurbsSurface& a) { 
  PlNurbsSurface::operator=(a) ; 
  updateMaxUV() ; 
  return *this ; 
}

inline  PlNurbsSurfaceSP& 
PlNurbsSurfaceSP::operator=(const PlNurbsSurfaceSP& a) { 
  PlNurbsSurface::operator=(a) ;  
  maxU = a.maxU ; 
  maxV = a.maxV ; 
  maxAtU_ = a.maxAtU_ ; 
  maxAtV_ = a.maxAtV_ ;  
  return *this ; 
}

inline void 
PlNurbsSurfaceSP::resizeKeep(int Pu, int Pv, int DegU, int DegV) { 
  PlNurbsSurface::resizeKeep(Pu,Pv,DegU,DegV) ; 
  updateMaxUV() ; 
}

inline void 
PlNurbsSurfaceSP::refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) {
  PlNurbsSurface::refineKnots(nU,nV) ;
  updateMaxUV() ;
}
  
inline void 
PlNurbsSurfaceSP::refineKnotU(const Vector_FLOAT& X){
  PlNurbsSurface::refineKnotU(X) ;
  updateMaxU() ;
}

inline void 
PlNurbsSurfaceSP::refineKnotV(const Vector_FLOAT& X){
  PlNurbsSurface::refineKnotV(X) ;
  updateMaxV() ;
}

  
inline void 
PlNurbsSurfaceSP::mergeKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) {
  PlNurbsSurface::mergeKnots(nU,nV) ;
  updateMaxUV() ;
}


inline void 
PlNurbsSurfaceSP::mergeKnotU(const Vector_FLOAT& X){
  PlNurbsSurface::mergeKnotU(X) ;
  updateMaxU() ;
}


inline void 
PlNurbsSurfaceSP::mergeKnotV(const Vector_FLOAT& X){
  PlNurbsSurface::mergeKnotV(X) ;
  updateMaxV() ;
}

inline int PlNurbsSurfaceSP::read(ifstream &fin) {
  int r = PlNurbsSurface::read(fin) ;
  updateMaxUV() ; 
  return r ; 
}

inline void PlNurbsSurfaceSP::degreeElevateU(int tU) {
  PlNurbsSurface::degreeElevateU(tU);
  updateMaxU();
}

inline void PlNurbsSurfaceSP::degreeElevateV(int tV) {
  PlNurbsSurface::degreeElevateV(tV);
  updateMaxV();
}


#endif
