/*=============================================================================
        File: nurbs_sp.hh
     Purpose:       
    Revision: $Id: nurbs_sp.hh,v 2.5 1998/05/25 21:08:22 lavoie Exp $
  Created by: Philippe Lavoie          (7 May, 1998)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#ifndef _nurbs_nurbs_sp_h_
#define _nurbs_nurbs_sp_h_

#include "nurbs.hh"

/**  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **
        Class: PlNurbsCurveSP --- A NURBS curve with surface point
  Description: A Nurbs curve with surface point manipulators. This allows 
               someone to modify the point on a curve for which a ControlPoint
	       has maximal influence over it. This might provide a more
	       intuitive method to modify a curve. 

   Created by: Philippe Lavoie (7 May, 1998)
  Modified by:
 **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **  **/
class PlNurbsCurveSP : public PlNurbsCurve{
public:
  PlNurbsCurveSP() : PlNurbsCurve() { ; }
  PlNurbsCurveSP(const PlNurbsCurve& nurb) ;
  PlNurbsCurveSP(const PlNurbsCurveSP& nurb) ;
  PlNurbsCurveSP(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int degree=3) ;
  PlNurbsCurveSP(const Vector_Point3D& P1, const Vector_FLOAT &W, const Vector_FLOAT &U1, int degree=3) ;


  virtual void reset(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int degree);
  virtual PlNurbsCurve& operator=(const PlNurbsCurve& a);
  PlNurbsCurveSP& operator=(const PlNurbsCurveSP& a);

  virtual void modKnot(const Vector_FLOAT& knot) ;

  virtual void removeKnot(int r, int s, int num) ;
  virtual void removeKnotsBound(const Vector_FLOAT& ub, Vector_FLOAT& ek, float E);

  virtual void refineKnotVector(const Vector_FLOAT& X) ;
  virtual void mergeKnotVector(const Vector_FLOAT &Um) ;
  virtual void knotInsertion(float u, int r,PlNurbsCurveSP& nc);


  virtual void degreeElevate(int t);

  int read(ifstream &fin) ;

  void modSurfCPby(int i, const CtrlPnt4D& a) 
    { P[i] +=  a / maxU[i] ;  }
  void modSurfCP(int i, const CtrlPnt4D& a) 
    { modSurfCPby(i,a-surfP(i)) ;  }

  float maxAt(int i) const 
    { return maxAt_[i] ; }

  CtrlPnt4D surfP(int i) const 
    { return point4D(maxAt_[i]) ; }

  void updateMaxU() ;

  int okMax() 
    { return (maxU.n<=1)?0:1 ; }

protected:

  Vector_FLOAT maxU ;
  Vector_FLOAT maxAt_ ;
};


inline PlNurbsCurveSP::PlNurbsCurveSP(const PlNurbsCurve& nurb) : 
  PlNurbsCurve(nurb) 
{
  updateMaxU(); 
}

inline PlNurbsCurveSP::PlNurbsCurveSP(const PlNurbsCurveSP& nurb) : 
  PlNurbsCurve(nurb) 
{ 
  maxU = nurb.maxU ;  
  maxAt_ = nurb.maxAt_ ; 
}

inline PlNurbsCurveSP::PlNurbsCurveSP(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int degree=3) : 
  PlNurbsCurve(P1,U1,degree) 
{ 
  updateMaxU(); 
}

inline PlNurbsCurveSP::PlNurbsCurveSP(const Vector_Point3D& P1, const Vector_FLOAT &W, const Vector_FLOAT &U1, int degree=3) : 
  PlNurbsCurve(P1,W,U1,degree) 
{ 
  updateMaxU(); 
}


inline void PlNurbsCurveSP::reset(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int degree) { 
  PlNurbsCurve::reset(P1,U1,degree); 
  updateMaxU() ; 
}

inline PlNurbsCurve& PlNurbsCurveSP::operator=(const PlNurbsCurve& a) { 
  PlNurbsCurve::operator=(a); 
  updateMaxU() ; 
  return *this; 
}

inline PlNurbsCurveSP& PlNurbsCurveSP::operator=(const PlNurbsCurveSP& a) { 
  PlNurbsCurve::operator=(a); 
  maxU = a.maxU ; 
  maxAt_ = a.maxAt_ ; 
  return *this; 
}

inline void PlNurbsCurveSP::modKnot(const Vector_FLOAT& knot) { 
  PlNurbsCurve::modKnot(knot) ; 
  updateMaxU() ; 
}

inline void PlNurbsCurveSP::removeKnot(int r, int s, int num) { 
  PlNurbsCurve::removeKnot(r,s,num); 
  updateMaxU() ; 
}

inline void PlNurbsCurveSP::removeKnotsBound(const Vector_FLOAT& ub, Vector_FLOAT& ek, float E) { 
  PlNurbsCurve::removeKnotsBound(ub,ek,E); 
  updateMaxU() ; 
}

inline void PlNurbsCurveSP::refineKnotVector(const Vector_FLOAT& X) {
  PlNurbsCurve::refineKnotVector(X); 
  updateMaxU() ; 
}

inline void PlNurbsCurveSP::mergeKnotVector(const Vector_FLOAT &Um) { 
  PlNurbsCurve::mergeKnotVector(Um); 
  updateMaxU() ; 
}

inline void PlNurbsCurveSP::knotInsertion(float u, int r,PlNurbsCurveSP& nc){ 
  PlNurbsCurve::knotInsertion(u,r,nc) ; 
  nc.updateMaxU() ; 
}

inline int PlNurbsCurveSP::read(ifstream &fin) {
  int r = PlNurbsCurve::read(fin) ;
  updateMaxU() ; 
  return r ;
}

inline void PlNurbsCurveSP::degreeElevate(int t){
  PlNurbsCurve::degreeElevate(t) ;
  updateMaxU() ; 
}


#endif
