/*=============================================================================
        File: curve.cc
     Purpose:       
    Revision: $Id: curve.cc,v 2.1 1998/05/22 16:35:33 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#include <curve.hh>

inline float abs2(float x){
  return x*x ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: minDist2 --- find the minimal distance between a point and the
                             curve
  Description: This is an iterative method to find the closest point to a 
               curve. 
        Input: p --> the minimal distance from that point
           guess <-> a starting value for the parameter $u$, on exit this 
	             will be set to the value
		     of the point on the curve closest to $p$.
	   error --> when iterations have an error smaller than this value, 
	             the function exits
	       s --> the size of the search in the parametric space.
             sep --> the number of points initially looked at to find a 
	             minimal distance
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value
              uM --> the maximal parametric value
       Output: The value of the minimal distance between $p$ and the curve. 
               The variable guess now holds the parametric value of the curve 
               point closest to $p$. 
 Restrictions: It has not been tested with closed loop curves.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlParaCurve::minDist2(const Point3D& p, float& guess,float error,float s,int sep,int maxIter, float um, float uM) const {
  if(um<0)
    um = minKnot() ;
  if(uM<0)
    uM = maxKnot() ;
  if(s<0)
    s = uM-um ;
  float d,d1,d2 ;
  Point3D p2 ;
  p2 = point3D(guess) ;
  d = abs2(p-p2) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float step ;
  step = s/(float)sep ;
  u1 = guess-s ;
  u2 = guess+s ;
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    float u = u1 ;
    d2 = d1 ;
    for(;u<u2;u+=step){
      p2 = point3D(u) ;
      d1 = abs2(p-p2) ;
      if(d1<d){
	d = d1 ;
	guess = u ;
      }
    }
    s /= 2.0 ;
    u1 = guess - s ;
    u2 = guess + s ;
    step = 2.0*s/(float)sep ;
    if(d-d2==0) niter = maxIter ;
    if(step<error) niter = maxIter ;
    niter++;
  }
  return d ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: minDistY --- find the closest point on the curve to the $y$ 
                            coordinate
  Description: This is an iterative method to find the closest point on the 
               curve to the y coordinate.
        Input: y --> the $y$ coordinate to be close too.
           guess <-> a starting value for the parameter $u$, on exit this 
	             will be set to the value of the point on the curve 
		     closest to $y$.
           error --> when iterations have an error smaller than this value, 
	             the function exits
               s --> the size of the search in the parametric space.
             sep --> the number of points initially looked at to find a minimal
	             distance
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value
              uM --> the maximal parametric value
       Output: The value of the minimal distance between $p$ and the curve. 
               The variable guess now holds
	       the parametric value of the curve point closest to $p$. 
 Restrictions: It has not been tested with closed loop curves.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Point3D PlParaCurve::minDistY(float y, float& guessU, float error, float s, int sep, int maxIter, float um, float uM) const {
  if(um<0)
    um = minKnot() ;
  if(uM<0)
    uM = maxKnot() ;
  if(s<0)
    s = uM-um ;
  float d,d1,d2 ;
  Point3D p2,result ;
  p2 = point3D(guessU) ;
  result = p2 ;
  d = abs2(y-p2.y) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float step ;
  step = s/(float)sep ;
  u1 = guessU-s ;
  u2 = guessU+s ;
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    float u = u1 ;
    d2 = d1 ;
    for(;u<u2;u+=step){
      p2 = point3D(u) ;
      d1 = abs2(y-p2.y) ;
      if(d1<d){
	d = d1 ;
	guessU = u ;
	result = p2 ;
      }
    }
    s /= 2.0 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    step = 2.0*s/(float)sep ;
    if(d-d2==0) niter = maxIter ;
    if(step<error) niter = maxIter ;
    niter++;
  }
  return result ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: minDistX --- find the closest point on the curve to the $x$ 
                            coordinate
  Description: This is an iterative method to find the closest point on the 
               curve to the x coordinate.
        Input: x --> the $x$ coordinate to be close too.
           guess <-> a starting value for the parameter $u$, on exit this 
	             will be set to the value of the point on the curve 
		     closest to $x$.
           error --> when iterations have an error smaller than this value, 
	             the function exits
               s --> the size of the search in the parametric space.
             sep --> the number of points initially looked at to find a minimal
	             distance
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value
              uM --> the maximal parametric value
       Output: The value of the minimal distance between $p$ and the curve. 
               The variable guess now holds
	       the parametric value of the curve point closest to $p$. 
 Restrictions: It has not been tested with closed loop curves.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Point3D PlParaCurve::minDistX(float x, float& guessU, float error, float s, int sep, int maxIter, float um, float uM) const {
  if(um<0)
    um = minKnot() ;
  if(uM<0)
    uM = maxKnot() ;
  if(s<0)
    s = uM-um ;
  float d,d1,d2 ;
  Point3D p2,result ;
  p2 = point3D(guessU) ;
  result = p2 ;
  d = abs2(x-p2.x) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float step ;
  step = s/(float)sep ;
  u1 = guessU-s ;
  u2 = guessU+s ;
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    float u = u1 ;
    d2 = d1 ;
    for(;u<u2;u+=step){
      p2 = point3D(u) ;
      d1 = abs2(x-p2.x) ;
      if(d1<d){
	d = d1 ;
	guessU = u ;
	result = p2 ;
      }
    }
    s /= 2.0 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    step = 2.0*s/(float)sep ;
    if(d-d2==0) niter = maxIter ;
    if(step<error) niter = maxIter ;
    niter++;
  }
  return result ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: minDistX --- find the closest point on the curve to the $x$ 
                            coordinate
  Description: This is an iterative method to find the closest point on the 
               curve to the x coordinate.
        Input: z --> the $x$ coordinate to be close too.
           guess <-> a starting value for the parameter $u$, on exit this 
	             will be set to the value of the point on the curve 
		     closest to $x$.
           error --> when iterations have an error smaller than this value, 
	             the function exits
               s --> the size of the search in the parametric space.
             sep --> the number of points initially looked at to find a minimal
	             distance
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value
              uM --> the maximal parametric value
       Output: The value of the minimal distance between $p$ and the curve. 
               The variable guess now holds
	       the parametric value of the curve point closest to $p$. 
 Restrictions: It has not been tested with closed loop curves.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Point3D PlParaCurve::minDistZ(float z, float& guessU, float error, float s, int sep, int maxIter, float um, float uM) const {
  if(um<0)
    um = minKnot() ;
  if(uM<0)
    uM = maxKnot() ;
  if(s<0)
    s = uM-um ;
  float d,d1,d2 ;
  Point3D p2,result ;
  p2 = point3D(guessU) ;
  result = p2 ;
  d = abs2(z-p2.z) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float step ;
  step = s/(float)sep ;
  u1 = guessU-s ;
  u2 = guessU+s ;
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    float u = u1 ;
    d2 = d1 ;
    for(;u<u2;u+=step){
      p2 = point3D(u) ;
      d1 = abs2(z-p2.z) ;
      if(d1<d){
	d = d1 ;
	guessU = u ;
	result = p2 ;
      }
    }
    s /= 2.0 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    step = 2.0*s/(float)sep ;
    if(d-d2==0) niter = maxIter ;
    if(step<error) niter = maxIter ;
    niter++;
  }
  return result ;
}

inline float coordValue(PlParaCurve::CoordinateType coord, const Point3D& p){
  switch(coord){
  case PlParaCurve::coordX: return p.x ; break ;
  case PlParaCurve::coordY: return p.y ; break ;
  case PlParaCurve::coordZ: return p.z ; break ;
  }
  return 0.0 ; // elliminates warning messages
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: extremum --- finds the minimal or maximal value on the curve
                            of the x,y or z coordinate.
  Description: Finds the minimal or maximal value on the curve of the x,y or 
               z coordinate.
        Input: findMin --> a flag indicatinf if we're looking for the minimal
	                   value or the maximal value.
	   coord --> Which coordinate to find: x,y or z.
           minDu --> The minimal distance between iterations in the parametric
	             space.
             sep --> the number of points initially looked at to find a minimal
	             distance
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value
              uM --> the maximal parametric value
       Output: The minimal value of $z$ along the curve
 Restrictions: It has not been tested with closed loop curves.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlParaCurve::extremum(int findMin, CoordinateType coord, float minDu, int sep, int maxIter, float um, float uM) const {
  if(um<0)
    um = minKnot() ;
  if(uM<0)
    uM = maxKnot() ;

  float c,d,d1,d2,result,guessU ;
  Point3D p2 ;
  p2 = point3D(um) ;
  c = coordValue(coord,p2) ;
  p2 = point3D(uM) ;
  if(findMin)
    c = minimum(c,coordValue(coord,p2)) ;
  else
    c = maximum(c,coordValue(coord,p2)) ;
  result = c ;
  d = minDu*10.0 ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float step ;
  float s ;
  s = uM - um ;
  step = s/(float)(sep+1) ;
  u1 = um ;
  u2 = uM ;
  guessU = um ;
  while(d>minDu && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    float u = u1 ;
    d2 = c ;
    d = guessU ;
    for(;u<=u2;u+=step){
      p2 = point3D(u) ;
      if(findMin){
	d1 = minimum(c,coordValue(coord,p2)) ;
	if(d1<c){
	  c = d1 ;
	  guessU = u ;
	  result = d1 ;
	}
      }
      else{
	d1 = maximum(c,coordValue(coord,p2)) ;
	if(d1>c){
	  c = d1 ;
	  guessU = u ;
	  result = d1 ;
	}
      }
    }
    s /= 2.0 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    step = 2.0*s/(float)sep ;
    if((c-d2)==0.0) niter = maxIter ;
    if(step<minDu) niter = maxIter ;
    d = absolute(guessU-d) ;
    niter++;
  }
  return result ;
}

