/*=============================================================================
        File: hnurbsS.cc
     Purpose:       
    Revision: $Id: hnurbsS.cc,v 2.10 1998/06/03 15:43:41 lavoie Exp $
  Created by: Philippe Lavoie    (7 October 1997)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1998 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#include <hnurbsS.hh>
#include <string.h>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlHNurbsSurface --- the basic constructor
  Description: The basic constructor
        Input: 
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::PlHNurbsSurface():PlNurbsSurface(), level(level_){

  baseLevel_ = 0 ;
  nextLevel_ = 0 ;
  firstLevel_ = lastLevel_ = this ;

  rU.resize(0) ;
  rV.resize(0) ;
  updateN = 0 ;
  baseUpdateN = 0 ;

  level_ = 0 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlHNurbsSurface --- constructor with a base level
  Description: Constructor with a base level
        Input: base --> the base level
	       xU --> the U knots to insert in this level
	       xV --> the V knots to insert in this level
       Output: 
 Restrictions: The base pointer must be pointing to a valid object
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::PlHNurbsSurface(PlHNurbsSurface* base, const PlVector<float>& xU, const PlVector<float>& xV):PlNurbsSurface(),level(level_){

  if(!base){
    PlError error("PlHNurbsSurface constructor") ;
    error << "Initializing a PlHNurbsSurface with a null base pointer!" ;
    error.fatal() ;
  }
  if(base->nextLevel_){
    PlError error("PlHNurbsSurface constructor") ;
    error << "You're trying to replace an existing level, this is not allowed." ;
    error.fatal() ;
  }


  nextLevel_ = 0 ;
  firstLevel_ = base->firstLevel_ ;
  lastLevel_ = this ;
  baseLevel_ = (PlHNurbsSurface*)base ;

  // update the information in the previous levels
  baseLevel_->nextLevel_ = this ;
  PlHNurbsSurface* l ;
  l = baseLevel_ ;
  while(l){
    l->lastLevel_ = this ;
    l = l->baseLevel_ ;
  }


  level_ = base->level_ + 1 ;

  rU = xU ;
  rV = xV ;

  updateN = 0 ;
  baseUpdateN = baseLevel_->modifiedN()-1 ; // Set it so that initBaseLevel will run

  initBase() ;
  offset.resize(baseSurf.CtrlP) ;

  P = baseSurf.CtrlP ;
  U = baseSurf.KnotU ;
  V = baseSurf.KnotV ;
  degU = baseSurf.degreeU ;
  degV = baseSurf.degreeV ;

  //updateSurface() ;

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlHNurbsSurface --- constructor with a base level
  Description: Constructor with a base level
        Input: base --> the base level
	       xU --> the U knots to insert in this level
	       xV --> the V knots to insert in this level
       Output: 
 Restrictions: The base pointer must be pointing to a valid object
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::PlHNurbsSurface(PlHNurbsSurface* base):PlNurbsSurface(),level(level_){

  if(!base){
    PlError error("PlHNurbsSurface constructor") ;
    error << "Initializing a PlHNurbsSurface with a null base pointer!" ;
    error.fatal() ;
  }
  if(base->nextLevel_){
    PlError error("PlHNurbsSurface constructor") ;
    error << "You're trying to replace an existing level, this is not allowed." ;
    error.fatal() ;
  }

  nextLevel_ = 0 ;
  firstLevel_ = base->firstLevel_ ;
  lastLevel_ = this ;
  baseLevel_ = (PlHNurbsSurface*)base ;

  // update the information in the previous levels
  baseLevel_->nextLevel_ = this ;
  PlHNurbsSurface* l ;
  l = baseLevel_ ;
  while(l){
    l->lastLevel_ = this ;
    l = l->baseLevel_ ;
  }

  level_ = base->level_ + 1 ;

  updateN = 0 ;
  rU.resize(0) ;
  rV.resize(0) ;

  baseUpdateN = baseLevel_->modifiedN()-1 ; // Set it so that initBase will run
  initBase() ;
  offset.resize(baseSurf.CtrlP) ;
  P = baseSurf.CtrlP ;
  U = baseSurf.KnotU ;
  V = baseSurf.KnotV ;
  degU = baseSurf.degreeU ;
  degV = baseSurf.degreeV ;
  //updateSurface() ;

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlHNurbsSurface --- constructs a base HNURBS
  Description: Constructs a base HNURBS. This HNURBS surface is set to level 0.
               And it corresponds to the NURBS surface. 
	       
	       This constructor does not transform the NURBS surface into
	       a HNURBS surface. It only copies the values from the NURBS
	       surface as it's base offset values.
        Input: S --> the NURBS surface at level 0
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::PlHNurbsSurface(const PlNurbsSurface& S):PlNurbsSurface(S),level(level_){

  baseLevel_ = 0 ;
  nextLevel_ = 0 ; 
  firstLevel_ = lastLevel_ = this ;

  level_ = 0 ;
  updateN = 0 ;
  baseUpdateN = 0 ;

  rU.resize(0) ;
  rV.resize(0) ;

  offset = P ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlHNurbsSurface --- the copy constructor
  Description: A copy constructor
        Input: S --> the HNURBS surface to copy
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::PlHNurbsSurface(const PlHNurbsSurface& S):PlNurbsSurface(S),level(level_) {

  baseLevel_ = nextLevel_ = 0 ;
  firstLevel_ = lastLevel_ = this ;

  level_ = 0 ;
  updateN = 0 ;
  baseUpdateN = 0 ;

  rU = S.rU ; 
  rV = S.rV ; 

  offset = S.offset ;

  this->copy(S) ;

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlHNurbsSurface --- the copy constructor
  Description: A copy constructor
        Input: S --> the HNURBS surface to copy
       Output: 
 Restrictions: The base pointer must be pointing to a valid object
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::PlHNurbsSurface(PlHNurbsSurface *base, const PlHNurbsSurface &surf):PlNurbsSurface(surf),level(level_) {

  if(!base){
    PlError error("PlHNurbsSurface constructor") ;
    error << "Initializing a PlHNurbsSurface with a null base pointer!" ;
    error.fatal() ;
  }
  if(base->nextLevel_){
    PlError error("PlHNurbsSurface constructor") ;
    error << "You're trying to replace an existing level, this is not allowed." ;
    error.fatal() ;
  }

  baseLevel_ = (PlHNurbsSurface*)base ;
  nextLevel_ = 0 ;
  lastLevel_ = this ;

  // update the information in the previous levels
  baseLevel_->nextLevel_ = this ;
  PlHNurbsSurface* l ;
  l = baseLevel_ ;
  while(l){
    l->lastLevel_ = this ;
    l = l->baseLevel_ ;
  }

  firstLevel_ = base->firstLevel_ ;
  level_ = base->level+1 ;
  baseUpdateN = base->modifiedN()-1 ;

  initBase() ; 

  updateN = 0 ;

  this->copy(surf) ;

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: copy --- copies a HNurbs Surface and all it children
  Description: Copies a HNurbs Surface and all it children
        Input: ns --> the HNurbs surface to copy
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::copy(const PlHNurbsSurface& nS){
  PlHNurbsSurface *levelP ;
  levelP = nS.nextLevel() ;

  PlNurbsSurface::operator=(nS) ;
  rU = nS.rU ;
  rV = nS.rV ;
  offset = nS.offset ;

  firstLevel_ = this ;

  if(levelP){
    PlHNurbsSurface *newLevel = new PlHNurbsSurface(this,*levelP) ;
    nextLevel_ = newLevel ;
    lastLevel_ = nextLevel_->lastLevel() ;
  }
  else{
    lastLevel_ = this ;
  }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateSurface --- updates the NURBS surface
  Description: Updates the NURBS surface according to the offset values
               and its base level. You can update only one control point 
	       from the surface if you specify a value for i and j or you 
	       can update all the points if i0 or j0 is below 0.
        Input: i0 --> the row of the control point to update [-1]
	       j0 --> the column of the control point to update [-1]
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by: Philippe Lavoie (15 April 1998)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::updateSurface(int i0, int j0){
  if(i0>=0 && j0>=0){
    if(offset(i0,j0).x()==0.0 && offset(i0,j0).y()==0.0 && offset(i0,j0).z()==0.0)
      return ;
  }
  if(baseLevel_){
    if(initBase()){
      P = baseSurf.CtrlP ;
      U = baseSurf.KnotU ;
      V = baseSurf.KnotV ;
      degU = baseSurf.degreeU ;
      degV = baseSurf.degreeV ;
    }
    if(i0>=0 && j0>=0){
      Point3D vecOffset ;
      vecOffset = offset(i0,j0).x()*ivec(i0,j0) +
	offset(i0,j0).y()*jvec(i0,j0) +
	offset(i0,j0).z()*kvec(i0,j0) ;
      P(i0,j0).x() = baseSurf.CtrlP(i0,j0).x()+vecOffset.x ;
      P(i0,j0).y() = baseSurf.CtrlP(i0,j0).y()+vecOffset.y ;
      P(i0,j0).z() = baseSurf.CtrlP(i0,j0).z()+vecOffset.z ;
    }
    else{
      for(int i=0;i<P.rows;++i)
	for(int j=0;j<P.cols;++j){
	  if(offset(i,j).x() != 0 || 
	     offset(i,j).y() != 0 || offset(i,j).z() || 0){
	    Point3D vecOffset ;
	    vecOffset = offset(i,j).x()*ivec(i,j) +
	      offset(i,j).y()*jvec(i,j) +
	      offset(i,j).z()*kvec(i,j) ;
	    P(i,j).x() = baseSurf.CtrlP(i,j).x()+vecOffset.x ;
	    P(i,j).y() = baseSurf.CtrlP(i,j).y()+vecOffset.y ;
	    P(i,j).z() = baseSurf.CtrlP(i,j).z()+vecOffset.z ;
	  }
	}
    }
  }
  else{
    if(i0>=0 && j0>=0)
      P(i0,j0) = offset(i0,j0) ;
    else{
      for(int i=0;i<P.rows;++i)
	for(int j=0;j<P.cols;++j){
	  P(i,j) = offset(i,j) ;
	}
    }
  }

  ++updateN ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: initBase --- initialize the base surface 
  Description: Initialize the base surface from the previous level if it has 
               been modified.
        Input: force --> if set, this forces an update of the base surface
       Output: 1 if the base surface is modified, 0 otherwise.
 Restrictions: 
   Created by: Philippe Lavoie (15 April 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::initBase(int force=0){
  if(!baseLevel_){
    return 0; // this means that the surface is the base level
  }

  // make sure none of the base level need updating
  if(baseLevel_->initBase())
    force = 1 ;
  

  // Only initialize the base level if it isn't up to date
  if(baseLevel_->modifiedN() == baseUpdateN && !force){
    return 0;
  }

  baseUpdateN = baseLevel_->modifiedN() ;

  baseSurf = *baseLevel_ ;

  if(rU.n>0)
    baseSurf.refineKnotU(rU) ;
  if(rV.n>0)
    baseSurf.refineKnotV(rV) ;

  PlVector<float> maxU,maxV ;
  int i,j ;
  
  if(baseSurf.degreeU>3)
    averagingKnots(baseSurf.KnotU,baseSurf.degreeU,maxU) ;
  else{
    maxU.resize(baseSurf.CtrlP.rows) ;
    for(i=0;i<baseSurf.CtrlP.rows;++i)
      if(!maxInfluence(i,baseSurf.KnotU,baseSurf.degreeU,maxU[i]))
	cerr << "Problem in maxInfluence U!\n" ;
  }

  if(baseSurf.degreeV>3)
    averagingKnots(baseSurf.KnotV,baseSurf.degreeV,maxV) ;
  else{
    maxV.resize(baseSurf.CtrlP.cols) ;
    for(i=0;i<baseSurf.CtrlP.cols;++i)
      if(!maxInfluence(i,baseSurf.KnotV,baseSurf.degreeV,maxV[i]))
	cerr << "Problem in maxInfluence V!\n" ;
  }

  ivec.resize(maxU.n,maxV.n) ;
  jvec.resize(maxU.n,maxV.n) ;
  kvec.resize(maxU.n,maxV.n) ;

  Matrix_Point3D ders ;

  for(i=0;i<maxU.n;++i)
    for(j=0;j<maxV.n;++j){
      baseSurf.deriveAt(maxU[i],maxV[j],1,ders) ;
      // It is possible that there are no valid derivative at that point
      // we then move to a nearby point to get a valid value
      Point3D norm = crossProduct(ders(0,1),ders(1,0)) ;
      if(norm.x == float(0) &&
	    norm.y == float(0) &&
	    norm.z == float(0)){
	const float delta = 0.00001 ;
	float nt = 1.0 ; 
	Matrix_Point3D dersT(ders) ;
	while(norm.x == float(0) && 
	      norm.y == float(0) &&
	      norm.z == float(0)){
	  if( nt*delta >(baseSurf.KnotU[baseSurf.KnotU.n-1]-baseSurf.KnotU[0])){
	      PlError error("initBase");
	      error << "Can't compute the derivative.\n" ;
	      error.fatal() ;
	    }
	  float nd ;

	  ders.reset(0) ;

	  nd = 0 ;
	  if(i != 0){
	    baseSurf.deriveAt(maxU[i] - nt*delta, maxV[j],1, dersT) ;
	    ders += dersT ;
	    ++nd ;
	  }
	  if(i != maxU.n-1){
	    baseSurf.deriveAt(maxU[i] + nt*delta, maxV[j],1, dersT) ;
	    ders += dersT ;
	    ++nd ;
	  }
	  if(j != 0){
	    baseSurf.deriveAt(maxU[i], maxV[j] - nt*delta,1, dersT) ;
	    ders += dersT ;
	    ++nd ;
	  }
	  if(j != maxV.n-1){
	    baseSurf.deriveAt(maxU[i], maxV[j] + nt*delta,1, dersT) ;
	    ders += dersT ;
	    ++nd ;
	  }

	  if(nd==0){
	    PlError error("initBase");
	    error << "Can't compute the derivative.\n" ;
	    error.fatal() ;
	  }

	  ders /= nd ;

	  norm = crossProduct(ders(0,1),ders(1,0)) ;
	  nt *= 10.0 ; 
	}
      }	

      ivec(i,j) = ders(0,1).unitLength() ;
      kvec(i,j) = crossProduct(ders(0,1),ders(1,0)).unitLength() ;
      jvec(i,j) = crossProduct(kvec(i,j),ivec(i,j)).unitLength() ;
    }
    

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: modifies --- specifies the level that modifies the region
  Description: Specifies what level modifies the region $(u,v)$
        Input: u --> the $u$ parametric value
	       v --> the $v$ parametric value
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::modifies(float u, float v){
  if(nextLevel_){
    int mod = nextLevel_->modifies(u,v) ;
    if(mod>=0)
      return mod ;
  }

  if(u<KnotU[0] || u>KnotU[KnotU.n-1])
    return -1 ;
  if(v<KnotV[0] || v>KnotU[KnotV.n-1])
    return -1 ;

  int su = findSpanU(u) ;
  int sv = findSpanV(v) ;

  for(int i=0;i<=degU;++i)
    for(int j=0;j<=degV;++j){
      if(offset(su-degU+i,sv+degV+j) != CtrlPnt4D(0,0,0,0))
	return level_ ;
    }

  return -1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: pointAt --- finds the 4D point at (u,v) for a certain level of
                           detail.
  Description: Finds the 4D point at (u,v) for a certain level of detail.
        Input: u --> the $u$ parametric value
	       v --> the $v$ parametric value
	     lod --> the level of detail3
       Output: the 4D point at (u,v) and at the level of detail specivied by
               lod.
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
CtrlPnt4D PlHNurbsSurface::pointAt(float u, float v, int lod) const {
  if(level_==lod)
    return PlNurbsSurface::operator()(u,v) ;

  if(nextLevel_)
    return nextLevel_->pointAt(u,v,lod) ; 
  
  return PlNurbsSurface::operator()(u,v) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: maxLevel --- the maximum level of detail
  Description: Finds the maximum level of detail available from this HNURBS
               surface
        Input: 
       Output: the maximum level of detail.
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::maxLevel() const{
  if(lastLevel_)
    return lastLevel_->level_ ;
  return level_ ; 
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: deriveAt --- Finds the derivative of the point $(u,v)$
  Description: Computes the matrix of derivatives at $u,v$. 
                 The value of skl(k,l) represents the 
		 derivative of the surface $S(u,v)$ with respect to 
		 $u$ $k$ times and to $v$ $l$ times.
        Input: u --> the $u$ parametric value
	       v --> the $v$ parametric value
	    ders --> the Matrix of derivatives.
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::deriveAt(float u, float v, int d, Matrix_CtrlPnt4D& ders,int lod) const {
  if(level_==lod){
    PlNurbsSurface::deriveAt(u,v,d,ders) ;
    return ;
  }

  if(nextLevel_){
    nextLevel_->deriveAt(u,v,d,ders,lod) ; 
    return ;
  }
  
  PlNurbsSurface::deriveAt(u,v,d,ders) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: deriveAt --- Finds the derivative of the point $(u,v)$
  Description: Computes the matrix of derivatives at $u,v$. 
                 The value of skl(k,l) represents the 
		 derivative of the surface $S(u,v)$ with respect to 
		 $u$ $k$ times and to $v$ $l$ times.
        Input: u --> the $u$ parametric value
	       v --> the $v$ parametric value
	    ders --> the Matrix of derivatives.
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::deriveAt(float u, float v, int d, Matrix_Point3D& ders, int lod) const {
  if(level_==lod){
    PlNurbsSurface::deriveAt(u,v,d,ders) ;
    return ;
  }

  if(nextLevel_){
    nextLevel_->deriveAt(u,v,d,ders,lod) ; 
    return ;
  }
  
  PlNurbsSurface::deriveAt(u,v,d,ders) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: ~PlHNurbsSurface --- destructor
  Description: Deletes all the levels.
        Input: 
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface::~PlHNurbsSurface(){
  if(nextLevel_)
    delete nextLevel_ ;
  lastLevel_ = 0 ;
  if(baseLevel_){
    baseLevel_->nextLevel_ = 0 ;
    baseLevel_->lastLevel_ = baseLevel_ ;
  }
  baseLevel_ = 0 ;
  nextLevel_ = 0 ;
  firstLevel_ = 0 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateLevels --- update the surface for all the levels
  Description: Update the surface for all the levels
        Input: upLevel --> updates the levels up to this level of detail
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::updateLevels(int upLevel){
  if(upLevel>=0){
    if(level_ <= upLevel){
      this->updateSurface() ;
    }
  }
  else{
    this->updateSurface() ;
  }

  if(upLevel>level || upLevel<0){
    if(nextLevel_)
      nextLevel_->updateLevels(upLevel) ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: splitUV --- insert n knots betwen each knots
  Description: Insert nu knots betwen each knots in the U vector and
               nv knots between each knots in the V vector. 

	       This doesn't not perform a split. It just generates
	       a suitable rU and rV vector. It is suggested that 
	       splitting should be done for the level above, not
	       the local level.

        Input: nu --> the number of new knots between each knots in U.
	       nv --> the number of new knots between each knots in V.
	       nU <-- the new refinement knot vector in U
	       nV <-- the new refinement knot vector in V
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::splitUV(int nu, int nv, PlVector<float> &nU, PlVector<float> &nV){

  nU.resize(KnotU.n*nu) ;
  nV.resize(KnotV.n*nv) ;
  
  int i,j,n ;

  n = 0 ; 
  for(i=1;i<KnotU.n;++i){
    if(KnotU[i] >KnotU[i-1]){
      float a = KnotU[i-1] ;
      float b = KnotU[i] ;


      for(j=0;j<nu;++j){
	nU[n] = a + (b-a)*float(j+1)/float(nu+1) ;
	++n ;
      }
    }
  }  
  nU.resize(n) ;

  n = 0 ;
  for(i=1;i<KnotV.n;++i){
    if(KnotV[i] > KnotV[i-1]){
      float a = KnotV[i-1] ;
      float b = KnotV[i] ;

      for(j=0;j<nv;++j){
	nV[n] = a + (b-a)*float(j+1)/float(nv+1) ;
	++n ;
      }
    }
  }  
  nV.resize(n) ;

}
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: splitUV --- insert n knots betwen each knots
  Description: Insert nu knots betwen each knots in the U vector and
               nv knots between each knots in the V vector. 

	       This doesn't not perform a split. It just generates
	       a suitable rU and rV vector. It is suggested that 
	       splitting should be done for the level above, not
	       the local level.

        Input: nu --> the number of new knots between each knots in U.
	       su --> the multiplicity of the each new knot in U
	       nv --> the number of new knots between each knots in V.
	       sv --> the multiplicity of the each new knot in V
	       nU <-- the new refinement knot vector in U
	       nV <-- the new refinement knot vector in V
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::splitUV(int nu, int su, int nv, int sv, PlVector<float> &nU, PlVector<float> &nV){

  int i,j,n ;

  if(su<=0)
    su = degU  ;
  if(sv<=0)
    sv = degV  ;
  if(su>degU+1)
    su = degU+1 ;
  if(sv>degV+1)
    sv = degV+1 ;

  nU.resize(KnotU.n*nu*su) ;
  nV.resize(KnotV.n*nv*sv) ;
  
  n = 0 ; 
  for(i=1;i<KnotU.n;++i){
    if(KnotU[i] >KnotU[i-1]){
      float a = KnotU[i-1] ;
      float b = KnotU[i] ;


      for(j=0;j<nu;++j){
	float u = a + (b-a)*float(j+1)/float(nu+1) ;
	for(int l=0;l<su;++l){
	  nU[n] = u ;
	  ++n ;
	}
      }
    }
  }  
  nU.resize(n) ;

  n = 0 ;
  for(i=1;i<KnotV.n;++i){
    if(KnotV[i] > KnotV[i-1]){
      float a = KnotV[i-1] ;
      float b = KnotV[i] ;

      for(j=0;j<nv;++j){
	float v = a + (b-a)*float(j+1)/float(nv+1) ;
	for(int l=0;l<sv;++l){
	  nV[n] = v ;
	  ++n ;
	}
      }
    }
  }  
  nV.resize(n) ;

}
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: addLevel --- adds a level to this HNURBS surface
  Description: Adds a level to this HNURBS surface
        Input: n --> the number of new knots between each knots.
       Output: a pointer to the new level or 0 if there was an error.
 Restrictions: returns 0, if there is already a nextlevel.
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface* PlHNurbsSurface::addLevel(int n) {
  PlHNurbsSurface *newLevel ;

  if(nextLevel_)
    return 0 ;

  Vector_FLOAT newU,newV ;
  
  splitUV(n,n,newU,newV) ;

  newLevel = new PlHNurbsSurface(this,newU,newV) ;

  return newLevel ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: addLevel --- adds a level to this HNURBS surface
  Description: Adds a level to this HNURBS surface
        Input: n --> the number of new knots between each knots.
       Output: a pointer to the new level or 0 if there was an error.
 Restrictions: returns 0, if there is already a nextlevel.
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurface* PlHNurbsSurface::addLevel() {
  PlHNurbsSurface *newLevel ;

  if(nextLevel_)
    return 0 ;

  newLevel = new PlHNurbsSurface(this) ;

  return newLevel ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: isoCurveU --- generates an iso curve in the U direction
  Description: Generates an iso-parametric curve which goes through the
               parametric value u along the U direction.
        Input: u --> the U parametric value
	       c <-- the iso-parametric curve
	     lod --> the level of detail to draw the curve with [-1]
       Output: 0 if an error occured, 1 otherwise.
 Restrictions: the parametric value $u$ must be in a valid range
   Created by: Philippe Lavoie          (7 October, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::isoCurveU(float u, PlNurbsCurve& c,int lod) const {
  if(lod>=0 && level_>lod)
    return 0;

  if(lod==level_ || lod<0){
    PlNurbsSurface::isoCurveU(u,c) ; 
    return 1 ;
  }
  
  if(nextLevel_){
    return nextLevel_->isoCurveU(u,c,lod) ;
  }

  return 0 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: isoCurveV --- generates an iso curve in the V direction
  Description: Generates an iso-parametric curve which goes through the
               parametric value v along the V direction.
        Input: v --> the V parametric value
	       c <-- the iso-parametric curve
	     lod --> the level of detail to draw the curve with [-1]
       Output: 0 if an error occured, 1 otherwise
 Restrictions: the parametric value $v$ must be in a valid range
   Created by: Philippe Lavoie          (7 October, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::isoCurveV(float v, PlNurbsCurve& c,int lod) const {
  if(lod>=0 && level_>lod)
    return 0;

  if(lod==level_ || lod<0){
    PlNurbsSurface::isoCurveV(v,c) ; 
    return 1 ;
  }
  
  if(nextLevel_){
    return nextLevel_->isoCurveV(v,c,lod) ;
  }

  return 0 ;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: read --- read a HNURBS surface from a file
  Description: Read a HNURBS surface from an input file stream.
        Input: fin --> the input file stream
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie        (7 October 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::read(ifstream &fin){
  if(!fin) {
    return 0 ;
  }
  int nu,nv,du,dv;
  char *type ;
  type = new char[4] ;
  if(!fin.read(type,sizeof(char)*4)) { delete []type ; return 0 ;}
  int r1 = strncmp(type,"hns3",4) ;
  int r2 = strncmp(type,"hns4",4) ;
  int r3 = strncmp(type,"hnso",4) ;
  if(!(r1 || r2 || r3)) 
    return 0 ;
  float *p,*p2 ;
  if(!r1 || !r2){
    if(!fin.read((char*)&nu,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&nv,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&du,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&dv,sizeof(int))) { delete []type ; return 0 ;}
    
    resize(nu,nv,du,dv) ;
    
    if(!fin.read((char*)U.memory(),sizeof(float)*U.n)) { delete []type ; return 0 ;}
    if(!fin.read((char*)V.memory(),sizeof(float)*V.n)) { delete []type ; return 0 ;}
    
    if(!r1){
      p = new float[3*nu*nv] ;
      if(!fin.read((char*)p,sizeof(float)*3*nu*nv)) { delete []type ; return 0 ;}
      p2 = p ;
      for(int i=0;i<nu;i++)
	for(int j=0;j<nv;j++){
	  P(i,j).x() = *(p++) ;
	  P(i,j).y() = *(p++) ;
	  P(i,j).z() = *(p++) ;
	  P(i,j).w() = 1.0 ;
	}
      delete []p2 ;
    }
    else{
      p = new float[4*nu*nv] ;
      if(!fin.read((char*)p,sizeof(float)*4*nu*nv)) { delete []type ; return 0 ;}
      p2 = p ;
      for(int i=0;i<nu;i++)
	for(int j=0;j<nv;j++){
	  P(i,j).x() = *(p++) ;
	  P(i,j).y() = *(p++) ;
	  P(i,j).z() = *(p++) ;
	  P(i,j).w() = *(p++) ;
	}
      delete []p2 ;
    }
    offset = P ;
    this->updateSurface() ;
  }
  else { // reading the offset information
    int ru,rv ;
    if(!fin.read((char*)&ru,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&rv,sizeof(int))) { delete []type ; return 0 ;}
    rU.resize(ru) ;
    rV.resize(rv) ;
    if(rU.n>0)
      if(!fin.read((char*)rU.memory(),sizeof(float)*rU.n)) { delete []type ; return 0 ;}
    if(rV.n>0)
      if(!fin.read((char*)rV.memory(),sizeof(float)*rV.n)) { delete []type ; return 0 ;}
    
    if(!fin.read((char*)&nu,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&nv,sizeof(int))) { delete []type ; return 0 ;}

    p = new float[4*nu*nv] ;
    if(!fin.read((char*)p,sizeof(float)*4*nu*nv)) { delete []type ; return 0 ;}
    p2 = p ;
    offset.resize(nu,nv) ;
    for(int i=0;i<nu;i++)
      for(int j=0;j<nv;j++){
	offset(i,j).x() = *(p++) ;
	offset(i,j).y() = *(p++) ;
	offset(i,j).z() = *(p++) ;
	offset(i,j).w() = *(p++) ;
      }
    delete []p2 ;    
    --baseUpdateN ;
    this->updateSurface() ;
  }

  // now we must read all the level information (if any)
  char *ptxt ;
  ptxt = new char[7] ;
  // the following line is used so that purify doesn't give a warning
  ptxt[0] = ptxt[1] = ptxt[2] = ptxt[3] = ptxt[4] = ptxt[5] = ptxt[6] = 0 ;
  int mark  = fin.tellg() ;
  if(!fin.read(ptxt,sizeof(char)*5))
    { delete []ptxt ; delete []type ; return 1 ; }
  char *pat ;
  pat = strstr(ptxt,"level") ;
  if(pat){
    // insert a new level
    PlHNurbsSurface *newPatch = this->addLevel() ;
    if(newPatch){
      if(!newPatch->read(fin)) 
	return 0 ;      
    }
    else
      return 0 ;
  }
  else{ 
    // something else is after the HNurbs
    // restore were the file was before the read
    fin.seekg(mark);
  }

  delete []ptxt ;
  delete []type ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: read --- read a HNURBS surface from a file
  Description: Read a HNURBS surface from a file.
        Input: filename --> the filename to read from 
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie   (7 October 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::read(const char* filename){
  ifstream fin(filename) ;
  if(!fin) {
    return 0 ;
  }

  return this->read(fin) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: write --- write a HNURBS surface to a file stream
  Description: Write a HNURBS surface to an output file stream.
        Input: fout <-- the output filestream to write to.
       Output: 1 on success, 0 on failure
 Restrictions: 
   Created by: Philippe Lavoie     (7 October 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::write(ofstream &fout) const {
  if(!fout)
    return 0 ;
  if(!baseLevel_){
    if(!fout.write((char*)&"hns4",sizeof(char)*4)) return 0 ;
    if(!fout.write((char*)&P.rows,sizeof(int))) return 0 ;
    if(!fout.write((char*)&P.cols,sizeof(int))) return 0 ;
    if(!fout.write((char*)&degU,sizeof(int))) return 0 ;
    if(!fout.write((char*)&degV,sizeof(int))) return 0 ;
    if(!fout.write((char*)U.memory(),sizeof(float)*U.n)) return 0 ;
    if(!fout.write((char*)V.memory(),sizeof(float)*V.n)) return 0 ;
    
    float *p,*p2 ;
    p = new float[P.rows*P.cols*4] ;
    p2 = p ;
    for(int i=0;i<P.rows;i++) 
      for(int j=0;j<P.cols;j++){
	*p = offset(i,j).x() ; p++ ;
	*p = offset(i,j).y() ; p++ ;
	*p = offset(i,j).z() ; p++ ;
	*p = offset(i,j).w() ; p++ ;
      }
    if(!fout.write((char*)p2,sizeof(float)*P.rows*P.cols*4)) return 0 ;
    delete []p2 ;
  }
  else{
    if(!fout.write((char*)&"hnso",sizeof(char)*4)) return 0 ;
    if(!fout.write((char*)&rU.n,sizeof(int))) return 0 ;
    if(!fout.write((char*)&rV.n,sizeof(int))) return 0 ;
    if(rU.n>0)
      if(!fout.write((char*)rU.memory(),sizeof(float)*rU.n)) return 0 ;
    if(rV.n>0)
      if(!fout.write((char*)rV.memory(),sizeof(float)*rV.n)) return 0 ;
    if(!fout.write((char*)&offset.rows,sizeof(int))) return 0 ;
    if(!fout.write((char*)&offset.cols,sizeof(int))) return 0 ;
    float *p,*p2 ;

    p = new float[offset.rows*offset.cols*4] ;
    p2 = p ;
    for(int i=0;i<offset.rows;i++) 
      for(int j=0;j<offset.cols;j++){
	*p = offset(i,j).x() ; p++ ;
	*p = offset(i,j).y() ; p++ ;
	*p = offset(i,j).z() ; p++ ;
	*p = offset(i,j).w() ; p++ ;
      }
    if(!fout.write((char*)p2,sizeof(float)*offset.rows*offset.cols*4)) 
      return 0 ;
    delete []p2 ;    
  }
  if(nextLevel_){
    if(!fout.write((char*)&"level",sizeof(char)*5)) return 0 ;
    if(!nextLevel_->write(fout)) return 0 ;
  }
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: write --- write a HNURBS surface to a file
  Description: Write a surface to a file.
        Input: filename <-- the filename to write to
       Output: 1 on success, 0 on failure
 Restrictions: 
   Created by: Philippe Lavoie    (7 October 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::write(const char* filename) const {
  ofstream fout(filename) ;    
  if(!fout)
    return 0 ;
  return write(fout);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: toNurbs --- generates a NURBS surface
  Description: Generates a NURBS surface.
        Input: surf <-- the NURBS surface corresponding to the HNURBS surface
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie    (7 October 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
//void PlHNurbsSurface::toNurbs(PlNurbsSurface& surf) const {
  
//}

inline void mergeInto(const Vector_FLOAT &a, Vector_FLOAT &b) {
  // merge a in b ;
  int n = b.n ;
  b.resize(b.n+a.n) ;  
  for(int i=0;i<a.n;++i)
    b[n+i] = a[i] ;
  b.qSort() ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: refineKnots --- refine both knot vectors
  Description: Refine both knot vectors.
        Input: nU --> the U knot vector to refine from 
               nV --> the V knot vector to refine from 
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV){
  refineKnotU(nU) ;
  mergeInto(nU,rU) ;

  initBase(1) ;

  refineKnotV(nV) ;  
  mergeInto(nV,rV) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: refineKnotU ---  refines the U knot vector
  Description: Refines the U knot vector.
        Input: X --> the knot vector to refine from
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::refineKnotU(const Vector_FLOAT& X) {
  updateSurface() ;
  Vector_FLOAT Xu(X) ;
  int i,j ;
  j = 0 ;
  for(i=0;i<X.n;++i){
    if(X[i]>=U[0] && X[i]<=U[U.n-1]){
      Xu[j] = X[i] ;
      ++j ;
    }
  }
  Xu.resize(j) ;

  if(Xu.n>0){
    if(nextLevel_){
      nextLevel_->refineKnotU(Xu) ;
    }
    
    PlNurbsSurface osurf(degU,degV,U,V,offset) ;
    
    osurf.refineKnotU(Xu) ;
    
    offset.resize(osurf.CtrlP.rows,osurf.CtrlP.cols) ;
    for(i=0;i<offset.rows;++i)
      for(j=0;j<offset.cols;++j)
	offset(i,j) = osurf.CtrlP(i,j) ;

    if(!baseLevel_){
      PlNurbsSurface::refineKnotU(Xu) ;
    }
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: refineKnotV ---  refines the V knot vector
  Description: Refines the V knot vector.
        Input: X --> the knot vector to refine from
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurface::refineKnotV(const Vector_FLOAT& X) {
  updateSurface() ;
  Vector_FLOAT Xv(X) ;
  int i,j ;
  j = 0 ;
  for(i=0;i<X.n;++i){
    if(X[i]>=V[0] && X[i]<=V[V.n-1]){
      Xv[j] = X[i] ;
      ++j ;
    }
  }
  Xv.resize(j) ;

  if(Xv.n>0){
    if(nextLevel_){
      nextLevel_->refineKnotV(Xv) ;
    }
    
    PlNurbsSurface osurf(degU,degV,U,V,offset) ;
    
    osurf.refineKnotV(Xv) ;
    
    offset.resize(osurf.CtrlP.rows,osurf.CtrlP.cols) ;
    for(i=0;i<offset.rows;++i)
      for(j=0;j<offset.cols;++j)
	offset(i,j) = osurf.CtrlP(i,j) ;
    if(!baseLevel_){
      PlNurbsSurface::refineKnotV(Xv) ;
    }
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: movePointOffset ---  move a point on the surface
  Description: This moves the point $s(u,v)$ by delta. As this is a HNURBS
               surface. It moves the offset surface by delta, it doesn't move 
	       the surface point per say.
        Input: u --> the parameter in the u direction
	       v --> the parameter in the v direction
	   delta --> the displacement of the point $s(u,v)$
       Output: 1 if the operation was succesfull, 0 otherwise
 Restrictions: u and v must be in a valid range.
   Created by: Philippe Lavoie          (3 June 1998)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurface::movePointOffset(float u, float v, const Point3D& delta){
  P = offset ; 

  // by definition the offset has w = 0 , but this isn't valid for
  // the control points, increasing the w by 1, will generate a valid surface
  for(int i=0;i<P.rows;++i)
    for(int j=0;j<P.cols;++j){
      P(i,j).w() += float(1) ; 
    }
  if(PlNurbsSurface::movePoint(u,v,delta)){
    offset = P ;
    P = baseSurf.CtrlP ; 
    updateSurface() ; 
    return 1 ;
  }
  updateSurface() ; 
  return 0 ;
}
