/*=====================================================================
        File: hnurbsS_sp.cc
     Purpose:       
    Revision: $Id: hnurbsS_sp.cc,v 2.2 1998/06/01 19:11:49 lavoie Exp $
  Created by: Philippe Lavoie          (14 May, 1998)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=====================================================================*/
#include <hnurbsS_sp.hh>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateMaxU --- updates the basis value
  Description: Updates the basis value at which a control point has 
               maximal influence. It also finds where the control point
	       has maximal influence. 
        Input:
       Output:
 Restrictions: The degrre in U of the surface must be of 3 or less.
   Created by: Philippe Lavoie (14 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurfaceSP::updateMaxU() {
  if(degU>3){
    PlError error("PlHNurbsSurfaceSP::updateMaxU()") ;
    error << "This class doesn't support surfaces having a degree of 4 and higher.\n" ;
    error.fatal() ;
  }
  else{
    maxU.resize(P.rows) ;
    maxAtU_.resize(P.rows) ;
    for(int i=0;i<P.rows;++i){
      if(!maxInfluence(i,U,degU,maxAtU_[i]))
	cerr << "Problem in maxInfluence U!\n" ;
      maxU[i] = nurbsBasisFun(maxAtU_[i],i,degU,U) ;
    }
    
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateMaxV --- updates the basis value
  Description: Updates the basis value at which a control point has 
               maximal influence. It also finds where the control point
	       has maximal influence. 
        Input:
       Output:
 Restrictions: The degrre in V of the surface must be of 3 or less.
   Created by: Philippe Lavoie (14 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurfaceSP::updateMaxV() {
  if(degV>3){
    PlError error("PlHNurbsSurfaceSP::updateMaxV()") ;
    error << "This class doesn't support surfaces having a degree of 4 and higher.\n" ;
    error.fatal() ;
  }
  else{
    maxV.resize(P.cols) ;
    maxAtV_.resize(P.cols) ;
    for(int i=0;i<P.cols;++i){
      if(!maxInfluence(i,V,degV,maxAtV_[i]))
	cerr << "Problem in maxInfluence V!\n" ;
      maxV[i] = nurbsBasisFun(maxAtV_[i],i,degV,V) ;
    }
    
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: modCPby --- modifies the surface point by
  Description: Modifies the surface point by a certain value.
        Input: i --> the row of the surface point
	       j --> the column of the surface point
	       a --> modify the surface point by this value
       Output:
 Restrictions: The degree in U and V of the surface must be of 3 or less.
   Created by: Philippe Lavoie (14 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurfaceSP::modSurfCPby(int i, int j, const CtrlPnt4D& a) {
  offset(i,j) +=  a / (maxU[i]*maxV[j]) ; 
  if(baseLevel_){
    Point3D vecOffset ; 
    vecOffset = offset(i,j).x()*ivec(i,j) +
      offset(i,j).y()*jvec(i,j) +
      offset(i,j).z()*kvec(i,j) ;
    P(i,j).x() = baseSurf.CtrlP(i,j).x()+vecOffset.x ;
    P(i,j).y() = baseSurf.CtrlP(i,j).y()+vecOffset.y ;
    P(i,j).z() = baseSurf.CtrlP(i,j).z()+vecOffset.z ;
  }
  else
    P(i,j) = offset(i,j) ; 
}

/*
  not defined for now. I'm not sure we need it/want it.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: modCP --- modifies the surface point to a value
  Description: Modifies the surface point to a certain value.
        Input: i --> the row of the surface point
	       j --> the column of the surface point
	       a --> modify the surface point to this value
       Output:
 Restrictions: The degree in U and V of the surface must be of 3 or less.
   Created by: Philippe Lavoie (14 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*void PlHNurbsSurfaceSP::modSurfCP(int i, int j, const CtrlPnt4D& a) {
  modSurfCPby(i,j,a-surfP(i,j)) ;
}
*/

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: addLevel --- adds a level to this HNURBS surface
  Description: Adds a level to this HNURBS surface
        Input: n --> the number of new knots between each knots.
	       s --> the multiplicity of each of these knots
       Output: a pointer to the new level or 0 if there was an error.
 Restrictions: returns 0, if there is already a nextlevel.
   Created by: Philippe Lavoie (14 May 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurfaceSP* PlHNurbsSurfaceSP::addLevel(int n, int s) {
  PlHNurbsSurfaceSP *newLevel ;

  if(nextLevel_)
    return 0 ;

  Vector_FLOAT newU,newV ;
  
  splitUV(n,s,n,s,newU,newV) ;

  newLevel = new PlHNurbsSurfaceSP(this,newU,newV) ;

  return newLevel ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: addLevel --- adds a level to this HNURBS surface
  Description: Adds a level to this HNURBS surface
        Input: n --> the number of new knots between each knots.
       Output: a pointer to the new level or 0 if there was an error.
 Restrictions: returns 0, if there is already a nextlevel.
   Created by: Philippe Lavoie (14 May 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlHNurbsSurfaceSP* PlHNurbsSurfaceSP::addLevel() {
  PlHNurbsSurfaceSP *newLevel ;

  if(nextLevel_)
    return 0 ;

  newLevel = new PlHNurbsSurfaceSP(this) ;

  return newLevel ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: copy --- copies a HNurbs Surface and all it children
  Description: Copies a HNurbs Surface and all it children
        Input: ns --> the HNurbs surface to copy
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (14 May 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurfaceSP::copy(const PlHNurbsSurface& nS){
  PlHNurbsSurface *levelP ;
  levelP = nS.nextLevel() ;

  PlNurbsSurface::operator=(nS) ;
  rU = nS.rU ;
  rV = nS.rV ;
  offset = nS.offset ;

  updateMaxUV() ; 

  firstLevel_ = this ;

  if(levelP){
    PlHNurbsSurfaceSP *newLevel ;
    newLevel =  new PlHNurbsSurfaceSP(this) ; 
    newLevel->copy(*levelP) ;
    nextLevel_ = newLevel ;
    lastLevel_ = nextLevel_->lastLevel() ;
  }
  else{
    lastLevel_ = this ;
  }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateSurface --- updates the NURBS surface
  Description: Updates the NURBS surface according to the offset values
               and its base level. You can update only one control point 
	       from the surface if you specify a value for i and j or you 
	       can update all the points if i0 or j0 is below 0.
        Input: i0 --> the row of the control point to update [-1]
	       j0 --> the column of the control point to update [-1]
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by: Philippe Lavoie (15 April 1998)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurfaceSP::updateSurface(int i0, int j0){
  if(i0>=0 && j0>=0){
    if(offset(i0,j0).x()==0.0 && offset(i0,j0).y()==0.0 && offset(i0,j0).z()==0.0)
      return ;
  }
  if(baseLevel_){
    if(initBase()){
      P = baseSurf.CtrlP ;
      U = baseSurf.KnotU ;
      V = baseSurf.KnotV ;
      degU = baseSurf.degreeU ;
      degV = baseSurf.degreeV ;
      updateMaxUV() ; 
    }
    if(i0>=0 && j0>=0){
      Point3D vecOffset ;
      vecOffset = offset(i0,j0).x()*ivec(i0,j0) +
	offset(i0,j0).y()*jvec(i0,j0) +
	offset(i0,j0).z()*kvec(i0,j0) ;
      P(i0,j0).x() = baseSurf.CtrlP(i0,j0).x()+vecOffset.x ;
      P(i0,j0).y() = baseSurf.CtrlP(i0,j0).y()+vecOffset.y ;
      P(i0,j0).z() = baseSurf.CtrlP(i0,j0).z()+vecOffset.z ;
    }
    else{
      for(int i=0;i<P.rows;++i)
	for(int j=0;j<P.cols;++j){
	  if(offset(i,j).x() != 0 || 
	     offset(i,j).y() != 0 || offset(i,j).z() || 0){
	    Point3D vecOffset ;
	    vecOffset = offset(i,j).x()*ivec(i,j) +
	      offset(i,j).y()*jvec(i,j) +
	      offset(i,j).z()*kvec(i,j) ;
	    P(i,j).x() = baseSurf.CtrlP(i,j).x()+vecOffset.x ;
	    P(i,j).y() = baseSurf.CtrlP(i,j).y()+vecOffset.y ;
	    P(i,j).z() = baseSurf.CtrlP(i,j).z()+vecOffset.z ;
	  }
	}
    }
  }
  else{
    if(i0>=0 && j0>=0)
      P(i0,j0) = offset(i0,j0) ;
    else{
      for(int i=0;i<P.rows;++i)
	for(int j=0;j<P.cols;++j){
	  P(i,j) = offset(i,j) ;
	}
    }
  }

  ++updateN ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateLevels --- update the surface for all the levels
  Description: Update the surface for all the levels
        Input: upLevel --> updates the levels up to this level of detail
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (7 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlHNurbsSurfaceSP::updateLevels(int upLevel){
  if(!okMax())
    updateMaxUV() ; 
  if(upLevel>=0){
    if(level<=upLevel){
      this->updateSurface() ;
    }
  }
  else{
    this->updateSurface() ;
  }

  if(upLevel>level || upLevel<0){
    if(nextLevel_)
      ((PlHNurbsSurfaceSP*)nextLevel_)->updateLevels(upLevel) ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: read --- read a HNURBS surface from a file
  Description: Read a HNURBS surface from an input file stream.
        Input: fin --> the input file stream
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie        (7 October 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlHNurbsSurfaceSP::read(ifstream &fin){
  if(!fin) {
    return 0 ;
  }
  int nu,nv,du,dv;
  char *type ;
  type = new char[4] ;
  if(!fin.read(type,sizeof(char)*4)) { delete []type ; return 0 ;}
  int r1 = strncmp(type,"hns3",4) ;
  int r2 = strncmp(type,"hns4",4) ;
  int r3 = strncmp(type,"hnso",4) ;
  if(!(r1 || r2 || r3)) 
    return 0 ;
  float *p,*p2 ;
  if(!r1 || !r2){
    if(!fin.read((char*)&nu,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&nv,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&du,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&dv,sizeof(int))) { delete []type ; return 0 ;}
    
    resize(nu,nv,du,dv) ;
    
    if(!fin.read((char*)U.memory(),sizeof(float)*U.n)) { delete []type ; return 0 ;}
    if(!fin.read((char*)V.memory(),sizeof(float)*V.n)) { delete []type ; return 0 ;}
    
    if(!r1){
      p = new float[3*nu*nv] ;
      if(!fin.read((char*)p,sizeof(float)*3*nu*nv)) { delete []type ; return 0 ;}
      p2 = p ;
      for(int i=0;i<nu;i++)
	for(int j=0;j<nv;j++){
	  P(i,j).x() = *(p++) ;
	  P(i,j).y() = *(p++) ;
	  P(i,j).z() = *(p++) ;
	  P(i,j).w() = 1.0 ;
	}
      delete []p2 ;
    }
    else{
      p = new float[4*nu*nv] ;
      if(!fin.read((char*)p,sizeof(float)*4*nu*nv)) { delete []type ; return 0 ;}
      p2 = p ;
      for(int i=0;i<nu;i++)
	for(int j=0;j<nv;j++){
	  P(i,j).x() = *(p++) ;
	  P(i,j).y() = *(p++) ;
	  P(i,j).z() = *(p++) ;
	  P(i,j).w() = *(p++) ;
	}
      delete []p2 ;
    }
    offset = P ;
    this->updateSurface() ;
  }
  else { // reading the offset information
    int ru,rv ;
    if(!fin.read((char*)&ru,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&rv,sizeof(int))) { delete []type ; return 0 ;}
    rU.resize(ru) ;
    rV.resize(rv) ;
    if(rU.n>0)
      if(!fin.read((char*)rU.memory(),sizeof(float)*rU.n)) { delete []type ; return 0 ;}
    if(rV.n>0)
      if(!fin.read((char*)rV.memory(),sizeof(float)*rV.n)) { delete []type ; return 0 ;}
    
    if(!fin.read((char*)&nu,sizeof(int))) { delete []type ; return 0 ;}
    if(!fin.read((char*)&nv,sizeof(int))) { delete []type ; return 0 ;}

    p = new float[4*nu*nv] ;
    if(!fin.read((char*)p,sizeof(float)*4*nu*nv)) { delete []type ; return 0 ;}
    p2 = p ;
    offset.resize(nu,nv) ;
    for(int i=0;i<nu;i++)
      for(int j=0;j<nv;j++){
	offset(i,j).x() = *(p++) ;
	offset(i,j).y() = *(p++) ;
	offset(i,j).z() = *(p++) ;
	offset(i,j).w() = *(p++) ;
      }
    delete []p2 ;    
    --baseUpdateN ;
    this->updateSurface() ;
  }

  // now we must read all the level information (if any)
  char *ptxt ;
  ptxt = new char[7] ;
  // the following line is used so that purify doesn't give a warning
  ptxt[0] = ptxt[1] = ptxt[2] = ptxt[3] = ptxt[4] = ptxt[5] = ptxt[6] = 0 ;
  int mark  = fin.tellg() ;
  if(!fin.read(ptxt,sizeof(char)*5))
    { delete []ptxt ; delete []type ; return 1 ; }
  char *pat ;
  pat = strstr(ptxt,"level") ;
  if(pat){
    // insert a new level
    PlHNurbsSurfaceSP *newPatch = new PlHNurbsSurfaceSP(this) ; 
    if(newPatch){
      if(!newPatch->read(fin)) 
	return 0 ;      
    }
    else
      return 0 ;
  }
  else{ 
    // something else is after the HNurbs
    // restore were the file was before the read
    fin.seekg(mark);
  }

  delete []ptxt ;
  delete []type ;
  return 1 ;
}

