/*=============================================================================
        File: matrixRT.cc
     Purpose:       
    Revision: $Id: matrixRT.cc,v 2.0 1998/04/15 18:11:16 lavoie Exp $
  Created by: Philippe Lavoie          (25 July, 1997)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/

#include <matrixRT.hh>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: MatrixRT ---  constructor with the angles and translation 
       			     parameters specified
  Description: Sets the matrixRT to be a rotation of ax,ay,az then a 
               translation x,y,z around the origin.

        Input: ax --> specifies the rotation around the $x$-axis
	       ay --> specifies the rotation around the $y$-axis
	       az --> specifies the rotation around the $z$-axis
	        x --> specifies the translation along the $x$-axis
	        y --> specifies the translation along the $y$-axis
	        z --> specifies the translation along the $z$-axis
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>::MatrixRT(T ax, T ay, T az, T x, T y, T z): PlMatrix<T>(4,4) {
  // the following is the same as
  // *this = C.translate(x,y,z)*B.rotate(ax,ay,az) ;
  rotate(ax,ay,az) ;
#ifdef COLUMN_ORDER
  m[12] = x ;
  m[13] = y ;
  m[14] = z ;  
#else
  m[3] = x ;
  m[7] = y ;
  m[11] = z ;
#endif
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: MatrixRT ---  default constructor
  Description: Default constructor
        Input:     
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>::MatrixRT() : PlMatrix<T>(4,4) {
  reset(0) ;
  diag(1.0) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: MatrixRT ---  default constructor
  Description: Default constructor
        Input:     
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>::MatrixRT(T* p) : PlMatrix<T>(p,4,4) {
  // nothing to do, p should have all the proper data.
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: rotate -- creates a rotation matrix of angle (ax,ay,az)
  Description: The rotation are clockwise around the main axes. The rotation
               is performed in that order: a rotation around the $z$-axis, 
	       then the $y$-axis and finally around the $x$-axis.
        Input: ax --> specifies the rotation around the $x$-axis
	       ay --> specifies the rotation around the $y$-axis
	       az --> specifies the rotation around the $z$-axis
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>& MatrixRT<T>::rotate(T ax,T ay, T az){
  T t1,t2,t4,t6,t7,t8,t10,t13 ;
  t1 = cos(az);
  t2 = cos(ay);
  t4 = sin(az);
  t6 = sin(ay);
  t7 = t1*t6;
  t8 = sin(ax);
  t10 = cos(ax);
  t13 = t4*t6;
#ifdef COLUMN_ORDER
  m[0] = t1*t2;
  m[4] = -t4*t2;
  m[8] = t6;
  m[12] = 0 ;
  m[1] = t7*t8+t4*t10;
  m[5] = -t13*t8+t1*t10;
  m[9] = -t2*t8;
  m[13] = 0 ;
  m[2] = -t7*t10+t4*t8;
  m[6] = t13*t10+t1*t8;
  m[10] = t2*t10;
  m[14] = m[3] = m[7] = m[11] = 0.0 ;
  m[15] = 1.0 ;
#else
  m[0] = t1*t2;
  m[1] = -t4*t2;
  m[2] = t6;
  m[3] = 0 ;
  m[4] = t7*t8+t4*t10;
  m[5] = -t13*t8+t1*t10;
  m[6] = -t2*t8;
  m[7] = 0 ;
  m[8] = -t7*t10+t4*t8;
  m[9] = t13*t10+t1*t8;
  m[10] = t2*t10;
  m[11] = m[12] = m[13] = m[14] = 0 ;
  m[15] = 1.0 ;
#endif
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: rotateXYZ -- creates a rotation matrix of angle (ax,ay,az)
  Description: The rotation are clockwise around the main axes. The rotation
               is performed in that order: a rotation around the $x$-axis, 
	       then the $y$-axis and finally around the $z$-axis.
        Input: ax --> specifies the rotation around the $x$-axis
	       ay --> specifies the rotation around the $y$-axis
	       az --> specifies the rotation around the $z$-axis
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>& MatrixRT<T>::rotateXYZ(T ax,T ay, T az){
  T t1,t2,t4,t5,t7,t8,t9,t17 ;
  t1 = (T)cos((double)az);
  t2 = (T)cos((double)ay);
  t4 = (T)sin((double)az);
  t5 = (T)cos((double)ax);
  t7 = (T)sin((double)ay);
  t8 = t1*t7;
  t9 = (T)sin((double)ax);
  t17 = t4*t7;
#ifdef COLUMN_ORDER
  m[0] = t1*t2;
  m[4] = -t4*t5+t8*t9;
  m[8] = t4*t9+t8*t5;
  m[12] = 0.0 ;
  m[1] = t4*t2;
  m[5] = t1*t5+t17*t9;
  m[9] = -t1*t9+t17*t5;
  m[13] = 0.0 ;
  m[2] = -t7;
  m[6] = t2*t9;
  m[10] = t2*t5;
  m[14] = m[3] = m[7] = m[11] = 0 ;
  m[15] = 1.0 ;
#else
  m[0] = t1*t2;
  m[1] = -t4*t5+t8*t9;
  m[2] = t4*t9+t8*t5;
  m[3] = 0.0 ;
  m[4] = t4*t2;
  m[5] = t1*t5+t17*t9;
  m[6] = -t1*t9+t17*t5;
  m[7] = 0.0 ;
  m[8] = -t7;
  m[9] = t2*t9;
  m[10] = t2*t5;
  m[11] = m[12] = m[13] = m[14] = 0 ;
  m[15] = 1.0 ;
#endif
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: translate -- generates a translation matrix
  Description: Generates a translation matrix 
        Input:  x --> specifies the translation along the $x$-axis
	        y --> specifies the translation along the $y$-axis
	        z --> specifies the translation along the $z$-axis
       Output:
 Restrictions: This resets the rotation part of the matrix
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>& MatrixRT<T>::translate(T x, T y, T z){
  reset(0) ;
  diag(1.0) ;
#ifdef COLUMN_ORDER
  m[12] = x ;
  m[13] = y ;
  m[14] = z ;
#else
  m[3] = x ;
  m[7] = y ;
  m[11] = z ;
#endif
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: scale -- generates a scaling matrix
  Description: Generates a scaling matrix  
        Input:  x --> specifies the scaling along the $x$-axis
	        y --> specifies the scaling along the $y$-axis
	        z --> specifies the scaling along the $z$-axis
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (22 August, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>& MatrixRT<T>::scale(T x, T y, T z){
  reset(0) ;
  m[0] = x ;
  m[5] = y ;
  m[10] = z ;
  m[15] = 1.0 ;
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* -- multiplies a matrixRT with a CtrlPnt4D
  Description: Multiplies a matrixRT with a CtrlPnt4D.
        Input: M -- the matrix
               P -- the control point
       Output: the output control point rotated and trsnalted by M
 Restrictions: 
   Created by:    Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
CtrlPnt4D operator*(const MatrixRT<T>& M, const CtrlPnt4D& P){
  CtrlPnt4D P2 ;

  P2.x() = M(0,0)*(T)P.x() + M(0,1)*(T)P.y() + M(0,2)*(T)P.z() + M(0,3)*(T)P.w() ;
  P2.y() = M(1,0)*(T)P.x() + M(1,1)*(T)P.y() + M(1,2)*(T)P.z() + M(1,3)*(T)P.w() ;
  P2.z() = M(2,0)*(T)P.x() + M(2,1)*(T)P.y() + M(2,2)*(T)P.z() + M(2,3)*(T)P.w() ;
  P2.w() = M(3,0)*(T)P.x() + M(3,1)*(T)P.y() + M(3,2)*(T)P.z() + M(3,3)*(T)P.w() ;

  return P2 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* -- multiplies a matrixRT with a point3D
  Description: Multiplies a matrixRT with a point3D.
        Input: M -- the matrix
               P -- the 3D point
       Output: the output 3D point rotatated and translated by M
 Restrictions: 
   Created by:    Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
Point3D operator*(const MatrixRT<T>& M, const Point3D& P){
  Point3D P2 ;

  P2.x = M(0,0)*(T)P.x + M(0,1)*(T)P.y + M(0,2)*(T)P.z + M(0,3) ;
  P2.y = M(1,0)*(T)P.x + M(1,1)*(T)P.y + M(1,2)*(T)P.z + M(1,3) ;
  P2.z = M(2,0)*(T)P.x + M(2,1)*(T)P.y + M(2,2)*(T)P.z + M(2,3) ;

  return P2 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: operator* -- multiplies a matrixRT with another matrixRT
  Description: Multiplies a matrixRT with a MatrixRT.
        Input: m1 -- the first matrix
               m2 -- the second matrix
       Output: a new matrixRT
 Restrictions: Be aware of the way C++ handles function calls. If you want
               to generate a matrixRT from different ones, you can't do
	       {\tt A = A.translate(x,y,z)*A.rotate(ax,ay,az)}, since the
	       translate and rotate return *this, when reaching the {\tt *} 
	       operator, the function will be called with the same argument
	       on its left and right. Instead, use something like this:
	       {\tt A = A.translate(x,y,z)*B.rotate(ax,ay,az)}.
   Created by: Philippe Lavoie   (22 August, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T> operator*(const MatrixRT<T>& M1, const MatrixRT<T>& M2){
  MatrixRT<T> M ;
  T *m1,*m2,*m ;
  m1 = M1.m ;
  m2 = M2.m ;
  m = M.m ;
#ifdef COLUMN_ORDER
  m[0] = m1[0]*m2[0] + m1[4]*m2[1] + m1[8]*m2[2] + m1[12]*m2[3] ;
  m[4] = m1[0]*m2[4] + m1[4]*m2[5] + m1[8]*m2[6] + m1[12]*m2[7] ;
  m[8] = m1[0]*m2[8] + m1[4]*m2[9] + m1[8]*m2[10] + m1[12]*m2[11] ;
  m[12] = m1[0]*m2[12] + m1[4]*m2[13] + m1[8]*m2[14] + m1[12]*m2[15] ;
  
  m[1] = m1[1]*m2[0] + m1[5]*m2[1] + m1[9]*m2[2] + m1[13]*m2[3] ;
  m[5] = m1[1]*m2[4] + m1[5]*m2[5] + m1[9]*m2[6] + m1[13]*m2[7] ;
  m[9] = m1[1]*m2[8] + m1[5]*m2[9] + m1[9]*m2[10] + m1[13]*m2[11] ;
  m[13] = m1[1]*m2[12] + m1[5]*m2[13] + m1[9]*m2[14] + m1[13]*m2[15] ;
  
  m[2] = m1[2]*m2[0] + m1[6]*m2[1] + m1[10]*m2[2] + m1[14]*m2[3] ;
  m[6] = m1[2]*m2[4] + m1[6]*m2[5] + m1[10]*m2[6] + m1[14]*m2[7] ;
  m[10] = m1[2]*m2[8] + m1[6]*m2[9] + m1[10]*m2[10] + m1[14]*m2[11] ;
  m[14] = m1[2]*m2[12] + m1[6]*m2[13] + m1[10]*m2[14] + m1[14]*m2[15] ;
  
  m[3] = m1[3]*m2[0] + m1[7]*m2[1] + m1[11]*m2[2] + m1[15]*m2[3] ;
  m[7] = m1[3]*m2[4] + m1[7]*m2[5] + m1[11]*m2[6] + m1[15]*m2[7] ;
  m[11] = m1[3]*m2[8] + m1[7]*m2[9] + m1[11]*m2[10] + m1[15]*m2[11] ;
  m[15] = m1[3]*m2[12] + m1[7]*m2[13] + m1[11]*m2[14] + m1[15]*m2[15] ;
  
#else
  m[0] = m1[0]*m2[0] + m1[1]*m2[4] + m1[2]*m2[8] + m1[3]*m2[12] ;
  m[1] = m1[0]*m2[1] + m1[1]*m2[5] + m1[2]*m2[9] + m1[3]*m2[13] ;
  m[2] = m1[0]*m2[2] + m1[1]*m2[6] + m1[2]*m2[10] + m1[3]*m2[14] ;
  m[3] = m1[0]*m2[3] + m1[1]*m2[7] + m1[2]*m2[11] + m1[3]*m2[15] ;
  
  m[4] = m1[4]*m2[0] + m1[5]*m2[4] + m1[6]*m2[8] + m1[7]*m2[12] ;
  m[5] = m1[4]*m2[1] + m1[5]*m2[5] + m1[6]*m2[9] + m1[7]*m2[13] ;
  m[6] = m1[4]*m2[2] + m1[5]*m2[6] + m1[6]*m2[10] + m1[7]*m2[14] ;
  m[7] = m1[4]*m2[3] + m1[5]*m2[7] + m1[6]*m2[11] + m1[7]*m2[15] ;
  
  m[8] = m1[8]*m2[0] + m1[9]*m2[4] + m1[10]*m2[8] + m1[11]*m2[12] ;
  m[9] = m1[8]*m2[1] + m1[9]*m2[5] + m1[10]*m2[9] + m1[11]*m2[13] ;
  m[10] = m1[8]*m2[2] + m1[9]*m2[6] + m1[10]*m2[10] + m1[11]*m2[14] ;
  m[11] = m1[8]*m2[3] + m1[9]*m2[7] + m1[10]*m2[11] + m1[11]*m2[15] ;
  
  m[12] = m1[12]*m2[0] + m1[13]*m2[4] + m1[14]*m2[8] + m1[15]*m2[12] ;
  m[13] = m1[12]*m2[1] + m1[13]*m2[5] + m1[14]*m2[9] + m1[15]*m2[13] ;
  m[14] = m1[12]*m2[2] + m1[13]*m2[6] + m1[14]*m2[10] + m1[15]*m2[14] ;
  m[15] = m1[12]*m2[3] + m1[13]*m2[7] + m1[14]*m2[11] + m1[15]*m2[15] ;
#endif
  return M ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator= --- the assignment operator with a matrix
  Description: The assignment operator with a matrix
        Input: M -- the matrix
       Output: the output 3D point rotatated and translated by M
 Restrictions: The matrix {\em must} be of size $4 \times 4$
   Created by:    Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>& MatrixRT<T>::operator=(const PlMatrix<T>& M) {
  if(M.rows != 4 || M.cols != 4){
    PlError error("MatrixRT<T>::operator=") ;
    error << "Trying to assign with a matrix of dimensions" << 
      M.rows << " " << M.cols << endl ;
    error.fatal() ;
  }
  T *a,*b ;
  a = m-1 ;
  b = M[0] - 1 ;
  for(int i=0;i<16;++i){
    *(++a) = *(++b) ;
  }
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator= --- the assignment operator with a matrixRT
  Description: The assignment operator with a matrixRT.
        Input: M -- the matrix
       Output: the output 3D point rotatated and translated by M
 Restrictions: The matrix {\em must} be of size $4 \times 4$
   Created by:    Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
template <class T>
MatrixRT<T>& MatrixRT<T>::operator=(const MatrixRT<T>& M) {
  T *a,*b ;
  a = m-1 ;
  b = M.m - 1 ;
  for(int i=0;i<16;++i){
    *(++a) = *(++b) ;
  }
  return *this ;
}



#ifdef __GNUG__

template class MatrixRT<float> ;
template class MatrixRT<double> ;

template Point3D operator*(const MatrixRT<float>& M, const Point3D& P) ;
template Point3D operator*(const MatrixRT<double>& M, const Point3D& P) ;
template CtrlPnt4D operator*(const MatrixRT<float>& M, const CtrlPnt4D& P) ;
template CtrlPnt4D operator*(const MatrixRT<double>& M, const CtrlPnt4D& P) ;
template MatrixRT<float> operator*(const MatrixRT<float>& M1, const MatrixRT<float>& M2) ;
template MatrixRT<double> operator*(const MatrixRT<double>& M1, const MatrixRT<double>& M2) ;

#else

void stupidSparcMatrix(){
  Point3D a ;
  CtrlPnt4D b ;
  MatrixRT<float> A ;
  MatrixRT<double> B ;
  a = A*a ;
  b = A*b ;
  a = B*a ;
  b = B*b ;
}

#endif 
