/*=====================================================================
        File: nurbs.cc
     Purpose:       
    Revision: $Id: nurbs.cc,v 2.9 1998/05/29 18:26:20 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=====================================================================*/
#include <nurbs.hh>
#include <fstream.h>
#include <string.h>
#include <nurbsS.hh>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlNurbsCurve --- default constructor
  Description: Default constructor.
        Input:
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsCurve::PlNurbsCurve(): 
  degree(deg), CtrlP(P), Knot(U),P(1),U(1),deg(0)
{
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlNurbsCurve --- copy constructor
  Description: A copy constructor.
        Input: nurb <-- the NURBS curve to copy
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsCurve::PlNurbsCurve(const PlNurbsCurve& nurb): 
  PlParaCurve(), degree(deg), CtrlP(P), Knot(U),
  P(nurb.P),U(nurb.U),deg(nurb.deg)
{
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: reset --- resets a NURBS curve to new values
  Description: Resets a NURBS curve to new values
        Input: P1 --> the new values for the control points
	       U1 --> the new values for the knot vector
	   Degree --> the new degree of the curve
       Output:
 Restrictions: The size of P1,U1 and Degree must agree: P.n+degree+1=U.n
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::reset(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int Degree) {
  int nSize = P1.n ;
  int mSize = U1.n ;
  deg = Degree ;
  if(nSize != mSize-deg-1){
    PlError err("reset");
    err << "Invalid input size for the control points and the knot vector when reseting a Nurbs Curve.\n";
    err << nSize << " control points  and " << mSize << " knots\n" ;
    err.fatal() ;
  }
  P.resize(P1.n) ;
  U.resize(U1.n) ;
  P = P1 ;
  U = U1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlNurbsCurve --- constructor with control points in 4D
  Description: Constructor with control points in 4D
        Input: P1 --> the control points
	       U1 --> the knot vector
	   Degree --> the degree of the curve
       Output:
 Restrictions: The size of P1,U1 and Degree must agree: P.n+degree+1=U.n
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsCurve::PlNurbsCurve(const Vector_CtrlPnt4D& P1, const Vector_FLOAT &U1, int Degree):
  degree(deg), CtrlP(P), Knot(U), P(P1), U(U1), deg(Degree) 
{

  if(P.n != U.n-deg-1){
    PlError err("PlNurbsCurve(P1,U1,Degree)");
    err << "Invalid input size for the control points and the knot vector.\n";
    err << P.n << " control points  and " << U.n << " knots\n" ;
    err.fatal() ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: PlNurbsCurve --- constructor with control points in 3D
  Description: Constructor with control points in 3D
        Input: P1 --> the control point vector
	        W --> the weight for each control points
	       U1 --> the knot vector
	   Degree --> the degree of the curve
       Output:
 Restrictions: The size of P1,U1 and Degree must agree: P.n+degree+1=U.n
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsCurve::PlNurbsCurve(const Vector_Point3D& P1, const Vector_FLOAT& W, const Vector_FLOAT& U1, int Degree): 
  degree(deg), CtrlP(P), Knot(U), P(P1.n), U(U1), deg(Degree)
{
  int nSize = P1.n ;
  int mSize = U1.n ;

  if(nSize != mSize-deg-1){
    PlError err("PlNurbsCurve(P1,W,U1,Degree)") ;
    err << "Invalid input size for the control points and the knot vector.\n" ;
    err << nSize << " control points  and " << mSize << " knots\n" ;
    err.fatal() ;
  }
  if(nSize != W.n){
    PlError err("PlNurbsCurve(P1,W,U1,Degree)") ;
    err << "Size mismatched between the control points and the weights\n" ;
    err << "ControlPoints size = " << nSize << ", Weight size = " << W.n << endl ;
    err.fatal() ;
  }

  for(int i = 0 ;i<nSize;i++){
    const Point3D& pt = P1[i] ; // This makes the SGI compiler happy
    for(int j=0;j<3;j++)
      P[i].data[j] = pt.data[j] * W[i] ;
    P[i].w() = W[i] ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator= --- the assignment operator for a NURBS curve
  Description: The assignment operator for a NURBS curve
        Input: curve --> the NURBS curve to copy
       Output: a reference to itself
 Restrictions: The curve being copied must be valid, otherwise strange results might occur.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsCurve& PlNurbsCurve::operator=(const PlNurbsCurve& curve) {
  if(curve.U.n != curve.P.n+curve.deg+1){
    PlError err("operator=") ;
    err << "Invalid assignment... the curve being assigned to isn't valid\n" ;
    err.fatal() ;
  }
  deg = curve.deg ;
  U = curve.U ;
  P = curve.P ;
  if(U.n!=P.n+deg+1){
    PlError err("operator=") ;
    err << "Error in assignment... couldn't assign properly the vectors\n" ;
    err.fatal() ;
  }
  return *this ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawImg --- draws a NURBS curve on an image
  Description: This will draw very primitively the NURBS curve on an image.
               The drawing assumes the line is only in the xy plane (the z 
	       is not used for now). 

	       The algorithm finds the points on the curve at a $step$ 
	       parametric intervall between them and join them by a line.
	       No fancy stuff.

        Input: Img <-- draws the nurbs curve to this Image
             color --> the line is drawn in this color
              step --> the parametric distance between two computed points.
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::drawImg(Image_UBYTE& Img,unsigned char color,float step){
  Point3D a1,a2 ;
  float u_max = U[U.n-1] ;
  if(step<=0)
    step = 0.01 ;
  a1 = this->point3D(U[deg]) ;
  float u ;
  int i1,j1,i2,j2 ;
  getCoordinates(a1,i1,j1,Img.rows,Img.cols) ;
  for(u=U[deg]+step ; u < u_max+(step/2.0) ; u+=step){ // the <= u_max doesn't work
    a2 = this->point3D(u) ;
    if(!getCoordinates(a2,i2,j2,Img.rows,Img.cols))
      continue ;
    Img.drawLine(i1,j1,i2,j2,color) ;
    i1 = i2 ;
    j1 = j2 ;
  }
  a2 = this->point3D(U[P.n]) ;
  if(getCoordinates(a2,i2,j2,Img.rows,Img.cols))
     Img.drawLine(i1,j1,i2,j2,color) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawImg --- draws a NURBS curve on an image
  Description: This will draw very primitively the NURBS curve on an image.
               The drawing assumes the line is only in the xy plane (the z 
	       is not used for now). 

	       The algorithm finds the points on the curve at a $step$ 
	       parametric intervall between them and join them by a line.
	       No fancy stuff.

        Input: Img <-- draws the nurbs curve to this Image
             color --> the line is drawn in this color
              step --> the parametric distance between two computed points.
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::drawImg(Image_Color& Img,const Color& color,float step){
  Point3D a1,a2 ;
  float u_max = U[U.n-1] ;
  if(step<=0)
    step = 0.01 ;
  a1 = this->point3D(U[deg]) ;
  int i1,j1,i2,j2 ;
  getCoordinates(a1,i1,j1,Img.rows,Img.cols) ;
  float u ;
  for(u=U[deg]+step ; u < u_max+(step/2.0) ; u+=step){ // the <= u_max doesn't work
    a2 = this->point3D(u) ;
    if(!getCoordinates(a2,i2,j2,Img.rows,Img.cols))
      continue ;
    Img.drawLine(i1,j1,i2,j2,color) ;
    i1 = i2 ;
    j1 = j2 ;
  }
  a2 = this->point3D(U[P.n]) ;
  if(getCoordinates(a2,i2,j2,Img.rows,Img.cols))
    Img.drawLine(i1,j1,i2,j2,color) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawAaImg --- draws an anti-aliased NURBS curve on an image
  Description: This will draw the NURBS by using a circular brush profile. The
               drawing is performed by averaging the intensity of the 
	       profile at the pixels.
	      
        Input: Img <-- draws the nurbs curve to this Image
             color --> the line is drawn in this color
	 precision --> this number influences the number of points used for
	               averaging purposes.
	     alpha --> a flag indicating if the profile is used as an alpha
	               chanel. If so, the line doesn't overwrite, it blends
		       the line with the image already present in Img.
       Output:
 Restrictions: This routine is very {\em slow}; use aliased drawing for speed.
   Created by: Philippe Lavoie (25 July 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::drawAaImg(Image_Color& Img, const Color& color, int precision, int alpha){
  PlNurbsCurve profile ;

  profile.makeCircle(Point3D(0,0,0),Point3D(1,0,0),Point3D(0,0,1),1.0,0,M_PI) ;
  drawAaImg(Img,color,profile,precision,alpha) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawAaImg --- draws an anti-aliased NURBS curve on an image
  Description: This will draw the NURBS by using a user-defined brush profile.
               The drawing is performed by averaging the intensity of the 
	       profile at the pixels.
	      
        Input: Img <-- draws the nurbs curve to this Image
             color --> the line is drawn in this color
	   profile --> the profile of the NURBS curve to draw
	 precision --> this number influences the number of points used for
	               averaging purposes.
	     alpha --> a flag indicating if the profile is used as an alpha
	               chanel. If so, the line doesn't overwrite, it blends
		       the line with the image already present in Img.
       Output:
 Restrictions: This routine is very {\em slow}; use aliased drawing for speed.
   Created by: Philippe Lavoie (22 August 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::drawAaImg(Image_Color& Img, const Color& color, const PlNurbsCurve& profile, int precision, int alpha){
  Vector_CtrlPnt4D sPts(2) ;
  sPts[0] = sPts[1] = CtrlPnt4D(1,1,1,1) ;
  Vector_FLOAT sKnot(4) ;
  sKnot[0] = sKnot[1] = 0.0 ;
  sKnot[2] = sKnot[3] = 1.0 ;
  
  PlNurbsCurve scaling(sPts,sKnot,1) ;  

  drawAaImg(Img,color,profile,scaling,precision,alpha) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: drawAaImg --- draws an anti-aliased NURBS curve on an image
  Description: This will draw the NURBS by using a brush profile. The
               drawing is performed by averaging the intensity of the 
	       profile at the pixels. 
	       
	       This function generates a sweep surface by using the profile
	       given in its argument. The sweep is always performed by 
	       following the $y$-axis of the profile. A scaling function
	       is also used when sweeping. This is used to vary the shape
	       of the profile while it's being swept (see the sweep member
	       function of PlNurbsSurface for more details).

        Input: Img <-- draws the nurbs curve to this Image
             color --> the line is drawn in this color
	   profile --> the profile of the NURBS curve to draw
	   scaling --> the scaling to give the profile while drawing the curve
	 precision --> this number influences the number of points used for
	               averaging purposes.
	     alpha --> a flag indicating if the profile is used as an alpha
	               chanel. If so, the line doesn't overwrite, it blends
		       the line with the image already present in Img.
       Output:
 Restrictions: This routine is very {\em slow}; use aliased drawing for speed 
               or lower the precision factor.
   Created by: Philippe Lavoie (25 July 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface PlNurbsCurve::drawAaImg(Image_Color& Img, const Color& color, const PlNurbsCurve& profile, const PlNurbsCurve& scaling, int precision, int alpha){
  Matrix_FLOAT addMatrix ;
  Matrix_INT nMatrix ;

  addMatrix.resize(Img.rows,Img.cols) ;
  nMatrix.resize(Img.rows,Img.cols) ;

  int i,j ;

  float du,dv ;
  // compute a coarse distance for the curve
  Point3D a,b,c ;
  a = point3D(0.0) ;
  b = point3D(0.5) ;
  c = point3D(1.0) ;

  float distance = abs(b-a) + abs(c-b) ;

  dv = distance*float(precision) ;
  dv = (U[U.n-1]-U[0])/dv ;

  // compute a coarse distance for the trajectory
  a = profile.point3D(0.0) ;
  b = profile.point3D(0.5) ;
  c = profile.point3D(1.0) ;
  distance = abs(b-a) + abs(c-b) ;
  du = distance*float(precision) ;
  du = (profile.U[profile.U.n-1]-profile.U[0])/du ;

  PlNurbsSurface drawCurve ;

  drawCurve.sweep(*this,profile,scaling,P.n-1) ;

  float u,v ;

  for(u=U[0];u<U[U.n-1];u+=du)
    for(v=profile.U[0];v<profile.U[profile.U.n-1];v+=dv){
      Point3D p ;
      p = drawCurve.point3D(u,v) ;
      if(getCoordinates(p,i,j,Img.rows,Img.cols)){
	addMatrix(i,j) += p.z ;
	nMatrix(i,j) += 1 ;
      }
    }

  float maxP = 1.0 ;
  for(i=0;i<Img.rows;++i)
    for(j=0;j<Img.cols;++j){
      addMatrix(i,j) /= float(nMatrix(i,j)) ;
      if(addMatrix(i,j)>maxP)
	maxP = addMatrix(i,j) ;
    }

  for(i=0;i<Img.rows;++i)
    for(j=0;j<Img.cols;++j){
      if(nMatrix(i,j)){
	double mean = double(addMatrix(i,j))/double(maxP) ;
	if(alpha){
	  Img(i,j).r = (unsigned char)(mean*double(color.r)+(1.0-mean)*double(Img(i,j).r)) ;
	  Img(i,j).g = (unsigned char)(mean*double(color.g)+(1.0-mean)*double(Img(i,j).g)) ;
	  Img(i,j).b = (unsigned char)(mean*double(color.b)+(1.0-mean)*double(Img(i,j).b)) ;
	}
	else{
	  Img(i,j) = mean*color ;
	}
      }
    }
  return drawCurve ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: transform --- performs geometrical modifications
  Description: Each control points will be modified by a rotation-translation
               matrix.
	      
        Input: A --> the rotation-translation matrix
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (22 August 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::transform(const MatrixRT<float>& A){
  for(int i=0;i<P.n;++i)
    P[i] = A*P[i] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator()  --- evaluates the curve in 4D at parameter $u$
  Description: It evaluates the NURBS curve in 4D at the parametric point 
               $u$.  Using the following equation
	       \begin{equation}
	         C(u) = \sum_{i=0}^n N_{i,p} P_i \hspace{0.5in} a \leq u \leq b
	       \end{equation}
	        where $P_i$ are the control points and $N_{i,p}$ are the 
	       $p$th degree B-spline basis functions.

	       For more details on the algorithm, see A4.1 on page 124 of 
	       the Nurbs Book.

        Input: u --> the parametric value at which the curve is evaluated
       Output: the 4D point at $C(u)$
 Restrictions: the parametric value must be in a valid range
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
CtrlPnt4D PlNurbsCurve::operator()(float u) const{
  Vector_FLOAT Nb(deg+1) ;
  CtrlPnt4D p ;
  int span ;

  p = 0 ;
  span = findSpan(u) ;
  basisFuns(u,span,Nb) ;
  for(int i=0;i<=deg;++i) {
    p += Nb[i] * P[span-deg+i] ;
  }
  return p ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: derive3D --- Computes the derivative of degree $d$ of the 
                            curve at parameter $u$
  Description: For more information on the algorithm used, see A3.2 on p 93 
               of the NurbsBook.
        Input: u --> the parametric value to evaluate at
               d --> the degree of the derivative
       Output: The derivative $d$ in 3D at the parameter $u$
 Restrictions: $u$ and $d$ must be in a valid range.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Point3D PlNurbsCurve::derive3D(float u, int d) const {
  Vector_Point3D ders ;
  deriveAt(u,d,ders) ;
  return ders[d] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: derive --- Computes the derivative of degree $d$ of the 
                            curve at parameter $u$
  Description: For more information on the algorithm used, see A3.2 on p 93 
               of the NurbsBook.
        Input: u --> the parametric value to evaluate at
               d --> the degree of the derivative
       Output: The derivative $d$ in 4D at the parameter $u$
 Restrictions: $u$ and $d$ must be in a valid range.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
CtrlPnt4D PlNurbsCurve::derive(float u, int d) const {
  Vector_CtrlPnt4D ders ;
  deriveAt(u,d,ders) ;
  return ders[d] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: deriveAt --- Computes the derivative of degree $d$ of the 
                              curve at parameter $u$
  Description: For more information on the algorithm used, see A3.2 on p 93 
               of the NurbsBook.
        Input: u --> the parametric value to evaluate at
               d --> the degree of the derivative
           ders <-- a vector containing the derivatives of the curve at $u$.
       Output: 
 Restrictions: $u$ and $d$ must be in a valid range.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::deriveAt(float u,int d, Vector_CtrlPnt4D& ders) const{
  int du = minimum(d,degree) ;
  int span ;
  Matrix_FLOAT derF(du+1,degree+1) ;
  ders.resize(d+1) ;
  ders.reset(0) ;

  span = findSpan(u) ;
  dersBasisFuns(du,u,span,derF) ;
  for(int k=0;k<=du;k++){
    ders[k] = 0 ;
    for(int j=0;j<=degree;j++){
      ders[k] = ders[k] + derF(k,j)*P[span-degree+j] ;
    }
  }
}


// Setup the binomial coefficients into th matrix Bin
// Bin(i,j) = (i  j)
// The binomical coefficients are defined as follow
//   (n)         n!
//   (k)  =    k!(n-k)!       0<=k<=n
// and the following relationship applies 
// (n+1)     (n)   ( n )
// ( k ) =   (k) + (k-1)
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: binomialCoef --- setup a matrix containing binomial coefficients
  Description: Setup the binomial coefficients into th matrix Bin
               \[ Bin(i,j) = \left( \begin{array}{c}i \\ j\end{array} \right)\]
	       The binomical coefficients are defined as follow
	       \[ \left(\begin{array}{c}   n \\ k \end{array} \right)= \frac{ n!}{k!(n-k)!} \mbox{for $0\leq k \leq n$} \]
	       and the following relationship applies 
	       \[ \left(\begin{array}{c} n+1 \\ k \end{array} \right) = 
	       \left(\begin{array}{c} n \\ k \end{array} \right) +
	       \left(\begin{array}{c} n \\ k-1 \end{array} \right) \]

        Input: Bin <-- the binomial matrix
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void binomialCoef(Matrix_FLOAT& Bin){
  int n,k ;
  // Setup the first line
  Bin(0,0) = 1.0 ;
  for(k=1;k<Bin.cols;k++)
    Bin(0,k) = 0.0 ;
  // Setup the other lines
  for(n=0;n<Bin.rows-1;n++){
    Bin(n+1,0) = 1.0 ;
    for(k=1;k<Bin.cols;k++)
      if(n+1<k)
	Bin(n,k) = 0.0 ;
      else
	Bin(n+1,k) = Bin(n,k) + Bin(n,k-1) ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: deriveAt --- Computes the derivative at the parameter $u$
  Description: Computes the derivative of the curve at the parameter $u$
        Input: u --> the parameter at which the derivative is computed
               d --> the degree of derivation
           ders <-- the vector containing the derivatives of the point at $u$.
       Output:
 Restrictions: $u$ and $d$ must be in a valid range.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::deriveAt(float u, int d, Vector_Point3D& ders) const{
  Vector_CtrlPnt4D dersW ;
  deriveAt(u,d,dersW) ;
  Point3D v ;
  int k,i ;
  ders.resize(d+1) ;
  
  Matrix_FLOAT Bin(d+1,d+1) ;
  binomialCoef(Bin) ;

  // Compute the derivative at the parmeter u

  for(k=0;k<=d;k++){
    v.x = dersW[k].x() ;
    v.y = dersW[k].y() ;
    v.z = dersW[k].z() ;
    for(i=1 ;i<=k ;i++){
      v = v - Bin(k,i)*dersW[i].w()*ders[k-i] ;
    }
    ders[k] = v/dersW[0].w() ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: normal --- Computes the normal of the curve at $u$ from a vector
  Description: Computes the normal of the curve at $u$ from a vector.

               If the curve lies only in the $xy$-plane, then calling the
	       function with the vector $v = (0,0,1)$ (the $z$-axis) will 
	       yield a proper normal for this curve.

        Input: u --> the parameter at which the normal is computed
               v --> the vector to compute the normal with
       Output: the normal vector in 3D.
 Restrictions: $u$ must be in a valid range.
   Created by: Philippe Lavoie          (2 September, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Point3D PlNurbsCurve::normal(float u, const Point3D& v) const{
  Vector_Point3D ders;
  deriveAt(u,1,ders) ;
  return crossProduct(ders[1],v) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: basisFun --- Computes the basis function of the curve
  Description:  Computes the $i$ basis function of degree $p$ of the curve at 
               parameter $u$. 
               This is often noted as $N_{ip}(u)$.

	       The B-spline basis function of $p$-degree is defined as
	       \begin{eqnarray}
	       N_{i,0}(u) & = & \left\{ \begin{array}{ll} 1 & \mbox{if $u_i \leq u < u_{i+1}$} \\ 0 & \mbox{otherwise}\end{array}\right. \nonumber \\
	       N_{i,p}(u) & = & \frac{u-u_i}{u_{i+p}-u_i}N_{i,p-1}(u)+\frac{u_{i+p+1}-u}{u_{i+p+1}-u_{i+1}}N_{i+1,p-1}(u) \nonumber
	       \end{eqnarray}
 
	       where the $u_i$ define the knot vector $U = \{u_0,\ldots,u_m\}$
	       as a nondecreasing sequence of real numbers, {\em i.e.}, 
	       $u_i \leq u_{i+1}$ for $i=0,\ldots,m-1$. And $m$ is related
	       to the number of control points $n$ and the degree of the curve
	       $p$ with the relation $m = n + p + 1$. The knot vector has
	       the form

	       \begin{equation}
	       U=\{\underbrace{a,\ldots,a}_{p+1},u_{p+1},\ldots,u_{m-p-1},\underbrace{b,\ldots,b}_{p+1} \} 
	       \end{equation}

        Input: u --> the parametric variable
               i --> specifies which basis function to compute
               p --> the degree to which the basis function is computed
       Output: the value of  $N_{ip}(u)$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlNurbsCurve::basisFun(float u, int i, int p) const{
  float Nip ;
  float saved,Uleft,Uright,temp ;
  
  if(p<1)
    p = deg ;

  Vector_FLOAT N(p+1) ;

  if((i==0 && u == U[0]) ||
     (i == U.n-p-2 && u==U[U.n-1])){
    Nip = 1.0 ;
    return Nip ;
  }
  if(u<U[i] || u>=U[i+p+1]){
    Nip = 0.0 ;
    return Nip;
  }
  int j ;
  for(j=0;j<=p;j++){
    if(u>=U[i+j] && u<U[i+j+1]) 
      N[j] = 1.0 ;
    else
      N[j] = 0.0 ;
  }
  for(int k=1; k<=p ; k++){
    if(N[0] == 0.0)
      saved = 0.0 ;
    else
      saved = ( (u-U[i])*N[0])/(U[i+k]-U[i]) ;
    for(j=0;j<p-k+1;j++){
      Uleft = U[i+j+1] ;
      Uright = U[i+j+k+1] ;
      if(N[j+1]==0.0){
	N[j] = saved ;
	saved = 0.0 ;
      }
      else {
	temp = N[j+1]/(Uright-Uleft) ;
	N[j] = saved+(Uright-u)*temp ;
	saved = (u-Uleft)*temp ;
      }
    }
  }
  Nip = N[0] ;

  return Nip ;  
}

// it returns the matrix ders, where ders(n+1,deg+1) and the C'(u) = ders(1,span-deg+j) ;
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  dersBasisFuns --- Compute the derivatives functions at $u$ of the NURBS curve
  Description: For information on the algorithm, see A2.3 on p72 of the NURBS 
               book.
	       The result is stored in the ders matrix, where ders is of 
	       size $(n+1,deg+1)$ and the derivative 
	       $N'_i(u) = ders(1,i=span-deg+j)$ where $j=0\ldots deg+1$.

        Input:   n -->  the degree of the derivation
                 u -->  the parametric value
              span --> the span for the basis functions
              ders <-- A matrix containing the derivatives of the curve.
       Output:
 Restrictions: $n$, $u$ and $span$ must be valid values.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::dersBasisFuns(int n,float u, int span, Matrix_FLOAT& ders) const {
  Vector_FLOAT left(degree+1),right(degree+1) ;
  Matrix_FLOAT ndu(degree+1,degree+1) ;
  float saved,temp ;
  int j,r ;

  ders.resize(n+1,degree+1) ;

  ndu(0,0) = 1.0 ;
  for(j=1; j<= degree ;j++){
    left[j] = u-U[span+1-j] ;
    right[j] = U[span+j]-u ;
    saved = 0.0 ;
    
    for(r=0;r<j ; r++){
      // Lower triangle
      ndu(j,r) = right[r+1]+left[j-r] ;
      temp = ndu(r,j-1)/ndu(j,r) ;
      // Upper triangle
      ndu(r,j) = saved+right[r+1] * temp ;
      saved = left[j-r] * temp ;
    }

    ndu(j,j) = saved ;
  }

  for(j=0;j<=degree;j++)
    ders(0,j) = ndu(j,degree) ;

  // Compute the derivatives
  Matrix_FLOAT a(degree+1,degree+1) ;
  for(r=0;r<=degree;r++){
    int s1,s2 ;
    s1 = 0 ; s2 = 1 ; // alternate rows in array a
    a(0,0) = 1.0 ;
    // Compute the kth derivative
    for(int k=1;k<=n;k++){
      float d ;
      int rk,pk,j1,j2 ;
      d = 0.0 ;
      rk = r-k ; pk = degree-k ;

      if(r>=k){
	a(s2,0) = a(s1,0)/ndu(pk+1,rk) ;
	d = a(s2,0)*ndu(rk,pk) ;
      }

      if(rk>=-1){
	j1 = 1 ;
      }
      else{
	j1 = -rk ;
      }

      if(r-1 <= pk){
	j2 = k-1 ;
      }
      else{
	j2 = degree-r ;
      }

      for(j=j1;j<=j2;j++){
	a(s2,j) = (a(s1,j)-a(s1,j-1))/ndu(pk+1,rk+j) ;
	d += a(s2,j)*ndu(rk+j,pk) ;
      }
      
      if(r<=pk){
	a(s2,k) = -a(s1,k-1)/ndu(pk+1,r) ;
	d += a(s2,k)*ndu(r,pk) ;
      }
      ders(k,r) = d ;
      j = s1 ; s1 = s2 ; s2 = j ; // Switch rows
    }
  }

  // Multiply through by the correct factors
  r = degree ;
  for(int k=1;k<=n;k++){
    for(j=0;j<=degree;j++)
      ders(k,j) *= r ;
    r *= degree-k ;
  }
}

// Computes the non-zero basis functions into N of size deg+1
// The following relationship applies   N[i] <= N[span-deg+i]  for i = 0..deg
// A2.1 on p68 of the Nurbs Book
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: basisFuns --- computes the non-zero basis functions of the curve
  Description: Computes the non-zero basis functions and puts the result 
               into N. N has a size of $deg+1$. To relate N to the basis 
	       functions, $N_{all}[span -deg +i] = N[i]$ for $i=0\ldots deg$.

	       The B-spline basis function of $p$-degree is defined as
	       \begin{eqnarray}
	       N_{i,0}(u) & = & \left\{ \begin{array}{ll} 1 & \mbox{if $u_i \leq u < u_{i+1}$} \\ 0 & \mbox{otherwise}\end{array}\right. \nonumber \\
	       N_{i,p}(u) & = & \frac{u-u_i}{u_{i+p}-u_i}N_{i,p-1}(u)+\frac{u_{i+p+1}-u}{u_{i+p+1}-u_{i+1}}N_{i+1,p-1}(u) \nonumber
	       \end{eqnarray}
 
	       where the $u_i$ define the knot vector $U = \{u_0,\ldots,u_m\}$
	       as a nondecreasing sequence of real numbers, {\em i.e.}, 
	       $u_i \leq u_{i+1}$ for $i=0,\ldots,m-1$. And $m$ is related
	       to the number of control points $n$ and the degree of the curve
	       $p$ with the relation $m = n + p + 1$. The knot vector has
	       the form

	       \begin{equation}
	       U=\{\underbrace{a,\ldots,a}_{p+1},u_{p+1},\ldots,u_{m-p-1},\underbrace{b,\ldots,b}_{p+1} \} 
	       \end{equation}

	       The B-spline basis function are non-zero for at most $p+1$ of
	       the $N_{i,p}$. The relationship between the non-zero basis
	       functions in N and $N_{i,p}$ is as follows, 
	       $N_{span - deg + j, p} = N[j]$ for $j=0,\ldots,deg$. Where
	       span is the non-zero span of the basis functions. This
	       non-zero span for $u$ can be found by calling 
	       {\tt findSpan(u)}.

        Input: u --> the parametric value
               i --> the non-zero span of the basis functions
               N <-- the non-zero basis functions
       Output:
 Restrictions: $u$ and $i$ must be valid values
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::basisFuns(float u, int i, Vector_FLOAT& N) const{
  Vector_FLOAT left(deg+1), right(deg+1) ;
  float temp,saved ;
   
  N.resize(deg+1) ;

  N[0] = 1.0 ;
  for(int j=1; j<= deg ; j++){
    left[j] = u-U[i+1-j] ;
    right[j] = U[i+j]-u ;
    saved = 0.0 ;
    for(int r=0 ; r<j; r++){
      temp = N[r]/(right[r+1]+left[j-r]) ;
      N[r] = saved+right[r+1] * temp ;
      saved = left[j-r] * temp ;
    }
    N[j] = saved ;
  }
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: findSpan --- Determines the knot span index
  Description: Determines the knot span for which their exists non-zero basis 
               functions. The span is the index $k$ for which the parameter 
	       $u$ is valid in the $[u_k,u_{k+1}]$ range.
        Input: u --> the parametric value
       Output: the span index at $u$.
 Restrictions: $u$ must be in a valid range
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::findSpan(float u) const{
  if(u>=U[P.n]) 
    return P.n-1 ;
  if(u<=U[0])
    return deg ;

  int low = deg ;
  int high = P.n+1 ; 
  int mid = (low+high)/2 ;

  while(u<U[mid] || u>= U[mid+1]){
    if(u<U[mid])
      high = mid ;
    else
      low = mid ;
    mid = (low+high)/2 ;
  }
  return mid ;
 
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: findKnot --- Finds the knot $k$ for which $u$ is in the range $[u_k,u_{k+1})$
  Description: Finds the knot $k$ for which $u$ is in the range $[u_k,u_{k+1})$
        Input: u --> parametric value
       Output: the index $k$
 Restrictions: $u$ must be in a valid range.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::findKnot(float u) const{
  if(u==U[P.n])
    return P.n ;
  for(int i=deg+1; i<P.n ; i++)
    if(U[i]>u){
      return i-1 ;
    }
  return -1 ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: findMult --- finds the multiplicity of a knot on a curve
  Description: Finds the multiplicity of a knot
        Input: r --> the knot to observe
       Output: the multiplicity of the knot
 Restrictions: $r$ must be a valid knot index
   Created by:  Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::findMult(int r) const {
  int s=1 ;
  for(int i=r;i>deg+1;i--)
    if(U[i]==U[i-1])
      s++ ;
    else
      return s ;
  return s ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: findMultSpan --- finds the multiplicity of a knot at a parametric value
  Description: Finds the index of the knot at parametric value $u$ and 
               returns its multiplicity.

        Input: u --> the parametric value
               r <-- the knot of interest
               s <-- the multiplicity of this knot
       Output:
 Restrictions: $u$ must be in a valid range.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::findMultSpan(float u, int& r, int& s) const {
  r = findKnot(u) ;
  if(u==U[r]){
    s = findMult(r) ;
  }
  else
    s = 0 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: resize --- resizes a NURBS curve
  Description: Resizes a NURBS curve. The old values are lost and new ones
               have to be created.
        Input: n --> the new number of control points for the curve
             Deg --> the new degree for the curve
       Output:
 Restrictions: 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::resize(int n, int Deg){
  deg = Deg ;
  P.resize(n) ;
  U.resize(n+deg+1) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: leastSquares ---  A least squares curve approximation
  Description: This routine solves the following problem: find the NURBS curve
               $C$ satisfying
	       \begin{itemize}
	       \item $Q_0 = C(0)$ and $Q_m = C(1)$
	       \item the remaining $Q_k$ are approximated in the least squares
	           sense, {\em i.e.}
		   \[ \sum_{k=1}^{m-1} | Q_k-C(\bar{u}_k)|^2 \]
		   in a minimum with respect to the $n$ variable $P_i$; the
		   $\bar{u}$ are the parameter values computed with the 
		   chord length method.
	       \end{itemize}

	       The resulting curve will generally not pass through $Q_k$ and
	       $C(\bar{u}_k)$ is not the closest point on $C(u)$ to $Q_k$.

	       For more details, see section 9.4.1 on page 491 of the NURBS 
	       book.
        Input:     Q --> the vector of 3D points
	        degC --> the degree of the curve
		   n --> the number of control points in the new curve.
       Output:
 Restrictions: $deg$ must be smaller than Q.n.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::leastSquares(const Vector_Point3D& Q, int degC, int n){
  Vector_FLOAT ub(Q.n) ;

  chordLengthParam(Q,ub) ;

  return leastSquares(Q,degC,n,ub) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: leastSquares ---  A least squares curve approximation
  Description: This routine solves the following problem: find the NURBS curve
               $C$ satisfying
	       \begin{itemize}
	       \item $Q_0 = C(0)$ and $Q_m = C(1)$
	       \item the remaining $Q_k$ are approximated in the least squares
	           sense, {\em i.e.}
		   \[ \sum_{k=1}^{m-1} | Q_k-C(\bar{u}_k)|^2 \]
		   in a minimum with respect to the $n$ variable $P_i$; the
		   $\bar{u}$ are the precomputed parameter values.
	       \end{itemize}

	       The resulting curve will generally not pass through $Q_k$ and
	       $C(\bar{u}_k)$ is not the closest point on $C(u)$ to $Q_k$.

	       For more details, see section 9.4.1 on page 491 of the NURBS 
	       book.
        Input:  Q --> the vector of 3D points
	     degC --> the degree of the curve
	        n --> the number of control points in the new curve
	       ub --> the knot coefficients
       Output:
 Restrictions: the variable curve {\bf must} contain a valid knot vector.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::leastSquares(const Vector_Point3D& Q, int degC, int n, const Vector_FLOAT& ub){
  int i,j;
  float d,a ;

  if(ub.n != Q.n){
    PlError err("leastSquares");
    err << "leastSquaresCurve\n" ;
    err << "ub size is different than Q's\n" ;
    err.fatal() ;
  }

  deg = degC ;
  U.resize(n+deg+1) ;

  // Changing the method to generate a U compare to the one 
  // described by Piegl and Tiller in the NURBS book (eq 9.69)

  U.reset(1.0) ;
  d = (float)(Q.n)/(float)(n) ; 
  for(j=0;j<=deg;j++)
    U[j] = 0 ;

  for(j=1;j<n-deg;++j){
    U[deg+j] = 0.0 ;
    for(int k=j;k<j+deg;++k){
      i = (int)(k*d) ;
      a = float(k*d)-float(i) ;
      int i2 = (int)((k-1)*d) ;
      U[deg+j] += a*ub[i2]+(1-a)*ub[i] ;
      }
    U[deg+j] /= deg ;
  }

  return leastSquares(Q, degC, n, ub, U) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: leastSquares ---  A least squares curve approximation
  Description: This routine solves the following problem: find the NURBS curve
               $C$ satisfying
	       \begin{itemize}
	       \item $Q_0 = C(0)$ and $Q_m = C(1)$
	       \item the remaining $Q_k$ are approximated in the least squares
	           sense, {\em i.e.}
		   \[ \sum_{k=1}^{m-1} | Q_k-C(\bar{u}_k)|^2 \]
		   in a minimum with respect to the $n$ variable $P_i$; the
		   $\bar{u}$ are the precomputed parameter values.
	       \end{itemize}

	       The resulting curve will generally not pass through $Q_k$ and
	       $C(\bar{u}_k)$ is not the closest point on $C(u)$ to $Q_k$.

	       For more details, see section 9.4.1 on page 491 of the NURBS 
	       book.
        Input:  Q --> the vector of 4D points
	     degC --> the degree of the curve
	        n --> the number of control points in the new curve
	       ub --> the knot coefficients
       Output:
 Restrictions: the variable curve {\bf must} contain a valid knot vector.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::leastSquares(const Vector_CtrlPnt4D& Q, int degC, int n, const Vector_FLOAT& ub){
  int i,j;
  float d,a ;

  if(ub.n != Q.n){
    PlError err("leastSquares");
    err << "leastSquaresCurve\n" ;
    err << "ub size is different than Q's\n" ;
    err.fatal() ;
  }

  deg = degC ;
  U.resize(n+deg+1) ;

  // Changing the method to generate a U compare to the one 
  // described by Piegl and Tiller in the NURBS book (eq 9.69)

  U.reset(1.0) ;
  d = (float)(Q.n)/(float)(n) ; 
  for(j=0;j<=deg;j++)
    U[j] = 0 ;

  for(j=1;j<n-deg;++j){
    U[deg+j] = 0.0 ;
    for(int k=j;k<j+deg;++k){
      i = (int)(k*d) ;
      a = float(k*d)-float(i) ;
      int i2 = (int)((k-1)*d) ;
      U[deg+j] += a*ub[i2]+(1-a)*ub[i] ;
      }
    U[deg+j] /= deg ;
  }

  return leastSquares(Q, degC, n, ub, U) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: leastSquares ---  A least squares curve approximation
  Description: This routine solves the following problem: find the NURBS curve
               $C$ satisfying
	       \begin{itemize}
	       \item $Q_0 = C(0)$ and $Q_m = C(1)$
	       \item the remaining $Q_k$ are approximated in the least squares
	           sense, {\em i.e.}
		   \[ \sum_{k=1}^{m-1} | Q_k-C(\bar{u}_k)|^2 \]
		   in a minimum with respect to the $n$ variable $P_i$; the
		   $\bar{u}$ are the precomputed parameter values.
	       \end{itemize}

	       The resulting curve will generally not pass through $Q_k$ and
	       $C(\bar{u}_k)$ is not the closest point on $C(u)$ to $Q_k$.

	       For more details, see section 9.4.1 on page 491 of the NURBS 
	       book.
        Input:  Q --> the vector of 3D points
	     degC --> the degree of the curve
	        n --> the number of control points in the new curve
	       ub --> the knot coefficients
	     knot --> the knot vector to use for the curve
       Output: 1 if succesfull, 0 it the number of points to approximate the curve with
               is too big compared to the number of points.
 Restrictions: the variable curve {\bf must} contain a valid knot vector.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::leastSquares(const Vector_Point3D& Q, int degC, int n, const Vector_FLOAT& ub, const Vector_FLOAT& knot){
  int i,j,span,m ;

  m = Q.n ;

  if(ub.n != Q.n){
    PlError err("leastSquares");
    err << "leastSquaresCurve\n" ;
    err << "ub size is different than Q's\n" ;
    err.fatal();
  }

  if(knot.n != n+degC+1){
    PlError err("leastSquares");
    err << "The knot vector supplied doesn't have the proper size.\n" ;
    err << "It should be n+degC+1 = " << n+degC+1 << " and it is " << knot.n << endl ;
    err.fatal() ;
  }

  deg = degC ;

  U = knot ;

  P.resize(n) ;

  Vector_Point3D R(n),rk(m) ; 
  Vector_FLOAT funs(deg+1) ;
  Matrix_DOUBLE N(m,n) ;
  R[0] = Q[0] ;
  R[n-1] = Q[m-1] ;
  N(0,0) = 1.0 ;
  N(m-1,n-1) = 1.0 ;

  // Set up N 
  N(0,0) = 1.0 ;
  N(m-1,n-1) = 1.0 ;

  //  for(i=1;i<m-1;i++){
  for(i=0;i<m;i++){
     span = findSpan(ub[i]) ;
     basisFuns(ub[i],span,funs);
     for(j=0;j<=deg;j++){
       //if(span-deg+j>0)
	 N(i,span-deg+j) = (double)funs[j] ;
     }
     rk[i] = Q[i]-N(i,0)*Q[0]-N(i,n-1)*Q[m-1] ;  

  }

  // Set up R
  //  for(i=1;i<n-1;i++){
  for(i=0;i<n;i++){
    R[i] = 0.0 ;
    //    for(j=1;j<m-1;j++)
    for(j=0;j<m;j++)
      R[i] += N(j,i)*rk[j] ;
    if(R[i].x*R[i].x<1e-10 && 
       R[i].y*R[i].y<1e-10 &&
       R[i].z*R[i].z<1e-10)
      return 0 ; 
  }

  // Solve      N^T*N*P = R

  // must check for the case where we want a curve of degree 1 having
  // only 2 points.
  if(n-2>0){ 
    Matrix_DOUBLE X(n-2,3),B(n-2,3),Ns(m-2,n-2) ;
    for(i=0;i<B.rows;i++){
      for(j=0;j<3;j++)
	B(i,j) = (double)R[i+1].data[j] ;
    }
    Ns = N.get(1,1,m-2,n-2) ;
    
    //LAPack.lse(transpose(Ns)*Ns,B,X) ;
    solve(transpose(Ns)*Ns,B,X) ;
    
    for(i=0;i<X.rows;i++){
      for(j=0;j<X.cols;j++)
	P[i+1].data[j] = (float)X(i,j) ;
      P[i+1].w() = 1.0 ;
    }
  }
  P[0] = Q[0] ;
  P[n-1] = Q[m-1] ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: leastSquares ---  A least squares curve approximation
  Description: This routine solves the following problem: find the NURBS curve
               $C$ satisfying
	       \begin{itemize}
	       \item $Q_0 = C(0)$ and $Q_m = C(1)$
	       \item the remaining $Q_k$ are approximated in the least squares
	           sense, {\em i.e.}
		   \[ \sum_{k=1}^{m-1} | Q_k-C(\bar{u}_k)|^2 \]
		   in a minimum with respect to the $n$ variable $P_i$; the
		   $\bar{u}$ are the precomputed parameter values.
	       \end{itemize}

	       The resulting curve will generally not pass through $Q_k$ and
	       $C(\bar{u}_k)$ is not the closest point on $C(u)$ to $Q_k$.

	       For more details, see section 9.4.1 on page 491 of the NURBS 
	       book.
        Input:  Q --> the vector of 4D points
	     degC --> the degree of the curve
	        n --> the number of control points in the new curve
	       ub --> the knot coefficients
	     knot --> the knot vector to use for the curve
       Output: 1 if succesfull, 0 it the number of points to approximate the curve with
               is too big compared to the number of points.
 Restrictions: the variable curve {\bf must} contain a valid knot vector.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::leastSquares(const Vector_CtrlPnt4D& Q, int degC, int n, const Vector_FLOAT& ub, const Vector_FLOAT& knot){
  int i,j,span,m ;

  m = Q.n ;

  if(ub.n != Q.n){
    PlError err("leastSquares");
    err << "leastSquaresCurve\n" ;
    err << "ub size is different than Q's\n" ;
    err.fatal();
  }

  if(knot.n != n+degC+1){
    PlError err("leastSquares");
    err << "The knot vector supplied doesn't have the proper size.\n" ;
    err << "It should be n+degC+1 = " << n+degC+1 << " and it is " << knot.n << endl ;
    err.fatal() ;
  }

  deg = degC ;

  U = knot ;

  P.resize(n) ;

  Vector_CtrlPnt4D R(n),rk(m) ; 
  Vector_FLOAT funs(deg+1) ;
  Matrix_DOUBLE N(m,n) ;
  R[0] = Q[0] ;
  R[n-1] = Q[m-1] ;
  N(0,0) = 1.0 ;
  N(m-1,n-1) = 1.0 ;

  // Set up N 
  N(0,0) = 1.0 ;
  N(m-1,n-1) = 1.0 ;

  //  for(i=1;i<m-1;i++){
  for(i=0;i<m;i++){
     span = findSpan(ub[i]) ;
     basisFuns(ub[i],span,funs);
     for(j=0;j<=deg;j++){
       //if(span-deg+j>0)
	 N(i,span-deg+j) = (double)funs[j] ;
     }
     rk[i] = Q[i]-N(i,0)*Q[0]-N(i,n-1)*Q[m-1] ;  

  }

  // Set up R
  //  for(i=1;i<n-1;i++){
  for(i=0;i<n;i++){
    R[i] = 0.0 ;
    //    for(j=1;j<m-1;j++)
    for(j=0;j<m;j++)
      R[i] += N(j,i)*rk[j] ;
    if(R[i].x()*R[i].x()<1e-10 && 
       R[i].y()*R[i].y()<1e-10 &&
       R[i].z()*R[i].z()<1e-10)
      return 0 ; 
  }

  // Solve      N^T*N*P = R

  // must check for the case where we want a curve of degree 1 having
  // only 2 points.
  if(n-2>0){ 
    Matrix_DOUBLE X(n-2,4),B(n-2,4),Ns(m-2,n-2) ;
    for(i=0;i<B.rows;i++){
      for(j=0;j<4;j++)
	B(i,j) = (double)R[i+1].data[j] ;
    }
    Ns = N.get(1,1,m-2,n-2) ;
    
    //LAPack.lse(transpose(Ns)*Ns,B,X) ;
    solve(transpose(Ns)*Ns,B,X) ;
    
    for(i=0;i<X.rows;i++){
      for(j=0;j<X.cols;j++)
	P[i+1].data[j] = (float)X(i,j) ;
      P[i+1].w() = 1.0 ;
    }
  }
  P[0] = Q[0] ;
  P[n-1] = Q[m-1] ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: getRemovalBnd --- Get the knot removal error bound for 
                                 an internal knot 
  Description: Get the knot removal error bound for an internal knot r 
               (non-rational). For more information on the algorithm, see 
	       A9.8 from the Nurbs book on page 428.
        Input: curve --> a NURBS curve
                   r --> the index of the internal knot to check
                   s --> the multiplicity of that knot
       Output: The maximum distance between the new curve and the old one
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlNurbsCurve::getRemovalBnd(int r, int s ) const{
  Vector_CtrlPnt4D temp(U.rows) ;
  int ord = degree+1 ;
  int last = r-s ;
  int first = r-degree ;
  int off ; 
  int i,j,ii,jj ;
  float alfi,alfj ;
  float u ;

  u = U[r] ;

  off = first-1;
  temp[0] = P[off] ;
  temp[last+1-off] = P[last+1] ;

  i=first ; j=last ;
  ii=1 ; jj=last-off ;

  while(j-i>0){
    alfi = (u-U[i])/(U[i+ord]-U[i]) ;
    alfj = (u-U[j])/(U[j+ord]-U[j]) ;
    temp[ii] = (P[i]-(1.0-alfi)*temp[ii-1])/alfi ; 
    temp[jj] = (P[j]-alfj*temp[jj+1])/(1.0-alfj) ;
    ++i ; ++ii ;
    --j ; --jj ;
  }
  if(j-i<0){
    return distance3D(temp[ii-1],temp[jj+1]) ;
  }
  else{
    alfi=(u-U[i])/(U[i+ord]-U[i]) ;
    return distance3D(P[i],alfi*temp[ii+1]+(1.0-alfi)*temp[ii-1]) ;
  }
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  removeKnot --- removes an internal knot from a curve.
  Description:  This is A5.8 on p185 from the NURB book modified to not check for tolerance before removing the knot.
        Input:  r --> the knot to remove
                s <-- the multiplicity of the knot
              num <-- the number of times to try to remove the knot
       Output:
 Restrictions: r {\bf must} be an internal knot.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::removeKnot(int  r, int s, int num)
{
  int m = U.n ;
  int ord = degree+1 ;
  int fout = (2*r-s-degree)/2 ;
  int last = r-s ;
  int first = r-degree ;
  float alfi, alfj ;
  int i,j,k,ii,jj,off ;
  float u ;

  Vector_CtrlPnt4D temp( 2*deg+1 ) ;

  u = U[r] ;

  if(num<1){
    PlError err("removeKnot");
    err << "A knot can only be removed a positive number of times!\n" ;
    err << "num = " << num << endl ;
    err.fatal() ;
  }

  int t;
  for(t=0;t<num;++t){
    off = first-1 ;
    temp[0] = P[off] ;
    temp[last+1-off] = P[last+1] ;
    i = first; j = last ;
    ii = 1 ; jj = last-off ;
    while(j-i > t){
      alfi = (u-U[i])/(U[i+ord+t]-U[i]) ;
      alfj = (u-U[j-t])/(U[j+ord]-U[j-t]) ;
      temp[ii] = (P[i]-(1.0-alfi)*temp[ii-1])/alfi ;
      temp[jj] = (P[j]-alfj*temp[jj+1])/(1.0-alfj) ;
      ++i ; ++ii ;
      --j ; --jj ;
    }
    i = first ; j = last ;
    while(j-i>t){
      P[i] = temp[i-off] ;
      P[j] = temp[j-off] ;
      ++i; --j ;
    }
    --first ; ++last ;
  }
  if(t==0) {
    cerr << "Major error happening... t==0\n" ;
    return ;
  }

  for(k=r+1; k<m ; ++k)
    U[k-t] = U[k] ;
  j = fout ; i=j ; // Pj thru Pi will be overwritten
  for(k=1; k<t; k++)
    if( (k%2) == 1)
      ++i ;
    else
      --j ;
  for(k=i+1; k<P.n ; k++) {// Shift
    P[j++] = P[k] ; 
  }

  resize(P.n-t,deg) ;

  return ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  removeKnotsBound --- Remove knots from a curve without exceeding an error bound
  Description: For more information about the algorithm, see A9.9 on p429 of the NURB book.
        Input: ub --> the knot coefficients
               ek --> the error after removing
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::removeKnotsBound(const Vector_FLOAT& ub,
				    Vector_FLOAT& ek, float E){
  Vector_FLOAT Br(U.n) ;
  Vector_INT S(U.n) ;
  Vector_INT Nl(U.n);
  Vector_INT Nr(U.n);
  Vector_FLOAT NewError(ub.n) ;
  Vector_FLOAT temp(ub.n) ;
  int i,BrMinI ;
  int r,s,Rstart,Rend,k ;
  float BrMin,u,Infinity=1e20 ;

  if(ek.n != ub.n){
    PlError err("removeKnotsBound");
    err << "Error in removeKnotsBoundCurve\n" ;
    err << "The size of ub and ek mismatch\n" ;
    err.fatal() ;
  }

  Br.reset(Infinity) ;
  S.reset(0) ;
  s = 1 ;
  for(i=deg+1;i<P.n;i++){
    if(U[i]<U[i+1]){
      Br[i] = getRemovalBnd(i,s) ;
      S[i] = s ;
      s = 1 ;
    }
    else {
      Br[i] = Infinity ;
      S[i] = 1 ;
      s++ ;
    }
  }

  
  // This might NOT be ok
  Nl[0] = 0 ;
  Nr.reset(ub.n-1) ;
  for(i=0;i<ub.n;i++){
    int span = findSpan(ub[i]);
    if(!Nl[span]){
      Nl[span] = i ;
    }
    if(i+1<ub.n)
      Nr[span] = i+1 ;
  }

  // set up N
  while(1) {
    BrMinI = Br.minIndex() ;
    BrMin = Br[BrMinI] ;

    if(BrMin == Infinity)
      break ;

    s = S[BrMinI] ;
    r = BrMinI ;

    // Verify the Error for the affected range
    Rstart = maximum(r-deg,deg+1) ;
    Rend = minimum(r+deg-S[r+deg]+1,P.n-1) ;
    Rstart = Nl[Rstart] ;
    Rend = Nr[Rend] ;

    int removable ;
    removable = 1 ;
    for(i=Rstart;i<=Rend ; i++){
      // Using eqn 9.81, 9.83 and A2.4 to compute NewError[i]
      float a ;
      s = S[r] ;
      if((deg+s)%2){
	u = ub[i] ;
	k = (deg+s+1)/2 ;
	a = U[r]-U[r-k+1] ;
	a /= U[r-k+deg+2]-U[r-k+1] ;
	NewError[i] = (1.0-a)*Br[r] * basisFun(u,r-k+1);
      }
      else{
	u = ub[i] ;
	k = (deg+s)/2 ;
	NewError[i] = Br[r] * basisFun(u,r-k);
      }
      temp[i] = NewError[i] + ek[i];
      if(temp[i]>E){
	removable = 0 ;
	Br[r] = Infinity ;
	break ;
      }
    }
    if(removable){
      // Remove the knot
      removeKnot(r,S[r],1) ;

      // update the error
      for(i=Rstart; i<=Rend ; i++)
	ek[i] = temp[i] ;

      // Break if there is no more interior knot
      if(P.n<=deg+1){
	break ;
      }

      // Update the new index range for some of the knots 
      Rstart = Nl[r-deg-1] ;
      Rend = Nr[r-S[r]] ;
      int span, oldspan ;
      oldspan = -1 ;
      for(k=Rstart;k<=Rend;k++){
	span = findSpan(ub[k]);
	if(span != oldspan){
	  Nl[span] = k ;
	}
	if(k+1<ub.n) 
	  Nr[span] = k+1 ; 
	oldspan = span ;
      }
      for(k=r-S[r]+1;k<Nl.n-1;k++){ 
      	Nl[k] = Nl[k+1] ; 
       	Nr[k] = Nr[k+1] ; 
      } 
      Nl.resize(Nl.n-1) ; // Shrink Nl 
      Nr.resize(Nr.n-1) ; // Shrink Nr
      // Update the new error bounds for some of the knots
      Rstart = maximum(r-deg,deg+1) ;
      Rend = minimum(r+deg-S[r]+1,P.n) ;
      s = S[Rstart] ;
      for(i=Rstart;i<=Rend;i++){
	if(U[i]<U[i+1]){
	  Br[i] = getRemovalBnd(i,s) ;
	  S[i] = s ;
	  s = 1;
	}
	else {
	  Br[i] = Infinity ;
	  S[i] = 1 ;
	  s++ ;
	}
      }
      for(i=Rend+1;i<Br.n-1;i++){
      	Br[i] = Br[i+1] ;
	S[i] = S[i+1] ;
      }
      Br.resize(Br.n-1) ; // Shrink Br
    }
    else{
      Br[r] = Infinity ;
    }
    
  }

} 
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: chordLengthParam --- chord length parameterization
  Description: Performs chord length parameterization from a vector of points.
               The chord length parameterization works as follows:
	       \begin{itemize} 
	       \item The total chord length is defined as
	             \begin{equation} d = \sum_{k=1}^{n-1} | Q_k-Q_{k-1}| 
		     \end{equation}
	       \item $\bar{u}_0 = 0$ and $\bar{u}_{n-1} = 1$
	       \item \begin{equation}
	              \bar{u}_k = \bar{u}_{k-1}+\frac{| Q_k-Q_{k-1}| }{d}
		      \hspace{0.5in} k=1,\ldots,n-2
	             \end{equation}
	       \end{itemize}
	       Where $n$ is the size of the vector $Q$.
        Input:  Q --> a vector of 3D points 
               ub <-- the result of chord length parameterization
       Output: the total chord length of the points.
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float chordLengthParam(const Vector_Point3D& Q, Vector_FLOAT &ub){
  int i ;
  float d = 0.0 ;

  ub.resize(Q.n) ;
  ub[0] = 0 ; 
  for(i=1;i<ub.n;i++){
    d += abs(Q[i]-Q[i-1]) ;
  }
  if(d>0){
    for(i=1;i<ub.n-1;++i){
      ub[i] = ub[i-1] + abs(Q[i]-Q[i-1])/d ;
    }
    ub[ub.n-1] = 1.0 ; // In case there is some addition round-off
  }
  else{
    for(i=1;i<ub.n-1;++i)
      ub[i] = float(i)/float(ub.n-1) ;
    ub[ub.n-1] = 1.0 ;
  }
  return d ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: chordLengthParam --- chord length parameterization
  Description: Performs chord length parameterization from a vector of points.
               The chord length parameterization works as follows:
	       \begin{itemize} 
	       \item The total chord length is defined as
	             \begin{equation} d = \sum_{k=1}^{n-1} | Q_k-Q_{k-1}| 
		     \end{equation}
	       \item $\bar{u}_0 = 0$ and $\bar{u}_{n-1} = 1$
	       \item \begin{equation}
	              \bar{u}_k = \bar{u}_{k-1}+\frac{| Q_k-Q_{k-1}| }{d}
		      \hspace{0.5in} k=1,\ldots,n-2
	             \end{equation}
	       \end{itemize}
	       Where $n$ is the size of the vector $Q$.
        Input:  Q --> a vector of 4D points 
               ub <-- the result of chord length parameterization
       Output: the total chord length of the points.
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float chordLengthParam(const Vector_CtrlPnt4D& Q, Vector_FLOAT &ub){
  int i ;
  float d = 0.0 ;

  ub.resize(Q.n) ;
  ub[0] = 0 ; 
  for(i=1;i<ub.n;i++){
    d += abs(Q[i]-Q[i-1]) ;
  }
  for(i=1;i<ub.n-1;i++){
    ub[i] = ub[i-1] + abs(Q[i]-Q[i-1])/d ;
  }
  ub[ub.n-1] = 1.0 ; // In case there is some addition round-off
  return d ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalApproxErrBnd --- Approximation of a curve bounded 
                                      to a certain error
  Description: It is a type II approximation: it starts with a lot of control
	       points then tries to eliminate as much as it can as long as
	       the curve stays within a certain error bound.

	       The method uses least squares fitting along with knot
	       removal techniques. It is the algorithm A9.10 on p 431 of 
	       the NURBS book.
               
        Input:  Q --> the points to approximate
	   degree --> the degree of the approximation curve
	        E --> the maximum error allowed
       Output:
 Restrictions: 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by: Philippe lavoie (17 September 1997)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalApproxErrBnd(Vector_Point3D& Q, int degC, float E){
  Vector_FLOAT ub(Q.n) ;
  chordLengthParam(Q,ub) ;

  globalApproxErrBnd(Q,ub,degC,E) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalApproxErrBnd --- Approximation of a curve bounded 
                                      to a certain error
  Description: It is a type II approximation: it starts with a lot of control
	       points then tries to eliminate as much as it can as long as
	       the curve stays within a certain error bound.

	       The method uses least squares fitting along with knot
	       removal techniques. It is the algorithm A9.10 on p 431 of 
	       the NURBS book.
               
        Input:  Q --> the points to approximate
	       ub <-> the vector of parameters where the points are located
	   degree --> the degree of the approximation curve
	        E --> the maximum error allowed
       Output:
 Restrictions: ub and Q must be of the same size
   Created by: Philippe Lavoie (24 January 1997)
  Modified by: Philippe lavoie (17 September 1997)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalApproxErrBnd(Vector_Point3D& Q, Vector_FLOAT& ub, int degC, float E){
  Vector_FLOAT ek(Q.n) ;
  Vector_FLOAT Uh(Q.n) ;
  PlNurbsCurve tcurve ;
  int i,j,degL ; 

  if(ub.n != Q.n){
    PlError err("globalApproxErrBnd");
    err << "The data vector and the parameter vectors are not of the same size!\n" ;
    err << "Q.n = " << Q.n << ", ub.n = " << ub.n << endl ;
    err.fatal() ;
  }

  resize(Q.n,1) ;

  // Initialize U
  deg = 1 ;
  for(i=0;i<ub.n;i++){
    U[i+deg] = ub[i] ;
  }
  U[0] = 0 ;
  U[U.n-1] = 1.0 ;
  // Initialize P
  for(i=0;i<P.n;i++)
    P[i] = Q[i] ;

  for(degL=1; degL<=degC+1 ; degL++){
    removeKnotsBound(ub,ek,E) ;

    if(degL==degC)
      break ;

    if(degL<degC){

      // Find the degree elevated knot vector
      Uh.resize(U.n*2) ;

      Uh[0] = U[0] ;
      j = 1 ;
      for(i=1;i<U.n;++i){
	if(U[i]>U[i-1])
	  Uh[j++] = U[i-1] ;
	Uh[j++] = U[i] ;
      }
      Uh[j++] = U[U.n-1] ;
      Uh.resize(j) ;
      tcurve = *this ;
      if(!leastSquares(Q,degL+1,Uh.n-degL-1-1,ub,Uh)){
	*this = tcurve ;
	degreeElevate(1);
      }
    }
    else{
      tcurve = *this ;
      if(!leastSquares(Q,degL,P.n,ub,U)){
	*this = tcurve ;
      }
    }


    // Project the points from curve to Q and update ek and ub
    //    for(i=0;i<Q.n;i++){
    for(i=0;i<Q.n;i++){
      float u_i ;
      Point3D r_i ;
      projectTo(Q[i],ub[i],u_i,r_i) ;
      ek[i] = abs(r_i-Q[i]) ;
      ub[i] = u_i ;
    }
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalApproxErrBnd2 --- Approximation of a curve bounded to 
                                       a certain error
  Description: The algorithm is quite simplistic but in some cases gives
               better result than globalApproxErrBnd (when the error allowed
	       is low and the data is very close to each other).

	       This algorithm generates a first degree interpolation of the
	       data points, degree elevates the curve by 1 degree recomputes
	       the error around each points, removes the knots which are 
	       within a certain error range then repeats the process until
	       the desired degree is reached.
        Input:  Q --> the points to approximate
	     degC --> the degree of the approximation curve
	        E --> the maximum error allowed
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalApproxErrBnd2(Vector_Point3D& Q,
				       int degC,
				       float E){
  Vector_FLOAT ub(Q.n) ;
  Vector_FLOAT ek(Q.n) ;
  Vector_FLOAT Uh(Q.n) ;
  PlNurbsCurve tcurve ;
  int i,degL ; 

  resize(Q.n,1) ;

  chordLengthParam(Q,ub) ;

  // Initialize U
  deg = 1 ;
  for(i=0;i<ub.n;i++){
    U[i+deg] = ub[i] ;
  }
  U[0] = 0 ;
  U[U.n-1] = 1.0 ;
  // Initialize P
  for(i=0;i<P.n;i++)
    P[i] = Q[i] ;

  for(degL=1; degL<degC ; degL++){
    degreeElevate(1);

    // Project the points from curve to Q and update ek and ub
    //    for(i=0;i<Q.n;i++){
    for(i=0;i<Q.n;i++){
      float u_i ;
      Point3D r_i ;
      projectTo(Q[i],ub[i],u_i,r_i) ;
      ek[i] = abs(r_i-Q[i]) ;
      ub[i] = u_i ;
    }
    removeKnotsBound(ub,ek,E) ;    
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalApproxErrBnd3 --- Approximation of a curve bounded 
                                       to a certain error
  Description: The algorithm is quite simplistic but in some cases gives
               better result than globalApproxErrBnd (when the error allowed
	       is low and the data is very close to each other).

	       This algorithm generates a first degree interpolation of the
	       data points, degree elevates it to the degree requested then
	       removes all the control points which are within the error
	       bound.
        Input:  Q --> the points to approximate
	     degC --> the degree of the approximation curve
	        E --> the maximum error allowed
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalApproxErrBnd3(Vector_Point3D& Q,int degC,float E){
  //PlNurbsCurve tCurve(1) ;
  Vector_FLOAT ub(Q.n) ;
  Vector_FLOAT ek(Q.n) ;
  int i ; 

  resize(Q.n,1) ;

  chordLengthParam(Q,ub) ;

  // Initialize U
  deg = 1 ;
  for(i=0;i<ub.n;i++){
    U[i+deg] = ub[i] ;
  }
  U[0] = 0 ;
  U[U.n-1] = 1.0 ;

  // Initialize P
  for(i=0;i<P.n;i++)
    P[i] = Q[i] ;
  
  //removeKnotsBoundCurve(curve,ub,ek,E/10.0,curve) ;
  degreeElevate(degC-1) ;
  removeKnotsBound(ub,ek,E) ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalApproxErrBnd3 --- Approximation of a curve bounded 
                                       to a certain error
  Description: The algorithm is quite simplistic but in some cases gives
               better result than globalApproxErrBnd (when the error allowed
	       is low and the data is very close to each other).

	       This algorithm generates a first degree interpolation of the
	       data points, degree elevates it to the degree requested then
	       removes all the control points which are within the error
	       bound.
        Input:  Q --> the points to approximate
	       ub --> the parameter vector
	     degC --> the degree of the approximation curve
	        E --> the maximum error allowed
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalApproxErrBnd3(Vector_Point3D& Q, 
				       const Vector_FLOAT &ub,
				       int degC,
				       float E){
  Vector_FLOAT ek(Q.n) ;
  int i ; 

  resize(Q.n,1) ;

  // Initialize U
  deg = 1 ;
  for(i=0;i<ub.n;i++){
    U[i+deg] = ub[i] ;
  }
  U[0] = 0 ;
  U[U.n-1] = 1.0 ;

  // Initialize P
  for(i=0;i<P.n;i++)
    P[i] = Q[i] ;
    
  degreeElevate(degC-1) ;
  removeKnotsBound(ub,ek,E) ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: projectTo --- project a point onto the curve
  Description: It finds the closest point in the curve to a point $p$.
               For more information, see A6.4 and A6.5 on p 231 of the 
	       NURBS book
        Input: 
                p --> the point $p$ being projected
            guess --> initial guess
                u <-- the parametric value for which $C(u)$ is closest to $p$.
                r <-- the point at $C(u)$
               e1 --> the minimal error
	       e2 --> the maximal error
           maxTry --> the maximum number of times to try before returning from the function
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::projectTo(const Point3D& p, float guess, float& u, Point3D& r, float e1, float e2,int maxTry) const{
  float un ;
  float c1, c2;
  Vector_Point3D Cd ;
  Point3D c, cd,cdd ;
  int t = 0 ;
  u = guess ;

  if(u<U[0]) u = U[0] ;
  if(u>U[U.n-1]) u = U[U.n-1] ;

  while(1) {
    ++t ;
    if(t>maxTry){
      r = c ;
      return ;
    }
    c = point3D(u) ;
    Cder(u,*this,2,Cd) ;
    cd = Cd[1] ;
    cdd = Cd[2] ;
    c1 = abs2(c-p) ;
    if(c1<e1*e1){
      r = c ;
      return ;
    }
    c2 = abs(cd*(c-p)) ;
    //c2 *= c2 ;
    c2 /= abs(cd)*abs(c-p) ;
    //if(c2<e2*e2){
    if(c2<e2){
      r =c ;
      return ;
    }
    
    un = u- cd*(c-p)/(cdd*(c-p)+abs2(cd)) ;
    
    if(un<U[0]) un = U[0] ;
    if(un>U[U.n-1]) un = U[U.n-1] ;

    if(abs2((un-u)*cd)<e1*e1){
      r = c ;
      return ;
    }
    u = un ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: degreeElevate  --- degree elevate a curve a number of times
  Description: For more information, see A5.9 on p 206 of the NURBS book
        Input: t --> the number of times to increase the degree of the curve
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::degreeElevate(int t){
  if(t<=0){
    return ;
  }

  PlNurbsCurve c(*this) ;

  int i,j,k ;
  int n = c.CtrlP.n-1;
  int p = c.degree ;
  int m = n+p+1;
  int ph = p+t ;
  int ph2 = ph/2 ;
  Matrix_FLOAT bezalfs(p+t+1,p+1) ; // coefficients for degree elevating the Bezier segment
  Vector_CtrlPnt4D bpts(p+1) ; // pth-degree Bezier control points of the current segment
  Vector_CtrlPnt4D ebpts(p+t+1) ; // (p+t)th-degree Bezier control points of the  current segment
  Vector_CtrlPnt4D Nextbpts(p-1) ; // leftmost control points of the next Bezier segment
  Vector_FLOAT alphas(p-1) ; // knot instertion alphas.

  // Compute the binomial coefficients
  Matrix_FLOAT Bin(ph+1,ph2+1) ;
  binomialCoef(Bin) ;

  // Compute Bezier degree elevation coefficients
  float inv,mpi ;
  bezalfs(0,0) = bezalfs(ph,p) = 1.0 ;
  for(i=1;i<=ph2;i++){
    inv= 1.0/Bin(ph,i) ;
    mpi = minimum(p,i) ;
    for(j=maximum(0,i-t); j<=mpi; j++){
      bezalfs(i,j) = inv*Bin(p,j)*Bin(t,i-j) ;
    }
  }

  for(i=ph2+1;i<ph ; i++){
    mpi = minimum(p,i) ;
    for(j=maximum(0,i-t); j<=mpi ; j++)
      bezalfs(i,j) = bezalfs(ph-i,p-j) ;
  }

  resize(c.P.n+c.P.n*t,ph) ; // Allocate more control points than necessary

  int mh = ph ;
  int kind = ph+1 ;
  float ua = c.U[0] ;
  float ub = 0.0 ;
  int r=-1 ; 
  int oldr ;
  int a = p ;
  int b = p+1 ; 
  int cind = 1 ;
  int rbz,lbz = 1 ; 
  int mul,save,s;
  float alf;
  int first, last, kj ;
  float den,bet,gam,numer ;

  P[0] = c.P[0] ;
  for(i=0; i <= ph ; i++){
    U[i] = ua ;
  }

  // Initialize the first Bezier segment

  for(i=0;i<=p ;i++) 
    bpts[i] = c.P[i] ;

  while(b<m){ // Big loop thru knot vector
    i=b ;
    while(b<m && c.U[b] >= c.U[b+1]) // for some odd reasons... == doesn't work
      b++ ;
    mul = b-i+1 ; 
    mh += mul+t ;
    ub = c.U[b] ;
    oldr = r ;
    r = p-mul ;
    if(oldr>0)
      lbz = (oldr+2)/2 ;
    else
      lbz = 1 ;
    if(r>0) 
      rbz = ph-(r+1)/2 ;
    else
      rbz = ph ;
    if(r>0){ // Insert knot to get Bezier segment
      numer = ub-ua ;
      for(k=p;k>mul;k--){
	alphas[k-mul-1] = numer/(c.U[a+k]-ua) ;
      }
      for(j=1;j<=r;j++){
	save = r-j ; s = mul+j ;
	for(k=p;k>=s;k--){
	  bpts[k] = alphas[k-s] * bpts[k]+(1.0-alphas[k-s])*bpts[k-1] ;
	}
	Nextbpts[save] = bpts[p] ;
      }
    }
    
    for(i=lbz;i<=ph;i++){ // Degree elevate Bezier,  only the points lbz,...,ph are used
      ebpts[i] = 0.0 ;
      mpi = minimum(p,i) ;
      for(j=maximum(0,i-t); j<=mpi ; j++)
	ebpts[i] += bezalfs(i,j)*bpts[j] ;
    }

    if(oldr>1){ // Must remove knot u=c.U[a] oldr times
      // if(oldr>2) // Alphas on the right do not change
      //	alfj = (ua-U[kind-1])/(ub-U[kind-1]) ;
      first = kind-2 ; last = kind ;
      den = ub-ua ;
      bet = (ub-U[kind-1])/den ;
      for(int tr=1; tr<oldr; tr++){ // Knot removal loop
	i = first ; j = last ;
	kj = j-kind+1 ;
	while(j-i>tr){ // Loop and compute the new control points for one removal step
	  if(i<cind){
	    alf=(ub-U[i])/(ua-U[i]) ;
	    P[i] = alf*P[i] + (1.0-alf)*P[i-1] ;
	  }
	  if( j>= lbz){
	    if(j-tr <= kind-ph+oldr){
	      gam = (ub-U[j-tr])/den ;
	      ebpts[kj] = gam*ebpts[kj] + (1.0-gam)*ebpts[kj+1] ;
	    }
	    else{
	      ebpts[kj] = bet*ebpts[kj]+(1.0-bet)*ebpts[kj+1] ;
	    }
	  }
	  ++i ; --j; --kj ;
	}
	--first ; ++last ;
      }
    }

    if(a!=p) // load the knot u=c.U[a]
      for(i=0;i<ph-oldr; i++){
	U[kind++] = ua ; 
      }
    for(j=lbz; j<=rbz ; j++) { // load control points onto the curve
      P[cind++] = ebpts[j] ; 
    }

    if(b<m){ // Set up for next pass thru loop
      for(j=0;j<r;j++)
	bpts[j] = Nextbpts[j] ;
      for(j=r;j<=p;j++)
	bpts[j] = c.P[b-p+j] ;
      a=b ; 
      b++ ;
      ua = ub ;
    }
    else{
      for(i=0;i<=ph;i++)
	U[kind+i] = ub ;
    }
  }
  resize(mh-ph,ph) ; // Resize to the proper number of control points
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: knotInsertion ---  it inserts a knot a number of times
  Description: It inserts the knot u, r times and generates the curve nc.
               For more information, see A5.1 on page 151 of the NURBS book
        Input: u --> the knot to insert
               r --> the number of times to insert $u$.
              nc <-- the resulting NURBS curve
       Output:
 Restrictions: c and nc must be different; {\em i.e.} don't call it with 
               {\tt curveKnotInsertion(c,0.1,1,c) ; }
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::knotInsertion(float u, int r,PlNurbsCurve& nc){
  // Compute k and s      u = [ u_k , u_k+1)  with u_k having multiplicity s
  int k=0,s=0 ;
  int i,j ;
  int p = degree ;
  
  for(i=0;i<U.n;i++){
    if(U[i]>u) {
      k = i-1 ;
      break ;
    }
  }
  if(u<=U[k]){
    s = 1 ;
    for(i=k;i>deg;i--){
      if(U[i]<=U[i-1])
	s++ ;
      else
	break ;
    }
  }
  else{
    s=0 ;
  }

  if((r+s)>p){
    PlError err("knotInsertion");
    err << "In Knot insertion...\n" ;
    err << "The number of insertion required will make the number of knots higher\n";
    err << "than the degree of the curve at the knot " <<  u <<  ".\n" ;
    err.fatal();
  }

  nc.resize(P.n+r,deg) ;

  // Load new knot vector
  for(i=0;i<=k;i++)
    nc.U[i] = U[i] ;
  for(i=1;i<=r;i++)
    nc.U[k+i] = u ;
  for(i=k+1;i<U.n; i++)
    nc.U[i+r] = U[i] ;

  // Save unaltered control points
  Vector_CtrlPnt4D R(p+1) ;

  for(i=0; i<=k-p ; i++)
    nc.P[i] = P[i] ;
  for(i=k-s ; i< P.n ; i++)
    nc.P[i+r] = P[i] ;
  for(i=0; i<=p-s; i++)
    R[i] = P[k-p+i] ;

  // Insert the knot r times
  int L=0 ;
  float alpha ;
  for(j=1; j<=r ; j++){
    L = k-p+j ;
    for(i=0;i<=p-j-s;i++){
      alpha = (u-U[L+i])/(U[i+k+1]-U[L+i]) ;
      R[i] = alpha*R[i+1] + (1.0-alpha)*R[i] ;
    }
    nc.P[L] = R[0] ;
    nc.P[k+r-j-s] = R[p-j-s] ;
  }

  // Load remaining control points
  for(i=L+1; i<k-s ; i++){
    nc.P[i] = R[i-L] ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: refineKnotVector --- refine the curve knot vector
  Description: For more information, see A5.4 on page 164 of the NURBS book
        Input:      
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::refineKnotVector(const Vector_FLOAT& X){
  int n = P.n-1 ;
  int p = deg ;
  int m = n+p+1 ;
  int a,b ;
  int r = X.n-1 ;

  PlNurbsCurve c(*this) ;

  resize(r+1+n+1,p) ;

  a = c.findSpan(X[0]) ;
  b = c.findSpan(X[r]) ;
  ++b ;
  int j ;
  for(j=0; j<=a-p ; j++)
    P[j] = c.P[j] ;
  for(j = b-1 ; j<=n ; j++)
    P[j+r+1] = c.P[j] ;
  for(j=0; j<=a ; j++)
    U[j] = c.U[j] ;
  for(j=b+p ; j<=m ; j++)
    U[j+r+1] = c.U[j] ;
  int i = b+p-1 ; 
  int k = b+p+r ;
  for(j=r; j>=0 ; j--){
    while(X[j] <= c.U[i] && i>a){
      P[k-p-1] = c.P[i-p-1] ;
      U[k] = c.U[i] ;
      --k ;
      --i ;
    }
    P[k-p-1] = P[k-p] ;
    for(int l=1; l<=p ; l++){
      int ind = k-p+l ;
      float alpha = U[k+l] - X[j] ;
      if(alpha==0.0)
	P[ind-1] = P[ind] ;
      else
	alpha /= U[k+l]-c.U[i-p+l] ;
      P[ind-1] = alpha*P[ind-1] + (1.0-alpha)*P[ind] ;
    }
    U[k] = X[j] ;
    --k ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalInterp --- global curve interpolation with points in 3D
  Description: Global curve interpolation with points in 3D.
        Input:   Q --> the 3D points to interpolate
	         d --> the degree of the interpolation
       Output:
 Restrictions: The number of points to interpolate must be greater than
               the degree specified for the curve.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalInterp(const Vector_Point3D& Q, int d){
  Vector_FLOAT ub ;
  chordLengthParam(Q,ub) ;
  globalInterp(Q,ub,d) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalInterp --- global curve interpolation with points in 3D
  Description: Global curve interpolation with points in 3D and with the parametric values
               specified.
        Input:   Q --> the 3D points to interpolate
		ub --> the parametric values 
	         d --> the degree of the interpolation
       Output:
 Restrictions: The number of points to interpolate must be greater than
               the degree specified for the curve.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalInterp(const Vector_Point3D& Q, const Vector_FLOAT& ub, int d){
  int i,j ;

  if(d<=0){
    PlError err("globalInterp");
    err << "The degree specified is equal or smaller than 0\n" ;
    err << "deg = " << deg << endl ;
    err.fatal() ;
  }
  if(d>=Q.n){
      PlError err("globalInterp");
      err << "The degree specified is greater then Q.n+1\n" ;
      err << "Q.n = " << Q.n << ", and deg = " << deg << endl ;
      err.warning() ;
      d = Q.n-1 ;
  }


  resize(Q.n,d) ;
  Matrix_DOUBLE A(Q.n,Q.n) ;

  knotAveraging(ub,d,U) ;

  // Initialize the basis matrix A
  Vector_FLOAT N(deg+1) ;

  for(i=1;i<Q.n-1;i++){
    int span = findSpan(ub[i]);
    basisFuns(ub[i],span,N) ;
    for(j=0;j<=deg;j++) 
	A(i,span-deg+j) = (double)N[j] ;
  }
  A(0,0)  = 1.0 ;
  A(Q.n-1,Q.n-1) = 1.0 ;

  // Init matrix for LSE
  Matrix_DOUBLE qq(Q.n,3) ;
  Matrix_DOUBLE xx(Q.n,3) ;
  for(i=0;i<Q.n;i++){
    const Point3D& qp = Q[i] ; // this makes the SGI compiler happy
    for(j=0; j<3;j++)
      qq(i,j) = (double)qp.data[j] ;
  }

  //LAPack.lse(A,qq,xx) ;
  solve(A,qq,xx) ;

  // Store the data
  for(i=0;i<xx.rows;i++){
    for(j=0;j<3;j++)
      P[i].data[j] = (float)xx(i,j) ;
    P[i].w() = 1.0 ;
  }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalInterpD --- global curve interpolation with the 
                                 1st derivatives of the points specified.
  Description: Global curve interpolation with the 1st degree derivative
               specified for each of the points to interpolate.
	       This will generate a number of control points 2 times greater 
	       than the number of interpolation points.

	       If the derivative specified is of unit length, {\em i.e.} the
	       tangent vectors are given, then the derivative at each point
	       will be multiplied by the chord length. A second multiplicative
	       factor can be specified to make the derivative even greater.
	       This second number should be close to 1.0. 

	       For more information about the algorithm, refer to section
	       9.2.4 on page 373 of the NURBS book.
        Input: Q --> the points to interpolate
	       D --> the first derivative at these points
               d --> the degree of the curve
	   unitD --> a flag specifying if the derivatives are unit vectors
	       a --> a multiplicative factor for the tangent (must be greater 
	             than 0 or the function aborts).
       Output:
 Restrictions: The number of points to interpolate must be greater than
               the degree specified for the curve. 
   Created by: Philippe Lavoie          (3 September, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalInterpD(const Vector_Point3D& Q, const Vector_Point3D& D, int d, int unitD, float a){
  int i,j,n ;

  if(d<=0){
    PlError err("globalInterpD");
    err << "The degree specified is equal or smaller than 0\n" ;
    err << "deg = " << deg << endl ;
    err.fatal() ;
  }
  if(d>=Q.n){
      PlError err("globalInterpD");
      err << "The degree specified is greater then Q.n+1\n" ;
      err << "Q.n = " << Q.n << ", and deg = " << deg << endl ;
      err.warning() ;
      d = Q.n-1 ;
  }
  if(a<=0){
    PlError err("globalInterpD");
    err << "The a value must be greater than 0\n" ;
    err << "It is presently " << a << endl ;
    err.fatal() ;
  }
  if(Q.n != D.n){
    PlError err("globalInterpD") ;
    err << "The number of points to interpolate is different than\n the number of derivative points.\n" ;
    err << "Q.n = " << Q.n << ", D.n = " << D.n << endl ;
    err.fatal() ;
  }

  deg = d ;
  n = 2*Q.n ; 

  resize(n,deg) ;

  Vector_FLOAT ub(Q.n) ;
  Vector_FLOAT ub2(2*Q.n) ;

  float chordLength ;

  chordLength = chordLengthParam(Q,ub) ;

  if(unitD)
    chordLength *= a ;

  // setup ub2 
  for(i=0;i<ub.n-1;++i){
    ub2[2*i] = ub[i] ;
    ub2[2*i+1] = (ub[i]+ub[i+1])/2.0 ;
  }
  ub2[ub2.n-1] = ub[ub.n-1] ;
  ub2[ub2.n-2] = (ub2[ub2.n-1]+ub2[ub2.n-3])/2.0 ;

  knotAveraging(ub2,deg,U) ;

  // Initialize the basis matrix A
  Matrix_DOUBLE A(n,n) ;
  Vector_FLOAT N(deg+1) ;
  Matrix_FLOAT Nd(1,1) ;

  for(i=1;i<Q.n-1;i++){
    int span = findSpan(ub[i]);
    basisFuns(ub[i],span,N) ;
    dersBasisFuns(1,ub[i],span,Nd) ;
    for(j=0;j<=deg;j++) {
	A(2*i,span-deg+j) = (double)N[j] ;
	A(2*i+1,span-deg+j) = (double)Nd(1,j) ;
    }
  }
  A(0,0)  = 1.0 ;
  A(1,0) = -1.0 ;
  A(1,1) = 1.0 ;
  A(A.rows-2,A.cols-2) = -1.0 ;
  A(A.rows-2,A.cols-1) = 1.0 ;
  A(A.rows-1,A.cols-1) = 1.0 ;

  // Init matrix for LSE
  Matrix_DOUBLE qq(n,3) ;
  Matrix_DOUBLE xx(n,3) ;
  for(i=0;i<Q.n;i++){
    const Point3D& qp = Q[i] ; // this makes the SGI compiler happy
    const Point3D& dp = D[i] ;
    for(j=0; j<3;j++){
      qq(2*i,j) = (double)qp.data[j] ;
      qq(2*i+1,j) = (double)dp.data[j] ;
      if(unitD)
	qq(2*i+1,j) *= chordLength ;
    }
  }

  float d0Factor = U[deg+1]/deg ;
  float dnFactor = (1-U[U.n-deg-2])/deg ;

  // the following three assignment must be made to make the SGI compiler
  // a happy compiler
  const Point3D& dp0 = D[0] ;
  const Point3D& dpn = D[D.n-1] ;
  const Point3D& qpn = Q[Q.n-1] ;
  for(j=0;j<3;++j){
    qq(1,j) = d0Factor*double(dp0.data[j]) ;
    qq(A.cols-2,j) = dnFactor*double(dpn.data[j]) ;
    if(unitD){
      qq(1,j) *= chordLength ;
      qq(A.cols-2,j) *= chordLength ;
    }
    qq(A.cols-1,j) = double(qpn.data[j]) ;
  }

  //LAPack.lse(A,qq,xx) ;
  solve(A,qq,xx) ;

  // Store the data
  for(i=0;i<xx.rows;i++){
    for(j=0;j<3;j++)
      P[i].data[j] = (float)xx(i,j) ;
    P[i].w() = 1.0 ;
  }

  P[0] = Q[0] ;
  P[P.n-1] = Q[Q.n-1] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalInterp   --- global curve interpolation with points in 4D
  Description: Global curve interpolation with points in 4D
        Input: Q --> the points in 4D to interpolate
	       d --> the degree of the curve interpolation
       Output:
 Restrictions: The number of points to interpolate must be greater than
               the degree specified for the curve.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalInterp(const Vector_CtrlPnt4D& Q, int d){
  int i,j ;

  resize(Q.n,d) ;
  Matrix_DOUBLE A(Q.n,Q.n) ;
  Vector_FLOAT ub(Q.n) ;

  chordLengthParam(Q,ub) ;

  // Setup the Knot Vector for the curve
  for(i=0;i<=deg;i++)
    U[i] = 0 ;
  for(i=P.n; i<U.n; i++)
    U[i] = 1.0 ;
  for(j=1; j<Q.n-deg ; j++){
    float t=0 ;
    for(i=j; i< j+deg ; i++)
      t += ub[i] ;
    U[j+deg] = t/(float)deg ;
  }
  
  // Initialize the basis matrix A
  Vector_FLOAT N(deg+1) ;

  for(i=1;i<Q.n-1;i++){
    int span = findSpan(ub[i]);
    basisFuns(ub[i],span,N) ;
    for(j=0;j<=deg;j++) 
	A(i,span-deg+j) = (double)N[j] ;
  }
  A(0,0)  = 1.0 ;
  A(Q.n-1,Q.n-1) = 1.0 ;

  // Init matrix for LSE
  Matrix_DOUBLE qq(Q.n,4) ;
  Matrix_DOUBLE xx(Q.n,4) ;
  for(i=0;i<Q.n;i++)
    for(j=0; j<4;j++)
      qq(i,j) = (double)Q[i].data[j] ;

  //LAPack.lse(A,qq,xx) ;
  solve(A,qq,xx) ;

  // Store the data
  for(i=0;i<xx.rows;i++){
    for(j=0;j<4;j++)
      P[i].data[j] = (float)xx(i,j) ;
  }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalInterp --- global curve interpolation with 4D points 
                                and a knot vector defined.
  Description: Global curve interpolation with 4D points and a knot vector 
               defined.
        Input:   Q --> the 3D points to interpolate
	        Uc --> the knot vector to set the curve to
	         d --> the degree of the interpolation   
       Output:
 Restrictions: The number of points to interpolate must be greater than
               the degree specified for the curve.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalInterp(const Vector_CtrlPnt4D& Q, const Vector_FLOAT& Uc, int d){
  int i,j ;

  resize(Q.n,d) ;
  Matrix_DOUBLE A(Q.n,Q.n) ;
  Vector_FLOAT ub(Q.n) ;

  if(Uc.n != U.n){
    PlError err("globalInterp");
    err << "Invalid dimension for the given Knot vector.\n" ;
    err << "U required = " << U.n << ", U given = " << Uc.n << endl ;
    err.fatal() ;
  }
  U = Uc ;
  chordLengthParam(Q,ub) ;
  
  // Initialize the basis matrix A
  Vector_FLOAT N(deg+1) ;

  for(i=1;i<Q.n-1;i++){
    int span = findSpan(ub[i]);
    basisFuns(ub[i],span,N) ;
    for(j=0;j<=deg;j++) 
	A(i,span-deg+j) = (double)N[j] ;
  }
  A(0,0)  = 1.0 ;
  A(Q.n-1,Q.n-1) = 1.0 ;

  // Init matrix for LSE
  Matrix_DOUBLE qq(Q.n,4) ;
  Matrix_DOUBLE xx(Q.n,4) ;
  for(i=0;i<Q.n;i++)
    for(j=0; j<4;j++)
      qq(i,j) = (double)Q[i].data[j] ;

  //LAPack.lse(A,qq,xx) ;
  solve(A,qq,xx) ;

  // Store the data
  for(i=0;i<xx.rows;i++){
    for(j=0;j<4;j++)
      P[i].data[j] = (float)xx(i,j) ;
  }

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: globalInterp --- global curve interpolation with 4D points, 
                                a knot vector defined and the parametric value
				vector defined.
  Description: Global curve interpolation with 4D points, a knot vector
               defined and the parametric value vector defined.
        Input:   Q --> the 3D points to interpolate
	        ub --> the parametric values vector
		Uc --> the knot vector to set the curve to
	         d --> the degree of the interpolation  		 
       Output: 
 Restrictions: The number of points to interpolate must be greater than
               the degree specified for the curve. Uc must be compatible with 
	       the values given for Q.n, ub.n and d.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::globalInterp(const Vector_CtrlPnt4D& Q, const Vector_FLOAT& ub, const Vector_FLOAT& Uc, int d){
  int i,j ;

  resize(Q.n,d) ;
  Matrix_DOUBLE A(Q.n,Q.n) ;

  if(Uc.n != U.n){
    PlError err("globalInterp");
    err << "Invalid dimension for the given Knot vector.\n" ;
    err << "U required = " << U.n << ", U given = " << Uc.n << endl ;
    err.fatal() ;
  }
  U = Uc ;
  
  // Initialize the basis matrix A
  Vector_FLOAT N(deg+1) ;

  for(i=1;i<Q.n-1;i++){
    int span = findSpan(ub[i]);
    basisFuns(ub[i],span,N) ;
    for(j=0;j<=deg;j++) 
	A(i,span-deg+j) = (double)N[j] ;
  }
  A(0,0)  = 1.0 ;
  A(Q.n-1,Q.n-1) = 1.0 ;

  // Init matrix for LSE
  Matrix_DOUBLE qq(Q.n,4) ;
  Matrix_DOUBLE xx(Q.n,4) ;
  for(i=0;i<Q.n;i++)
    for(j=0; j<4;j++)
      qq(i,j) = (double)Q[i].data[j] ;

  //LAPack.lse(A,qq,xx) ;
  solve(A,qq,xx) ;

  // Store the data
  for(i=0;i<xx.rows;i++){
    for(j=0;j<4;j++)
      P[i].data[j] = (float)xx(i,j) ;
  }

}

inline float pow2(float a){
  return a*a ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: intersectLine --- find the intersection point of two lines
  Description: This routines finds the intersection point of two lines.
               The algorithm generates a plane from one of the lines 
	       and finds the intersection point between this plane and the 
	       other line.
	       
        Input: p1 --> a point in the first line
	       t1 --> the tangent at p0t along the line 2
	       p2 --> a point in the second line
	       t2 --> the tangent at p1 along the line 2
	        p <-- the intersection point
       Output: 1 if the lines intersect, 0 if they don't. 
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int intersectLine(const Point3D& p1, const Point3D& t1, const Point3D& p2, const Point3D& t2, Point3D& p){
  // a line is written like 
  // L(t) = Q + u*t
  // u is the parametric value P is a point in the line and T is the tangent
  // a plane is of the form
  // (X-P).v = 0 
  // where P is a point where the plane goes through and v is the normal to it
  // and X is (x,y,z)
  // solving for u

  Point3D v,px ;

  px = crossProduct(t1,p1-p2) ;
  v = crossProduct(px,t1) ;
  
  float t = (p1-p2)*v ;
  float vw = v*t2 ;
  if(pow2(vw)<1e-7)
    return 0 ;
  t /= vw ;
  p = p2+(((p1-p2)*v)/vw)*t2 ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: makeCircle --- generates a circular curve
  Description: Generates parts of a circle, starting at angle $as$ and 
   	       finishing at $ae$ with a radius $r$ and having the origin 
	       located at $O$. The $X$ and $Y$ vector describe the local
	       $x$-axis and the local $y$-axis of the circle.

	       The degrees are specified in radians.
        Input:  O --> the center of the circle
	        X --> unit length vector lying in the plane of the circle
	        Y --> unit length vector lying in the plane of the circle
		r --> the radius of the circle
		as --> start angle measured with respect to $X$
		ae --> end angle measured with respect to $X$
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::makeCircle(const Point3D& O, const Point3D& X, const Point3D& Y, float r, double as, double ae){
  double theta,angle,dtheta ;
  int narcs ;
  while(ae<as)
    ae += 2*M_PI ;
  theta = ae-as ;
  if(theta<M_PI/2.0)	
    narcs = 1 ;
  else{
    if(theta<M_PI)
      narcs = 2 ;
    else{
      if(theta<1.5*M_PI)
	narcs = 3 ;
      else
	narcs = 4 ;
    }
  }
  dtheta = theta/(double)narcs ;
  int n = 2*narcs+1 ; // n control points ;
  double w1 = cos(dtheta/2.0) ; // dtheta/2.0 is base angle

  Point3D P0,T0,P2,T2,P1 ;
  P0 = O + r*cos(as)*X + r*sin(as)*Y ; 
  T0 = -sin(as)*X + cos(as)*Y ; // initialize start values
  resize(n,2) ;
  
  P[0] = P0 ;
  int i ;
  int index = 0 ;
  angle = as ;
  for(i=1;i<=narcs;++i){
    angle += dtheta ;
    P2 = O+ r*cos(angle)*X + r*sin(angle)*Y ;  
    P[index+2] = P2 ;
    T2 = -sin(angle)*X + cos(angle)*Y ;
    intersectLine(P0,T0,P2,T2,P1) ;
    P[index+1] = P1 ;
    P[index+1] *= w1 ;
    index += 2 ;
    if(i<narcs){
      P0 = P2 ;
      T0 = T2 ;
    }
  }
  int j = 2*narcs+1 ; // load the knot vector
  for(i=0;i<3;++i){
      U[i] = 0.0 ;
      U[i+j] = 1.0 ;
  }
  switch(narcs){
  case 1: break ;
  case 2: 
    U[3] = U[4] = 0.5 ;
    break ;
  case 3:
    U[3] = U[4] = 1.0/3.0 ;
    U[5] = U[6] = 2.0/3.0 ;
    break ;
  case 4:
    U[3] = U[4] = 0.25 ;
    U[5] = U[6] = 0.50 ;  
    U[7] = U[8] = 0.75 ;
    break ;    
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: makeCircle --- generates a circular curve
  Description: Generates a circular curve of radius $r$ at origin $O$. The 
               curve is drawn in the $xy$-axis.
        Input:  O --> the center of the circle
		r --> the radius of the circle
		as --> start angle measured with respect to the $x$-axis
		ae --> end angle measured with respect to the $y$-axis
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::makeCircle(const Point3D& O, float r, double as, double ae){
  makeCircle(O,Point3D(1,0,0),Point3D(0,1,0),r,as,ae) ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: knotUnion --- finds the union of two knot vectors
  Description: Finds the union between two knot vectors
        Input: Ua --> knot vector A
               Ub --> knot vector B
       Output: the union of Ua and Ub
 Restrictions: The result is useless unless the knot vectors being compared are from a NURBS curve of a same degree
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Vector_FLOAT knotUnion(const Vector_FLOAT& Ua, const Vector_FLOAT& Ub) {
  Vector_FLOAT U(Ua.n+Ub.n) ;
  int done = 0 ;
  int i,ia,ib ;  
  float t ;

  i = ia = ib = 0 ;
  while(!done){
    if(Ua[ia] == Ub[ib]){
      t = Ua[ia] ;
      ++ia ; ++ib ;
    }
    else{
      if(Ua[ia]<Ub[ib]){
	t = Ua[ia] ;
	++ia ;
      }
      else{
	t = Ub[ib] ;
	++ib ;
      }
    }
    U[i++] = t ;
    done = (ia>=Ua.n || ib>=Ub.n) ;
  }
  U.resize(i) ;
  return U ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
        Member: mergeKnotVector --- Merges the knot vector of a curve with another knot vector
  Description: Will merge the Knot vector U with the one from the curve  
               and it will refine the curve appropriately.
        Input: Um --> the knot vector to merge with
       Output:
 Restrictions: the knot U must be common with the one from the curve c
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::mergeKnotVector(const Vector_FLOAT &Um){
  int i,ia,ib ;
  // Find the knots to insert
  Vector_FLOAT I(Um.n) ;
  int done = 0 ;
  i = ia = ib = 0 ;
  while(!done) {
    if(Um[ib] == U[ia]){
      ++ib ; ++ia ;
    }
    else{
      I[i++] = Um[ib] ;
      ib++ ;
    }
    done = (ia>=U.n || ib >= Um.n) ;
  }
  I.resize(i) ;

  if(I.n>0){
    // Refine the curve
    refineKnotVector(I) ;
  }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: generateCompatibleCurves ---  Generate compatible curves from 
                                             an array of curves
  Description: This routine will put to the same degree all the curves in 
               the array and it will ensure that they have the same knot 
	       vector.
        Input: ca <-> the array of Nurbs curves 
       Output:
 Restrictions: the knot vector of all the curves must be in the range $[0,1]$
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void generateCompatibleCurves(PlNurbsCurveArray &ca){
  int i;
  PlNurbsCurve tc ;

  if(ca.n<=1) // Nothing to do... only 1 curve in the array
    return ;

  // Increase all the curves to the highest degree
  int p = 1 ;
  for(i=0;i<ca.n;i++)
    if(p<ca[i].deg)
      p = ca[i].deg ;
  for(i=0;i<ca.n;i++){
    ca[i].degreeElevate(p-ca[i].deg);
  }

  // Find a common Knot vector
  Vector_FLOAT Uc(ca[0].U) ;
  for(i=1;i<ca.n;i++){
    Uc = knotUnion(Uc,ca[i].U) ;
  }

  // Refine the knot vectors
  for(i=0;i<ca.n;i++)
    ca[i].mergeKnotVector(Uc) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: read --- reads a PlNurbsCurve from a file
  Description: Reads a PlNurbsCurve from an input stream.
        Input: fin --> an input file stream
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::read(ifstream &fin){
  if(!fin) {
    return 0 ;
  }
  int np,d;
  char *type ;
  type = new char[3] ;
  if(!fin.read(type,sizeof(char)*3)) { delete []type ; return 0 ;}
  int r1 = strncmp(type,"nc3",3) ;
  int r2 = strncmp(type,"nc4",3) ;
  if(!(r1==0 || r2==0)) {
    delete []type ;
    return 0 ;
  }
  if(!fin.read((char*)&np,sizeof(int))) { delete []type ; return 0 ;}
  if(!fin.read((char*)&d,sizeof(int))) { delete []type ; return 0 ;}

  resize(np,d) ;
  
  if(!fin.read((char*)U.memory(),sizeof(float)*U.n)) { delete []type ; return 0 ;}
     

  float *p,*p2 ;
  if(!r1){
    p = new float[3*np] ;
    if(!fin.read((char*)p,sizeof(float)*3*np)) { delete []type ; return 0 ;}
    p2 = p ;
    for(int i=0;i<np;i++){
      P[i].x() = *(p++) ;
      P[i].y() = *(p++) ;
      P[i].z() = *(p++) ;
      P[i].w() = 1.0 ;
    }
    delete []p2 ;
  }
  else{
    p = new float[4*np] ;
    if(!fin.read((char*)p,sizeof(float)*4*np)) { delete []type ; return 0 ;}
    p2 = p ;
    for(int i=0;i<np;i++){
      P[i].x() = *(p++) ;
      P[i].y() = *(p++) ;
      P[i].z() = *(p++) ;
      P[i].w() = *(p++) ;
    }
    delete []p2 ;
  }

  delete []type ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: read --- reads a PlNurbsCurve from a file
  Description: Reads a PlNurbsCurve from a file.
        Input: filename --> the filename to read the curve from 
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::read(const char* filename){
  ifstream fin(filename) ;
  if(!fin) {
    return 0 ;
  }
  return read(fin) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: write --- writes a PlNurbsCurve to a file
  Description: Writes a PlNurbsCurve to a file.
        Input: filename --> the filename to write to.
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::write(const char* filename) const {
  ofstream fout(filename) ;  

  if(!fout)
    return 0 ;
  return write(fout) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: write --- writes a PlNurbsCurve to a file
  Description: Writes a PlNurbsCurve to an output stream.
        Input: fout --> the output stream
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::write(ofstream &fout) const {
  if(!fout)
    return 0 ;
  if(!fout.write((char*)&"nc4",sizeof(char)*3)) return 0 ;
  if(!fout.write((char*)&P.n,sizeof(int))) return 0 ;
  if(!fout.write((char*)&deg,sizeof(int))) return 0 ;
  if(!fout.write((char*)U.memory(),sizeof(float)*U.n)) return 0 ;
  
  float *p,*p2 ;
  p = new float[P.n*4] ;
  p2 = p ;
  for(int i=0;i<P.n;i++) {
    *p = P[i].x() ; p++ ;
    *p = P[i].y() ; p++ ;
    *p = P[i].z() ; p++ ;
    *p = P[i].w() ; p++ ;
  }
  if(!fout.write((char*)p2,sizeof(float)*P.n*4)) return 0 ;
  delete []p2 ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: nurbsbasisFuns --- computes the non-zero basis functions of the curve
  Description: Computes the non-zero basis functions and puts the result 
               into N. N has a size of deg+1. To relate N to the basis 
	       functions, $N_{all}[span -deg +i] = N[i]$ for $i=0\ldots deg$.
        Input: u --> the parametric value
            span --> the non-zero span of the basis functions
	     deg --> the degree of the curve
	       U --> the knot vector on which the Basis functions are computed
               N <-- the non-zero basis functions
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void nurbsBasisFuns(float u, int span, int deg, const Vector_FLOAT& U, Vector_FLOAT& N) {
  Vector_FLOAT left(deg+1), right(deg+1) ;
  float temp,saved ;
   
  N.resize(deg+1) ;

  N[0] = 1.0 ;
  for(int j=1; j<= deg ; j++){
    left[j] = u-U[span+1-j] ;
    right[j] = U[span+j]-u ;
    saved = 0.0 ;
    for(int r=0 ; r<j; r++){
      temp = N[r]/(right[r+1]+left[j-r]) ;
      N[r] = saved+right[r+1] * temp ;
      saved = left[j-r] * temp ;
    }
    N[j] = saved ;
  }
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine:  nurbsDersBasisFuns --- Compute the derivatives functions at $u$ of the NURBS curve
  Description: For information on the algorithm, see A2.3 on p72 of 
               the NURBS book. The result is stored in the ders matrix, where 
	       ders is of size $(n+1,deg+1)$ and the derivative 
	       $C'(u) = ders(1,span-deg+j)$ where $j=0\ldots deg+1$.

        Input:  n --> the degree of the derivation
                u --> the parametric value
             span --> the span for the basis functions
	      deg --> the degree of the curve
	        U --> the knot vector on which the Basis functions must be
		      computed.
             ders <-- A matrix containing the derivatives of the curve.
       Output:
 Restrictions: $n$, $u$ and $span$ must be in a valid range.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void nurbsDersBasisFuns(int n,float u, int span,  int deg, const Vector_FLOAT& U, Matrix_FLOAT& ders) {
  Vector_FLOAT left(deg+1),right(deg+1) ;
  Matrix_FLOAT ndu(deg+1,deg+1) ;
  float saved,temp ;
  int j,r ;

  ders.resize(n+1,deg+1) ;

  ndu(0,0) = 1.0 ;
  for(j=1; j<= deg ;j++){
    left[j] = u-U[span+1-j] ;
    right[j] = U[span+j]-u ;
    saved = 0.0 ;
    
    for(r=0;r<j ; r++){
      // Lower triangle
      ndu(j,r) = right[r+1]+left[j-r] ;
      temp = ndu(r,j-1)/ndu(j,r) ;
      // Upper triangle
      ndu(r,j) = saved+right[r+1] * temp ;
      saved = left[j-r] * temp ;
    }

    ndu(j,j) = saved ;
  }

  for(j=0;j<=deg;j++)
    ders(0,j) = ndu(j,deg) ;

  // Compute the derivatives
  Matrix_FLOAT a(deg+1,deg+1) ;
  for(r=0;r<=deg;r++){
    int s1,s2 ;
    s1 = 0 ; s2 = 1 ; // alternate rows in array a
    a(0,0) = 1.0 ;
    // Compute the kth derivative
    for(int k=1;k<=n;k++){
      float d ;
      int rk,pk,j1,j2 ;
      d = 0.0 ;
      rk = r-k ; pk = deg-k ;

      if(r>=k){
	a(s2,0) = a(s1,0)/ndu(pk+1,rk) ;
	d = a(s2,0)*ndu(rk,pk) ;
      }

      if(rk>=-1){
	j1 = 1 ;
      }
      else{
	j1 = -rk ;
      }

      if(r-1 <= pk){
	j2 = k-1 ;
      }
      else{
	j2 = deg-r ;
      }

      for(j=j1;j<=j2;j++){
	a(s2,j) = (a(s1,j)-a(s1,j-1))/ndu(pk+1,rk+j) ;
	d += a(s2,j)*ndu(rk+j,pk) ;
      }
      
      if(r<=pk){
	a(s2,k) = -a(s1,k-1)/ndu(pk+1,r) ;
	d += a(s2,k)*ndu(r,pk) ;
      }
      ders(k,r) = d ;
      j = s1 ; s1 = s2 ; s2 = j ; // Switch rows
    }
  }

  // Multiply through by the correct factors
  r = deg ;
  for(int k=1;k<=n;k++){
    for(j=0;j<=deg;j++)
      ders(k,j) *= r ;
    r *= deg-k ;
  }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: decompose --- decompose the curve into Bzier segments
  Description: This function decomposes the curve into an array of 4D Bzier 
               segments.
        Input: c <-- an array of Bzier segments
       Output:
 Restrictions:
   Created by: Philippe Lavoie (16 February 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurve::decompose(PlNurbsCurveArray& c) const {
  int i,m,a,b,nb,mult,j,r,save,s,k ;
  float numer,alpha ;
  Vector_FLOAT alphas(deg+1) ;

  // all the curves will have the same vector
  Vector_FLOAT nU ;
  nU.resize(2*(deg+1)) ;
  for(i=0;i<nU.n/2;i++)
    nU[i] = 0 ;
  for(i=nU.n/2;i<nU.n;i++)
    nU[i] = 1 ;
  
  c.resize(P.n-deg) ;
  for(i=0;i<c.n;i++){
    c[i].resize(deg+1,deg) ;
    c[i].U = nU ;
  }
  
  m = P.n+deg ;
  a = deg ;
  b = deg+1 ;
  nb = 0 ;

  for(i=0;i<=deg;i++)
    c[nb].P[i] = P[i] ;
  while(b<m){
    i = b ;
    while(b<m && U[b+1] <= U[b]) b++ ;
    mult = b-i+1 ;
    if(mult<deg){
      numer = U[b]-U[a] ; // th enumerator of the alphas
      for(j=deg;j>mult;j--) // compute and store the alphas
	alphas[j-mult-1] = numer/(U[a+j]-U[a]) ;
      r = deg-mult ; // insert knot r times
      for(j=1;j<=r;j++){
	save=r-j;
	s=mult+j; // this many new points
	for(k=deg;k>=s;k--){
	  alpha = alphas[k-s] ;
	  c[nb].P[k] = alpha*c[nb].P[k] + (1.0-alpha)*c[nb].P[k-1] ;	  
	}
	if(b<m) // control point of
	  c[nb+1].P[save] = c[nb].P[deg]; // next segment
      }
    }
    nb++ ;
    if(b<m){ // initialize for next segment
      for(i=deg-mult; i<= deg ; i++)
	c[nb].P[i] = P[b-deg+i];
      a=b ;
      b++ ;
    }
  }
  c.resize(nb) ;
}

inline Point3D project2D(const CtrlPnt4D& p){
  Point3D pnt ;
  if(p.z()==0.0){
    pnt.x = p.x()/p.w() ;
    pnt.y = p.y()/p.w() ;
  }
  else{
    pnt.x = (p.x()/p.w())/abs(p.z()) ;
    pnt.y = (p.y()/p.w())/abs(p.z()) ;
  }
  return pnt ;
}

const float offX = 50 ;
const float offY = 70 ;
inline void movePsP(Point3D &p, float magFact){
  p = p*magFact+Point3D(offX,offY,0)  ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writePS --- writes a post-script file representing the curve
  Description: Writes the curve in the postscript format to a file. 
        Input: filename <-- the file to write the postscript file to
                    cp --> a flag indicating if the control points should be 
	                   drawn, 0 = no and 1 = yes
	       magFact --> a magnification factor, the 2D point of the control
                           points will be magnified by this value. The size is
			   measured in postscript points. If the magFact is 
			   set to a value smaller or equal to 0, than the 
			   program will try to guess a magnification factor 
			   such that the curve is large enough to fill the 
			   page.
	          dash --> the size of the dash in postscript points . A size 
		           smaller or equal to 0 indicates that 
			   the line joining the control points is plain.
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: If the weights of the curve are not all at 1, the result might 
               not be representative of the true NURBS curve.
   Created by: Philippe Lavoie (16 February 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::writePS(const char* filename,int cp,float magFact, float dash) const {
  ofstream fout(filename) ;  

  if(!fout)
    return 0 ;

  // extract the Bezier segments 
  PlNurbsCurveArray Ca ;
  decompose(Ca) ;

  int guess =0 ;
  if(magFact<=0){
    magFact = 1.0 ;
    guess = 1 ;
  }
  
  Matrix_Point3D pnts(Ca.n,deg+1) ;
  int i,j ;
  for(i=0;i<Ca.n;i++){
    for(j=0;j<deg+1;j++){
      pnts(i,j) = project2D(Ca[i].P[j]) ;
    }
  }


  // find bounding box parameters
  float mx,my,Mx,My ;
  mx = Mx = pnts(0,0).x ;
  my = My = pnts(0,0).y ;

  for(i=0;i<Ca.n;i++){
    Point3D p ;
    int step ;
    step = 8 ;
    for(j=0;j<=step;j++){
      float u ;
      u = (float)j/(float)step ;
      p = project2D(Ca[i](u)) ;
      if(p.x < mx)
	mx = p.x ;
      if(p.x > Mx)
	Mx = p.x ;
      if(p.y < my)
	my = p.y ;
      if(p.y > My) 
	My = p.y ;      
    }
  }

  if(guess){
    magFact = minimum((float)500/(float)(Mx-mx),(float)700/(float)(My-my)) ;
  }

  mx = mx*magFact+offX;
  my = my*magFact+offY;
  Mx = Mx*magFact+offX;
  My = My*magFact+offY;
  for(i=0;i<Ca.n;i++)
    for(j=0;j<deg+1;j++)
	movePsP(pnts(i,j),magFact) ;

  /*  mx -= 10 ;
  Mx += 10 ;
  my -= 10 ;
  My += 10 ;
  */
  
  fout << "%!PS-Adobe-2.1\n%%Title: " << filename << endl ;
  fout << "%%Creator: PlNurbsCurve::writePS\n" ;
  fout << "%%BoundingBox: " << mx << " " << my << " " << Mx << " " << My << endl ;
  fout << "%%Pages: 0" << endl ;
  fout << "%%EndComments" << endl ;
  fout << "0 setlinewidth\n" ;
  fout << "0 setgray\n" ;
  fout << "\n" ;

  fout << "newpath\n" ;
  fout << pnts(0,0).x << " " << pnts(0,0).y << " moveto\n" ;
  for(i=0;i<Ca.n;i++){
    for(j=1;j<deg+1;j++){
      fout << pnts(i,j).x << " " << pnts(i,j).y << " " ;
    }
    fout << "curveto\n" ;
  }
  fout << "stroke\n" ;

  if(cp>0){ // draw the control points of the original curve
    Vector_Point3D pts(P.n) ;
    for(i=0;i<P.n;i++){
      pts[i] = project2D(P[i]) ;
      movePsP(pts[i],magFact) ;
      fout << "newpath\n" ;
      fout << pts[i].x << " " << pts[i].y << "  3 0 360 arc\nfill\n" ;
    }
    if(dash>0)
      fout << "[" << dash << "] " << dash << " setdash\n" ;
    fout << "newpath\n" ;
    
    fout << pts[0].x << " " << pts[0].y << " moveto\n" ;
    for(i=1;i<P.n;i++)
      fout << pts[i].x << " " << pts[i].y << " lineto\n" ;
    fout << "stroke\n" ;
  }
  else{
    if(cp<0){
      Vector_Point3D pts(P.n*Ca.n) ;
      int k=0 ;
      for(i=0;i<Ca.n;i++)
	for(j=0;j<deg+1;j++){
	  pts[k] = project2D(Ca[i].P[j]) ;
	  movePsP(pts[k],magFact) ;
	  fout << "newpath\n" ;
	  fout << pts[k].x << " " << pts[k].y << "  3 0 360 arc\nfill\n" ;
	  k++ ;
	}
      if(dash>0)
	fout << "[" << dash << "] " << dash << " setdash\n" ;
      fout << "newpath\n" ;
      
      fout << pts[0].x << " " << pts[0].y << " moveto\n" ;
      for(i=1;i<k;i++)
	fout << pts[i].x << " " << pts[i].y << " lineto\n" ;
      fout << "stroke\n" ;      
    }
  }

  fout << "showpage\n%%EOF\n" ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writePSp --- writes a post-script file representing the curve
  Description: Writes the curve in the postscript format to a file, it also
               draws the points defined in $points$ with their associated
	       vectors if $vector$ is used.
        Input: filename <-- the file to write the postscript file to
	         points --> draws these additional points as empty circles
		vectors --> specify a vector associated with the points
		            (this can be an empty vector)
                     cp --> a flag indicating if the control points should be 
		            drawn, 0 = no and 1 = yes
		magFact --> a magnification factor, the 2D point of the control
		            points will be magnified by this value. The size 
			    is measured in postscript points. If the magFact 
			    is set to a value smaller or equal to 0, than the 
			    program will try to guess a magnification factor 
			    such that the curve is large enough to fill the 
			    page.
                   dash --> the size of the dash in postscript points . A size
		            smaller or equal to 0 indicates that 
			    the line joining the control points is plain.
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: If the weights of the curve are not all at 1, the result might 
               not be representative of the true NURBS curve. If vectors is
	       used, then it must be of the same size as points. If a vector
	       element is (0,0,0) it will not be drawn.
   Created by: Philippe Lavoie (16 February 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::writePSp(const char* filename,const Vector_Point3D& points, const Vector_Point3D& vectors, int cp, float magFact, float dash) const {
  ofstream fout(filename) ;  

  if(!fout)
    return 0 ;

  // extract the Bezier segments 
  PlNurbsCurveArray Ca ;
  decompose(Ca) ;

  int guess =0 ;
  if(magFact<=0){
    magFact = 1.0 ;
    guess = 1 ;
  }
  
  Matrix_Point3D pnts(Ca.n,deg+1) ;
  int i,j ;
  for(i=0;i<Ca.n;i++){
    for(j=0;j<deg+1;j++){
      pnts(i,j) = project2D(Ca[i].P[j]) ;
    }
  }


  // find bounding box parameters
  float mx,my,Mx,My ;
  mx = Mx = pnts(0,0).x ;
  my = My = pnts(0,0).y ;

  for(i=0;i<Ca.n;i++){
    Point3D p ;
    int step ;
    step = 8 ;
    for(j=0;j<=step;j++){
      float u ;
      u = (float)j/(float)step ;
      p = project2D(Ca[i](u)) ;
      if(p.x < mx)
	mx = p.x ;
      if(p.x > Mx)
	Mx = p.x ;
      if(p.y < my)
	my = p.y ;
      if(p.y > My) 
	My = p.y ;      
    }
  }

  if(guess){
    magFact = minimum((float)500/(float)(Mx-mx),(float)700/(float)(My-my)) ;
  }

  mx = mx*magFact+offX;
  my = my*magFact+offY;
  Mx = Mx*magFact+offX;
  My = My*magFact+offY;
  for(i=0;i<Ca.n;i++)
    for(j=0;j<deg+1;j++)
	movePsP(pnts(i,j),magFact) ;

  /*  mx -= 10 ;
  Mx += 10 ;
  my -= 10 ;
  My += 10 ;
  */
  
  fout << "%!PS-Adobe-2.1\n%%Title: " << filename << endl ;
  fout << "%%Creator: PlNurbsCurve::writePS\n" ;
  fout << "%%BoundingBox: " << mx << " " << my << " " << Mx << " " << My << endl ;
  fout << "%%Pages: 0" << endl ;
  fout << "%%EndComments" << endl ;
  fout << "0 setlinewidth\n" ;
  fout << "0 setgray\n" ;
  fout << "\n" ;

  fout << "newpath\n" ;
  fout << pnts(0,0).x << " " << pnts(0,0).y << " moveto\n" ;
  for(i=0;i<Ca.n;i++){
    for(j=1;j<deg+1;j++){
      fout << pnts(i,j).x << " " << pnts(i,j).y << " " ;
    }
    fout << "curveto\n" ;
  }
  fout << "stroke\n" ;

  if(cp>0){ // draw the control points of the original curve
    Vector_Point3D pts(P.n) ;
    for(i=0;i<P.n;i++){
      pts[i] = project2D(P[i]) ;
      movePsP(pts[i],magFact) ;
      fout << "newpath\n" ;
      fout << pts[i].x << " " << pts[i].y << "  3 0 360 arc\nfill\n" ;
    }
    if(dash>0)
      fout << "[" << dash << "] " << dash << " setdash\n" ;
    fout << "newpath\n" ;
    
    fout << pts[0].x << " " << pts[0].y << " moveto\n" ;
    for(i=1;i<P.n;i++)
      fout << pts[i].x << " " << pts[i].y << " lineto\n" ;
    fout << "stroke\n" ;
  }
  else{
    if(cp<0){
      Vector_Point3D pts(P.n*Ca.n) ;
      int k=0 ;
      for(i=0;i<Ca.n;i++)
	for(j=0;j<deg+1;j++){
	  pts[k] = project2D(Ca[i].P[j]) ;
	  movePsP(pts[k],magFact) ;
	  fout << "newpath\n" ;
	  fout << pts[k].x << " " << pts[k].y << "  3 0 360 arc\nfill\n" ;
	  k++ ;
	}
      if(dash>0)
	fout << "[" << dash << "] " << dash << " setdash\n" ;
      fout << "newpath\n" ;
      
      fout << pts[0].x << " " << pts[0].y << " moveto\n" ;
      for(i=1;i<k;i++)
	fout << pts[i].x << " " << pts[i].y << " lineto\n" ;
      fout << "stroke\n" ;      
    }
  }

  for(i=0;i<points.n;++i){
    Point3D p ;
    p = points[i] ;
    movePsP(p,magFact) ;
    fout << "newpath\n" ;
    fout << p.x << " " << p.y << "  3 0 360 arc\nfill\n" ;
  }

  if(vectors.n==points.n){
    for(i=0;i<points.n;++i){
      Point3D p,p2 ;
      p = points[i] ;
      p2 = points[i] + vectors[i] ;
      movePsP(p,magFact) ;
      movePsP(p2,magFact) ;
      fout << "newpath\n" ;
      fout << p.x << " " << p.y << " moveto\n" ;
      if(dash>0)
	fout << "[" << dash/2.0 << "] " << dash/2.0 << " setdash\n" ;
      fout << p2.x << " " << p2.y << " lineto\n" ;
      fout << "stroke\n" ;
    }
  }

  fout << "showpage\n%%EOF\n" ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writeVRML --- writes the curve to a VRML file
  Description: A circle is swept around the trajectory made by the curve.
               The resulting surface is saved as a VRML file.
        Input: filename --> the name of the VRML file to save to
	         radius --> the radius of the line
		      K --> the minimum number of interpolation
		  color --> the color of the line
		     Nu --> the number of points for the circle
		     Nv --> the number of points along the path
		    u_s --> the starting parametric value for $u$
		    u_e --> the end parametric value for $u$
       Output: returns 1 on success, 0 otherwise.
 Restrictions:
   Created by: Philippe Lavoie (25 July 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::writeVRML(const char* filename,float radius,int K, const Color& color,int Nu,int Nv, float u_s, float u_e) const{
  PlNurbsSurface S ;
  PlNurbsCurve C ;
  
  C.makeCircle(Point3D(0,0,0),Point3D(1,0,0),Point3D(0,0,1),radius,0,2*M_PI);
  S.sweep(*this,C,K) ;
  return S.writeVRML(filename,color,Nu,Nv,0,1,u_s,u_e) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: knotAveraging --- generates a knot vector using the averaging technique
  Description: Generates a knot vector using the averaging technique.
               The technique is as follows:
	       \begin{itemize}
	       \item $u_0 = \cdots = u_{deg} = 0$
	       \item $u_{m-deg} = \cdots = u_{m-1} = 1$
	       \item \begin{equation}
	             u_{j+deg} = \frac{1}{deg}\sum_{i=j}^{j+deg+1}\bar{u}_i
		     \hspace{0.5in} j= 1,\ldots,n-deg-1
	             \end{equation}
	       \end{itemize}
	       where $n$ is the size of the $\bar{u}$ knot coefficient vector,
	       $m=n+deg+1$ is the size of the knot vector and $deg$ is the 
	       degree of the curve.
        Input: uk --> the knot coefficients
	      deg --> the degree of the curve associated with the knot vector
                U <-- an average knot vector
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void knotAveraging(const Vector_FLOAT& uk, int deg, Vector_FLOAT& U){
  U.resize(uk.n+deg+1) ;

  int j ;
  for(j=1;j<uk.n-deg;++j){
    U[j+deg] = 0.0 ;
    for(int i=j;i<j+deg;++i)
      U[j+deg] += uk[i] ;
    U[j+deg] /= (float)deg ;
  }
  for(j=0;j<=deg;++j)
    U[j] = 0.0 ;
  for(j=U.n-deg-1;j<U.n;++j)
    U[j] = 1.0 ;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: averagingKnots --- compute the parameters by averaging the knots
  Description: Generates a parameter vector by averaging the knots from the 
               knot vector. 
        Input:   U --> the knot vector
	       deg --> the degree of the curve associated with the knot vector
	        uk <-- the parameter vector
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (26 August 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void averagingKnots(const Vector_FLOAT& U, int deg, Vector_FLOAT& uk){
  uk.resize(U.n-deg-1) ;

  int i,k ;

  uk[0] = U[0] ;
  uk[uk.n-1] = U[U.n-1] ;

  for(k=1;k<uk.n-1;++k){
    uk[k] = 0.0 ;
    for(i=k+1;i<k+deg+1;++i)
      uk[k] += U[i] ;
    uk[k] /= deg ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  movePoint --- moves a point in the NURBS curve
  Description:  This modifies the curve such that the point $C(u)$ is moved
                by delta. 

		See section 11.5.1 of the NURBS book for an explanation of 
		the algorithm.
        Input:   u --> the point to move
	     delta --> the movement in 3D of the point at $C(u)$
       Output:
 Restrictions: 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::movePoint(float u, const Point3D& delta) {
  Vector_Point3D d(1) ;
  d[0] = delta ;
  return movePoint(u,d) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  movePoint --- moves a point in the NURBS curve
  Description:  This modifies the curve such that the point $C(u)$ is moved
                by delta. Delta is a vector containing the movement as
		$D^{(k)}$ where $(k)$ specifies the derivative. Thus
		at D[0], this specifies the 0th derivative movement, at 
		D[1] it specifies the 1st derivative movement of the point.
		{\em i.e.} Suppose that $C(u) = (10,20,3)$ then a
		D[0] = (10,10,10) will move the point to $C(u) = (20,30,13)$ 

		See section 11.5.1 of the NURBS book for an explanation of 
		the algorithm.
        Input:   u --> the point to move
	     delta --> the vector of movement
       Output:
 Restrictions: 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::movePoint(float u, const Vector_Point3D& delta) {
  int i,j ;

  // setup B
  Matrix_DOUBLE B ;
  
  int n,m ; // n is the number of rows, m the number of columns

  m = deg + 1 ;
  n = delta.n ;

  B.resize(n,m) ;
  
  int span = findSpan(u) ;

  n = 0 ;
  
  Matrix_FLOAT R ;

  dersBasisFuns(delta.n-1,u,span,R) ;


  for(i=0;i<delta.n;++i){
    if(delta[i].x==0.0 && delta[i].y==0.0 && delta[i].z==0.0)
      continue ;
    for(j=0;j<m;++j){
      B(n,j) = (double)R(i,j) ;
    }
    ++n ;
  }

  Matrix_DOUBLE A  ;
  Matrix_DOUBLE Bt(transpose(B)) ;
  Matrix_DOUBLE BBt ;

  //BBt = LAPack.inverse(B*Bt) ;
  BBt = inverse(B*Bt) ;
  /*
  if(LAPack.last_error>0){
    return 0 ; // matrix B*Bt is singular 
  }
  */
  A = Bt*BBt ;

  Matrix_DOUBLE dD ;

  dD.resize(delta.n,3) ;
  for(i=0;i<delta.n;++i){
    const Point3D& d = delta[i] ; // this makes the SGI compiler happy
    for(j=0;j<3;++j)
      dD(i,j) = (double)d.data[j] ;
  }

  Matrix_DOUBLE dP ;

  dP = A*dD ;

  for(i=0;i<m;++i){
    P[span-deg+i].x() += dP(i,0)*P[span-deg+i].w() ;
    P[span-deg+i].y() += dP(i,1)*P[span-deg+i].w() ;
    P[span-deg+i].z() += dP(i,2)*P[span-deg+i].w() ;
  }

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point with some constraint
  Description: This will modify the NURBS curve by respecting a certain 
               number of constraints. $u_r$ specifies the parameters on which
	       the constraints should be applied. 
	       The constraint are defined by $D$ which specifies the
	       vector by which the points should move.

	       For example, if you want to move the point $C(0.5)$ by 
	       $(10,0,10$ and fix the point $C(0.6)$ on the current curve
	       (a move of (0,0,0)). 
	       $u_r = 0.5, 0.6$ and $D = (10,0,10), (0,0,0)$
	       
	       The $u_r$ vector should be in an increasing order.

	       See section 11.5.1 of the NURBS book for an explanation of 
	       the algorithm.
        Input: ur --> the vector of parameters on which a constraint is applied
	        D --> a vector of the value of $D_r^{(0)}$

       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: ur and D must have the same size and the values inside ur
               should not repeat
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::movePoint(const Vector_FLOAT& ur, const Vector_Point3D& D) {
  Vector_INT fixCP(0) ;
  Vector_INT Dr(D.n) ;
  Vector_INT Dk(D.n) ;

  if(ur.n != D.n){
    PlError err("movePoint(ur,D)");
    err << "The two input vectors are not of the same size\n" ;
    err << "ur.n= " << ur.n << ", D.n = " << D.n << endl ;
    err.fatal() ;
  }
  
  for(int i=0;i<Dr.n;++i){
    Dr[i] = i ;
  }
  Dk.reset(0) ;
  return movePoint(ur,D,Dr,Dk,fixCP) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point with some constraint
  Description: This will modify the NURBS curve by respecting a certain 
               number of constraints. $u_r$ specifies the parameters on which
	       the constraints should be applied. 
	       The constraint are defined by $D_r^{(k)}$ which requires 3
	       vectors to fully qualify. $D$ specifies the value
	       of the constraint and $D_r$ and $D_k$ are used to specify
	       on which parameter the constraint is applied and of what degree.

	       For example, if you want to move the point $C(0.5)$ by 
	       $(10,0,10$ and fix the point $C(0.6)$ on the current curve
	       (a move of (0,0,0)) but change its 1st derivative by (0,20,0). 
	       Then the following values must be inputed to the routine.
	       $u_r = 0.5,0.6$, $D = (10,0,10), (0,0,0), (0,20,0)$, 
	       $D_r = 0, 1, 1$ and $D_k = 0, 0, 1$. 

	       The values in D should be ordered in respect with r and k.
	       {\em i.e.} for $D_i=D_{r_i}^{(k_i)}$, then $i < j$ implies 
	       that $r_i < r_j$ and that either $r_i < r_j$ or $k_i < k_j$.

	       See section 11.5.1 of the NURBS book for an explanation of 
	       the algorithm.

        Input: ur --> the vector of parameters on which a constraint is applied
	        D --> a vector of the value of $D_r^{(k)}$
	       Dr --> a vector specifying the value of r for D
	       Dk --> a vector specifying the value of k for D
       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: The values inside ur should {\em not} repeat. D,Dr and Dk must
               be of the same size.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::movePoint(const Vector_FLOAT& ur, const Vector_Point3D& D, const Vector_INT& Dr, const Vector_INT& Dk) {
  Vector_INT fixCP(0) ;

  if(D.n != Dr.n || D.n !=Dk.n){
    PlError err("movePoint(ur,D,Dr,Dk)");
    err << "The D,Dr,Dk vectors are not of the same size\n" ;
    err << "D.n= " << D.n << ", Dr.n = " << Dr.n 
	<< ", Dk.n = " << Dk.n << endl ;
    err.fatal() ;
  }  

  return movePoint(ur,D,Dr,Dk,fixCP) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point with some constraint
  Description: This will modify the NURBS curve by respecting a certain 
               number of constraints. $u_r$ specifies the parameters on which
	       the constraints should be applied. 
	       The constraint are defined by $D_r^{(k)}$ which requires 3
	       vectors to fully qualify. $D$ specifies the value
	       of the constraint and $D_r$ and $D_k$ are used to specify
	       on which parameter the constraint is applied and of what degree.
	       
	       A second constraint $fixCP$ consists of specifying which 
	       control points can not be moved by the routine.

	       For example, if you want to move the point $C(0.5)$ by 
	       $(10,0,10$ and fix the point $C(0.6)$ on the current curve
	       (a move of (0,0,0)) but change its 1st derivative by (0,20,0). 
	       Doing this without modifying control point 4 . 
	       Then the following values must be inputed to the routine.
	       $u_r = 0.5, 0.6$, $D = (10,0,10), (0,0,0), (0,20,0)$, 
	       $D_r = 0, 1, 1$, $D_k = 0, 0, 1$ and $fixCP= 4$. 

	       The values in D should be ordered in respect with r and k.
	       {\em i.e.} for $D_i=D_{r_i}^{(k_i)}$, then $i < j$ implies 
	       that $r_i < r_j$ and that either $r_i < r_j$ or $k_i < k_j$.

	       See section 11.5.1 of the NURBS book for an explanation of 
	       the algorithm.

        Input: ur --> the vector of parameters on which a constraint is applied
	        D --> a vector of the value of $D_r^{(k)}$
	       Dr --> a vector specifying the value of r for D
	       Dk --> a vector specifying the value of k for D
	    fixCP --> a vector specifying which control points {\em can not} be
	              modified.
       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: The values of ur should {\em not} repeat. D,Dr and Dk must
               be of the same size.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::movePoint(const Vector_FLOAT& ur, const Vector_Point3D& D, const Vector_INT& Dr, const Vector_INT& Dk, const Vector_INT& fixCP) {
  int i,j,n ;

  if(D.n != Dr.n || D.n !=Dk.n){
    PlError err("movePoint(ur,D,Dr,Dk,fixCP)");
    err << "The D,Dr,Dk vectors are not of the same size\n" ;
    err << "D.n= " << D.n << ", Dr.n = " << Dr.n 
	<< ", Dk.n = " << Dk.n << endl ;
    err.fatal() ;
  }  

  // setup B
  Matrix_DOUBLE B ;
  
  B.resize(D.n,P.n) ;
  
  int span ;

  Matrix_FLOAT R ;

  B.reset(0.0) ;

  for(i=0;i<D.rows;++i){
    span = findSpan(ur[Dr[i]]) ;
    dersBasisFuns(Dk[i],ur[Dr[i]],span,R) ;
    for(j=0;j<=deg;++j){
      B(i,span-deg+j) = (double)R(Dk[i],j) ;
    }
  }

  // optimize B
  Vector_INT remove(B.cols) ;
  Vector_INT map(B.cols) ;
  remove.reset((int)1.0) ;

  for(j=0;j<B.cols;++j){
    for(i=0;i<B.rows;++i)
      if((B(i,j)*B(i,j))>1e-10){
	remove[j] = 0 ;
	break ;
      }
  }

  for(i=0;i<fixCP.n;++i){
    remove[fixCP[i]] = 1 ;
  }

  n = 0 ;
  for(i=0;i<B.cols;++i){
    if(!remove[i]){
      map[n] = i ;
      ++n ;
    }
  }

  map.resize(n) ;

  Matrix_DOUBLE Bopt(B.rows,n) ;
  for(j=0;j<n;++j){
    for(i=0;i<B.rows;++i)
      Bopt(i,j) = B(i,map[j]) ;
  }

  Matrix_DOUBLE A  ;
  Matrix_DOUBLE Bt(transpose(Bopt)) ;
  Matrix_DOUBLE BBt ;

  //BBt = LAPack.inverse(Bopt*Bt) ;

  cerr << "B size = " << Bopt.rows << ", " << Bopt.cols << endl ;
  cerr << "Bt size= " << Bt.rows << ", " << Bt.cols << endl ;
  BBt = inverse(Bopt*Bt) ;
  /*
  if(LAPack.last_error>0){
    return 0 ; // matrix B*Bt is singular 
  }*/

  A = Bt*BBt ;

  Matrix_DOUBLE dD ;

  dD.resize(D.rows,3) ;
  for(i=0;i<D.rows;++i){
    const Point3D& d = D[i] ; // this makes the SGI compiler happy
    for(j=0;j<3;++j)
      dD(i,j) = (double)d.data[j] ;
  }

  Matrix_DOUBLE dP ;

  dP = A*dD ;

  for(i=0;i<map.n;++i){
    P[map[i]].x() += dP(i,0)*P[map[i]].w() ;
    P[map[i]].y() += dP(i,1)*P[map[i]].w() ;
    P[map[i]].z() += dP(i,2)*P[map[i]].w() ;
  }

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: splitAt --- splits the curve in two curves
  Description: Splits the curve into two curves. 
        Input:  u --> splits at this parametric value
	       cl --> the lower curve
	       cu --> the upper curve
       Output: 1 if the operation, 0 otherwise.
 Restrictions: You {\em must} make sure that you split at a valid parametric 
               value. You can't split a curve at its end points.
   Created by: Philippe Lavoie (16 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::splitAt(float u, PlNurbsCurve& cl, PlNurbsCurve& cu) const {
  if(u<= U[0])
    return 0 ;
  if(u>= U[U.n-1])
    return 0 ;

  // get the multiplicity at u
  int s,j,i ;
  int span = findSpan(u) ;

  if(absolute(u-U[span])<1e-6)
    s = findMult(span) ;
  else
    s = 0 ;

  PlVector<float> X(deg+1-s) ;
  X.reset(u) ;

  cl = *this ;
  if(X.n>0)
    cl.refineKnotVector(X) ;

  span = cl.findSpan(u)-deg ;
  // span is the begining of the upper curve
  cu.resize(cl.P.n-span,deg) ;
  for(i=cl.P.n-1,j=cu.P.n-1;j>=0;--j,--i){
    cu.P[j] = cl.P[i] ;
  }

  for(i=cl.U.n-1,j=cu.U.n-1;j>=0;--j,--i){
    cu.U[j] = cl.U[i] ;
  }
  cl.resize(span,deg) ;

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: mergeOf --- the curve is the result of mergin two curves
  Description: The curve is the result of mergin two curves
        Input: cl --> the lower curve
	       cu --> the upper curve
       Output: 1 if the operation is succesfull, 0 otherwise.
 Restrictions: You must make sure the the knot vectors are compatible, 
               {\em i.e.} that the end knots of the lower curve are the
	       same as the first knots of the upper curve. The curves
	       must also be of the same degree.
   Created by: Philippe Lavoie (16 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurve::mergeOf(const PlNurbsCurve& cl, const PlNurbsCurve &cu){
  if(cl.degree != cu.degree){
    PlError err("PlNurbsCurve::mergeOf");
    err << " The two curves are not of the same degree\n" ;
    err.warning() ;
    return 0 ;
  }
  if((cl.U[cl.U.n-1]-cu.U[0])*(cl.U[cl.U.n-1]-cu.U[0])>1e-8){
    PlError err("PlNurbsCurve::mergeOf");
    err << " The two knot vectors are not compatible.\n" ;
    err << "The first one is " << cl.U << endl ;
    err << "The second is    " << cu.U << endl ;
    err.warning() ;
    return 0 ;
  }
  if(abs2(cl.P[cl.P.n-1]-cu.P[0])>1e-8){
    PlError err("PlNurbsCurve::mergeOf");
    err << " The end control points are NOT the same.\n" ;
    err << " cl.P[n-1] = " << cl.P[cl.P.n-1] << endl ;
    err << " cu.P[0] = " << cu.P[0] << endl ;
    err.warning() ;
    return 0 ;
  }
  resize(cl.P.n+cu.P.n,cl.deg) ;

  int i ;
  for(i=0;i<cl.P.n;++i)
    P[i] = cl.P[i] ;
  for(;i<P.n;++i)
    P[i] = cu.P[i-cl.P.n] ;

  for(i=0;i<cl.U.n;++i)
    U[i] = cl.U[i] ;
  for(;i<U.n;++i)
    U[i] = cu.U[i-cl.U.n+deg+1] ;

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: findSpan --- Determines the knot span index
  Description: Determines the knot span for which their exists non-zero basis 
               functions. The span is the index $k$ for which the parameter 
	       $u$ is valid in the $[u_k,u_{k+1}]$ range.
        Input:   u --> the parametric value
	         U --> the knot vector
               deg --> the degree of the curve
       Output: the span index at $u$.
 Restrictions: $u$ must be in a valid range
   Created by: Philippe Lavoie (17 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int findSpan(float u, const Vector_FLOAT& U, int deg) {
  if(u>=U[U.n-1]) 
    return U.n-deg-1 ;
  if(u<=U[0])
    return deg ;

  int low = deg ;
  int high = U.n-deg ;
  int mid = (low+high)/2 ;

  while(u<U[mid] || u>= U[mid+1]){
    if(u<U[mid])
      high = mid ;
    else
      low = mid ;
    mid = (low+high)/2 ;
  }
  return mid ;
 
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: tesselate --- generates a list of points from the curve
  Description: Generates a list of points from the curve. The list is
               generated within a user specified tolerance.
        Input: tolerance --> the tolerance for the tesselation.
       Output: The list of points.
 Restrictions: 
   Created by: Philippe Lavoie (17 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
BasicList<Point3D> PlNurbsCurve::tesselate(float tolerance,BasicList<float> *uk) const {
  BasicList<Point3D> list,list2 ;

  PlNurbsCurveArray ca ;
  decompose(ca) ;

  if(ca.n==1){
    // get the number of steps 
    float u = 0 ;
    Point3D maxD(0,0,0) ;
    Point3D prev ;

    Vector_Point3D ders(2) ;

    deriveAt(u,1,ders) ;
    
    prev = ders[1] ;

    int i ;
    for(i=1;i<11;++i){
      u = float(i)/float(10) ;
      deriveAt(u,1,ders) ;
      Point3D delta = ders[1]-prev ;
      delta.x = absolute(delta.x) ;
      delta.y = absolute(delta.y) ;
      delta.z = absolute(delta.z) ;
      maxD = maximumRef(maxD,delta) ;
      prev = ders[1] ;
    }
    
    const float sqr2 = float(1.414241527) ;

    int n = (int)rint(sqr2*abs(maxD)/tolerance) ;

    n+=2 ;
    if(n<3) n = 3 ;

    for(i=0;i<n;++i){
      u = (U[U.n-1]-U[0])*float(i)/float(n-1) + U[0] ;
      list.add(point3D(u)) ;
      if(uk)
	uk->add(u) ;
    }

    return list ;
  }
  else{
    for(int i=0;i<ca.n;++i){
      list2 = ca[i].tesselate(tolerance,uk) ;

      // remove the last point from the list to elliminate
      list.erase((BasicNode<Point3D>*)list.last) ;
      list.addElements(list2) ;      
    }
  }
  return list ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: maxInfluence --- finds the parametric value of maximal influence
  Description: Finds the parametric value of maximal influence for a control
               point $i$. This finds the parametric value $u$ were the basis 
	       function $N_{i,p}(u)$ is maximal. 

	       This routine only works for $N_{i,p}(u)$ where $p=1,2,3$. Other
	       values of $p$ are not supported. The reason is that the routine
	       uses pre-computed equations to find the proper values.
        Input: i --> the $i$th control point
	       U --> the knot vector
	       p --> the degree of the basis function
	       u <-- the parametric value of maixmal influence
       Output: 1 if the operation was succesfull, 0 otherwise
 Restrictions: The knot vector must be properly constructed, i.e.
               \begin{equation*}
	       U=\{\underbrace{a,\ldots,a}_{p+1},u_{p+1},\ldots,u_{m-p-1},\underbrace{b,\ldots,b}_{p+1} \} 
	       \end{equation*}
   Created by: Philippe Lavoie (17 October 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int maxInfluence(int i, const Vector_FLOAT& U, int p, float &u){
  if(i>U.n-p-2)
    return 0 ;
  switch(p){
  case 1: u = U[i+1] ; return 1 ; 
  case 2: { 
    float A = U[i] + U[i+1] - U[i+2] - U[i+3] ;
    if(A>=0){
      u = U[i] ;
      return 1 ;
    }
    else{
      u = (U[i]*U[i+1] - U[i+2]*U[i+3])/A ;
      return 1 ;
    }
  }
  case 3:{
    double A = U[i]-U[i+3] ;
    if(A>=0){ // 4 knots at the same place from U[i] to U[i+3]
      u = U[i] ;
      return 1 ;
    }
    A = U[i+1]-U[i+3] ;
    if(A>=0){ // three knots are equal 
      u = U[i+1] ;
      return 1 ;
    }
    // there are 4 points of possible interest
    // the 'good' one lie between U[i+1] and U[i+3]
    double a,b,c,d,e,X ;
    a = U[i] ;
    b = U[i+1] ;
    c = U[i+2] ;
    d = U[i+3] ;
    e = U[i+4] ;

    double t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t15,t16,t18;
    double t21,t22,t24,t25,t27,t28,t31,t32,t34,t35,t45,t46 ,t49 ,t52 ,t56;
    double t63 ,t66 ,t69 ,t88 ,t107,t110 ;
    double t115,t116,t117,t118,t119,t120,t121,t122,t124,t125,t127;
    double t133,t135,t136,t143,t151,t154 ;
    double t14,t17,t19,t20,t26,t29,t30,t33,t36,t37,t38,t39,t47,t55,t57 ;
    double t59,t62,t64,t65,t67,t70,t72,t73,t75,t76,t77,t78,t79,t80 ;
    double t81,t82,t83,t84,t85,t86,t87,t90,t91,t92,t93,t95,t96 ;
    double t97,t99,t101 ;

    t1 = b*e;
    t2 = b*b;
    t3 = b*a;
    t4 = b*d;
    t5 = b*c;
    t6 = d*e;
    t7 = c*d;
    t8 = c*e;
    t9 = c*a;
    t10 = a*e;
    t11 = d*a;
    t12 = a*a;
    t13 = -t1+t2+t3-t4-t5+t6+t7+t8-t9-t10-t11+t12;
    t14 = 1/t13;
    t15 = t2*a;
    t16 = t3*e;
    t17 = t3*c;
    t18 = t7*e;
    t19 = t3*d;
    t20 = t12*b;
    t21 = t2*t12;
    t22 = e*e;
    t24 = c*c;
    t25 = t24*t22;
    t26 = t25*t11;
    t27 = t12*t22;
    t28 = t27*t5;
    t29 = t24*t12;
    t30 = t29*t6;
    t31 = t27*t7;
    t32 = d*d;
    t33 = t32*t12;
    t34 = t33*t5;
    t35 = t33*t8;
    t36 = t29*t4;
    t37 = t33*t1;
    t38 = t12*a;
    t39 = t38*b;
    t45 = t21*t22-t26-t28+t30+t31-t34+t35-t36-t37+t39*t7+
          t39*t8-t38*c*t6+t39*t6;
    t46 = t2*b;
    t47 = t46*a;
    t49 = t15*t18;
    t52 = t3*t22*c*d;
    t55 = t24*d*e*t3;
    t56 = t2*t22;
    t57 = t56*t7;
    t59 = c*t32*t16;
    t62 = t7*e*t12*b;
    t63 = t56*t9;
    t64 = t56*t11;
    t65 = t24*t32;
    t66 = t65*t3;
    t67 = t46*c;
    t69 = t2*t32;
    t70 = t69*t9;
    t72 = t69*t8;
    t73 = t47*t8-3.0*t49+2.0*t52+2.0*t55+t57+2.0*t59-3.0*t62-t63-t64+t66+
          t67*t11-t70+t47*t6+t72;
    t75 = t2*t24;
    t76 = t75*t10;
    t77 = t69*t10;
    t78 = t75*t11;
    t79 = t32*t22;
    t80 = t79*t9;
    t81 = t79*t3;
    t82 = t75*t6;
    t83 = t25*t3;
    t84 = t65*t10;
    t85 = t65*t1;
    t86 = t79*t5;
    t87 = t25*t4;
    t88 = t27*t4;
    t90 = -t76-t77-t78-t80+t81+t82+t83-t84-t85-t86-t87-t88-t67*t6;
    t91 = t29*t1;
    t92 = t21*t8;
    t93 = t21*t6;
    t95 = t21*t7;
    t96 = t21*t24;
    t97 = t46*t12;
    t99 = t2*t38;
    t101 = t65*t22;
    t107 = -t91+2.0*t92+2.0*t93+t46*t38+2.0*t95+t96-t97*c-t99*c+t101+t21*t32-
            t99*d-t99*e-t97*e-t97*d;
    t110 = sqrt(t45+t73+t90+t107);

    X = t14*(2.0*t15-2.0*t16-2.0*t17+2.0*t18-2.0*t19+2.0*t20+2.0*t110)/2;

    if(c-b > 0){      

      // X might be near U[i+2] but due to floating point error, it won't
      // be detected. It happens during tests, so it's possible...
      /*
      if(absolute(X-c)<0.0001*c){
	PlError error("maxInfluence");
	error << "A numerical error in computing the point of maximal influence" ;
	error.warning() ;
	u = X ;
	return 1 ;
      }
      */

      if(X> b && X<=c){
	u = (float)X ;
	return 1 ;
      }

      X = t14*(2.0*t15-2.0*t16-2.0*t17+2.0*t18-2.0*t19+2.0*t20-2.0*t110)/2;
      if(X>b && X<=c){
	u = (float)X ;
	return 1 ;
      }
    }

    t115 = -t9-t32-t6+t1-t3+t4+t10+t11+t8-t5-t22+t7;
    t116 = 1/t115;
    t117 = t6*a;
    t118 = t4*e;
    t119 = d*t22;
    t120 = t32*e;
    t121 = -t26+t28+t30-t31+t34-t35-t36-t37+2.0*t49-3.0*t52+2.0*t55-t57-3.0*
      t59;
    t122 = 2.0*t62+t63-t64+t66+t70-t72-t76-t77-t78+2.0*t80+2.0*t81+t82+t83-
      t84;
    t124 = t32*d;
    t125 = t124*b;
    t127 = t124*t22;
    t133 = t22*e;
    t135 = -t85+2.0*t86-t87-t88-t91-t125*t9-t127*c+t125*t8+t125*t10+t124*a*t8
      +t124*t133-t92+t93;
    t136 = t133*b;
    t143 = t32*t133;
    t151 = -t95+t136*t7-t136*t9+t133*c*t11+t136*t11+t69*t22+t96+t101-t143*c-b
      *t32*t133-t125*t22-t143*a-t127*a+t79*t12;
    t154 = sqrt(t121+t122+t135+t151);


    
    X=t116*(2.0*t117+2.0*t118-2.0*t17-2.0*t119-2.0*t120+2.0*t18+2.0*t154)/2.0;

    if(X>=c && X<d){
      u = (float)X ;
      return 1 ;
    }

    X=t116*(2.0*t117+2.0*t118-2.0*t17-2.0*t119-2.0*t120+2.0*t18-2.0*t154)/2.0;

    if(X>=c && X<d){
      u = (float)X ;
      return 1 ;
    }
  }
  default:
    PlError error("maxInfluence");
    error << "The point of maximal influence is only computed for a degree of 3 or lower." ;
    error.warning() ;
    return 0 ;
  }
  return 0;
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: nurbsBasisFun --- Computes the basis function 
  Description:  Computes the $i$ basis function of degree $p$ at 
               parameter $u$. 
               This is often noted as $N_{ip}(u)$.

	       The B-spline basis function of $p$-degree is defined as
	       \begin{eqnarray}
	       N_{i,0}(u) & = & \left\{ \begin{array}{ll} 1 & \mbox{if $u_i \leq u < u_{i+1}$} \\ 0 & \mbox{otherwise}\end{array}\right. \nonumber \\
	       N_{i,p}(u) & = & \frac{u-u_i}{u_{i+p}-u_i}N_{i,p-1}(u)+\frac{u_{i+p+1}-u}{u_{i+p+1}-u_{i+1}}N_{i+1,p-1}(u) \nonumber
	       \end{eqnarray}
 
	       where the $u_i$ define the knot vector $U = \{u_0,\ldots,u_m\}$
	       as a nondecreasing sequence of real numbers, {\em i.e.}, 
	       $u_i \leq u_{i+1}$ for $i=0,\ldots,m-1$. And $m$ is related
	       to the number of control points $n$ and the degree of the curve
	       $p$ with the relation $m = n + p + 1$. The knot vector has
	       the form

	       \begin{equation}
	       U=\{\underbrace{a,\ldots,a}_{p+1},u_{p+1},\ldots,u_{m-p-1},\underbrace{b,\ldots,b}_{p+1} \} 
	       \end{equation}

        Input: u --> the parametric variable
               i --> specifies which basis function to compute
               p --> the degree to which the basis function is computed
	       U --> the knot vector
       Output: the value of  $N_{ip}(u)$
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float nurbsBasisFun(float u, int i, int p, const Vector_FLOAT& U) {
  float Nip ;
  float saved,Uleft,Uright,temp ;
  
  if(p<1){
    PlError error("nurbsBasisFun") ;
    error << "You need to specify a valid degree for the basis function!\n" ;
    error << "p = " << p << " but it requires a value >0.\n" ;
    error.fatal() ; 
  }

  Vector_FLOAT N(p+1) ;

  if((i==0 && u == U[0]) ||
     (i == U.n-p-2 && u==U[U.n-1])){
    Nip = 1.0 ;
    return Nip ;
  }
  if(u<U[i] || u>=U[i+p+1]){
    Nip = 0.0 ;
    return Nip;
  }
  int j ;
  for(j=0;j<=p;j++){
    if(u>=U[i+j] && u<U[i+j+1]) 
      N[j] = 1.0 ;
    else
      N[j] = 0.0 ;
  }
  for(int k=1; k<=p ; k++){
    if(N[0] == 0.0)
      saved = 0.0 ;
    else
      saved = ( (u-U[i])*N[0])/(U[i+k]-U[i]) ;
    for(j=0;j<p-k+1;j++){
      Uleft = U[i+j+1] ;
      Uright = U[i+j+k+1] ;
      if(N[j+1]==0.0){
	N[j] = saved ;
	saved = 0.0 ;
      }
      else {
	temp = N[j+1]/(Uright-Uleft) ;
	N[j] = saved+(Uright-u)*temp ;
	saved = (u-Uleft)*temp ;
      }
    }
  }
  Nip = N[0] ;

  return Nip ;  
}


