/*=====================================================================
        File: nurbsArray.hh
     Purpose:       
    Revision: $Id: nurbsArray.cc,v 2.0 1998/04/15 18:11:17 lavoie Exp $
  Created by: Philippe Lavoie          (7 Oct, 1997)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=====================================================================*/

#include <nurbs.hh>
#include <nurbsS.hh>
#include <string.h>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  PlNurbsCurveArray --- constructor from a pointer to an array of curves
  Description: Constructor from a pointer to an array of curves.
        Input:
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsCurveArray::PlNurbsCurveArray(PlNurbsCurve* Ca, int s):n(sze){
  sze = rsize = 0 ;
  resize(s) ;
  for(int i=0;i<n;i++)
    C[i] = &Ca[i] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  init --- initialize the array of curves with a vector of nurbs curve
  Description: Initialize the array of curves with a vector of nurbs curve.
        Input: ca --> a pointer to a vector of NURBS curve
              size --> the size of the array
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurveArray::init(PlNurbsCurve* ca,int size){
  resize(size) ;
  for(int i=0;i<n;i++)
    C[i] = &ca[i] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: resize --- resize the PlNurbsCurveArray
  Description: Resize the PlNurbsCurveArray
        Input: size --> the new size
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsCurveArray::resize(int size) {
  int i ;
  if(size<=rsize){
    sze = size ;
    return ;
  }
  PlNurbsCurve** t;
  t = new PlNurbsCurve* [size] ;
  if(C){
    for(i=0;i<rsize;++i)
      t[i] = C[i] ;
    delete []C ;
  }
  for(i=rsize;i<size;++i)
    t[i] = new PlNurbsCurve ;
  C = t ;
  sze = size ;
  rsize = size ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: read --- reads a PlNurbsCurveArray from a file
  Description: Reads a PlNurbsCurveArray from a file.
        Input: filename --> the filename to read the curve array from 
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurveArray::read(const char* filename){
  ifstream fin(filename) ;
  if(!fin) {
    return 0 ;
  }
  int np,d;
  int na ;
  char *type ;
  type = new char[3] ;
  if(!fin.read(type,sizeof(char)*3)) return 0 ;
  int r1 = strncmp(type,"nca",3) ;
  if(!(r1==0))
    return 0 ;
  if(!fin.read((char*)&na,sizeof(int))) return 0 ;

  resize(na) ;

  int nread = 0 ;
  while(nread < na){
    if(!fin.read((char*)&np,sizeof(int))) return 0 ;
    if(!fin.read((char*)&d,sizeof(int))) return 0 ;
    
    operator[](nread).resize(np,d) ;
    
    if(!fin.read((char*)operator[](nread).Knot.memory(),sizeof(float)*operator[](nread).Knot.n)) return 0 ;
    
    float *p,*p2 ;
    p = new float[4*np] ;
    if(!fin.read((char*)p,sizeof(float)*4*np)) return 0 ;
    p2 = p ;
    for(int i=0;i<np;i++){
      CtrlPnt4D t ;
      t.x() = *(p++) ;
      t.y() = *(p++) ;
      t.z() = *(p++) ;
      t.w() = *(p++) ;
      operator[](nread).modCP(i,t) ;
    }
    delete []p2 ;
    ++nread ;
  }

  delete []type ;
  return 1 ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: write --- writes a PlNurbsCurveArray from a file
  Description: Writes a PlNurbsCurveArray from a file.
        Input: filename --> the filename to read the curve array from 
       Output: 0 if an error occurs, 1 otherwise
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsCurveArray::write(const char* filename){
  ofstream fout(filename) ;
  if(!fout) {
    return 0 ;
  }
  if(!fout.write((char*)&"nca",sizeof(char)*3)) return 0 ;
  if(!fout.write((char*)&n,sizeof(int))) return 0 ;

  int nwrote = 0 ;
  while(nwrote < n){
    if(!fout.write((char*)&(operator[](nwrote).CtrlP.n),sizeof(int))) return 0 ;
    if(!fout.write((char*)&(operator[](nwrote).degree),sizeof(int))) return 0 ;
    if(!fout.write((char*)operator[](nwrote).Knot.memory(),sizeof(float)*operator[](nwrote).Knot.n)) return 0 ;
        
    float *p,*p2 ;
    p = new float[4*operator[](nwrote).CtrlP.n] ;
    p2 = p ;
    for(int i=0;i<operator[](nwrote).CtrlP.n;i++){
      *p = operator[](nwrote).CtrlP[i].x() ; p++ ;
      *p = operator[](nwrote).CtrlP[i].y() ; p++ ;
      *p = operator[](nwrote).CtrlP[i].z() ; p++ ;
      *p = operator[](nwrote).CtrlP[i].w() ; p++ ;
    }
    if(!fout.write((char*)p2,sizeof(float)*4*operator[](nwrote).CtrlP.n)) return 0 ;
    delete []p2 ;

    ++nwrote ;
  }

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  PlNurbsSurfaceArray --- constructor from a pointer to an array of curves
  Description: Constructor from a pointer to an array of curves.
        Input:
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurfaceArray::PlNurbsSurfaceArray(PlNurbsSurface* Sa, int s):n(sze){
  sze = rsize = 0 ;
  resize(s) ;
  for(int i=0;i<n;i++)
    S[i] = &Sa[i] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  init --- initialize the array of curves with a vector of nurbs curve
  Description: Initialize the array of curves with a vector of nurbs curve.
        Input: ca --> a pointer to a vector of NURBS curve
              size --> the size of the array
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurfaceArray::init(PlNurbsSurface* Sa,int size){
  resize(size) ;
  for(int i=0;i<n;i++)
    S[i] = &Sa[i] ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: resize --- resize the PlNurbsSurfaceArray
  Description: Resize the PlNurbsSurfaceArray
        Input: size --> the new size
       Output:
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurfaceArray::resize(int size) {
  int i ;
  if(size<=rsize){
    sze = size ;
    return ;
  }
  PlNurbsSurface** t;
  t = new PlNurbsSurface* [size] ;
  if(S){
    for(i=0;i<rsize;++i)
      t[i] = S[i] ;
    delete []S ;
  }
  for(i=rsize;i<size;++i)
    t[i] = new PlNurbsSurface ;
  S = t ;
  sze = size ;
  rsize = size ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: operator= --- copy one surface array to another
  Description: Copy one surface array to another
        Input: S --> the array to copy
       Output: a reference to itself
 Restrictions:
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurfaceArray& PlNurbsSurfaceArray::operator=(const PlNurbsSurfaceArray& Sa){
  resize(Sa.n) ;

  for(int i=0;i<n;++i)
    *(S[i]) = Sa[i] ;
  return *this ;
}
