/*=============================================================================
        File: nurbsS.cc
     Purpose:       
    Revision: $Id: nurbsS.cc,v 2.7 1998/06/03 18:24:22 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/

#include <nurbsS.hh>
#include <string.h>
#include <matrixRT.hh>
#include <math.h>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: PlNurbsSurface ---  default constructor
  Description: Default constructor.
        Input:     
       Output:
 Restrictions: The surface is initialized to invalid values. Use a reset or
               a read function to set them to correct values.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface::PlNurbsSurface(): 
  PlParaSurface(),KnotU(U), KnotV(V), CtrlP(P), degreeU(degU), degreeV(degV),
  U(1),V(1),P(1,1),degU(0),degV(0)
{
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: PlNurbsSurface ---  the copy constructor
  Description: The copy constructor.
        Input: s --> the PlNurbsSurface to copy    
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface::PlNurbsSurface(const PlNurbsSurface& s): 
  PlParaSurface(), KnotU(U), KnotV(V), CtrlP(P), degreeU(degU), degreeV(degV),
  U(s.U), V(s.V), P(s.P), degU(s.degU), degV(s.degV)
{
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: PlNurbsSurface --- constructor with points in 4D
  Description: Constructor with points in 4D.
        Input: DegU --> the degree in the $u$ direction
               DegV --> the degree in the $v$ direction
                 Uk --> the $u$ knot vector
                 Vk --> the $v$ knot vector   
		 Cp --> the matrix of control points in 4D
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface::PlNurbsSurface(int DegU, int DegV, const Vector_FLOAT& Uk, const Vector_FLOAT& Vk, const Matrix_CtrlPnt4D& Cp) : 
  PlParaSurface(),KnotU(U), KnotV(V), CtrlP(P), degreeU(degU), degreeV(degV),
  U(Uk),V(Vk),P(Cp),degU(DegU),degV(DegV)
{
  int bad = 0 ;

  if(U.n != P.rows+degU+1){
    PlError err("PlNurbsSurface constructor") ;
    err << "The U knot vector and the number of rows of the control points are incompatible\n" ;
    err << "P.rows = " << P.rows << ", U.n = " << U.n << endl ;
    bad = 1 ;
    err.fatal() ;
  }
  if(V.n != P.cols+degV+1){
    PlError err("PlNurbsSurface constructor") ;    
    err << "The V knot vector and the number of columns of the control points are incompatible\n" ;
    err << "P.cols = " << P.cols << ", V.n = " << V.n << endl ;
    bad = 1 ;
    err.fatal() ;
  }
  if(bad)
    exit(-1) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: PlNurbsSurface --- constructor with points in 3D
  Description: Constructor with points in 3D.
        Input: DegU --> the degree of the surface in the U direction
               DegV --> the degree of the surface in the V direction
                Uk --> the U knot vector 
                Vk --> the V knot vector
                Cp --> the matrix of 3D control points
                 W --> the weight value for each control points
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface::PlNurbsSurface(int DegU, int DegV, Vector_FLOAT& Uk, Vector_FLOAT& Vk, Matrix_Point3D& Cp, Matrix_FLOAT& W) : 
  PlParaSurface(),KnotU(U), KnotV(V), CtrlP(P), degreeU(degU), degreeV(degV),
  U(Uk),V(Vk),P(Cp.rows,Cp.cols),degU(DegU),degV(DegV)
{
  int bad = 0 ;

  if(U.n != Cp.rows+degU+1){
    PlError err("PlNurbsSurface constructor") ;
    err << "The U knot vector and the number of rows of the control points are incompatible\n" ;
    err << "P.rows = " << P.rows << ", U.n = " << U.n << endl ;
    bad = 1 ;
    err.fatal() ;
  }
  if(V.n != Cp.cols+degV+1){
    PlError err("PlNurbsSurface constructor") ;
    err << "The V knot vector and the number of columns of the control points are incompatible\n" ;
    err << "P.cols = " << P.cols << ", V.n = " << V.n << endl ;
    bad = 1 ;
    err.fatal() ;
  }
  if(W.rows != Cp.rows || W.cols != Cp.cols){
    PlError err("PlNurbsSurface constructor") ;
    err << "The dimension of the weights are incompatible with the dimension of the control poitns\n" ;
    err << "W( " << W.rows << ", " << W.cols << ") and P( " << P.rows << ", " << P.cols << ") \n" ;
    bad = 1 ;
    err.fatal() ;
  }

  for(int i=0;i<Cp.rows;i++)
    if(bad)
      break ;
    else{
      for(int j=0;j<Cp.cols;j++)
	if(W(i,j)==0.0){
	  PlError err("PlNurbsSurface constructor") ;
	  err << "Error initializing a PlNurbsSurface.\n" ;
	  err << "The weight W(" << i << ", " << j << ") is equal to 0.\n" ;
	  bad = 1 ;
	  err.fatal() ;
	  break ;
	}
	else {
	  P(i,j) = Cp(i,j) ;
	  P(i,j) *= W(i,j) ;
	}
    }

  if(bad)
    exit(-1) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: operator= --- PlNurbsSurface assignment
  Description: PlNurbsSurface assignment.
        Input: nS --> the NURBS surface to copy
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface& PlNurbsSurface::operator=(const PlNurbsSurface& nS){
  P = nS.P ;
  U = nS.U ;
  V = nS.V ;
  degU = nS.degU ;
  degV = nS.degV ;
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: ok --- determines if the surface is valid
  Description: Determines if the surface is valid. The routine only verifies 
               if the number of control points in the U and V direction 
	       matches the length of the U and V knot vectors.
        Input:     
       Output: 1 if the surface is valid, 0 otherwise
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::ok() {
  if(P.rows <= degU)
    return 0 ;
  if(P.cols <= degV)
    return 0 ;
  if(P.rows != U.n+degU+1)
    return 0 ;
  if(P.cols != V.n+degV+1)
    return 0 ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: resize --- resize the surface
  Description: Resize the surface. Proper values must be assigned once this 
               function has been called since the resize operator is 
	       destructive.
        Input:  Pu --> the number of control points in the U direction
                Pv --> the number of control points in the V direction
              DegU --> the degree of the surface in the U direction
              DegV --> the degree of the surface in the V direction
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::resize(int Pu, int Pv, int DegU, int DegV){
  P.resize(Pu,Pv) ;
  degU = DegU ;
  degV = DegV ;
  U.resize(Pu+DegU+1) ;
  V.resize(Pv+DegV+1) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: resizeKeep --- resize the surface
  Description: Resize the surface while keeping the old values.
        Input:  Pu --> the number of control points in the U direction
                Pv --> the number of control points in the V direction
              DegU --> the degree of the surface in the U direction
              DegV --> the degree of the surface in the V direction
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::resizeKeep(int Pu, int Pv, int DegU, int DegV){
  P.resizeKeep(Pu,Pv) ;
  degU = DegU ;
  degV = DegV ;
  U.resize(Pu+DegU+1) ;
  V.resize(Pv+DegV+1) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: skinV --- generates a NURBS surface from  skinning
  Description: The NURBS surface is generates from skinning. 
               The skinning is performed in the V direction.
        Input:   ca <-> an array of NURBS curves
               degV --> the degree to skin in the V direction
               surf <-- the skinned surface
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: The number of curves to skin from must be greater than degV
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::skinV(PlNurbsCurveArray& ca, int dV) {
  Vector_FLOAT vk(ca.n) ;
  Vector_FLOAT d(ca.n) ;
  int i,k,N,K ;

  if(ca.n<dV){
    PlError err("PlNurbsSurface skinV") ;
    err << "The number of curves are insufficient for the degree of interpolation specified.\n" ;
    err << "n of curves = " << ca.n << ", degree of interpolation = " << dV << endl ;
    err.warning() ;
    return 0 ;
  }

  generateCompatibleCurves(ca) ;

  K = ca.n ;
  N = ca[0].CtrlP.n ;

  resize(N,K,ca[0].degree,dV) ;

  d.resize(ca[0].CtrlP.n) ;
  for(i=0;i<N;i++){
    d[i] = 0 ;
    for(k=1;k<K;k++){
      d[i] += abs(ca[k].CtrlP[i]-ca[k-1].CtrlP[i]) ;
    }
  }

  vk[0] = 0 ;
  for(k=1;k<K;k++){
    vk[k] = 0 ;
    for(i=0;i<N;i++)
      vk[k] += abs(ca[k].CtrlP[i]-ca[k-1].CtrlP[i])/d[i] ;
    vk[k] /= (float)N ;//+ 1.0 ;
    vk[k] += vk[k-1] ;
  }
  vk[vk.n-1] = 1.0 ;

  for(i=1;i<K-degV;i++){
    V[i+degV] = 0 ;
    for(k=i;k<i+degV;k++)
      V[i+degV] += vk[k] ;
    V[i+degV] /= (float)degV ;
  }
  for(i=0;i<=degV;i++) V[i] = 0.0 ;
  for(i=V.n-degV-1;i<V.n;i++) V[i] = 1.0 ;


  Vector_CtrlPnt4D Q(K) ;
  PlNurbsCurve i_curve ;


  for(i=0;i<N;i++){
    for(k=0;k<K;k++)
      Q[k] = ca[k].CtrlP[i] ;
    i_curve.globalInterp(Q,vk,V,degV);
    for(k=0;k<K;k++)
      P(i,k) = i_curve.CtrlP[k] ;
  }

  U = ca[0].Knot ;

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: skinU --- generates a NURBS surface from  skinning
  Description: The NURBS surface is generates from skinning. 
               The skinning is performed in the U direction.
        Input:    ca <-> an array of NURBS curves
                degU --> the degree to skin in the Udirection
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: The number of curves to skin from must be greater than degU
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::skinU(PlNurbsCurveArray& ca, int dU) {
  Vector_FLOAT uk(ca.n);
  Vector_FLOAT d(ca.n) ;
  int i,k,N,K ;

  if(ca.n<dU){
    PlError err("PlNurbsSurface skinU") ;
    err << "The number of curves are insufficient for the degree of interpolation specified.\n" ;
    err << "n of curves = " << ca.n << ", degree of interpolation = " << dU << endl ;
    err.warning() ;
    return 0 ;
  }

  generateCompatibleCurves(ca) ;

  K = ca.n ;
  N = ca[0].CtrlP.n ;

  resize(K,N,dU,ca[0].degree) ;

  d.resize(ca[0].CtrlP.n) ;
  for(i=0;i<N;i++){
    d[i] = 0 ;
    for(k=1;k<K;k++){
      d[i] += abs(ca[k].CtrlP[i]-ca[k-1].CtrlP[i]) ;
    }
  }

  uk[0] = 0 ;
  for(k=1;k<K;k++){
    uk[k] = 0 ;
    for(i=0;i<N;i++)
      uk[k] += abs(ca[k].CtrlP[i]-ca[k-1].CtrlP[i])/d[i] ;
    uk[k] /= (float)N ;//+ 1.0 ;
    uk[k] += uk[k-1] ;
  }
  uk[uk.n-1] = 1.0 ;

  for(i=1;i<K-degU;i++){
    U[i+degU] = 0 ;
    for(k=i;k<i+degU;k++)
      U[i+degU] += uk[k] ;
    U[i+degU] /= (float)degU ;
  }
  for(i=0;i<=degU;i++) U[i] = 0.0 ;
  for(i=U.n-degU-1;i<U.n;i++) U[i] = 1.0 ;


  Vector_CtrlPnt4D Q(K) ;
  PlNurbsCurve i_curve ;


  for(i=0;i<N;i++){
    for(k=0;k<K;k++)
      Q[k] = ca[k].CtrlP[i] ;
    i_curve.globalInterp(Q,uk,U,degU) ;
    for(k=0;k<K;k++)
      P(k,i) = i_curve.CtrlP[k] ;
  }

  V = ca[0].Knot ;

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: surfMeshParams --- Computes the parameters for global surface 
                                  interpolation
  Description: Computes the parameters for global surface interpolation. 
               For more information, see A9.3 on p377 on the NURBS book.
        Input:    Q --> the matrix of 3D points
                 uk <-- the knot coefficients in the U direction
                 vl <-- the knot coefficients in the V direction            
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int surfMeshParams(const Matrix_Point3D& Q, Vector_FLOAT& uk, Vector_FLOAT& vl){
  int n,m,k,l,num ;
  double d,total ;
  Vector_FLOAT cds(Q.rows) ;

  n = Q.rows ;
  m = Q.cols ;
  uk.resize(n) ;
  vl.resize(m) ;
  num = m ;
  

  // Compute the uk
  uk.reset(0) ;

  for(l=0;l<m;l++){
    total = 0.0 ;
    for(k=1;k<n;k++){
      cds[k] = abs(Q(k,l)-Q(k-1,l)) ;
      total += cds[k] ;
    }
    if(total==0.0) 
      num-- ;
    else {
      d = 0.0 ;
      for(k=1;k<n;k++){
	d += cds[k] ;
	uk[k] += d/total ;
      }
    }
  }
  if(num==0) 
    return 0 ;
  for(k=1;k<n-1;k++)
    uk[k] /= num ;
  uk[n-1] = 1.0 ;

  // Compute the vl
  vl.reset(0) ;
  cds.resize(m) ;

  num = n ;

  for(k=0;k<n;k++){
    total = 0.0 ;
    for(l=1;l<m;l++){
      cds[l] = abs(Q(k,l)-Q(k,l-1)) ;
      total += cds[l] ;
    }
    if(total==0.0) 
      num-- ;
    else {
      d = 0.0 ;
      for(l=1;l<m;l++){
	d += cds[l] ;
	vl[l] += d/total ;
      }
    }
  }
  if(num==0) 
    return 0 ;
  for(l=1;l<m-1;l++)
    vl[l] /= num ;
  vl[m-1] = 1.0 ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: surfMeshParams --- Computes the parameters for global surface 
                                  interpolation
  Description: Computes the parameters for global surface interpolation. 
               For more information, see A9.3 on p377 on the NURBS book.
        Input:    Q --> the matrix of 3D points
                 uk <-- the knot coefficients in the U direction
                 vl <-- the knot coefficients in the V direction            
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int surfMeshParams(const Matrix_CtrlPnt4D& Q, Vector_FLOAT& uk, Vector_FLOAT& vl){
  int n,m,k,l,num ;
  double d,total ;
  Vector_FLOAT cds(Q.rows) ;

  n = Q.rows ;
  m = Q.cols ;
  uk.resize(n) ;
  vl.resize(m) ;
  num = m ;
  

  // Compute the uk
  uk.reset(0) ;

  for(l=0;l<m;l++){
    total = 0.0 ;
    for(k=1;k<n;k++){
      cds[k] = abs(Q(k,l)-Q(k-1,l)) ;
      total += cds[k] ;
    }
    if(total==0.0) 
      num-- ;
    else {
      d = 0.0 ;
      for(k=1;k<n;k++){
	d += cds[k] ;
	uk[k] += d/total ;
      }
    }
  }
  if(num==0) 
    return 0 ;
  for(k=1;k<n-1;k++)
    uk[k] /= num ;
  uk[n-1] = 1.0 ;

  // Compute the vl
  vl.reset(0) ;
  cds.resize(m) ;

  num = n ;

  for(k=0;k<n;k++){
    total = 0.0 ;
    for(l=1;l<m;l++){
      cds[l] = abs(Q(k,l)-Q(k,l-1)) ;
      total += cds[l] ;
    }
    if(total==0.0) 
      num-- ;
    else {
      d = 0.0 ;
      for(l=1;l<m;l++){
	d += cds[l] ;
	vl[l] += d/total ;
      }
    }
  }
  if(num==0) 
    return 0 ;
  for(l=1;l<m-1;l++)
    vl[l] /= num ;
  vl[m-1] = 1.0 ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: globalInterp --- generates a surface using global interpolation
  Description: Generates a NURBS surface using global interpolation. 
        Input:    Q --> a matrix of 3D points
                 pU --> the degree of interpolation in the U direction
                 pV --> the degree of interpolation in the V direction
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::globalInterp(const Matrix_Point3D& Q, int pU, int pV){
  Vector_FLOAT vk,uk ;

  resize(Q.rows,Q.cols,pU,pV) ;

  surfMeshParams(Q,uk,vk) ;
  knotAveraging(uk,pU,U) ;
  knotAveraging(vk,pV,V) ;
  

  Vector_CtrlPnt4D Pts(Q.rows) ;
  PlNurbsCurve R ;
  
  int i,j ;

  for(j=0;j<Q.cols;j++){
    for(i=0;i<Q.rows;i++)
      Pts[i] = Q(i,j) ;
    R.globalInterp(Pts,uk,U,pU);
    for(i=0;i<Q.rows;i++)
      P(i,j) = R.CtrlP[i] ;
  }

  Pts.resize(Q.cols) ;
  for(i=0;i<Q.rows;i++){
    for(j=0;j<Q.cols;j++)
      Pts[j] = P(i,j) ;
    R.globalInterp(Pts,vk,V,pV) ;
    for(j=0;j<Q.cols;j++)
      P(i,j) = R.CtrlP[j] ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: globalInterp --- generates a surface using global interpolation
  Description: Generates a NURBS surface using global interpolation. 
        Input:    Q --> a matrix of 4D points
                 pU --> the degree of interpolation in the U direction
                 pV --> the degree of interpolation in the V direction
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::globalInterp(const Matrix_CtrlPnt4D& Q, int pU, int pV){
  Vector_FLOAT vk,uk ;

  resize(Q.rows,Q.cols,pU,pV) ;

  surfMeshParams(Q,uk,vk) ;
  knotAveraging(uk,pU,U) ;
  knotAveraging(vk,pV,V) ;
  

  Vector_CtrlPnt4D Pts(Q.rows) ;
  PlNurbsCurve R ;
  
  int i,j ;

  for(j=0;j<Q.cols;j++){
    for(i=0;i<Q.rows;i++)
      Pts[i] = Q(i,j) ;
    R.globalInterp(Pts,uk,U,pU);
    for(i=0;i<Q.rows;i++)
      P(i,j) = R.CtrlP[i] ;
  }

  Pts.resize(Q.cols) ;
  for(i=0;i<Q.rows;i++){
    for(j=0;j<Q.cols;j++)
      Pts[j] = P(i,j) ;
    R.globalInterp(Pts,vk,V,pV) ;
    for(j=0;j<Q.cols;j++)
      P(i,j) = R.CtrlP[j] ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: leastSquares --- generates a surface using global least
                                squares approximation.
  Description: Generates a NURBS surface using global least squares
               approximation.
        Input:    Q --> a matrix of 3D points
                 pU --> the degree of interpolation in the U direction
                 pV --> the degree of interpolation in the V direction
		 nU --> the number of points in the U direction
		 nV --> the number of poitns in the V direction
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::leastSquares(const Matrix_Point3D& Q, int pU, int pV, int nU, int nV){
  Vector_FLOAT vk,uk ;

  resize(nU,nV,pU,pV) ;

  surfMeshParams(Q,uk,vk) ;

  Vector_CtrlPnt4D Pts(Q.rows) ;
  PlNurbsCurve R ;
  
  int i,j ;

  Matrix_CtrlPnt4D P2 ;

  P2.resize(nU,Q.cols) ;

  for(j=0;j<Q.cols;j++){
    for(i=0;i<Q.rows;i++)
      Pts[i] = Q(i,j) ;
    R.leastSquares(Pts,pU,nU,uk);
    for(i=0;i<P.rows;i++)
      P2(i,j) = R.CtrlP[i] ;
    if(j==0)
      U = R.Knot ;
  }

  Pts.resize(Q.cols) ;
  for(i=0;i<P.rows;i++){
    for(j=0;j<Q.cols;j++)
      Pts[j] = P2(i,j) ;
    R.leastSquares(Pts,pV,nV,vk) ;
    for(j=0;j<P.cols;j++)
      P(i,j) = R.CtrlP[j] ;
    if(i==0)
      V = R.Knot ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: globalSurfInterpXY --- generates a surface using global interpolation
  Description: Generates a NURBS surface using global interpolation. 
               The data points are assumed to be part of grided points 
	       in x-y. {\em i.e.} the original data set as points covering
	       the xy plane at regular $x$ and $y$ intervals with only the $z$ 
	       being a free variable
        Input:    Q --> a matrix of 3D points
                 pU --> the degree of interpolation in the U direction
                 pV --> the degree of interpolation in the V direction
                  S <-- the interpolated surface
       Output:
 Restrictions: Q(0,0) in x-y should be the smallest corner and 
               Q(Q.rows-1,Q.cols-1) the biggest.
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void globalSurfInterpXY(const Matrix_Point3D& Q, int pU, int pV, PlNurbsSurface& S) {
  Vector_FLOAT uk,vk ;
  float um,uM ;
  float vm,vM ;

  um = Q(0,0).y ;
  vm = Q(0,0).x ;
  uM = Q(Q.rows-1,0).y ;
  vM = Q(0,Q.cols-1).x ;

  uk.resize(Q.rows) ;
  vk.resize(Q.cols) ;

  uk[0] = 0.0 ;
  vk[0] = 0.0 ;
  uk[uk.n-1] = 1.0 ;
  vk[vk.n-1] = 1.0 ;

  float dU = uM-um ;
  float dV = vM-vm ;

  int i ;
  for(i=1;i<uk.n-1;++i){
    uk[i] = Q(i,0).y/dU ;
  }
  
  for(i=1;i<vk.n-1;++i){
    vk[i] = Q(0,i).x/dV ;
  }
  
  globalSurfInterpXY(Q,pU,pV,S,uk,vk) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: globalSurfInterpXY --- generates a surface using global interpolation
  Description: Generates a NURBS surface using global interpolation. 
               The data points are assumed to be part of grided points 
	       in x-y. i.e. the original data set as points covering 
	       the xy plane at regular $x$ and $y$ intervals with only the $z$ 
	       being a free variable
        Input:    Q --> a matrix of 3D points
                 pU --> the degree of interpolation in the U direction
                 pV --> the degree of interpolation in the V direction
                  S <-- the interpolated surface
       Output:
 Restrictions:  Q(0,0) in x-y should be the smallest corner and 
                Q(Q.rows-1,Q.cols-1) the biggest.
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void globalSurfInterpXY(const Matrix_Point3D& Q, int pU, int pV, PlNurbsSurface& S, const Vector_FLOAT& uk, const Vector_FLOAT& vk){
  Vector_FLOAT V,U ;
  int i,j ;


  knotAveraging(uk,pU,U) ;
  knotAveraging(vk,pV,V) ;
  
  Vector_CtrlPnt4D P(Q.rows) ;
  PlNurbsCurve R ;
  
  S.resize(Q.rows,Q.cols,pU,pV) ;
  S.U = U ;
  S.V = V ;

  for(j=0;j<Q.cols;j++){
    for(i=0;i<Q.rows;i++)
      P[i] = Q(i,j) ;
    R.globalInterp(P,uk,U,pU) ;
    for(i=0;i<Q.rows;i++)
      S.P(i,j) = R.CtrlP[i] ;
  }

  P.resize(Q.cols) ;
  for(i=0;i<Q.rows;i++){
    for(j=0;j<Q.cols;j++)
      P[j] = S.P(i,j) ;
    R.globalInterp(P,vk,V,pV) ;
    for(j=0;j<Q.cols;j++)
      S.P(i,j) = R.CtrlP[j] ;
  }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: globalSurfApprox --- generates a surface using global approximation
  Description: Generates a NURBS surface using global approximation.
        Input:    Q --> a matrix of 3D points
                 pU --> the degree of interpolation in the U direction
                 pV --> the degree of interpolation in the V direction
                  S <-- the interpolated surface
       Output:
 Restrictions: This routine is still in a research phase
   Created by: Philippe Lavoie      (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void globalSurfApprox(const Matrix_Point3D& Q, int pU, int pV, PlNurbsSurface& S, double error){

  PlNurbsCurveArray R ; 
  Vector_Point3D P ;

  Matrix_CtrlPnt4D St ;
  Vector_FLOAT Ut,Vt ;

  Vector_FLOAT vk,uk ;

  surfMeshParams(Q,uk,vk) ;
  
  R.resize(Q.cols) ;
  P.resize(Q.rows) ;

  int i,j ;

  for(j=0;j<Q.cols;j++){
    for(i=0;i<Q.rows;i++)
      P[i] = Q(i,j) ;
    R[j].globalApproxErrBnd(P,uk,pU,error) ;
  }


  generateCompatibleCurves(R) ;

  Ut.resize(R[0].Knot.n) ;
  Ut = R[0].Knot ;

  St.resize(R[0].CtrlP.n,R.n) ;

  for(i=0;i<R[0].CtrlP.n;i++){
    for(j=0;j<R.n;j++)
      St(i,j) = R[j].CtrlP[i] ;
  }

  P.resize(St.cols) ;
  R.resize(St.rows) ;
  for(i=0;i<St.rows;i++){
    for(j=0;j<St.cols;j++)
      P[j] = project(St(i,j)) ;
    R[i].globalApproxErrBnd(P,vk,pV,error) ;
  }

  generateCompatibleCurves(R) ;

  Vt.resize(R[0].Knot.n) ;
  Vt = R[0].Knot ;
  
  S.resize(St.rows,R[0].CtrlP.n,pU,pV) ;
  for(i=0;i<S.CtrlP.rows;i++)
    for(j=0;j<S.CtrlP.cols;j++){
      S.P(i,j) = R[i].CtrlP[j] ;
    }
  S.U = Ut ;
  S.V = Vt ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: degreeElevate --- degree elevate the surface in the U and V 
                                 direction
  Description: Degree elevate the surface in the U and V direction.
        Input: tU --> elevate the degree of the surface in the u direction 
	              by this amount.
               tV --> elevate the degree of the surface in the v direction 
	              by this amount.
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::degreeElevate(int tU, int tV) {
  degreeElevateU(tU) ;
  degreeElevateV(tV) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: degreeElevateU --- degree elevate the surface in the U direction
  Description: Degree elevate the surface in the U direction.
        Input: t --> elevate the degree in the u direction by this amount.
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::degreeElevateU(int t) {
  if(t<=0){
    return ;
  }

  PlNurbsSurface S(*this) ;
  
  int i,j,k ;
  int n = S.CtrlP.rows-1;
  int p = S.degreeU ;
  int m = n+p+1;
  int ph = p+t ;
  int ph2 = ph/2 ;
  Matrix_FLOAT bezalfs(p+t+1,p+1) ; // coefficients for degree elevating the Bezier segment
  Matrix_CtrlPnt4D bpts(p+1,S.P.cols) ; // pth-degree Bezier control points of the current segment
  Matrix_CtrlPnt4D ebpts(p+t+1,S.P.cols) ; // (p+t)th-degree Bezier control points of the  current segment
  Matrix_CtrlPnt4D Nextbpts(p-1,S.P.cols) ; // leftmost control points of the next Bezier segment
  Vector_FLOAT alphas(p-1) ; // knot instertion alphas.
  
  // Compute the Binomial coefficients
  Matrix_FLOAT Bin(ph+1,ph2+1) ;
  binomialCoef(Bin) ;
  
  // Compute Bezier degree elevation coefficients
  float inv,mpi ;
  bezalfs(0,0) = bezalfs(ph,p) = 1.0 ;
  for(i=1;i<=ph2;i++){
    inv= 1.0/Bin(ph,i) ;
    mpi = minimum(p,i) ;
    for(j=maximum(0,i-t); j<=mpi; j++){
      bezalfs(i,j) = inv*Bin(p,j)*Bin(t,i-j) ;
    }
  }
  
  for(i=ph2+1;i<ph ; i++){
    mpi = minimum(p,i) ;
    for(j=maximum(0,i-t); j<=mpi ; j++)
      bezalfs(i,j) = bezalfs(ph-i,p-j) ;
  }
  
  // Allocate more control points than necessary
  resize(S.P.rows+S.P.rows*t,S.P.cols,ph,S.degreeV) ; 
  
  int colJ ;
  int mh = ph ;
  int kind = ph+1 ;
  float ua = S.U[0] ;
  float ub = 0.0 ;
  int r=-1 ; 
  int oldr ;
  int a = p ;
  int b = p+1 ; 
  int cind = 1 ;
  int rbz,lbz = 1 ; 
  int mul,save,s;
  float alf ;
  int first, last, kj ;
  float den,bet,gam,numer ;
  
  for(i=0;i<S.P.cols;i++)
    P(0,i) = S.P(0,i) ;
  for(i=0; i <= ph ; i++){
    U[i] = ua ;
  }
  
  // Initialize the first Bezier segment
  
  for(i=0;i<=p ;i++) 
    for(j=0;j<S.P.cols;j++)
      bpts(i,j) = S.P(i,j);
  
  while(b<m){ // Big loop thru knot vector
    i=b ;
    while(b<m && S.U[b] >= S.U[b+1]) // for some odd reasons... == doesn't work
      b++ ;
    mul = b-i+1 ; 
    mh += mul+t ;
    ub = S.U[b] ;
    oldr = r ;
    r = p-mul ;
    if(oldr>0)
      lbz = (oldr+2)/2 ;
    else
      lbz = 1 ;
    if(r>0) 
      rbz = ph-(r+1)/2 ;
    else
      rbz = ph ;
    if(r>0){ // Insert knot to get Bezier segment
      numer = ub-ua ;
      for(k=p;k>mul;k--){
	alphas[k-mul-1] = numer/(S.U[a+k]-ua) ;
      }
      for(j=1;j<=r;j++){
	save = r-j ; s = mul+j ;
	for(k=p;k>=s;k--){
	  for(colJ=0;colJ<S.P.cols;colJ++){
	    bpts(k,colJ) = alphas[k-s]*bpts(k,colJ)+(1.0-alphas[k-s])*bpts(k-1,colJ) ;}
	}
	if(Nextbpts.rows>0)
	  for(colJ=0;colJ<S.P.cols;colJ++){
	    Nextbpts(save,colJ) = bpts(p,colJ) ;}
      }
    }
    
    for(i=lbz;i<=ph;i++){ // Degree elevate Bezier,  only the points lbz,...,ph are used
      for(colJ=0;colJ<S.P.cols;colJ++){
	ebpts(i,colJ) = 0.0 ;}
      mpi = minimum(p,i) ;
      for(j=maximum(0,i-t); j<=mpi ; j++)
	for(colJ=0;colJ<S.P.cols;colJ++){
	  ebpts(i,colJ) += bezalfs(i,j)*bpts(j,colJ) ;}
    }
    
    if(oldr>1){ // Must remove knot u=c.U[a] oldr times
      // if(oldr>2) // Alphas on the right do not change
      //	alfj = (ua-nc.U[kind-1])/(ub-nc.U[kind-1]) ;
      first = kind-2 ; last = kind ;
      den = ub-ua ;
      bet = (ub-U[kind-1])/den ;
      for(int tr=1; tr<oldr; tr++){ // Knot removal loop
	i = first ; j = last ;
	kj = j-kind+1 ;
	while(j-i>tr){ // Loop and compute the new control points for one removal step
	  if(i<cind){
	    alf=(ub-U[i])/(ua-U[i]) ;
	    for(colJ=0;colJ<S.P.cols;colJ++){
	      P(i,colJ) = alf*P(i,colJ) + (1.0-alf)*P(i-1,colJ) ;}
	  }
	  if( j>= lbz){
	    if(j-tr <= kind-ph+oldr){
	      gam = (ub-U[j-tr])/den ;
	      for(colJ=0;colJ<S.P.cols;colJ++){
		ebpts(kj,colJ) = gam*ebpts(kj,colJ) + (1.0-gam)*ebpts(kj+1,colJ) ;}
	    }
	    else{
	      for(colJ=0;colJ<S.P.cols;colJ++){
		ebpts(kj,colJ) = bet*ebpts(kj,colJ)+(1.0-bet)*ebpts(kj+1,colJ) ;}
	    }
	  }
	  ++i ; --j; --kj ;
	}
	--first ; ++last ;
      }
    }
  
    if(a!=p) // load the knot u=c.U[a]
      for(i=0;i<ph-oldr; i++){
	U[kind++] = ua ; 
      }
    
    for(j=lbz; j<=rbz ; j++) { // load control points onto nc
      for(colJ=0;colJ<S.P.cols;colJ++)
	P(cind,colJ) = ebpts(j,colJ) ; 
      ++cind ;
    }
    
    if(b<m){ // Set up for next pass thru loop
      for(colJ=0;colJ<S.P.cols;colJ++){
	for(j=0;j<r;j++)
	  bpts(j,colJ) = Nextbpts(j,colJ) ;
	for(j=r;j<=p;j++)
	  bpts(j,colJ) = S.P(b-p+j,colJ) ;
      }
      a=b ; 
      b++ ;
      ua = ub ;
    }
    else{
      for(i=0;i<=ph;i++)
	U[kind+i] = ub ;
    }
  }
  // Resize to the proper number of control points  
  resizeKeep(mh-ph,S.P.cols,ph,S.degreeV) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: degreeElevateV --- degree elevate the surface in the V direction
  Description: Degree elevate the surface in the V direction.
        Input: t --> elevate the degree in the v direction by this amount.
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::degreeElevateV(int t) {
  if(t<=0){
    return ;
  }

  PlNurbsSurface S(*this) ;

  int i,j,k ;
  int n = S.CtrlP.cols-1;
  int p = S.degreeV ;
  int m = n+p+1;
  int ph = p+t ;
  int ph2 = ph/2 ;
  Matrix_FLOAT bezalfs(p+t+1,p+1) ; // coefficients for degree elevating the Bezier segment
  Matrix_CtrlPnt4D bpts(p+1,S.P.rows) ; // pth-degree Bezier control points of the current segment
  Matrix_CtrlPnt4D ebpts(p+t+1,S.P.rows) ; // (p+t)th-degree Bezier control points of the  current segment
  Matrix_CtrlPnt4D Nextbpts(p-1,S.P.rows) ; // leftmost control points of the next Bezier segment
  Vector_FLOAT alphas(p-1) ; // knot instertion alphas.
  
  // Compute the Binomial coefficients
  Matrix_FLOAT Bin(ph+1,ph2+1) ;
  binomialCoef(Bin) ;
  
  // Compute Bezier degree elevation coefficients
  float inv,mpi ;
  bezalfs(0,0) = bezalfs(ph,p) = 1.0 ;
  for(i=1;i<=ph2;i++){
    inv= 1.0/Bin(ph,i) ;
    mpi = minimum(p,i) ;
    for(j=maximum(0,i-t); j<=mpi; j++){
      bezalfs(i,j) = inv*Bin(p,j)*Bin(t,i-j) ;
    }
  }
  
  for(i=ph2+1;i<ph ; i++){
    mpi = minimum(p,i) ;
    for(j=maximum(0,i-t); j<=mpi ; j++)
      bezalfs(i,j) = bezalfs(ph-i,p-j) ;
  }
  
  resize(S.P.rows,S.P.cols+S.P.cols*t,S.degreeU,ph) ; // Allocate more control points than necessary
  
  int rowJ ;
  int mh = ph ;
  int kind = ph+1 ;
  float va = S.V[0] ;
  float vb = 0.0 ;
  int r=-1 ; 
  int oldr ;
  int a = p ;
  int b = p+1 ; 
  int cind = 1 ;
  int rbz,lbz = 1 ; 
  int mul,save,s;
  float alf ;
  int first, last, kj ;
  float den,bet,gam,numer ;
  
  for(i=0;i<S.P.rows;i++)
    P(i,0) = S.P(i,0) ;
  for(i=0; i <= ph ; i++){
    V[i] = va ;
  }
  
  // Initialize the first Bezier segment
  
  for(i=0;i<=p ;i++) 
    for(j=0;j<S.P.rows;j++)
      bpts(i,j) = S.P(j,i);
  
  while(b<m){ // Big loop thru knot vector
    i=b ;
    while(b<m && S.V[b] >= S.V[b+1]) // for some odd reasons... == doesn't work
      b++ ;
    mul = b-i+1 ; 
    mh += mul+t ;
    vb = S.V[b] ;
    oldr = r ;
    r = p-mul ;
    if(oldr>0)
      lbz = (oldr+2)/2 ;
    else
      lbz = 1 ;
    if(r>0) 
      rbz = ph-(r+1)/2 ;
    else
      rbz = ph ;
    if(r>0){ // Insert knot to get Bezier segment
      numer = vb-va ;
      for(k=p;k>mul;k--){
	alphas[k-mul-1] = numer/(S.V[a+k]-va) ;
      }
      for(j=1;j<=r;j++){
	save = r-j ; s = mul+j ;
	for(k=p;k>=s;k--){
	  for(rowJ=0;rowJ<S.P.rows;rowJ++){
	    bpts(k,rowJ) = alphas[k-s]*bpts(k,rowJ)+(1.0-alphas[k-s])*bpts(k-1,rowJ) ;}
	}
	if(Nextbpts.rows>0)
	  for(rowJ=0;rowJ<S.P.rows;rowJ++){
	    Nextbpts(save,rowJ) = bpts(p,rowJ) ;}
      }
    }
    
    for(i=lbz;i<=ph;i++){ // Degree elevate Bezier,  only the points lbz,...,ph are used
      for(rowJ=0;rowJ<S.P.rows;rowJ++){
	ebpts(i,rowJ) = 0.0 ;}
      mpi = minimum(p,i) ;
      for(j=maximum(0,i-t); j<=mpi ; j++)
	for(rowJ=0;rowJ<S.P.rows;rowJ++){
	  ebpts(i,rowJ) += bezalfs(i,j)*bpts(j,rowJ) ;}
    }
    
    if(oldr>1){ // Must remove knot V=S.V[a] oldr times
      // if(oldr>2) // Alphas on the right do not change
      //	alfj = (va-nc.U[kind-1])/(vb-V[kind-1]) ;
      first = kind-2 ; last = kind ;
      den = vb-va ;
      bet = (vb-V[kind-1])/den ;
      for(int tr=1; tr<oldr; tr++){ // Knot removal loop
	i = first ; j = last ;
	kj = j-kind+1 ;
	while(j-i>tr){ // Loop and compute the new control points for one removal step
	  if(i<cind){
	    alf=(vb-V[i])/(va-V[i]) ;
	    for(rowJ=0;rowJ<S.P.rows;rowJ++){
	      P(rowJ,i) = alf*P(rowJ,i) + (1.0-alf)*P(rowJ,i-1) ;}
	  }
	  if( j>= lbz){
	    if(j-tr <= kind-ph+oldr){
	      gam = (vb-V[j-tr])/den ;
	      for(rowJ=0;rowJ<S.P.rows;rowJ++){
		ebpts(kj,rowJ) = gam*ebpts(kj,rowJ) + (1.0-gam)*ebpts(kj+1,rowJ) ;}
	    }
	    else{
	      for(rowJ=0;rowJ<S.P.rows;rowJ++){
		ebpts(kj,rowJ) = bet*ebpts(kj,rowJ)+(1.0-bet)*ebpts(kj+1,rowJ) ;}
	    }
	  }
	  ++i ; --j; --kj ;
	}
	--first ; ++last ;
      }
    }
  
    if(a!=p) // load the knot v=S.V[a]
      for(i=0;i<ph-oldr; i++){
	V[kind++] = va ; 
      }
    
    for(j=lbz; j<=rbz ; j++) { // load control points onto nc
      for(rowJ=0;rowJ<S.P.rows;rowJ++)
	P(rowJ,cind) = ebpts(j,rowJ) ; 
      ++cind ;
    }
    
    if(b<m){ // Set up for next pass thru loop
      for(rowJ=0;rowJ<S.P.rows;rowJ++){
	for(j=0;j<r;j++)
	  bpts(j,rowJ) = Nextbpts(j,rowJ) ;
	for(j=r;j<=p;j++)
	  bpts(j,rowJ) = S.P(rowJ,b-p+j) ;
      }
      a=b ; 
      b++ ;
      va = vb ;
    }
    else{
      for(i=0;i<=ph;i++)
	V[kind+i] = vb ;
    }
  }
  // Resize to the proper number of control points  
  resizeKeep(S.P.rows,mh-ph,S.degreeU,ph) ; 
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: findMultU --- finds the multiplicity of a knot o
  Description: Finds the multiplicity of a knot in the U knot
        Input: r --> the knot to observe
       Output: the multiplicity of the knot
 Restrictions: $r$ must be a valid U knot index
   Created by:  Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::findMultU(int r) const {
  int s=1 ;
  for(int i=r;i>degU+1;i--)
    if(U[i]<=U[i-1])
      s++ ;
    else
      return s ;
  return s ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: findMultV --- finds the multiplicity of a knot 
  Description: Finds the multiplicity of a knot in the V knot
        Input: r --> the knot to observe
       Output: the multiplicity of the V knot
 Restrictions: $r$ must be a valid knot index
   Created by:  Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::findMultV(int r) const {
  int s=1 ;
  for(int i=r;i>degV+1;i--)
    if(V[i]<=V[i-1])
      s++ ;
    else
      return s ;
  return s ;
}




/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: findSpan --- finds the span in the U and V direction
  Description: Finds the span in the U and V direction. The spanU is the index
               $k$ for which the parameter $u$ is valid in the $[u_k,u_{k+1}]$ 
	       range. The spanV is the index $k$ for which the parameter $v$ is 
	       valid in the $[v_k,v_{k+1}]$ range.
        Input:     u --> find the U span for this parametric value 
                   v --> find the V span for this parametric value
	       spanU <-- the U span
	       spanV <-- the V span
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::findSpan(float u, float v, int& spanU, int& spanV) const{
  spanU = findSpanU(u) ;
  spanV = findSpanV(v) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: findSpanU --- finds the span in the U direction
  Description: Finds the span in the U direction. The span is the index
               $k$ for which the parameter $u$ is valid in the $[u_k,u_{k+1}]$ 
	       range.
        Input: u --> find the span for this parametric value
       Output: the span for $u$.
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::findSpanU(float u) const{
  if(u>=U[P.rows]) 
    return P.rows-1 ;
  if(u<=U[0])
    return degU ;

  int low = degU ;
  int high = P.rows+1 ; 
  int mid = (low+high)/2 ;

  while(u<U[mid] || u>= U[mid+1]){
    if(u<U[mid])
      high = mid ;
    else
      low = mid ;
    mid = (low+high)/2 ;
  }
  return mid ;  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: findSpanV --- finds the span in the V direction
  Description: Finds the span in the V direction. The span is the index
               $k$ for which the parameter $v$ is valid in the $[v_k,v_{k+1}]$ 
	       range. 
        Input: v --> find the span for this parametric value    
       Output: the span for $v$
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::findSpanV(float v) const{
  if(v>=V[P.cols]) 
    return P.cols-1 ;
  if(v<=V[0])
    return degV ;

  int low = degV ;
  int high = P.cols+1 ; 
  int mid = (low+high)/2 ;

  while(v<V[mid] || v>= V[mid+1]){
    if(v<V[mid])
      high = mid ;
    else
      low = mid ;
    mid = (low+high)/2 ;
  }
  return mid ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: basisFuns --- find the non-zero basis functions in the U and V direction
  Description: Find the basis functions in the U and V direction.
        Input:   u --> the u parametric value
                 v --> the v parametric value 
             spanU --> the span of u
             spanV --> the span of v
                Nu <-- the vector containing the U non-zero basis functions
                Nv <-- the vector containing the V non-zero basis functions
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::basisFuns(float u, float v, int spanU, int spanV, Vector_FLOAT& Nu, Vector_FLOAT &Nv) const{
  basisFunsU(u,spanU,Nu) ;
  basisFunsV(v,spanV,Nv) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: basisFunsU --- the non-zero basis function in the U direction
  Description: Finds the non-zero basis function in the U direction.
        Input:     u --> the u parametric value
                span --> the span of u
                   N <-- the vector containing the basis functions
       Output:
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::basisFunsU(float u, int span, Vector_FLOAT& N) const {
  Vector_FLOAT left(degU+1), right(degU+1) ;
  float temp,saved ;
   

  N.resize(degU+1) ;

  N[0] = 1.0 ;
  for(int j=1; j<= degU ; j++){
    left[j] = u-U[span+1-j] ;
    right[j] = U[span+j]-u ;
    saved = 0.0 ;
    for(int r=0 ; r<j; r++){
      temp = N[r]/(right[r+1]+left[j-r]) ;
      N[r] = saved+right[r+1]*temp ;
      saved = left[j-r]*temp ;
    }
    N[j] = saved ;
  }  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: basisFunsV --- the non-zero basis function in the V direction
  Description: Fins the non-zero basis function in the V direction
        Input:    v --> the v parametric value
               span --> the span of v 
                  N <-- the vector containing the basis functions
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::basisFunsV(float v, int span, Vector_FLOAT& N) const {
  Vector_FLOAT left(degV+1), right(degV+1) ;
  float temp,saved ;
   

  N.resize(degV+1) ;

  N[0] = 1.0 ;
  for(int j=1; j<= degV ; j++){
    left[j] = v-V[span+1-j] ;
    right[j] = V[span+j]-v ;
    saved = 0.0 ;
    for(int r=0 ; r<j; r++){
      temp = N[r]/(right[r+1]+left[j-r]) ;
      N[r] = saved+right[r+1]*temp ;
      saved = left[j-r]*temp ;
    }
    N[j] = saved ;
  }  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: deriveAt --- computes the point and the derivatives of 
                            degree $d$ and below at $u,v$
  Description:   Computes the matrix of derivatives at $u,v$. 
                 The value of skl(k,l) represents the 
		 derivative of the surface $S(u,v)$ with respect to 
		 $u$ $k$ times and to $v$ $l$ times.
        Input:   u --> the u parametric value
                 v --> the v parametric value
		 d --> the derivative is computed up to and including to 
		       this value
	       skl <-- the matrix containing the derivatives
       Output: 
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::deriveAt(float u, float v, int d, Matrix_CtrlPnt4D &skl) const {
  int k,l,du,dv;
  skl.resize(d+1,d+1) ;

  du = minimum(d,degU) ;
  for(k=degU+1;k<=d;++k)
    for(l=0;l<=d-k;++l)
      skl(k,l) = 0.0 ;
  dv=minimum(d,degV) ;
  for(l=degV+1;l<=d;++l)
    for(k=0;k<=d-l;++k)
      skl(k,l) = 0.0 ;
  int uspan = findSpanU(u) ;
  int vspan = findSpanV(v) ;
  Matrix_FLOAT Nu,Nv ;
  nurbsDersBasisFuns(du,u,uspan,degU,U,Nu) ;
  nurbsDersBasisFuns(dv,v,vspan,degV,V,Nv) ;

  Vector_CtrlPnt4D temp(degV+1) ;
  int dd,r,s ;
  for(k=0;k<=du;++k){
    for(s=0;s<=degV;++s){
      temp[s] = 0.0 ;
      for(r=0;r<=degU;++r)
	temp[s] += Nu(k,r)*P(uspan-degU+r,vspan-degV+s) ;
    }
    dd = minimum(d-k,dv) ;
    for(l=0;l<=dd;++l){
      skl(k,l) = 0.0 ;
      for(s=0;s<=degV;++s)
	skl(k,l) += Nv(l,s)*temp[s] ;
    }
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: deriveAt --- computes the point and the derivatives of degree 
                            $d$ and below at $u,v$
  Description: Computes the matrix of derivatives at $u,v$. 
               The value of skl(k,l) represents the 
	       derivative of the surface $S(u,v)$ with respect to 
	       $u$ $k$ times and to $v$ $l$ times.
        Input:   u --> the u parametric value
                 v --> the v parametric value
		 d --> the derivative is computed up to and including to 
		       this value
	       skl <-- the matrix containing the derivatives
       Output: 
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::deriveAt(float u, float v, int d, Matrix_Point3D &skl) const {
  int k,l ;
  Matrix_CtrlPnt4D ders ;
  Point3D pv,pv2 ;
  
  skl.resize(d+1,d+1) ;

  deriveAt(u,v,d,ders) ;
  
  Matrix_FLOAT Bin(d+1,d+1) ;
  binomialCoef(Bin) ;
  int i,j ; 

  for(k=0;k<=d;++k){
    for(l=0;l<=d-k;++l){
      pv.x = ders(k,l).x() ;
      pv.y = ders(k,l).y() ;
      pv.z = ders(k,l).z() ;
      for(j=1;j<=l;j++)
	pv -= Bin(l,j)*ders(0,j).w()*skl(k,l-j) ;
      for(i=1;i<=k;i++){
	pv -= Bin(k,i)*ders(i,0).w()*skl(k-i,l) ;
	pv2 = 0.0 ;
	for(j=1;j<=l;j++)
	  pv2 += Bin(l,j)*ders(i,j).w()*skl(k-i,l-j) ;
	pv -= Bin(k,i)*pv2 ;
      }
      skl(k,l) = pv/ders(0,0).w() ;
    }
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: normal --- computes the normal of the surface at $u,v$
  Description: Computes the normal of the surface at $u,v$
        Input:   u --> the u parametric value
                 v --> the v parametric value
       Output: the normal at $(u,v)$.
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
Point3D PlNurbsSurface::normal(float u, float v) const {
  Matrix_Point3D ders ;

  deriveAt(u,v,1,ders) ;

  return crossProduct(ders(1,0),ders(0,1)) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: point --- returns the point on the surface at u,v
  Description: Returns the point on the surface at $u,v$
        Input:  u --> the u parametric value
                v --> the v parametric value
       Output: The 4D point at $u,v$
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
CtrlPnt4D PlNurbsSurface::operator()(float u, float v) const{
  int uspan = findSpanU(u) ;
  int vspan = findSpanV(v) ;
  Vector_FLOAT Nu,Nv ;
  Vector_CtrlPnt4D temp(degV+1)  ;
  CtrlPnt4D sp ;

  basisFuns(u,v,uspan,vspan,Nu,Nv) ;

  int l;
  for(l=0;l<=degV;l++){
    temp[l] =0.0 ;
    for(int k=0;k<=degU;k++)
      temp[l] += Nu[k]*P(uspan-degU+k,vspan-degV+l) ;
  }
  sp = CtrlPnt4D(0,0,0,0) ;
  for(l=0;l<=degV;l++){
    sp += Nv[l]*temp[l] ;
  }
  return sp ;
}

inline
int max3(int a,int b, int c){
  int m = a ;
  if(m <b)
    m = b ;
  if(m <c)
    m = c ;
  return m ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
      Routine: gordonSurface --- interpolation of a surface from 2 sets of orthogonal curves
  Description: Interpolation of a surface from 2 sets of orthogonal curves. 
               See A10.3 at page 494 on the NURBS book for more details about
	       the implementation.
        Input:  lU <-> an array of curves in the U direction
                lV <-> an array of curves in the V direction
     intersections --> a matrix of 3D points specifying where the the curves 
                       intersect
                gS <-- the Gordon surface
       Output:
 Restrictions: 
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void gordonSurface(PlNurbsCurveArray& lU, PlNurbsCurveArray& lV, const Matrix_Point3D& intersections, PlNurbsSurface& gS){
  PlNurbsSurface sU,sV,sI ;
  sU.skinU(lU,3) ;
  sV.skinV(lV,3) ;
  sI.globalInterp(intersections,3,3) ;

  int du = max3(sU.degreeU,sV.degreeU,sI.degreeU) ;
  int dv = max3(sU.degreeV,sV.degreeV,sI.degreeV) ;
  

  PlNurbsSurface SU,SV,SI ;
  degreeElevate(sU,du-sU.degreeU,dv-sU.degreeV,SU) ;
  degreeElevate(sV,du-sV.degreeU,dv-sV.degreeV,SV) ;
  degreeElevate(sI,du-sI.degreeU,dv-sI.degreeV,SI) ;


  Vector_FLOAT U,V ;
  U = knotUnion(SU.KnotU,SV.KnotU) ;
  U = knotUnion(U,SI.KnotU) ;
  V = knotUnion(SU.KnotV,SV.KnotV) ;
  V = knotUnion(V,SI.KnotV) ;

  SU.mergeKnots(U,V) ;
  SV.mergeKnots(U,V) ;
  SI.mergeKnots(U,V) ;
  
  gS = SU ;

  for(int i=0;i<gS.P.rows;i++)
    for(int j=0;j<gS.P.cols;j++){
      gS.P(i,j) += SV.P(i,j) ;
      gS.P(i,j) -= SI.P(i,j) ;
    }
}

inline void insert(float u, Vector_FLOAT& v){
  int i ;
  if(u<v[0] || u>v[v.n-1])
    return ;
  v.resize(v.n+1) ;
  i = v.n-1 ;
  while(v[i-1]>u){
    v[i] = v[i-1] ;
    --i ;
  }
  v[i] = u ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: sweep --- generates a surface by sweeping a curve along a 
                         trajectory
  Description: Sweeping consists of creating a surface by moving a curve
               profile through a trajectory. The method uses here consists of 
	       using skinning of $K$ instances of the curve $C(u)$ along 
	       $T(u)$. The $K$ value should be viewed as the minimum number of 
	       sections required.

	       The profile curve $C(u)$ should lie on the $xz$-plane.
	       It follows the trajectory curve $T(u)$ along its $y$-axis.

	       The scaling function is used to modify the shape of the curve
	       profile while it's being swept. It can scale in any of the
	       4 dimensions to obtain the desired effects on the profile.

	       See A10.1 on page 476 of the NURBS book for more details 
	       about the implementation.

	       You might have to play with the useAy and invAz variables
	       to obtain a satisfactory result for the sweep operation.
	       This is either because there is an error in the code or 
	       because it is the way it is supposed to work.

        Input:   T --> the trajectory of the curve 
                 C --> the curve profile to sweep
		Sv --> a scaling function
		 K --> the minimum number of insertion
	     useAy --> a 0 indicates that rotation angle around the $y$-axis 
	               is not computed, otherwise it is computed. Results
		       are usually better with a value of 0.
	     invAz --> a 1 indicates that the computed rotation around the
	               $z$-axis should be inversed, a 0 indicates it
		       stays as it is.
       Output: the NURBS surface representing the sweep of $C$ along $T$
 Restrictions: This will not yield a correct value for a closed trajectory 
               curve.
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::sweep(const PlNurbsCurve& T, const PlNurbsCurve& C, const PlNurbsCurve& Sv, int K, int useAy, int invAz) {
  int i,j,k,m,q,ktv,nsect ;
  q = T.degree ;
  ktv = T.Knot.n ;
  nsect = K  ;
  
  V.resize(T.Knot.n) ;
  V = T.Knot ;

  if(ktv <= nsect+q){
    m = nsect+q-ktv+1 ;
    // insert m knots into T(v)
    // locations are not critical so inserting in the middle of the 
    // biggest span is used.
    for(i=0;i<m;++i){
      float md,mt ;
      float mu = -1;
      md = 0 ;
      for(j=1;j<V.n;++j){
	mt = V[j]-V[j-1] ;
	if(mt>md){
	  md = mt ;
	  mu = (V[j]+V[j-1])/2.0 ;
	}
      }
      insert(mu,V) ;
    }
  }
  else{
    if(ktv>nsect){ // must increase the number of instances of C(u)
      nsect = ktv-q-1 ; 
    }
  }

  Vector_FLOAT v; 

  // Compute the parameters by averaging the knots
  v.resize(nsect) ;
  v[0] = T.Knot[0] ;
  v[nsect-1] = T.Knot[T.Knot.n-1] ;
  for(k=1;k<nsect-1;++k){ 
    v[k] = 0.0 ;
    for(i=k+1;i<k+q+1;++i) 
      v[k] += V[i] ;
    v[k] /= q ;
  }

  resize(C.CtrlP.n,nsect,C.degree,T.degree) ;

  U = C.Knot ;

  // setup Bv ;
  Vector_Point3D B ;
  PlNurbsCurve Bv ;
  Vector_Point3D Td ;

  B.resize(v.n) ;
  T.deriveAt(v[0],1,Td) ;

  B[0] = Td[1] ;
  if(Td[1].y ==0)
    B[0] = crossProduct(Point3D(0,1,0),B[0]) ;
  else
    B[0] = crossProduct(Point3D(1,0,0),B[0]) ;
  B[0] /= abs(B[0]) ;


  for(i=1;i<v.n;++i){
    T.deriveAt(v[i],1,Td);
    Point3D Ti(Td[1]) ;
    Ti = Ti/abs(Ti) ;
    Point3D bi ;
    bi = B[i-1]-(B[i-1]*Ti)*Ti ;
    B[i] = bi/abs(bi) ;
  }

  Bv.globalInterp(B,v,minimum(3,B.n-1)) ;

  Vector_CtrlPnt4D Q(C.CtrlP.n) ; 
  for(k=0;k<nsect;++k){
    // scale the control points by Sv(v[k])
    for(i=0;i<Q.n;++i){
      CtrlPnt4D sk(Sv(v[k])) ;
      Q[i].x() = sk.x()*C.CtrlP[i].x() ;
      Q[i].y() = sk.y()*C.CtrlP[i].y() ;
      Q[i].z() = sk.z()*C.CtrlP[i].z() ;
      Q[i].w() = sk.w()*C.CtrlP[i].w() ;
    }
    // compute o(v[k])
    Point3D o = T.point3D(v[k]) ;
    //float w = T(v[k]).w() ;

    // compute x(v[k])
    T.deriveAt(v[k],1,Td) ;

    Point3D x = Td[1]/abs(Td[1]) ;

    // compute z(v[k])
    Point3D z = Bv.point3D(v[k]) ; 
    z /= abs(z) ;

    // compute y(v[k]) 
    Point3D y = crossProduct(z,x) ;

    // compute the transform matrix
    double az = M_PI+atan2(y.y,y.x) ;
    double ax = -atan2(z.y,z.z) ;
    double ay = 0 ;
    if(useAy){
      ay = atan2(x.z,x.x) ;
    }
    if(invAz){
      az = -1.0*az ;
    }
    MatrixRT_DOUBLE A(ax,ay,az,o.x,o.y,o.z) ;

    for(i=0;i<Q.n;++i){
      P(i,k) = A*Q[i] ;
    }
  }

  // interpolate along V
  Q.resize(P.cols) ;
  PlNurbsCurve R;
  for(i=0;i<P.rows;++i){
    for(k=0;k<P.cols;++k)
      Q[k] = P(i,k) ;
    R.globalInterp(Q,v,V,degV) ;
    for(k=0;k<P.cols;++k)
      P(i,k) = R.CtrlP[k] ;
  }
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: sweep --- generates a surface by sweeping a curve along a 
                         trajectory
  Description: Sweeping consists of creating a surface by moving a curve
  	       through a trajectory. The method uses here consists of using
	       skinning of $K$ instances of the curve $C(u)$ along $T(u)$.
	       The $K$ value should be viewed as a minimum.

	       The cross-sectional curve $C(u)$ should lie on the $xz$-plane.
	       It follows the trajectory curve $T(u)$ along its $y$-axis.

	       You might have to play with the useAy and invAz variables
	       to obtain a satisfactory result for the sweep operation.
	       This is either because there is an error in the code or 
	       because it is the way it is supposed to work.

        Input:   T --> the trajectory of the curve 
                 C --> the curve to sweep
		 K --> the minimum number of insertion
	     useAy --> a 0 indicates that rotation angle around the $y$-axis 
	               is not computed, otherwise it is computed. 
	     invAz --> a 1 indicates that the computed rotation around the
	               $z$-axis should be inversed, a 0 indicates it
		       stays as it is.
       Output: the NURBS surface representing the sweep of $C$ along $T$
 Restrictions: This will not yield a correct value for a closed trajectory 
               curve.
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::sweep(const PlNurbsCurve& T, const PlNurbsCurve& C, int K, int useAy,int invAz) {
  // setup Sv 
  Vector_CtrlPnt4D p(2) ;
  p[0] = CtrlPnt4D(1,1,1,1) ;
  p[1] = CtrlPnt4D(1,1,1,1) ;
  Vector_FLOAT u(4) ;
  u[0] = u[1] = 0.0 ; u[2] = u[3] = 1.0 ;

  PlNurbsCurve Sv(p,u,1) ;
  
  sweep(T,C,Sv,K,useAy,invAz) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: transform --- performs geometrical modifications
  Description: Each control points will be modified by a rotation-translation
               matrix.
	      
        Input: A --> the rotation-translation matrix
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (22 August 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::transform(const MatrixRT<float>& A){
  for(int i=0;i<P.rows;++i)
    for(int j=0;j<P.cols;++j)
      P(i,j) = A*P(i,j) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: refineKnots --- refine both knot vectors
  Description: Refine both knot vectors.
        Input: nU --> the U knot vector to refine from 
               nV --> the V knot vector to refine from 
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::refineKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV){
  refineKnotU(nU) ;
  refineKnotV(nV) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: refineKnotU ---  refines the U knot vector
  Description: Refines the U knot vector.
        Input: X --> the knot vector to refine from
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::refineKnotU(const Vector_FLOAT& X) {
  if(X.n<=0)
    return ;
  int n = P.rows-1 ;
  int p = degU;
  int m = n+p+1 ;
  int a,b ;
  int r = X.n-1 ;
  PlNurbsSurface nS ;
  
  nS = *this ;
  nS.resize(r+1+n+1,P.cols,degU,degV) ;

  a = findSpanU(X[0]) ;
  b = findSpanU(X[r]) ;
  ++b ;

  int j,col ;
  for(col=0;col<P.cols;col++){
    for(j=0; j<=a-p ; j++)
      nS.P(j,col) = P(j,col);
    for(j = b-1 ; j<=n ; j++)
      nS.P(j+r+1,col) = P(j,col) ;
  }
  for(j=0; j<=a ; j++)
    nS.U[j] = U[j] ;
  for(j=b+p ; j<=m ; j++)
    nS.U[j+r+1] = U[j] ;
  int i = b+p-1 ; 
  int k = b+p+r ;
  for(j=r; j>=0 ; j--){
    while(X[j] <= U[i] && i>a){
      for(col=0;col<P.cols;col++)
	nS.P(k-p-1,col) = P(i-p-1,col) ;
      nS.U[k] = U[i] ;
      --k ;
      --i ;
    }
    for(col=0;col<P.cols;col++)
	nS.P(k-p-1,col) = nS.P(k-p,col) ;
    for(int l=1; l<=p ; l++){
      int ind = k-p+l ;
      float alpha = nS.U[k+l] - X[j] ;
      if(alpha==0.0)
	for(col=0;col<P.cols;col++)
	  nS.P(ind-1,col) = nS.P(ind,col) ;
      else
	alpha /= nS.U[k+l]-U[i-p+l] ;
      for(col=0;col<P.cols;col++)
	nS.P(ind-1,col) = alpha*nS.P(ind-1,col) + (1.0-alpha)*nS.P(ind,col) ;
    }
    nS.U[k] = X[j] ;
    --k ;
  }
  *this = nS ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: refineKnotV --- refines the V knot vector
  Description: Refines the V knot vector 
        Input: X --> the knot vector to refine from 
       Output:
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::refineKnotV(const Vector_FLOAT& X) {
  if(X.n<=0)
    return ;
  int n = P.cols-1 ;
  int p = degV;
  int m = n+p+1 ;
  int a,b ;
  int r = X.n-1 ;
  PlNurbsSurface nS ;
  
  nS = *this ;
  nS.resize(P.rows,r+1+n+1,degU,degV) ;

  a = findSpanV(X[0]) ;
  b = findSpanV(X[r]) ;
  ++b ;

  int j,row ;
  for(row=0;row<P.rows;row++){
    for(j=0; j<=a-p ; j++)
      nS.P(row,j) = P(row,j);
    for(j = b-1 ; j<=n ; j++)
      nS.P(row,j+r+1) = P(row,j) ;
  }
  for(j=0; j<=a ; j++)
    nS.V[j] = V[j] ;
  for(j=b+p ; j<=m ; j++)
    nS.V[j+r+1] = V[j] ;
  int i = b+p-1 ; 
  int k = b+p+r ;
  for(j=r; j>=0 ; j--){
    while(X[j] <= V[i] && i>a){
      for(row=0;row<P.rows;row++)
	nS.P(row,k-p-1) = P(row,i-p-1) ;
      nS.V[k] = V[i] ;
      --k ;
      --i ;
    }
    for(row=0;row<P.rows;row++)
	nS.P(row,k-p-1) = nS.P(row,k-p) ;
    for(int l=1; l<=p ; l++){
      int ind = k-p+l ;
      float alpha = nS.V[k+l] - X[j] ;
      if(alpha==0.0)
	for(row=0;row<P.rows;row++)
	  nS.P(row,ind-1) = nS.P(row,ind) ;
      else
	alpha /= nS.V[k+l]-V[i-p+l] ;
      for(row=0;row<P.rows;row++)
	nS.P(row,ind-1) = alpha*nS.P(row,ind-1) + (1.0-alpha)*nS.P(row,ind) ;
    }
    nS.V[k] = X[j] ;
    --k ;
  }
  *this = nS ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: mergeKnots --- merges a U and V knot vector with the surface knot vectors
  Description: Merges a U and V knot vector with the surface knot vectors.
        Input:  nU --> the U knot vector to merge with
                nV --> the V knot vector to merge with
       Output:
 Restrictions:  The nU and nV knot vectors must be compatible with the 
                current vectors
   Created by:    Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::mergeKnots(const Vector_FLOAT& nU, const Vector_FLOAT& nV) {
  mergeKnotU(nU) ;
  mergeKnotV(nV) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: mergeKnotU --- merges the U knot vector with another one
  Description: Merges the U knot vector with another one 
        Input: X --> a knot vector
       Output:
 Restrictions: The knot vector must be compatible with the U knot vector
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::mergeKnotU(const Vector_FLOAT& X){
  int i,ia,ib ;
  // Find the knots to insert
  Vector_FLOAT I(U.n) ;
  int done = 0 ;
  i = ia = ib = 0 ;
  while(!done) {
    if(X[ib] == U[ia]){
      ++ib ; ++ia ;
    }
    else{
      I[i++] = X[ib] ;
      ib++ ;
    }
    done = (ia>=U.n || ib >= X.n) ;
  }
  I.resize(i) ;

  if(I.n>0)
    refineKnotU(I) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: mergeKnotV --- merges the V knot vector with another one
  Description: Merges the V knot vector with another one.
        Input: X --> a knot vector
       Output:
 Restrictions: The knot vector must be compatible with the V knot vector
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::mergeKnotV(const Vector_FLOAT& X){
  int i,ia,ib ;
  // Find the knots to insert
  Vector_FLOAT I(V.n) ;
  int done = 0 ;
  i = ia = ib = 0 ;
  while(!done) {
    if(X[ib] == V[ia]){
      ++ib ; ++ia ;
    }
    else{
      I[i++] = X[ib] ;
      ib++ ;
    }
    done = (ia>=V.n || ib >= X.n) ;
  }
  I.resize(i) ;

  if(I.n>0)
    refineKnotV(I) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: read --- read a surface from a file
  Description: Read a surface from an input file stream.
        Input: fin --> the input file stream
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::read(ifstream &fin){
  if(!fin) {
    return 0 ;
  }
  int nu,nv,du,dv;
  char *type ;
  type = new char[3] ;
  if(!fin.read(type,sizeof(char)*3)) { delete []type ; return 0 ;}
  int r1 = strncmp(type,"ns3",3) ;
  int r2 = strncmp(type,"ns4",3) ;
  if(!(r1 || r2)) 
    return 0 ;
  if(!fin.read((char*)&nu,sizeof(int))) { delete []type ; return 0 ;}
  if(!fin.read((char*)&nv,sizeof(int))) { delete []type ; return 0 ;}
  if(!fin.read((char*)&du,sizeof(int))) { delete []type ; return 0 ;}
  if(!fin.read((char*)&dv,sizeof(int))) { delete []type ; return 0 ;}

  resize(nu,nv,du,dv) ;
  
  if(!fin.read((char*)U.memory(),sizeof(float)*U.n)) { delete []type ; return 0 ;}
  if(!fin.read((char*)V.memory(),sizeof(float)*V.n)) { delete []type ; return 0 ;}
     

  float *p,*p2 ;
  if(!r1){
    p = new float[3*nu*nv] ;
    if(!fin.read((char*)p,sizeof(float)*3*nu*nv)) { delete []type ; return 0 ;}
    p2 = p ;
    for(int i=0;i<nu;i++)
      for(int j=0;j<nv;j++){
	P(i,j).x() = *(p++) ;
	P(i,j).y() = *(p++) ;
	P(i,j).z() = *(p++) ;
	P(i,j).w() = 1.0 ;
      }
    delete []p2 ;
  }
  else{
    p = new float[4*nu*nv] ;
    if(!fin.read((char*)p,sizeof(float)*4*nu*nv)) { delete []type ; return 0 ;}
    p2 = p ;
    for(int i=0;i<nu;i++)
      for(int j=0;j<nv;j++){
	P(i,j).x() = *(p++) ;
	P(i,j).y() = *(p++) ;
	P(i,j).z() = *(p++) ;
	P(i,j).w() = *(p++) ;
      }
    delete []p2 ;
  }

  delete []type ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: read --- read a surface from a file
  Description: Read a surface from a file.
        Input: filename --> the filename to read from 
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::read(const char* filename){
  ifstream fin(filename) ;
  if(!fin) {
    return 0 ;
  }

  return read(fin) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: write --- write a surface to a file stream
  Description: Write a surface to an output file stream.
        Input: fout <-- the output filestream to write to.
       Output: 1 on success, 0 on failure
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::write(ofstream &fout) const {
  if(!fout)
    return 0 ;
  if(!fout.write((char*)&"ns4",sizeof(char)*3)) return 0 ;
  if(!fout.write((char*)&P.rows,sizeof(int))) return 0 ;
  if(!fout.write((char*)&P.cols,sizeof(int))) return 0 ;
  if(!fout.write((char*)&degU,sizeof(int))) return 0 ;
  if(!fout.write((char*)&degV,sizeof(int))) return 0 ;
  if(!fout.write((char*)U.memory(),sizeof(float)*U.n)) return 0 ;
  if(!fout.write((char*)V.memory(),sizeof(float)*V.n)) return 0 ;
  
  float *p,*p2 ;
  p = new float[P.rows*P.cols*4] ;
  p2 = p ;
  for(int i=0;i<P.rows;i++) 
    for(int j=0;j<P.cols;j++){
      *p = P(i,j).x() ; p++ ;
      *p = P(i,j).y() ; p++ ;
      *p = P(i,j).z() ; p++ ;
      *p = P(i,j).w() ; p++ ;
    }
  if(!fout.write((char*)p2,sizeof(float)*P.rows*P.cols*4)) return 0 ;
  delete []p2 ;
  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: write --- write a surface to a file
  Description: Write a surface to a file.
        Input: filename <-- the filename to write to
       Output: 1 on success, 0 on failure
 Restrictions: 
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::write(const char* filename) const {
  ofstream fout(filename) ;    
  if(!fout)
    return 0 ;
  return write(fout);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
       Member: transpose --- transpose the U and V coordinates of a surface
  Description: Transpose the U and V coordinates of a surface. After this
              operation the $(u,v)$ points correspond to $(v,u)$.
        Input:     
       Output: A reference 
 Restrictions: This is not completely tested
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurface& PlNurbsSurface::transpose(void){
  Vector_FLOAT t(U) ;
  int d ;
  U = V ;
  V = t ;
  d = degU ;
  degU = degV ;
  degV = d ;
  P = ::transpose(P) ;
  return *this ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point on the surface
  Description: This moves the point $s(u,v)$ by delta.	      
        Input: u --> the parameter in the u direction
	       v --> the parameter in the v direction
	   delta --> the displacement of the point at S(u,v)
       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::movePoint(float u, float v, const Point3D& delta){
  // setup B
  Matrix_DOUBLE B(1,(degU+1)*(degV+1)) ;
  int i,j,k ;
  
  int spanU,spanV ;

  Vector_FLOAT Ru,Rv ;

  B.reset(0.0) ;
  
  findSpan(u,v,spanU,spanV) ;
  nurbsBasisFuns(u,spanU,degU,U,Ru) ;
  nurbsBasisFuns(v,spanV,degV,V,Rv) ;
  for(j=0;j<=degU;++j){
    for(k=0;k<=degV;++k){
      B(0,j*(degV+1)+k) 
	= (double)Ru[j]*(double)Rv[k] ;
    }
  }
  
  Matrix_DOUBLE A  ;
  Matrix_DOUBLE Bt(::transpose(B)) ;
  Matrix_DOUBLE BBt ;

  BBt = inverse(B*Bt) ;
  A = Bt*BBt ;

  Matrix_DOUBLE dD(1,3) ;

  for(j=0;j<3;++j)
    dD(0,j) = (double)delta.data[j] ;

  Matrix_DOUBLE dP ;

  dP = A*dD ;

  i= 0 ;

  for(j=0;j<=degU;++j){
    for(k=0;k<=degV;++k){
      float w = P(spanU-degU+j,spanV-degV+k).w() ;
      P(spanU-degU+j,spanV-degV+k).x() += dP(i,0)*w ;
      P(spanU-degU+j,spanV-degV+k).y() += dP(i,1)*w ;
      P(spanU-degU+j,spanV-degV+k).z() += dP(i,2)*w ;
      ++i ;
    }
  }

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point with some constraint
  Description: This will modify the NURBS surface by respecting a certain 
               number of constraints. $u_r$ and $v_r$ specifies the parameters
	       on which the constraints should be applied. 
	       The constraint are defined by $D_i(u,v)$ which requires 3
	       vectors to fully qualify. $D$ specifies the value
	       of the constraint and $Du$ and $Dv$ are used to specify
	       on which parameter the constraint is applied.
	       
	       ur and vr should be in an increasing order.

        Input: ur --> the vector of parameters in the u direction
	       vr --> the vector of parameters in the v direction
	        D --> a vector of the value of $D_i^(k,l)(u,v)$
	       Du --> a vector specifying the index of the value of u for $D_i$
	       Dv --> a vector specifying the index of the value of v for $D_i$
       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: The vectors defining $D_i(u,v)$ should all be of the
               same size. 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::movePoint(const Vector_FLOAT& ur, const Vector_FLOAT& vr, const Vector_Point3D& D, const Vector_INT& Du, const Vector_INT& Dv) {
  Vector_INT Dk(Du.n),Dl(Dv.n) ;
  BasicArray<Coordinate> fixCP(0) ;
  Dk.reset(0) ;
  Dl.reset(0) ;
  return movePoint(ur,vr,D,Du,Dv,Dk,Dl,fixCP) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point with some constraint
  Description: This will modify the NURBS surface by respecting a certain 
               number of constraints. $u_r$ and $v_r$ specifies the parameters
	       on which the constraints should be applied. 
	       The constraint are defined by $D_i^{(k,l)}(u,v)$ which requires
	       5 vectors to fully qualify. $D$ specifies the value
	       of the constraint and $Du$ and $Dv$ are used to specify
	       on which parameter the constraint is applied and 
	       $Dk$ and $Dl$ specify the partial degree of the constraint.
	       
	       The values in D should be ordered in respect with i,k and l.
	       ur and vr should be in an increasing order.

        Input: ur --> the vector of parameters in the u direction
	       vr --> the vector of parameters in the v direction
	        D --> a vector of the value of $D_i^(k,l)(u,v)$
	       Du --> a vector specifying the index of the value of u for $D_i$
	       Dv --> a vector specifying the index of the value of v for $D_i$
	       Dk --> a vector specifying the value of k for $D_i$
	       Dl --> a vector specifying the value of l for $D_i$
       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: The vectors defining $D_i^{(k,l)}(u,v)$ should all be of the
               same size. 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::movePoint(const Vector_FLOAT& ur, const Vector_FLOAT& vr, const Vector_Point3D& D, const Vector_INT& Du, const Vector_INT& Dv, const Vector_INT& Dk, const Vector_INT& Dl) {
  BasicArray<Coordinate> fixCP(0) ;
  return movePoint(ur,vr,D,Du,Dv,Dk,Dl,fixCP) ;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: movePoint --- moves a point with some constraint
  Description: This will modify the NURBS surface by respecting a certain 
               number of constraints. $u_r$ and $v_r$ specifies the parameters
	       on which the constraints should be applied. 
	       The constraint are defined by $D_i^{(k,l)}(u,v)$ which requires
	       5 vectors to fully qualify. $D$ specifies the value
	       of the constraint and $Du$ and $Dv$ are used to specify
	       on which parameter the constraint is applied and 
	       $Dk$ and $Dl$ specify the partial degree of the constraint.
	       
	       A second constraint $fixCP$ consists of specifying which 
	       control points can not be moved by the routine.

	       The values in D should be ordered in respect with i,k and l.
	       ur and vr should be in an increasing order.

        Input: ur --> the vector of parameters in the u direction
	       vr --> the vector of parameters in the v direction
	        D --> a vector of the value of $D_i^(k,l)(u,v)$
	       Du --> a vector specifying the index of the value of u for $D_i$
	       Dv --> a vector specifying the index of the value of v for $D_i$
	       Dk --> a vector specifying the value of k for $D_i$
	       Dl --> a vector specifying the value of l for $D_i$
	    fixCP --> a vector specifying which control points {\em can not} be
	              modified.
       Output: 1 if the operation is possible, 0 if the problem is ill defined
               {\em i.e.} there isn't enough information to find a unique 
	       solution (the system is overdetermined) or that the system
	       has non-independant components.

 Restrictions: The vectors defining $D_i^{(k,l)}(u,v)$ should all be of the
               same size. 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::movePoint(const Vector_FLOAT& ur, const Vector_FLOAT& vr, const Vector_Point3D& D, const Vector_INT& Du, const Vector_INT& Dv, const Vector_INT& Dk, const Vector_INT& Dl, const BasicArray<Coordinate>& fixCP) {
  int i,j,k,n ;

  if(D.n != Du.n || D.n !=Du.n || D.n != Dv.n || D.n != Dv.n){
    PlError err("movePoint(ur,D,Dr,Dk,fixCP)");
    err << "The D,Dr,Dk vectors are not of the same size\n" ;
    err << "D.n= " << D.n << ", Du.n = " << Du.n 
	<< ", Dk.n = " << Dk.n << ", Dv.n = " << Dv.n 
	<< ", Dl.n = " << Dl.n << endl ;
    err.fatal() ;
  }  
  
  // setup B
  Matrix_DOUBLE B ;
  
  B.resize(D.n,P.rows*P.cols) ;
  
  int spanU,spanV ;

  Matrix_FLOAT Ru,Rv ;

  B.reset(0.0) ;
  
  for(i=0;i<D.rows;++i){
    findSpan(ur[Du[i]],vr[Dv[i]],spanU,spanV) ;
    nurbsDersBasisFuns(Dk[i],ur[Du[i]],spanU,degU,U,Ru) ;
    nurbsDersBasisFuns(Dl[i],vr[Dv[i]],spanV,degV,V,Rv) ;
    for(j=0;j<=degU;++j){
      for(k=0;k<=degV;++k){
	B(i,(spanU-degU+j)*P.cols+spanV-degV+k) 
	  = (double)Ru(Dk[i],j)*(double)Rv(Dl[i],k) ;
      }
    }
  }
  
  // optimize B
  Vector_INT remove(B.cols) ;
  BasicArray<Coordinate> map(B.cols) ;
  remove.reset((int)1.0) ;

  for(j=0;j<B.cols;++j){
    for(i=0;i<B.rows;++i)
      if((B(i,j)*B(i,j))>1e-10){
	remove[j] = 0 ;
	break ;
      }
  }

  for(i=0;i<fixCP.n;++i){
    remove[fixCP[i].i*P.cols+fixCP[i].j] = 1 ;
  }
  
  n = 0 ;
  for(i=0;i<B.cols;++i){
    if(!remove[i]){
      map[n].i = i/P.cols ;
      map[n].j = i%P.cols ;
      ++n ;
    }
  }

  map.resize(n) ;

  Matrix_DOUBLE Bopt(B.rows,n) ;
  for(j=0;j<n;++j){
    for(i=0;i<B.rows;++i)
      Bopt(i,j) = B(i,map[j].i*P.cols+map[j].j) ;
  }

  Matrix_DOUBLE A  ;
  Matrix_DOUBLE Bt(::transpose(Bopt)) ;
  Matrix_DOUBLE BBt ;

  //BBt = LAPack.inverse(Bopt*Bt) ;
  BBt = inverse(Bopt*Bt) ;
  /*
  if(LAPack.last_error>0){
    return 0 ; // matrix B*Bt is singular 
  }
  */
  A = Bt*BBt ;

  Matrix_DOUBLE dD ;

  dD.resize(D.rows,3) ;
  for(i=0;i<D.rows;++i){
    const Point3D& d = D[i] ; // this makes the SGI compiler happy
    for(j=0;j<3;++j)
      dD(i,j) = (double)d.data[j] ;
  }

  Matrix_DOUBLE dP ;

  dP = A*dD ;

  for(i=0;i<map.n;++i){
    P(map[i].i,map[i].j).x() += dP(i,0)*P(map[i].i,map[i].j).w() ;
    P(map[i].i,map[i].j).y() += dP(i,1)*P(map[i].i,map[i].j).w() ;
    P(map[i].i,map[i].j).z() += dP(i,2)*P(map[i].i,map[i].j).w() ;
  }

  return 1 ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
      Routine: projectToLine --- projects a point on a line
  Description: 
	       
        Input: S --> a point on the line
	       T --> the tangent vector of the line
	     pnt --> the point to project
	       p <-- the point projected on the line
       Output:  
 Restrictions: 
   Created by: Philippe Lavoie          (25 July, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void projectToLine(const Point3D& S, const Point3D& T, const Point3D& pnt, Point3D& p) {
  Point3D a = pnt-S ;
  
  p = S+ abs(a)*cos(angle(a,T))*T.unitLength() ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: makeFromRevolution --- generates a surface of revolution
  Description: Generates a surface of revolution of a profile around an
               arbitrary axis (specified by a starting point S and a 
	       tangent T) with a certain angle.

	       The angle is specified in radians.
        Input: profile --> the curves to rotate around the axis
	             S --> a point on the axis
		     T --> the tangent vector of the rotation axis
		 theta --> the angle of the revolution
       Output:
 Restrictions: If a point is within a distance of 1e-7 from the axis,
               it will be assumed to be on the axis.
   Created by: Philippe Lavoie          (7 October, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::makeFromRevolution(const PlNurbsCurve& profile, const Point3D& S, const Point3D& T, double theta){
  double angle,dtheta ;
  int narcs ;
  int i,j ;

  while(theta>2.0*M_PI)
    theta -= 2.0*M_PI ;

  if(theta<M_PI/2.0)	
    narcs = 1 ;
  else{
    if(theta<M_PI)
      narcs = 2 ;
    else{
      if(theta<1.5*M_PI)
	narcs = 3 ;
      else
	narcs = 4 ;
    }
  }
  dtheta = theta/(double)narcs ;

  int n = 2*narcs+1 ; // n control points ;
  resize(n,profile.CtrlP.n,2,profile.degree) ;

  switch(narcs){
  case 1: break ;
  case 2: 
    U[3] = U[4] = 0.5 ;
    break ;
  case 3:
    U[3] = U[4] = 1.0/3.0 ;
    U[5] = U[6] = 2.0/3.0 ;
    break ;
  case 4:
    U[3] = U[4] = 0.25 ;
    U[5] = U[6] = 0.50 ;  
    U[7] = U[8] = 0.75 ;
    break ;    
  }

  j = 3 + 2*(narcs-1) ;  // loading the end knots
  for(i=0;i<3;++j,++i){
    U[i] = 0.0 ;
    U[j] = 1.0 ;
  }

  V = profile.Knot ;

  double wm = cos(dtheta/2.0) ; // dtheta/2.0 is base angle

  Vector_DOUBLE cosines(narcs+1),sines(narcs+1) ;
  angle = 0 ;
  for(i=1;i<=narcs;++i){
    angle = dtheta*(double)i ;
    cosines[i] = cos(angle) ;
    sines[i] = sin(angle) ;
  }

  Point3D P0,T0,P2,T2,P1 ;

  for(j=0;j<P.cols;++j){
    Point3D O ;
    float wj = profile.CtrlP[j].w() ;

    projectToLine(S,T,project(profile.CtrlP[j]),O) ;
    Point3D X,Y ;
    X = project(profile.CtrlP[j])-O ;

    double r = abs(X) ;

    Y = crossProduct(T,X) ;

    if(X.x == float(0) && 
       X.y == float(0) &&
       X.z == float(0)){
      for(i=0;i<P.rows;++i){
	P(i,j) = O ;
	P(i,j) *= wj ;
      }
      continue ;
    }
    X = X.unitLength() ;
    Y = crossProduct(T,X) ;
    Y = Y.unitLength() ;

    P0 = project(profile.CtrlP[j]) ;
    P(0,j) = profile.CtrlP[j] ;

    T0 = Y ;
    int index = 0 ;


    for(i=1;i<=narcs;++i){
      angle = dtheta*(double)i ;
      //P2 = O+ r*cosines[i]*X + r*sines[i]*Y ;  
      P2 = O+ r*cos(angle)*X + r*sin(angle)*Y ;  
      P(index+2,j) = P2 ;
      P(index+2,j) *= wj ;
      //T2 = -sines[i]*X + cosines[i]*Y ;
      T2 = -sin(angle)*X + cos(angle)*Y ;
      intersectLine(P0,T0,P2,T2,P1) ;
      P(index+1,j) = P1 ;
      P(index+1,j) *= wm*wj ;
      index += 2 ;
      P0 = P2 ;
      T0 = T2 ;
    }
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: isoCurveU --- generates an iso curve in the U direction
  Description: Generates an iso-parametric curve which goes through the
               parametric value u along the U direction.
        Input: u --> the U parametric value
	       c <-- the iso-parametric curve
       Output:
 Restrictions: the parametric value $u$ must be in a valid range
   Created by: Philippe Lavoie          (7 October, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::isoCurveU(float u, PlNurbsCurve& c) const {
  c.resize(P.cols,degV) ;

  c.modKnot(V) ;

  if(u>U[U.n-1])
    u = U[U.n-1] ;
  if(u<U[0])
    u = U[0] ;



  int span = findSpanU(u) ;
  
  Vector_FLOAT N ;
  basisFunsU(u,span,N) ;

  CtrlPnt4D p ;
  for(int i=0;i<P.cols;++i){
    p = 0 ;
    for(int j=0;j<=degU;++j)
      p += N[j]*P(span-degU+j,i) ;
    c.modCP(i,p) ;
  }
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: isoCurveV --- generates an iso curve in the V direction
  Description: Generates an iso-parametric curve which goes through the
               parametric value v along the V direction.
        Input: v --> the V parametric value
	       c <-- the iso-parametric curve

       Output:
 Restrictions: the parametric value $v$ must be in a valid range
   Created by: Philippe Lavoie          (7 October, 1997)
  Modified by: 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::isoCurveV(float v, PlNurbsCurve& c) const {
  c.resize(P.rows,degU) ;

  c.modKnot(U) ;

  if(v>V[V.n-1])
    v = V[V.n-1] ;
  if(v<V[0])
    v = V[0] ;

  int span = findSpanV(v) ;
  
  Vector_FLOAT N ;
  basisFunsV(v,span,N) ;
  
  CtrlPnt4D p ;
  for(int i=0;i<P.rows;++i){
    p = 0  ;
    for(int j=0;j<=degV;++j)
      p += N[j]*P(i,span-degV+j) ;
    c.modCP(i,p) ;
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: decompose --- decompose the surface into Bzier patches
  Description: This function decomposes the curve into an array of 4D Bzier 
               patches.
        Input: S <-- an array of Bzier segments
       Output: The number of Bzier strips in the u direction.
 Restrictions:
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::decompose(PlNurbsSurfaceArray& S) const {
  int i,m,a,b,nb,mult,j,r,save,s,k,row,col ;
  float numer,alpha ;


  Vector_FLOAT alphas(degU+1) ;
  // all the surfaces will have the same knot vector in both the U and V
  // direction
  Vector_FLOAT nU ;
  nU.resize(2*(degU+1)) ;
  for(i=0;i<nU.n/2;++i)
    nU[i] = 0 ;
  for(i=nU.n/2;i<nU.n;++i)
    nU[i] = 1 ;
  
  Vector_FLOAT nV ;
  nV.resize(2*(degV+1)) ;
  for(i=0;i<nV.n/2;++i)
    nV[i] = 0 ;
  for(i=nV.n/2;i<nV.n;++i)
    nV[i] = 1 ;

  PlNurbsSurfaceArray Su ;
  Su.resize(P.rows-degU) ; // )*(P.cols-degV)) ;
  for(i=0;i<Su.n;i++){
    Su[i].resize(degU+1,P.cols,degU,degV) ;
    Su[i].U = nU ;
    Su[i].V = V ;
  }
  
  m = P.rows+degU ;
  a = degU ;
  b = degU+1 ;
  nb = 0 ;

  for(i=0;i<=degU;i++)
    for(col=0;col<P.cols;col++)
      Su[nb].P(i,col) = P(i,col) ;
  while(b<m){
    i = b ;
    while(b<m && U[b+1] <= U[b]) b++ ;
    mult = b-i+1 ;
    if(mult<degU){
      numer = U[b]-U[a] ; // the enumerator of the alphas
      for(j=degU;j>mult;j--) // compute and store the alphas
	alphas[j-mult-1] = numer/(U[a+j]-U[a]) ;
      r = degU-mult ; // insert knot r times
      for(j=1;j<=r;j++){
	save=r-j;
	s=mult+j; // this many new points
	for(k=degU;k>=s;k--){
	  alpha = alphas[k-s] ;
	  for(col=0;col<P.cols;++col)
	    Su[nb].P(k,col) = alpha*Su[nb].P(k,col)+(1.0-alpha)*Su[nb].P(k-1,col);
	}
	if(b<m) // control point of next patch
	  for(col=0;col<P.cols;++col)
	    Su[nb+1].P(save,col) = Su[nb].P(degU,col) ;
      }
    }
    ++nb ;
    if(b<m){ // initialize for next segment
      for(i=degU-mult; i<= degU ; ++i)
	for(col=0;col<P.cols;++col)
	  Su[nb].P(i,col) = P(b-degU+i,col) ;
      a=b ;
      ++b ;
    }
  }
  Su.resize(nb) ;
  
  alphas.resize(degV+1) ;

  S.resize(Su.n*(P.cols-degV)) ;

  for(i=0;i<S.n;i++){
    S[i].resize(degU+1,degV+1,degU,degV) ;
    S[i].U = nU ;
    S[i].V = nV ;
  }
  
  nb = 0 ;

  for(int np=0;np<Su.n;++np){
    for(i=0;i<=degU;i++)
      for(j=0;j<=degV;++j)
	S[nb].P(i,j) = Su[np].P(i,j) ;
    m = P.cols+degV ;
    a = degV ;
    b = degV+1 ;
    while(b<m){
      i = b ;
      while(b<m && V[b+1] <= V[b]) b++ ;
      mult = b-i+1 ;
      if(mult<degV){
	numer = V[b]-V[a] ; // the enumerator of the alphas
	for(j=degV;j>mult;j--) // compute and store the alphas
	  alphas[j-mult-1] = numer/(V[a+j]-V[a]) ;
	r = degV-mult ; // insert knot r times
	for(j=1;j<=r;j++){
	  save=r-j;
	  s=mult+j; // this many new points
	  for(k=degV;k>=s;k--){
	    alpha = alphas[k-s] ;
	    for(row=0;row<=degU;++row)
	      S[nb].P(row,k) = alpha*S[nb].P(row,k)+(1.0-alpha)*S[nb].P(row,k-1);
	  }
	  if(b<m) // control point of next patch
	    for(row=0;row<=degU;++row)
	      S[nb+1].P(row,save) = S[nb].P(row,degV) ;
	}
      }
      ++nb ;
      if(b<m){ // initialize for next patch
	for(i=degV-mult; i<= degV ; ++i)
	  for(row=0;row<=degU;++row)
	    S[nb].P(row,i) = Su[np].P(row,b-degV+i) ;
	a=b ;
	++b ;
      }
    }
  }

  S.resize(nb) ;
  return Su.n ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writePOVRAY --- writes a set of povray bicubic patches
  Description: Writes a set of povray bicubic patches to the ostream.
        Input: patch_type --> the type of the patch
	         flatness --> the flatness level
	      num_u_steps --> the minimum number of triangles in the U 
	                      direction
	      num_v_steps --> the minimum number of triangles in the V 
	                      direction
  		   povray --> the output stream
       Output: 1 on success, 0 on failure.
 Restrictions: POVRAY only accepts rational spline patches. Thus you can't
               have a value other then 1.0 for the weights of your surface.
	       A warning message will be generated if this is the case.

	       POVRAY only deals with surfaces of degree 3. If the surface 
	       as a degree below 3 either in the U or V direction it will be
	       elevated to be at 3 in both directions. If the surface as a
	       degree higher then 3, then the function aborts.
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writePOVRAY(ostream& povray, int patch_type, double flatness, int num_u_steps, int num_v_steps) const {
  if(degU>3 || degV>3){
    PlError err("PlNurbsSurface writePOVRAY") ;
    err << "The degree of the surface is higher than 3!\n"
	<< "A povrary file can not be generated.\n" ;
    err.warning() ;
    return 0;
  }


  PlNurbsSurface S(*this) ;

  S.degreeElevate(3-degU,3-degV) ;

  PlNurbsSurfaceArray Sa ;
  S.decompose(Sa) ;
  int bad = 0 ; 

  povray << "//\n//Generated for POV-Ray(tm) 3.0 by Phil's NURBS library\n" ;
  povray << "//   http://yukon.genie.uottawa.ca/info/soft/nurbs\n//\n" ;

  for(int i=0;i<Sa.n;++i){
    povray << "bicubic_patch {\n\ttype " << patch_type << endl ;
    povray << "\tflatness " << flatness << endl ;
    povray << "\tu_steps " << num_u_steps << endl ;
    povray << "\tv_steps " << num_v_steps << endl ;
    for(int j=0;j<4;++j){
      for(int k=0;k<4;++k){
	Point3D p = project(Sa[i].CtrlP(j,k)) ;
	if(Sa[i].CtrlP(j,k).w()>1.0001 || Sa[i].CtrlP(j,k).w()<0.9999)
	  bad = 1 ;
	povray << "\t<" << p.x << ", " << p.y << ", " << p.z << ">" ;
	if(j==3 && k==3)
	  povray << "\n}\n\n" ;
	else
	  povray << ",\n " ;
      }
      povray << "\n" ;
    }
  }
  if(bad){
    PlError err("PlNurbsSurface writePOVRAY") ;
    err << "Warning: at least one of the control point was not rational\n" ;
    err << "The resulting surface will NOT be the same as the one which\n" ;
    err << "generated it.\n" ;
    err.warning() ;
  }
  return bad ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writePOVRAY --- writes the surface as a mesh of triangles
  Description: Writes the surface as a mesh of triangles. You might have
               to change the values for the tolerance to get exactly what
	       you're looking for.
        Input: povray <-- the output stream
	       tolerance --> the tolerance when performing the tesselation
	       color --> the color of the object {\em [(250,250,250)]}
	       diffuse --> the diffuse factor {\em [0.6]}
	       ambient --> the ambient factor {\em [0.2]}
	       smooth --> flags which indicates if we generate smooth triangles {\em [0]}
       Output: 1 on success, 0 on failure
 Restrictions: 
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writePOVRAY(float tolerance, ostream& povray,const Color& color, int smooth, float ambient, float diffuse) const {
  BasicList<Point3D> points ;
  BasicList<int> connect ;
  BasicList<Point3D> norm ;

  if(smooth)
    tesselate(tolerance,points,connect,&norm) ;
  else
    tesselate(tolerance,points,connect) ;
    
  povray << "mesh {\n" ;
  
  BasicNode<int> *node ;

  BasicNode<Point3D> *nodeP ;
  BasicNode<Point3D> *nodeN ;
  nodeP = points.goToFirst() ;
  nodeN = 0 ;

  if(smooth)
    nodeN = norm.goToFirst() ;

  Vector_Point3D Pts(points.size()) ;
  Vector_Point3D Norm(norm.size()) ;
  for(int i=0;i<Pts.n;++i){
    Pts[i] = *nodeP->data ;
    nodeP = points.goToNext() ;
    if(smooth){
      Norm[i] = *nodeN->data ;
      nodeN = norm.goToNext() ;
    }
  }

  node = connect.goToFirst() ;
  while(node){
    if(smooth)
      povray << "\tsmooth_triangle { " ;
    else
      povray << "\ttriangle { " ;
    povray << "< " << Pts[*node->data].x << ", " << Pts[*node->data].y << ", " << Pts[*node->data].z << "> , " ;
    if(smooth)
      povray << "< " << Norm[*node->data].x << ", " << Norm[*node->data].y << ", " << Norm[*node->data].z << "> , " ;
    node = connect.goToNext() ; 
    povray << "< " << Pts[*node->data].x << ", " << Pts[*node->data].y << ", " << Pts[*node->data].z << "> , " ;
    if(smooth)
      povray << "< " << Norm[*node->data].x << ", " << Norm[*node->data].y << ", " << Norm[*node->data].z << "> , " ;
    node = connect.goToNext() ;
    povray << "< " << Pts[*node->data].x << ", " << Pts[*node->data].y << ", " << Pts[*node->data].z << "> " ;
    if(smooth)
      povray << ", < " << Norm[*node->data].x << ", " << Norm[*node->data].y << ", " << Norm[*node->data].z << "> " ;
    node = connect.goToNext() ; // skip the -1 value
    node = connect.goToNext() ;
    povray << "}\n" ;
  }
  
  povray << "\ttexture{ pigment { rgb < " << (double)color.r/255.0 << ", " << (double)color.g/255.0 << ", " << (double)color.b/255.0 << " >}\n" ;
  povray << "\t\tfinish { ambient " << ambient << " diffuse " << diffuse << " }\n\t}\n" ;
  
  povray << "}\n\n" ;
  return povray.good() ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writePOVRAY --- writes a set of povray bicubic patches
  Description: Writes a set of povray bicubic patches to the ostream.
        Input: patch_type --> the type of the patch
	         flatness --> the flatness level
	      num_u_steps --> the minimum number of triangles in the U 
	                      direction
	      num_v_steps --> the minimum number of triangles in the V 
	                      direction
       Output: an ostream containing the definition of the surface
 Restrictions: POVRAY only accepts rational spline patches. Thus you can't
               have a value other then 1.0 for the weights of your surface.
	       A warning message will be generated if this is the case.

	       POVRAY only deals with surfaces of degree 3. If the surface 
	       as a degree below 3 either in the U or V direction it will be
	       elevated to be at 3 in both directions. If the surface as a
	       degree higher then 3, then the function aborts.
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writePOVRAY(const char *filename, const Color& col, const Point3D& cView, const Point3D& up, int patch_type, double flatness, int num_u_steps, int num_v_steps) const {
  ofstream fout(filename) ;
  if(!fout)
    return 0 ;
  
  Point3D view(-1*cView) ;

  fout << "//\n//Generated for POV-Ray(tm) 3.0 by Phil's NURBS library\n//\n" ;
  fout << "\n#include \"colors.inc\"\n" ;

  // we want the camera to look at the center of the object
  // and be able to view the whole object
  // we use and angle of 36 to view the object
  // and position the rest according to this.
  Point3D minP, maxP ;
  minP.x = extremum(1,coordX) ;
  minP.y = extremum(1,coordY) ;
  minP.z = extremum(1,coordZ) ;
  maxP.x = extremum(0,coordX) ;
  maxP.y = extremum(0,coordY) ;
  maxP.z = extremum(0,coordZ) ;

  Point3D lookAt  ;
  lookAt.x = (minP.x+maxP.x)/2.0 ;
  lookAt.y = (minP.y+maxP.y)/2.0 ;
  lookAt.z = (minP.z+maxP.z)/2.0 ;

  Point3D camera1, camera2 ;

  Point3D q1 = minP-lookAt ;
  Point3D q2 = maxP-lookAt ;
  float D1 = absolute(dot(q1,view))/abs(view) ;
  float D2 = absolute(dot(q2,view))/abs(view) ;

  float a1 = abs(q1)*cos(angle(view,q1)) ;
  float a2 = abs(q2)*cos(angle(view,q2)) ;

  float b1 = D1/tan(18.0*M_PI/180.0) ;
  float b2 = D2/tan(18.0*M_PI/180.0) ; // this gives the 36 degree angle

  camera1 = lookAt+(a1+b1)*view.unitLength() ;
  camera2 = lookAt+(a2+b2)*view.unitLength() ;

  Point3D right ;

  right = crossProduct(view,up) ; // inversed because pov-ray uses a left-handed system

  fout << "camera {\n\tlocation <" ;
  if(abs2(camera1-lookAt)>abs2(camera2-lookAt))
    fout << camera1.x << ", " << camera1.y << ", " << camera1.z << ">\n" ;
  else
    fout << camera2.x << ", " << camera2.y << ", " << camera2.z << ">\n" ;
  fout << "\tup < " << up.x << ", " << up.y << ", " << up.z << ">\n" ;
  fout << "\tright < " << right.x << ", " << right.y << ", " << right.z << ">\n" ;
  fout << "\tlook_at < " << lookAt.x << ", " << lookAt.y << ", " << lookAt.z << ">\n\tangle 36\n}\n\n" ;

  fout << "union {\n" ;
  writePOVRAY(fout,patch_type,flatness,num_u_steps,num_v_steps) ;
  fout << " texture {\n\tpigment {\n\t\tcolor rgb < " << (double)col.r/255.0 << 
    ", " << (double)col.g/255.0 << ", " << (double)col.b/255.0 << "> \n" <<
    "\t}\n\tfinish { \n\t\tambient .2\n\t\tdiffuse .6\n\t}\n }\n" ;
  fout << "\n}\n" ;

  fout << "light_source { < " ;
  if(abs2(camera1-lookAt)>abs2(camera2-lookAt))
    fout << camera1.x+view.x << ", " << camera1.y+view.y << ", " << camera1.z+view.z << "> color White}\n\n" ;
  else
    fout << camera2.x+view.x << ", " << camera2.y+view.y << ", " << camera2.z+view.z << "> color White}\n\n" ;
  
  return fout.good() ;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writePOVRAY --- writes a set of povray bicubic patches
  Description: Writes a set of povray bicubic patches to the ostream.
        Input: patch_type --> the type of the patch
	         flatness --> the flatness level
	      num_u_steps --> the minimum number of triangles in the U 
	                      direction
	      num_v_steps --> the minimum number of triangles in the V 
	                      direction
       Output: an ostream containing the definition of the surface
 Restrictions: POVRAY only accepts rational spline patches. Thus you can't
               have a value other then 1.0 for the weights of your surface.
	       A warning message will be generated if this is the case.

	       POVRAY only deals with surfaces of degree 3. If the surface 
	       as a degree below 3 either in the U or V direction it will be
	       elevated to be at 3 in both directions. If the surface as a
	       degree higher then 3, then the function aborts.
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writePOVRAY(float tolerance, const char *filename, const Color& col, const Point3D& cView, const Point3D& up, int smooth, float ambient, float diffuse) const {
  ofstream fout(filename) ;
  if(!fout)
    return 0 ;
  
  Point3D view(-1*cView) ;

  fout << "//\n//Generated for POV-Ray(tm) 3.0 by Phil's NURBS library\n//\n" ;
  fout << "\n#include \"colors.inc\"\n" ;

  // we want the camera to look at the center of the object
  // and be able to view the whole object
  // we use and angle of 36 to view the object
  // and position the rest according to this.
  Point3D minP, maxP ;
  minP.x = extremum(1,coordX) ;
  minP.y = extremum(1,coordY) ;
  minP.z = extremum(1,coordZ) ;
  maxP.x = extremum(0,coordX) ;
  maxP.y = extremum(0,coordY) ;
  maxP.z = extremum(0,coordZ) ;

  Point3D lookAt  ;
  lookAt.x = (minP.x+maxP.x)/2.0 ;
  lookAt.y = (minP.y+maxP.y)/2.0 ;
  lookAt.z = (minP.z+maxP.z)/2.0 ;

  Point3D camera1, camera2 ;

  Point3D q1 = minP-lookAt ;
  Point3D q2 = maxP-lookAt ;
  float D1 = absolute(dot(q1,view))/abs(view) ;
  float D2 = absolute(dot(q2,view))/abs(view) ;

  float a1 = abs(q1)*cos(angle(view,q1)) ;
  float a2 = abs(q2)*cos(angle(view,q2)) ;

  float b1 = D1/tan(18.0*M_PI/180.0) ;
  float b2 = D2/tan(18.0*M_PI/180.0) ; // this gives the 36 degree angle

  camera1 = lookAt+(a1+b1)*view.unitLength() ;
  camera2 = lookAt+(a2+b2)*view.unitLength() ;

  Point3D right ;

  right = crossProduct(view,up) ; // inversed because pov-ray uses a left-handed system

  fout << "camera {\n\tlocation <" ;
  if(abs2(camera1-lookAt)>abs2(camera2-lookAt))
    fout << camera1.x << ", " << camera1.y << ", " << camera1.z << ">\n" ;
  else
    fout << camera2.x << ", " << camera2.y << ", " << camera2.z << ">\n" ;
  fout << "\tup < " << up.x << ", " << up.y << ", " << up.z << ">\n" ;
  fout << "\tright < " << right.x << ", " << right.y << ", " << right.z << ">\n" ;
  fout << "\tlook_at < " << lookAt.x << ", " << lookAt.y << ", " << lookAt.z << ">\n\tangle 36\n}\n\n" ;

  writePOVRAY(tolerance,fout,col,smooth,ambient,diffuse) ;
  

  fout << "light_source { < " ;
  if(abs2(camera1-lookAt)>abs2(camera2-lookAt))
    fout << camera1.x+view.x << ", " << camera1.y+view.y << ", " << camera1.z+view.z << "> color White}\n\n" ;
  else
    fout << camera2.x+view.x << ", " << camera2.y+view.y << ", " << camera2.z+view.z << "> color White}\n\n" ;
  
  return fout.good() ;
}




/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writeRIB --- writes a NuPatch for render man
  Description: Writes a stream which is compatible with Render Man 
               specifications of a NURBS surface.

	       \begin{center}
	       The RenderMan \Pisymbol{psy}{226} % the (R) symbol
	       Interface Procedures and RIB Protocol are:\\
	       Copyright 1988, 1989, Pixar.  All rights reserved.\\
	       RenderMan \Pisymbol{psy}{226} is a registered trademark of 
	       Pixar.\\
	       \end{center}


        Input: rib <-- the rib ostream
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writeRIB(ostream& rib) const {
  rib << "NuPatch " << P.rows << " " << degU+1 << " [ " ; 
  int k;
  // I have to loop since RIB is new line sensitive and 
  // by default a vector adds a new line at the end when using cout
  for(k=0;k<U.n;++k) 
    rib << U[k] << " " ;
  rib << " ] " << U[0] << " " << U[U.n-1] << " " << P.cols << " " << degV+1 << " [ " ;
  for(k=0;k<V.n;++k)
    rib << V[k] << " " ;
  rib << " ] " << V[0] << " " << V[V.n-1] << " \"Pw\" [ " ;
  for(int j=0;j<P.cols;++j)
    for(int i=0;i<P.rows;++i)
      rib << P(i,j) ;
  rib << " ]\n" ;
  return rib.good() ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writeRIB --- writes a NuPatch for render man
  Description: Writes a file whith follows the RIB protocol of RenderMan.
               It generates a file which views the whole object.
	       The material used for rendering is plastic.

	       \begin{center}
	       The RenderMan \Pisymbol{psy}{226} % the (R) symbol
	       Interface Procedures and RIB Protocol are:\\
	       Copyright 1988, 1989, Pixar.  All rights reserved.\\
	       RenderMan \Pisymbol{psy}{226} is a registered trademark of 
	       Pixar.\\
	       \end{center}

        Input: filename --> the file to write to
	            col --> the color of the object
		   view --> the view point
       Output: 0 if an error occurs, 1 otherwise
 Restrictions: 
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writeRIB(const char* filename, const Color& col, const Point3D& view) const {
  ofstream fout(filename) ;
  if(!fout)
    return 0;
  
  // The following code is based on Listing 8.2 from the RenderMan Companion
  // http://pete.cs.caltech.edu/RMR/Pixar/ch8/listing8_2.c
  int i,j ;

  // aimZ
  MatrixRT<double> Rx ;
  double xzlen, yzlen, yrot, xrot;

  //
  // The initial rotation about the y axis is given by the projection of
  // the direction vector onto the x,z plane: the x and z components
  // of the direction. 
  xzlen = sqrt(view.x*view.x+view.z*view.z) ;
  if(xzlen == 0)    
    yrot = (view.y < 0) ? M_PI : 0;
  else
    yrot = acos(view.y/xzlen);
  
  //
  // The second rotation, about the x axis, is given by the projection on
  // the y,z plane of the y-rotated direction vector: the original y
  // component, and the rotated x,z vector from above. 
  //
  yzlen = sqrt(view.y*view.y+xzlen*xzlen);
  xrot = acos(xzlen/yzlen);       /* yzlen should never be 0 */

  // A rotation around y first 
  if (view.y > 0){
    if(view.x>0)
      Rx.rotate(xrot,yrot,0.0) ;
    else
      Rx.rotate(xrot,-yrot,0.0) ;
  }
  else{
    if(view.x>0)
      Rx.rotate(-xrot,yrot,0.0) ;
    else
      Rx.rotate(-xrot,-yrot,0.0) ;
  }

  Point3D minP, maxP ;
  minP.x = extremum(1,coordX) ;
  minP.y = extremum(1,coordY) ;
  minP.z = extremum(1,coordZ) ;
  maxP.x = extremum(0,coordX) ;
  maxP.y = extremum(0,coordY) ;
  maxP.z = extremum(0,coordZ) ;

  Point3D lookAt  ;
  lookAt.x = (minP.x+maxP.x)/2.0 ;
  lookAt.y = (minP.y+maxP.y)/2.0 ;
  lookAt.z = (minP.z+maxP.z)/2.0 ;

  Point3D camera1, camera2 ;

  Point3D q1 = minP-lookAt ;
  Point3D q2 = maxP-lookAt ;
  float D1 = absolute(dot(q1,view))/abs(view) ;
  float D2 = absolute(dot(q2,view))/abs(view) ;

  float a1 = abs(q1)*cos(angle(view,q1)) ;
  float a2 = abs(q2)*cos(angle(view,q2)) ;

  float b1 = D1/tan(18.0*M_PI/180.0) ;
  float b2 = D2/tan(18.0*M_PI/180.0) ; // this gives the 36 degree angle

  camera1 = lookAt+(a1+b1)*view.unitLength() ;
  camera2 = lookAt+(a2+b2)*view.unitLength() ;

  Point3D camera ;

  if(abs(camera1-lookAt)>abs(camera2-lookAt))
    camera = camera1 ;
  else
    camera = camera2 ;

  char front[1024] ;

  char *ext ;
  ext = strstr(filename,".rib") ;
  if(ext){
    for(i=0;i<1024;++i){
      if(&filename[i] == ext)
	break ;
      else
	front[i] = filename[i] ;
      if(!front[i])
	break ;
    }
  }
  else{
    strcpy(front,filename) ;
  }
  

  fout << "##RenderMan RIB-Structure 1.0\n" ;
  fout << "#" << filename << endl;
  fout << "Format 400 400 1\n";
  fout << "Display \"" << front << ".tif\" \"file\" \"rgba\"\n" ;
  fout << "Projection \"perspective\" \"fov\" [36]\n" ;
  fout << "Translate 0 0 " << abs(camera-lookAt) << endl ;
  fout << "Option \"render\" \"prmanspecular\" [1]\n" ;

  fout << "\nWorldBegin\n" ;
  fout << "LightSource \"ambientlight\" 0 \"intensity\" [0.3]\n" ;
  fout << "LightSource \"distantlight\" 1 \"to\" [ " << view.x << " " << view.y << " " << view.z  << "]\n" ;
  //fout << "LightSource \"spotlight\" 1 \"intensity\" [300] \"from\" [ " ;
  //fout << camera.x+view.x << " " << camera.y+view.y << " " << camera.z+view.z << " ]\n\n" ;
  fout << "AttributeBegin\nSurface \"plastic\"\n" ;
  fout << "Color [ " << (double)col.r/255.0 << " " << (double)col.g/255.0 << " " << double(col.b)/255.0 << "]\n" ;

  fout << "Transform [ " ;

  for(j=0;j<4;++j)
    for(i=0;i<4;++i)
      fout << Rx(i,j) << " " ;
  fout << "]\n" ;
  fout << "Translate " << -lookAt.x << " " << -lookAt.y << " " << -lookAt.z << endl ;


  writeRIB(fout) ;

  fout << "AttributeEnd\nWorldEnd\n\n" ;

  return fout.good() ;
}

inline void borderTess(float tolerance, const PlNurbsCurve &c, BasicList<Point3D>& points, BasicList<int>& connect, BasicList<float> &uk, Vector_Point3D &border, Vector_Point3D& inPts, int start, int inSize, int mulF, int addF){
  int i,j ;
  BasicList<Point3D> list ;
  uk.reset() ;
  list = c.tesselate(tolerance/float(4),&uk) ;
  int startC = points.size() ;
  points.addElements(list) ;

  /*
  border.resize(list.size()) ;
  BasicNode<Point3D> *node = list.goToFirst() ;
  for(i=0;i<border.n;++i){
    border[i] = *node->data ;
    node = list.goToNext() ;
  }
  */

  double nf = double(inSize)/double(list.size()) ;
  if(nf<float(1)){
    j = 0 ;
    for(i=0;i<list.size()-1;++i){
      if(j>=inSize-1){
	for(;i<list.size()-1;++i){
	  connect.add(startC+i) ;
	  connect.add(start+j*mulF+addF);
	  connect.add(startC+i+1) ;
	  connect.add(-1) ;
	}
	break ;
	}
	if(abs2(inPts[j*mulF+addF] - *list[i]->data)>abs2(inPts[(j+1)*mulF+addF] - *list[i]->data)){
	  connect.add(startC+i) ;
	  connect.add(start+j*mulF+addF) ;
	  connect.add(start+(j+1)*mulF+addF) ;
	  connect.add(-1) ;
	  ++j ;
	}
	connect.add(startC+i) ;
	connect.add(start+j*mulF+addF) ;
	connect.add(startC+i+1) ;
	connect.add(-1) ;
      }
      for(;j<inSize-1;++j){
	  connect.add(startC+i) ;
	  connect.add(start+j*mulF+addF);
	  connect.add(start+(j+1)*mulF+addF) ;
	  connect.add(-1) ;
      }
    }
    else{
      i = 0 ;
      for(j=0;j<inSize-1;++j){
	if(i>=list.size()-1){
	  for(;j<inSize-1;++j){
	    connect.add(startC+i) ;
	    connect.add(start+j*mulF+addF);
	    connect.add(start+(j+1)*mulF+addF) ;
	    connect.add(-1) ;
	  }
	  break ;
	}
	if(abs2(inPts[(j+1)*mulF+addF] - *list[i]->data)>abs2(inPts[(j+1)*mulF+addF] - *list[i+1]->data)){
	  connect.add(startC+i) ;
	  connect.add(start+j*mulF+addF);
	  connect.add(startC+i+1) ;
	  connect.add(-1) ;
	  ++i ;
	}
	connect.add(startC+i) ;
	connect.add(start+j*mulF+addF) ;
	connect.add(start+(j+1)*mulF+addF) ;
	connect.add(-1) ;
      }
      for(;i<list.size()-1;++i){
	  connect.add(startC+i) ;
	  connect.add(start+(inSize-1)*mulF+addF) ;
	  connect.add(startC+i+1) ;
	  connect.add(-1) ;
      }
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: tesselate --- generates a list of triangles for a surface
  Description: Generates the list of triangles necessary to represent the
               surface. This generates the 3D points and the method
	       how to connect them. The connect list is a list of the form
	       $\{\ldots, j, k, l, -1, \dots\}$ where the $j$,$k$ and $l$
	       value specify the end points of a triangle, the -1 is used
	       to delimit triangle definitions.
        Input: tolerance --> the tolerance for the tesselation.
	       points <-- the list of points
	       connect <-- how the points should be connected
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie (8 October, 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::tesselate(float tolerance, BasicList<Point3D> &points, BasicList<int> &connect, BasicList<Point3D> *norm) const {
  PlNurbsSurfaceArray sa ;
  PlNurbsSurface surf(*this) ;
  surf.decompose(sa) ;
  if(sa.n==1){
    float u = 0 ;
    float v = 0 ;
    const int nSample = 1 ;

    Vector_Point3D maxDu(nSample+1) ;
    Point3D maxDU(0,0,0) ;
    Point3D maxDv(0,0,0),maxDV(0,0,0) ;
    Point3D prevV ;

    Matrix_Point3D ders(2,2) ;


    int i,j ;
    Vector_Point3D prevU(nSample+1) ;


    for(i=0;i<nSample+1;++i){
      u = float(i)/float(nSample) ;
      deriveAt(u,0,1,ders) ;
      prevU[i] = ders(1,0) ;
    }

    for(i=1;i<nSample+1;++i){
      u = v = float(0) ;
      deriveAt(u,v,1,ders) ;
      prevV = ders(0,1) ;
      u = float(i)/float(nSample) ;
      maxDv = Point3D(0,0,0) ;
      maxDu[0] = Point3D(0,0,0) ;
      for(j=0;j<nSample+1;++j){
	v = float(j)/float(nSample) ;
	deriveAt(u,v,1,ders) ;
	Point3D deltaU = ders(1,0)-prevU[j] ;
	Point3D deltaV = ders(0,1)-prevV ;
	deltaU.x = absolute(deltaU.x) ;
	deltaU.y = absolute(deltaU.y) ;
	deltaU.z = absolute(deltaU.z) ;
	deltaV.x = absolute(deltaV.x) ;
	deltaV.y = absolute(deltaV.y) ;
	deltaV.z = absolute(deltaV.z) ;
	maxDu[j] = maximumRef(maxDu[j],deltaU) ;
	maxDv = maximumRef(maxDv,deltaV) ;
	prevU[j] = ders(1,0) ;
	prevV = ders(0,1) ; 
      }
      maxDV = abs2(maxDv)>abs2(maxDV) ? maxDv: maxDV ;
    }

    for(i=1;i<maxDu.n;++i)
      maxDU = abs2(maxDu[i])>abs2(maxDU) ? maxDu[i] : maxDU ;

    const float sqr2 = float(1.414241527) ;

    int nu = int(sqr2*abs(maxDU)/tolerance) ;
    int nv = int(sqr2*abs(maxDV)/tolerance) ;

    nu+=2 ;
    nv+=2 ;

    if(nu<4)
      nu = 4 ;
    if(nv<4)
      nv = 4 ;
    
    int start = points.size() ;

    Vector_Point3D inPts((nu-2)*(nv-2)) ;

    int k = 0; 

    for(i=1;i<nu-1;++i)
      for(j=1;j<nv-1;++j){
	u = (U[U.n-1]-U[0])*float(i)/float(nu-1) + U[0] ;
	v = (V[V.n-1]-V[0])*float(j)/float(nv-1) + V[0] ;
	inPts[k] = point3D(u,v) ;
	points.add(inPts[k]) ;
	//points.add(point3D(u,v)) ;
	++k ;
	if(norm)
	  norm->add(normal(u,v)) ;
    }

    for(i=0;i<nu-3;++i)
      for(j=0;j<nv-3;++j){
	connect.add(start+i*(nv-2)+j) ;
	connect.add(start+i*(nv-2)+j+1) ;
	connect.add(start+(i+1)*(nv-2)+j) ;
	connect.add(-1) ;
	connect.add(start+i*(nv-2)+j+1) ;
	connect.add(start+(i+1)*(nv-2)+j+1) ;
	connect.add(start+(i+1)*(nv-2)+j) ;
	connect.add(-1) ;
      }

    // now we need to add the contours
    PlNurbsCurve c ;
    
    BasicList<float> uk ;
    Vector_Point3D border;

    isoCurveV(U[0],c) ;
    borderTess(tolerance,c,points,connect,uk,border,inPts,start,nu-2,nv-2,0);
    if(norm){
      BasicNode<float> *fnode = uk.goToFirst() ;
      for(i=0;i<uk.size();++i){
	norm->add(normal(U[0],*fnode->data)) ;
	fnode = uk.goToNext() ;
      }
    }

    isoCurveV(U[U.n-1],c) ;
    borderTess(tolerance,c,points,connect,uk,border,inPts,start,nu-2,nv-2,nv-3);
    if(norm){
      BasicNode<float> *fnode = uk.goToFirst() ;
      for(i=0;i<uk.size();++i){
	norm->add(normal(U[U.n-1],*fnode->data)) ;
	fnode = uk.goToNext() ;
      }
    }


    isoCurveU(V[0],c) ;
    borderTess(tolerance,c,points,connect,uk,border,inPts,start,nv-2,1,0);
    if(norm){
      BasicNode<float> *fnode = uk.goToFirst() ;
      for(i=0;i<uk.size();++i){
	norm->add(normal(*fnode->data,V[0])) ;
	fnode = uk.goToNext() ;
      }
    }


    isoCurveU(V[V.n-1],c) ;
    borderTess(tolerance,c,points,connect,uk,border,inPts,start,nv-2,1,(nv-2)*(nu-3)) ;
    if(norm){
      BasicNode<float> *fnode = uk.goToFirst() ;
      for(i=0;i<uk.size();++i){
	norm->add(normal(*fnode->data,V[V.n-1])) ;
	fnode = uk.goToNext() ;
      }
    }

  }
  else{
    for(int i=0;i<sa.n;++i){
      sa[i].tesselate(tolerance,points,connect,norm) ;
    }
  }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writeVRML --- write the NURBS surface to a VRML file
  Description: Writes a VRML file which represents the surface for the 
               parametric space $[uS,uE]$ and $[vS,vE]$.
               It does not optimize the number of points required to 
	       represent the surface. 
        Input: filename --> the file name for the output VRML file
	      tolerance --> the tolerance for the tesselation.
		  color --> the color of the object
       Output: 1 on success, 0 otherwise
 Restrictions: The parametric surface must be valid
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by:  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlNurbsSurface::writeVRML(const char* filename, float tolerance, const Color& color) const {
  ofstream fout(filename) ;
  if(!fout)
    return 0 ;

  BasicList<Point3D> points ;
  BasicList<int> connect ;

  tesselate(tolerance,points,connect) ;
  
  fout << "#VRML V1.0 ascii\n" ;
  fout << "#  Generated by Phil's NURBS library\n" ;
  fout << "\nSeparator {\n" << "\tMaterialBinding { value PER_VERTEX_INDEXED }\n"  ;
  //fout << "\tMaterial{ diffuseColor " << (int)color.r << " " << (int)color.g << " " << (int)color.b << "}\n" ;
  fout << "\tCoordinate3 {\n" ;
  fout << "\t\tpoint [\n" ;

  BasicNode<Point3D> *node ;
  node = points.goToFirst() ;
  while(node){
    fout << "\t\t\t" << node->data->x << " " << node->data->y << " " << node->data->z << ",\n" ;
    node = points.goToNext() ;
  }
  
  fout << "\t\t]\n" ; // point [
  fout << "\t}\n" ; // cordinate3

  fout << "\tIndexedFaceSet{\n" ;
  fout << "\t\tcoordIndex[\n" ;
  
  BasicNode<int> *nodeC ;
  nodeC = connect.goToFirst() ;
  while(nodeC){
    fout << "\t\t\t" << *nodeC->data << ", " ; // x
    nodeC = connect.goToNext() ;
    fout << *nodeC->data << ", " ; // y
    nodeC = connect.goToNext() ;
    fout << *nodeC->data << "," ;  // z
    nodeC = connect.goToNext() ;
    fout << *nodeC->data << ",\n" ; // -1
    nodeC = connect.goToNext() ;
  }
  fout << "\t\t]\n" ;
  fout << "\t}\n" ; // IndexedFaceSet

  fout << "}\n" ;

  return 1 ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: makeSphere --- generates a sphere
  Description: The NURBS surface is now a sphere of radius $r$ located
               at $O$.
        Input: O --> the location of the center of the sphere
	       r --> the radius of the sphere
       Output: 
 Restrictions: 
   Created by: Philippe Lavoie          (8 May, 1998)
  Modified by:  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurface::makeSphere(const Point3D& O, float r) {
  PlNurbsCurve c ;

  c.makeCircle(Point3D(0,0,0),r,0,M_PI) ;

  makeFromRevolution(c,Point3D(0,0,0),Point3D(1,0,0),2.0*M_PI) ; 
  MatrixRT<float> T ;
  T.translate(O.x,O.y,O.z) ;
  transform(T) ;
}
