/*=====================================================================
        File: nurbsS_sp.cc
     Purpose:       
    Revision: $Id: nurbsS_sp.cc,v 2.3 1998/06/01 20:21:07 lavoie Exp $
  Created by: Philippe Lavoie          (8 May, 1998)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=====================================================================*/
#include <nurbsS_sp.hh>
#include "nan.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateMaxU --- updates the basis value
  Description: Updates the basis value at which a control point has 
               maximal influence. It also finds where the control point
	       has maximal influence. 
        Input:
       Output:
 Restrictions: The degrre in U of the surface must be of 3 or less.
   Created by: Philippe Lavoie (8 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurfaceSP::updateMaxU() {
  if(degU>3){
    PlError error("PlNurbsSurfaceSP::updateMaxU()") ;
    error << "This class doesn't support surfaces having a degree of 4 and higher.\n" ;
    error.fatal() ;
  }
  else{
    maxU.resize(P.rows) ;
    maxAtU_.resize(P.rows) ;
    for(int i=0;i<P.rows;++i){
      if(!maxInfluence(i,U,degU,maxAtU_[i]))
	cerr << "Problem in maxInfluence U!\n" ;
      maxU[i] = nurbsBasisFun(maxAtU_[i],i,degU,U) ;
    }
    
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: updateMaxV --- updates the basis value
  Description: Updates the basis value at which a control point has 
               maximal influence. It also finds where the control point
	       has maximal influence. 
        Input:
       Output:
 Restrictions: The degrre in V of the surface must be of 3 or less.
   Created by: Philippe Lavoie (8 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void PlNurbsSurfaceSP::updateMaxV() {
  if(degV>3){
    PlError error("PlNurbsSurfaceSP::updateMaxV()") ;
    error << "This class doesn't support surfaces having a degree of 4 and higher.\n" ;
    error.fatal() ;
  }
  else{
    maxV.resize(P.cols) ;
    maxAtV_.resize(P.cols) ;
    for(int i=0;i<P.cols;++i){
      if(!maxInfluence(i,V,degV,maxAtV_[i]))
	cerr << "Problem in maxInfluence V!\n" ;
      maxV[i] = nurbsBasisFun(maxAtV_[i],i,degV,V) ;
    }
    
  }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: generateParallel --- generate a parallel surface
  Description: Generates an offset surface from this surface. An offset
               surface is a surface wich has its surface parallel to
	       an other one. There is a distance of $d$ between the two
	       parallel surfaces.

	       The algorithm used is very naive. It generates a surface
	       such that a point $s_2(u,v) = s(u,v) + d n(u,v)$ where
	       $s_2(u,v,)$ is the point on the parallel surface at $(u,v)$,
	       $s(u,v)$ is the point on the original surface at $(u,v)$,
	       $d$ is the offset between the two and $n(u,v)$ is the
	       normal on the surface at the point $(u,v)$.
        Input: d --> the distance between the surface and its offset
       Output: the parallel surface.
 Restrictions: 
   Created by: Philippe Lavoie (8 May, 1998)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
PlNurbsSurfaceSP PlNurbsSurfaceSP::generateParallel(float d) const {
  PlNurbsSurfaceSP p(*this) ;

  Vector_Point3D offset(P.rows*P.cols) ;
  Vector_FLOAT ur(P.rows*P.cols) ;
  Vector_FLOAT vr(P.rows*P.cols) ;
  Vector_INT Du(P.rows*P.cols) ;
  Vector_INT Dv(P.rows*P.cols) ;

  Du.reset(0) ;
  Dv.reset(0) ;

  int i,j,no ;

  no = 0 ;

  for(i=0;i<P.rows;++i)
    for(j=0;j<P.cols;++j){
      Point3D norm ;
      norm = normal(maxAtU_[i],maxAtV_[j]) ;
      if(norm.x == float(0) && 
	 norm.y == float(0) &&
	 norm.z == float(0)){
	// normal is undefined there...
	// compute an average and find a suitable normal
	const float delta = 0.00001 ;
	// must handle the corner cases
	int ok = 0 ; 
	if(i==0 && j==0){
	  norm = normal(maxAtU_[i]+delta,maxAtV_[j]) ;
	  norm += normal(maxAtU_[i],maxAtV_[j]+delta) ;
	  norm /= float(2) ;
	  ok = 1 ;
	}
	if(i==P.rows-1 && j==P.cols-1){
	  norm = normal(maxAtU_[i]-delta,maxAtV_[j]) ;
	  norm += normal(maxAtU_[i],maxAtV_[j]-delta) ;
	  norm /= float(2) ;
	  ok = 1 ;
	}
	if(i==0 && j==P.cols-1){
	  norm = normal(maxAtU_[i]-delta,maxAtV_[j]) ;
	  norm += normal(maxAtU_[i],maxAtV_[j]+delta) ;
	  norm /= float(2) ;
	  ok = 1 ;
	}
	if(i==P.rows-1 && j==0){
	  norm = normal(maxAtU_[i]-delta,maxAtV_[j]) ;
	  norm += normal(maxAtU_[i],maxAtV_[j]+delta) ;
	  norm /= float(2) ;
	  ok = 1 ;
	}
	if(!ok){
	  float nt = 1.0 ; 
	  while(norm.x == float(0) && 
	     norm.y == float(0) &&
	     norm.z == float(0)){
	    if( nt*d >(U[U.n-1]-U[0])){
	      PlError error("generateParallel");
	      error << "Can't compute a normal point.\n" ;
	      error.fatal() ;
	    }
	    float u1,u2,v1,v2 ;
	    if(i==0 || i==P.rows-1){
	      u1 = u2 = maxAtU_[i] ;
	      v1 = maxAtV_[j]+ nt*delta ;
	      v2 = maxAtV_[j]- nt*delta ;
	      if(v1>V[V.n-1]) v1 = V[V.n-1] ;
	      if(v2<V[0]) v2 = V[0] ;
	      norm = normal(u1,v1);
	      norm += normal(u2,v2) ;
	      norm /= 2 ; 
	    }
	    else{
	      u1 = maxAtU_[i]- nt*delta ;
	      u2 = maxAtU_[i]+ nt*delta ;
	      v1 = v2 = maxAtV_[j] ;
	      if(u1 < U[0]) u1 = U[0] ;
	      if(u2 > U[U.n-1]) u2 = U[U.n-1] ;

	      float u3,v3 ;
	      u3 = maxAtU_[i] ;
	      if(j==0)
		v3 = maxAtV_[j]+ nt*delta ;
	      else
		v3 = maxAtV_[j]- nt*delta ;

	      if(v3<V[0]) v3 = V[0] ;
	      if(v3>V[V.n-1]) v3 = V[V.n-1] ;

	      norm = normal(u1,v1);
	      norm += normal(u2,v2) ;
	      norm += normal(u3,v3) ;
	      norm /= 3 ; 
	    }
	    nt *= 10.0 ; 
	  }
	}
      }
      Point3D unit = norm.unitLength();
      unit *= d ;
      //CtrlPnt4D offset(unit ) ;
      //offset.w() = 0.0 ; 
      //p.modSurfCPby(i,j,offset) ;
      Du[no] = i ;
      Dv[no] = j ;
      offset[no] = unit ;
      ++no ;
    }

  p.movePoint(maxAtU_,maxAtV_,offset,Du,Dv) ;

  return p ;
}
