/*=============================================================================
        File: surface.cc
     Purpose:       
    Revision: $Id: surface.cc,v 2.3 1998/06/03 15:44:34 lavoie Exp $
  Created by: Philippe Lavoie          (3 Oct, 1996)
 Modified by: 

 Copyright notice:
          Copyright (C) 1996-1997 Philippe Lavoie
 
          This library is free software; you can redistribute it and/or
          modify it under the terms of the GNU Library General Public
          License as published by the Free Software Foundation; either
          version 2 of the License, or (at your option) any later version.
 
          This library is distributed in the hope that it will be useful,
          but WITHOUT ANY WARRANTY; without even the implied warranty of
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
          Library General Public License for more details.
 
          You should have received a copy of the GNU Library General Public
          License along with this library; if not, write to the Free
          Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
=============================================================================*/
#include <surface.hh>
#include MatrixLib(matrixMat.hh)

inline float abs2(float x){
  return x*x ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: minDist2 --- find the minimal distance between a point and 
                              the surface.
  Description: This is an iterative method to find the closest point to a 
               surface. 
        Input: p --> the minimal distance from that point
          guessU <-> a starting value for the parameter $u$, on exit this will 
	             be set to the value of the point on the surface closest 
		     to $p$.
          guessV <-> a starting value for the parameter $v$, on exit this 
	             will be set to the value of the point on the surface 
		     closest to $p$.
           error --> when iterations have an error smaller than this value,
	             the function exits
               s --> the size of the search in the parametric space.
             sep --> the number of points initially looked at to find a minimal 
	             distance.
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value for $u$
              uM --> the maximal parametric value for $u$
              vm --> the minimal parametric value for $v$
              vM --> the maximal parametric value for $v$
       Output: The value of the minimal distance between $p$ and the surface. 
               The variables guessU and guessV now holds the parametric value 
	       of the surface point closest to $p$. 
 Restrictions: It has not been tested with closed loop surfaces.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlParaSurface::minDist2(const Point3D& p, float& guessU, float& guessV,float error,float s,int sep,int maxIter, float um, float uM, float vm, float vM) const {
  float d,d1,d2 ;
  Point3D p2 ;
  p2 = point3D(guessU,guessV) ;
  d = abs2(p-p2) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float v1,v2 ;
  float step ;
  step = 2.0*s/(float)sep ;
  u1 = guessU-s ;
  u2 = guessU+s ;
  v1 = guessV-s ;
  v2 = guessV+s ;
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    if(v1<vm)
      v1 = vm ;
    if(v2>vM)
      v2 = vM ;
    float u,v ;
    d2 = d1 ;
    for(u=u1;u<u2;u+=step)
      for(v=v1;v<v2;v+=step){
	p2 = point3D(u,v) ;
	d1 = abs2(p-p2) ;
	if(d1<d){
	  d = d1 ;
	  guessU = u ;
	  guessV = v ;
	}
      }
    s /= 2.0 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    v1 = guessV - s ;
    v2 = guessV + s ;
    step = 2.0*s/(float)sep ;
    if(d-d2==0) niter = maxIter ;
    if(step<error) niter = maxIter ;
    ++niter;
  }
  return d ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  minDist2b --- find the minimal distance between a point and 
                              the surface.
  Description: This is an iterative method to find the closest point to a 
               surface. The method is slightly different than minDist2.

        Input: p --> the minimal distance from that point
          guessU <-> a starting value for the parameter $u$, on exit this
	             will be set to the value of the point on the surface 
		     closest to $p$.
 	  guessV <-> a starting value for the parameter $v$, on exit this 
	             will be set to the value of the point on the surface 
		     closest to $p$.
           error --> when iterations have an error smaller than this value, 
	             the function exits
               s --> the size of the search in the parametric space.
             sep --> the number of points initially looked at to find a 
	             minimal distance.
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value for $u$
              uM --> the maximal parametric value for $u$
              vm --> the minimal parametric value for $v$
              vM --> the maximal parametric value for $v$
       Output: The value of the minimal distance between $p$ and the surface. 
               The variables guessU and guessV now holds the parametric value 
	       of the surface point closest to $p$. 
 Restrictions: It has not been tested with closed loop surfaces.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlParaSurface::minDist2b(const Point3D& p, float& guessU, float& guessV,float error,float s,int sep,int maxIter, float um, float uM, float vm, float vM) const {
  float d,d1,d2 ;
  Point3D p2 ;
  p2 = point3D(guessU,guessV) ;
  d = abs2(p-p2) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float v1,v2 ;
  float step ;
  step = 2.0*s/(float)sep ;
  u1 = guessU-s ;
  u2 = guessU+s ;
  v1 = guessV-s ;
  v2 = guessV+s ;
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    if(v1<vm)
      v1 = vm ;
    if(v2>vM)
      v2 = vM ;
    float u,v ;
    d2 = d1 ;
    for(u=u1;u<=u2;u+=step)
      for(v=v1;v<=v2;v+=step){
	p2 = point3D(u,v) ;
	d1 = abs2(p-p2) ;
	if(d1<d){
	  d = d1 ;
	  guessU = u ;
	  guessV = v ;
	}
      }
    s = step ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    v1 = guessV - s ;
    v2 = guessV + s ;
    step = s/2.0 ; // *s/(float)sep ;
    if(d-d2==0) niter = maxIter ;
    if(step<error) niter = maxIter ;
    ++niter;
  }
  return d ;
  
}

inline float abs2xy(const Point3D &p){
  return (p.x*p.x)+(p.y*p.y) ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member:  minDist2xy --- find the minimal distance between a point and 
                               the surface in the x-y plane
  Description: This is an iterative method to find the closest point to a 
               surface. The distance is search in the x-y plane. The z 
               component is {\em not} taken into account for the search.
        Input: p --> the minimal distance from that point
          guessU <-> a starting value for the parameter $u$, on exit this 
	             will be set to the value
		     of the point on the surface closest to $p$.
          guessV <-> a starting value for the parameter $v$, on exit this
	             will be set to the value
		     of the point on the surface closest to $p$.
           error --> when iterations have an error smaller than this value, 
	             the function exits
              dU --> if a parametric delta is smaller than this value, 
	             the function stops.
               s --> the size of the search in the parametric space.
            sepU --> the number of points initially looked at to find a 
	             minimal distance in the $u$ direction
            sepV --> the number of points initially looked at to find a 
                     minimal distance in the $v$ direction
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value for $u$
              uM --> the maximal parametric value for $u$
              vm --> the minimal parametric value for $v$
              vM --> the maximal parametric value for $v$
       Output: The value of the minimal distance between $p$ and the surface. 
               The variables guessU and guessV now holds the parametric 
	       value of the surface point closest to $p$. 
 Restrictions: It has not been tested with closed loop surfaces.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlParaSurface::minDist2xy(const Point3D& p, float& guessU, float& guessV,float error,float dU,float s,int sepU,int sepV,int maxIter, float um, float uM, float vm, float vM) const {
  float d,d1,d2,dz ;
  Point3D p2 ;
  p2 = point3D(guessU,guessV) ;
  d = abs2xy(p-p2) ;
  dz = abs2(p.z-p2.z) ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float v1,v2 ;
  float stepU,stepV ;
  if(sepU>0){
    stepU = 2.0*s/(float)sepU ;
    u1 = guessU-s ;
    u2 = guessU+s ;
  }
  else{
    stepU = s ;
    u1 = guessU ;
    u2 = guessU ;
  }
  if(sepV>0){
    stepV = 2.0*s/(float)sepV ;
    v1 = guessV-s ;
    v2 = guessV+s ;
  }
  else{
    stepV = s ;
    v1 = guessV ;
    v2 = guessV ;
  }
  while(d>error && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    if(v1<vm)
      v1 = vm ;
    if(v2>vM)
      v2 = vM ;
    float u,v ;
    d2 = d1 ;
    for(u=u1;u<=u2;u+=stepU)
      for(v=v1;v<=v2;v+=stepV){
	p2 = point3D(u,v) ;
	d1 = abs2xy(p-p2) ;
	if(d1<d){
	  d = d1 ;
	  dz = abs2(p.z-p2.z) ;
	  guessU = u ;
	  guessV = v ;
	}
      }
    if(d-d2==0) niter = maxIter ;
    if(stepU<dU) niter = maxIter ;
    if(stepV<dU) niter = maxIter ;
    s = stepU*0.55 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    s = stepV*0.55 ;
    v1 = guessV - s ;
    v2 = guessV + s ;
    stepU *= 0.5 ;
    stepV *= 0.5 ;
    ++niter;
  }
  return dz ;
  
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: writeVRML --- write the NURBS surface to a VRML file
  Description: Writes a VRML file which represents the surface for the 
               parametric space $[uS,uE]$ and $[vS,vE]$.
               It does not optimize the number of points required to 
	       represent the surface. 
        Input: filename --> the file name for the output VRML file
                     Nu --> the number of points in the $u$ direction
		     Nv --> the number of points in the $v$ direction
		     uS --> the starting value of $u$
		     uE --> the end value of $u$
		     vS --> the starting value of $v$
		     vE --> the end value of $v$
       Output: 1 on success, 0 otherwise
 Restrictions: The parametric surface must be valid
   Created by: Philippe Lavoie          (24 January, 1997)
  Modified by:  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlParaSurface::writeVRML(const char* filename,const Color& color,int Nu,int Nv, float uS,float uE,float vS, float vE) const {
  ofstream fout(filename) ;

  if(!fout)
    return 0 ;
  
  fout << "#VRML V1.0 ascii\n" ;
  fout << "#  Generated by Phil's NURBS library\n" ;
  fout << "\nSeparator {\n" << "\tMaterialBinding { value PER_VERTEX_INDEXED }\n"  ;
  fout << "\tMaterial{ diffuseColor " << (int)color.r << " " << (int)color.g << " " << (int)color.b << "}\n" ;
  fout << "\tCoordinate3 {\n" ;
  fout << "\t\tpoint [\n" ;

  float u,v,du,dv ;

  if(Nu<=1)
    Nu = 2 ;  // Should I put a warning message ?
  if(Nv<=1)
    Nv = 2 ;  // Should I put a warning message ?

  u = uS ;
  v = vS ;
  du = (uE-uS)/(float)(Nu-1) ;
  dv = (vE-vS)/(float)(Nv-1) ;

  int i,j ;


  for(i=0;i<Nu;++i){
    v = vS ;
    for(j=0 ;j<Nv;++j){
      Point3D p ;
      p = point3D(u,v) ;
      fout << "\t\t\t" << p.x << " " << p.y << " " << p.z << ",\n" ;
      v += dv ;
    }
    u += du ;
  }
  
  fout << "\t\t]\n" ; // point [
  fout << "\t}\n" ; // cordinate3

  fout << "\tIndexedFaceSet{\n" ;
  fout << "\t\tcoordIndex[\n" ;
  
  for(i=0;i<Nu-1;++i){
    for(j=0;j<Nv-1;++j) {
      fout << "\t\t\t" << i*Nv+j << ", " << i*Nv+j+1 << ", " << (i+1)*Nv+j << ", -1,\n";
      fout << "\t\t\t" << i*Nv+j+1 << ", " << (i+1)*Nv+j+1<< ", " << (i+1)*Nv+j  << ", -1,\n";
    }
  }
  fout << "\t\t]\n" ;
  fout << "\t}\n" ; // IndexedFaceSet

  fout << "}\n" ;

  return 1 ;
}

inline float compare(int findMin, float a, float b){
  if(findMin)
    return minimum(a,b) ;
  return maximum(a,b) ;
}

inline float coordValue(PlParaSurface::CoordinateType coord, const Point3D& p){
  switch(coord){
  case PlParaSurface::coordX: return p.x ; break ;
  case PlParaSurface::coordY: return p.y ; break ;
  case PlParaSurface::coordZ: return p.z ; break ;
  }
  return 0.0 ; // elliminates warning messages
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: extremum --- finds the minimal or maximal value on the curve
                            of the x,y or z coordinate.
  Description: Finds the minimal or maximal value on the curve of the x,y or 
               z coordinate.
        Input: findMin --> a flag indicatinf if we're looking for the minimal
	                   value or the maximal value.
	   coord --> Which coordinate to find: x,y or z.
           minDu --> The minimal distance between iterations in the parametric
	             space.
            sepU --> the number of points initially looked at to find a minimal
	             distance in the U direction
            sepV --> the number of points initially looked at to find a minimal
	             distance in the U direction
         maxiter --> the maximal number of iterations
              um --> the minimal parametric value for $u$
              uM --> the maximal parametric value for $u$
              vm --> the minimal parametric value for $v$
              vM --> the maximal parametric value for $v$
       Output: The minimal value of $z$ along the curve
 Restrictions: It has not been tested with closed loop curves.
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
float PlParaSurface::extremum(int findMin, CoordinateType coord, float minDu, int sepU, int sepV, int maxIter, float um, float uM, float vm, float vM) const {
  float c,du,dv,d1,d2,result,guessU,guessV ;
  float minDv = minDu ;
  Point3D p2 ;

  guessU = 0 ;
  guessV = 0 ;

  // check for corner values...
  // because of the nature of += these value might be missed 
  // in the loop
  p2 = point3D(um,vm) ;
  c = coordValue(coord,p2) ;
  p2 = point3D(um,vM) ;
  c = compare(findMin,c,coordValue(coord,p2)) ;
  p2 = point3D(uM,vm) ;
  c = compare(findMin,c,coordValue(coord,p2)) ;
  p2 = point3D(uM,vM) ;
  c = compare(findMin,c,coordValue(coord,p2)) ; 
  result = c ;
  du = minDu*10.0 ;
  dv = minDv*10.0 ;
  d2 = d1 = 0 ;
  int niter = 0 ;
  float u1,u2 ;
  float v1,v2 ;
  float stepU,stepV ;
  float s ;
  s = uM - um ;
  stepU = s/(float)(sepU+1) ;
  stepV = s/(float)(sepV+1) ;
  u1 = um ;
  u2 = uM ;
  v1 = vm ;
  v2 = vM ;
  while((du>minDu || dv>minDv) && niter<maxIter) {
    if(u1<um)
      u1=um;
    if(u2>uM)
      u2 = uM ;
    if(v1<vm)
      v1 = vm ;
    if(v2>vM)
      v2 = vM ;
    float u,v ;
    d2 = c ;
    du = guessU ;
    dv = guessV ;
    for(u=u1;u<=u2;u+=stepU)
      for(v=v1;v<=v2;v+=stepV){
	p2 = point3D(u,v) ;
	if(findMin){
	  d1 = minimum(c,coordValue(coord,p2)) ;
	  if(d1<c){
	    c = d1 ;
	    guessU = u ;
	    guessV = v ;
	    result = d1 ;
	  }
	}
	else{
	  d1 = maximum(c,coordValue(coord,p2)) ;
	  if(d1>c){
	    c = d1 ;
	  guessU = u ;
	  guessV = v ;
	  result = d1 ;
	  }
	}
      }
    s /= 2.0 ;
    u1 = guessU - s ;
    u2 = guessU + s ;
    v1 = guessV - s ;
    v2 = guessV + s ;
    stepU = 2.0*s/(float)sepU ;
    stepV = 2.0*s/(float)sepV ;
    if((c-d2)==0.0) niter = maxIter ;
    if(stepU<minDu) niter = maxIter ;
    if(stepV<minDv) niter = maxIter ;
    du = absolute(guessU-du) ;
    dv = absolute(guessV-dv) ;
    niter++;
  }
  return result ;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
       Member: project --- projects a point onto the surface
  Description: Projects a point using newton-Raphson's method. We want to
               solve for
	       \[ X'_u(u_i,v_i)\cdot \delta u + X'_v(u_i,v_i)\cdot \delta v
	         = P - X(u_i,v_i) \]
	       This is an over-determined system and the least square 
	       approximation is taken:
	       \[ \left[ \begin{array}{cc} 
	       X'_u(u_i,v_i) \cdot X'_u(u_i,v_i) & 
	       X'_u(u_i,v_i) \cdot X'_v(u_i,v_i) \\
	       X'_v(u_i,v_i) \cdot X'_u(u_i,v_i) & 
	       X'_v(u_i,v_i) \cdot X'_v(u_i,v_i) \end{array} \right]
	       \left[ \begin{array}{c} \delta u \\ \delta v \end{array}
	       \right]
	        = 
		\left[ \begin{array}{c}
		(P-X(u_i,v_i)) \cdot X'_u(u_i,v_i) \\
		(P-X(u_i,v_i)) \cdot X'_v(u_i,v_i) 
		\end{array} \right]
		\]

		If $||X'_u|| = 0$ and $||X'_v|| = 0$, then the equation is
		singular. In that case the routine returns with 0. 
		Use one of the other iterative method to find a suitable 
		projection, if this happens.
		
        Input: P --> the point to project
 	       u <-> the u parametric value of the result
	       v <-> the v parametric value of the result
	    maxI --> the maximal number of iterations
              um --> the minimal parametric value for $u$
              uM --> the maximal parametric value for $u$
              vm --> the minimal parametric value for $v$
              vM --> the maximal parametric value for $v$
       Output: 1 on sucess, 0 if their was a singularity in computation
 Restrictions: 
   Created by: Philippe Lavoie (24 January 1997)
  Modified by:
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int PlParaSurface::projectOn(const Point3D& p, float& u, float& v, int maxI, const float um, const float uM, const float vm, const float vM) const {
  int i = 0 ;
  Point3D xu,xv,x,t ;
  Matrix_Point3D ders ;
  Matrix_DOUBLE A(6,2) ;
  Matrix_DOUBLE B(6,1) ;
  Matrix_DOUBLE X(2,1) ;



  while(i < maxI){
    // make sure u and v are in bounds
    boundTo(u,um,uM);
    boundTo(v,vm,vM);

    // compute the over determined solution to the equation
    deriveAt(u,v,1,ders) ;
    xu = ders(1,0) ;
    xv = ders(0,1) ;
    x = ders(0,0) ;
    
    t = xu*xu ;
    A(0,0) = t.x ;
    A(1,0) = t.y ;
    A(2,0) = t.z ;

    t = xu*xv ;
    A(0,1) = t.x ;
    A(1,1) = t.y ;
    A(2,1) = t.z ;

    t = xv*xu ;
    A(3,0) = t.x ;
    A(4,0) = t.y ;
    A(5,0) = t.z ;

    t = xv*xv ;
    A(3,1) = t.x ;
    A(4,1) = t.y ;
    A(5,1) = t.z ;

    t = (p-x)*xu ;
    B(0,0) = t.x ;
    B(1,0) = t.y ;
    B(2,0) = t.z ;

    t = (p-x)*xv ;
    B(3,0) = t.x ;
    B(4,0) = t.y ;
    B(5,0) = t.z ;
    

    SVDMatrix<double> svd(A) ;
    if(!svd.solve(B,X))  // the matrix is singular
      return 0 ;

    // X now contains the du and dv ;

    if(float(X(0,0)) == float(0) && (float)X(1,0) == float(0)){
      // we are hopefully done
      return 1 ;
    }
    
    u += (float)X(0,0) ;
    v += (float)X(1,0) ;
    ++i ;
  }

  return 0 ;
}
