/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: attribute.h,v 1.5 1998/07/23 20:15:57 source Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * basic attribute classes
 *
 */

#ifndef __ATTRIBUTE_H_
#define __ATTRIBUTE_H_

#include "3dlib/color.h"
#include "3dlib/g3d_math.h"
#include "3dlib/t_fields.h"
#include "3dlib/t_objects.h"


class G3DAReference;
class G3DAListIterator;
class G3DAttributeList;
class G3DShape;


/**This is the base class for all Attributes.
  There are 2 types of attributes. The first type includes attributes
  that are not object specific e.g.
     - texture
     - resolution
     - url
  The second type consists of attributes that can not be inherited, like
     - center
     - constraints

  A second classification is based on the number an object may have of
  the same kind of Attributes. There are two possibilities:
     - zero or one
     - any number

  A search for a specific attribute is started with getAttribute and
  can may be continued with getNextAttribute. The latter returns NULL if
  no more attributes of the specified type are available.
 */
class G3DAttribute
{
  friend class G3DAReference;
  friend class G3DAttributeList;
  friend class G3DAListIterator;

protected:
    ///
  G3DAttribute( void ) { next_ = suc_ = NULL; refCount_ = 1; };
  G3DAttribute( const G3DAttribute &orig )
    { next_ = suc_ = NULL; refCount_ = 1; type_ = orig.type_; }
  virtual ~G3DAttribute();

public:
    ///Delete the object if it is not referenced by other objects
  virtual void Dispose( void );
    ///
  virtual void activate( void ) const;
    ///
  void    addReference( void ) { ++refCount_; };
  virtual G3DAttribute* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const = 0;
    /**Store a list of parameter forming a complete description of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint  getParameter( Param* ) = 0;
  virtual const G3DAttribute* getReference( void ) const;
    ///
  t_type getType( void ) const { return type_; };
    ///Return a Descriptive name of the Attribute.
  virtual const char* getTypeName( void ) const = 0;
  virtual bool  isInheritable( void );
  virtual int   isMono( void ) = 0;
    ///Is attribute a reference to . 
  virtual bool  isReference( void ) const { return( false ); }
          bool  isReferenced( void ) const { return( refCount_ > 1 ); }
    ///Individualize the attribute.
  //void   setExist(void) { _ref = NULL; };
    ///The opposite of getParameter.
  virtual void  setParameter( const Param* );
    ///
  //void   setReference(const G3DAttribute *pRef) { _ref = pRef; };
    ///Set the user visible values of the attribute to the values of atr.
  virtual void  setValues( const G3DAttribute *atr ) = 0;

protected:
  t_type tnum() { return type_ & TA_MASK; }
  t_type type() { return type_ & (~TA_MASK); }

protected:
  G3DAttribute *next_, *suc_;
  t_type        type_;

private:
    ///Used for garbage Collection
  int   refCount_;

};


/**Attributes of this type can only appear once in each object.
   If you try to add another one it will replace the one already there.
   */
class G3DMonoAttribute : public G3DAttribute
{
public:
  G3DMonoAttribute( void ) {};
  G3DMonoAttribute( const G3DMonoAttribute &orig ) : G3DAttribute(orig) {}

  virtual int isMono( void ) { return 1; }
};


/**Attributes of this type can be assigned multiple times to an
object.*/
class G3DPolyAttribute : public G3DAttribute
{
public:
  G3DPolyAttribute( void ) {};
  G3DPolyAttribute( const G3DPolyAttribute &orig ) : G3DAttribute(orig) {}

  virtual int isMono( void ) { return 0; }
};


/**An Object of this class references another attribute. It is used
   to share attributes beween objects.
 */
class G3DAReference : public G3DAttribute
{
public:
//TODO make this private and provide a function that guarantees
// to return a pointer
    ///Create a reference to ref.
  G3DAReference( G3DAttribute *ref )
    {
      ref_ = ref;
      ref_->addReference();
      type_ = ref->type_;
    };

  virtual G3DAttribute* copy( void ) const;
    ///
  virtual void Dispose( void );
    ///
  virtual void activate( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint  getParameter( Param* );
    ///
  virtual const G3DAttribute* getReference( void ) const;
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  virtual int   isMono( void );
    ///Is attribute a reference to . 
  virtual bool  isReference( void ) const { return( true ); }
    ///Set the user visible values of the attribute to the values of atr.
  virtual void  setValues( const G3DAttribute *atr );

protected:
  G3DAttribute *ref_;
};


/**This attribute determens the number and/or size of the polygons
   in a tesselation of a 3d object.
   Quality values have to be in the range [0,1] with 1 being the
   highest quality (i.e. < 1 pixel deviation in the output).
   */
class G3DAResolution : public G3DMonoAttribute
{
public:
  G3DAResolution( void )
    { quality_ = 0.5; type_ = TA_RESOLUTION; }
  G3DAResolution( const G3DAResolution &orig ) : G3DMonoAttribute(orig)
    { quality_ = orig.quality_; } 

  virtual G3DAttribute* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint  getParameter( Param* );
    ///
  float getQuality( void ) const { return quality_; };
  static  const char* GetTypeName( void );
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
    ///
  void  setQuality( float q ) { quality_ = q; };
    ///Set the user visible values of the attribute to the values of atr.
  virtual void  setValues( const G3DAttribute *atr );

protected:
  float    quality_;
     ///1: make an equidistant mesh, even if not necessary
  //int      style_;
  static Field FieldList_[1];
};



/**Objects of this class are used to store an URL string.
 */
class G3DA_URL : public G3DMonoAttribute
{
public:
  G3DA_URL( void ) { url_ = NULL;  type_ = TA_URL; }
  G3DA_URL( const G3DA_URL &orig ) : G3DMonoAttribute(orig)
    { url_ = NULL; }
  ~G3DA_URL();

  virtual G3DAttribute* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint  getParameter( Param* );
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  const char* getURL( void ) const { return url_; };
  void        setURL( const char* );
    ///Set the user visible values of the attribute to the values of atr.
  virtual void setValues( const G3DAttribute *atr );

protected:
  char *url_;  //TODO use a KDE URL class object?
  static Field FieldList_[1];
};


#endif
