/*
 * Project : "MindsEye"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s):  F. Meier, M. Kuball
 * 
 * $Id: camera.cpp,v 1.17 1998/06/22 20:12:21 source Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "camera.h"


G3DCamera::G3DCamera( const G3DVector3d& pos, const G3DVector3d& lookAt,
		      const G3DVector3d& up )
{
  init();
  setUVN( pos, lookAt, up );
}

G3DCamera::G3DCamera() 
{
  init();
  setup();
}

G3DCamera::G3DCamera( const G3DCamera &orig )
{
  *this = orig;
}

G3DCamera::~G3DCamera()
{
  free( _name );
}


void G3DCamera::init()
{
  _name       = NULL;
  _eye        = G3DPoint3d( 0, 0, 6 );
  _focalPoint = G3DPoint3d( 0, 0, 0);
  _up         = G3DVector3d( 0, 1, 0 );
  _fov        = 40;
  _znear      = 0.1;
  _zfar       = 1000;
  _pmValid    = false;
  _tmValid    = false;
  _rollAngle  = 0;
  _width      = -1;  //TODO reasonable defaults?
  _height     = -1;

  _parallelScale   = 20;
  _windowCenter[0] = 0;
  _windowCenter[1] = 0;
  viewMode_        = Perspective;
}

void
G3DCamera::setup()
{
  _direction = _eye - _focalPoint;
  _distance = _direction.length();
  _direction.normalize();

  //project _up to be perpendicular with _direction vector
  float a = _up.dot( _direction );
  _up += (_direction * a) * (-1);
  assert( _up.length() != 0.0 );
  _up.normalize();

  _right = _up.cross( _direction ); // cross product

  invalidateTM();
}


G3DCamera&
G3DCamera::operator=( const G3DCamera &cam )
{
  _name = NULL;  //TODO
  _eye        = cam._eye;
  _focalPoint = cam._focalPoint;
  _up         = cam._up;
  _right      = cam._right;
  _direction  = cam._direction;
  _distance   = cam._distance;
  _fov        = cam._fov;
  _znear      = cam._znear;
  _zfar       = cam._zfar;
  _rollAngle  = cam._rollAngle; //TODO
  _width      = cam._width;
  _height     = cam._height;

  _parallelScale   = cam._parallelScale;
  viewMode_        = cam.viewMode_;
  _windowCenter[0] = cam._windowCenter[0];
  _windowCenter[1] = cam._windowCenter[1];

  _pmValid    = _tmValid = false;

  return *this;
}


////////////////////////////////////////
/*TODO
void  
G3DCamera::setDistance( double d )
{
  double minDist = 0.05;

  if (d < minDist) {
    assert(0);  // TODO remove this assertion? should never happen
    d = minDist;
  }

  _distance   = d;
  _focalPoint = (_direction * (-_distance)) + _eye;
  invalidateTM();
}
*/
void  
G3DCamera::setFOV( double f )
{
  double minFOV = 0, maxFOV = 179.9;  //TODO 0 as min value?
  
  if( f < minFOV )
    f = minFOV;
  else if( f > maxFOV )
    f = maxFOV;

  if( f != _fov )
    {
      _fov = f;
      invalidatePM();
    }
}

void  
G3DCamera::setLookAt( const TVector<double,3> &focalPnt )
{
  double tDistance, minDist = 0.05;

  G3DVector3d vt( focalPnt );
  tDistance = (vt - _eye).length();
  if( tDistance > minDist )
    {
      _focalPoint = focalPnt;
      setup();
    } 
}

void  
G3DCamera::setPosition( const TVector<double,3> &eye )
{
  double minDist   = 0.05;
  G3DVector3d tv( eye );
  double tDistance = ( tv - _focalPoint).length();
  if( tDistance > minDist )
    {
      _eye = eye;
      setup();
    }
}

void 
G3DCamera::setUVN( const TVector<double,3>& pos,
		   const TVector<double,3>& lookAt,
		   const TVector<double,3>& up )
{
  _eye        = pos;
  _focalPoint = lookAt;
  _up         = up;
  setup();
}
  
void  
G3DCamera::setZFar( double zf )
{ 
  _zfar = zf; 
  invalidatePM();
}

void  
G3DCamera::setZNear( double zn )
{ 
  _znear = zn;
  invalidatePM();
}



G3DVector3d  
G3DCamera::getDirection( void ) const
{
  if( isParallelMode() )
    return( G3DVector3d(0, 0, 0) ); //it's kind of 2-dimensional world
  else
    return G3DVector3d( _direction );
}

G3DVector3d
G3DCamera::getRight( void ) const
{
  switch( viewMode_ )
    {
    case Perspective:
      return G3DVector3d( _right );
    case Top:
    case Bottom:
    case Front:
    case Back:
      return( G3DVector3d(1.0, 0.0, 0.0) );
    case Left:
      return( G3DVector3d(0.0, 0.0, 1.0) );
    case Right:
      return( G3DVector3d(0.0, 0.0, -1.0) );
    }
}

G3DVector3d  
G3DCamera::getUp( void ) const
{
  switch( viewMode_ )
    {
    case Perspective:
      return G3DVector3d( _up );
      break;
    case Top:
      return( G3DVector3d(0.0, 0.0, 1.0) );
    case Bottom:
      return( G3DVector3d(0.0, 0.0, -1.0) );
    case Front:
    case Back:
    case Left:
    case Right:
      return( G3DVector3d(0.0, 1.0, 0.0) );
    }
}


////////////////////////////////////////


void 
G3DCamera::azimuth( double angle )
{
  G3DMatrix tMat, tR, tT;

  tT.translate( _focalPoint * (-1)  );
  // here: (0, 1, 0) is better than _up, if elevation & azimuth
  // are used together (*)
  tR.rotate( G3DVector3d(0, 1, 0), angle );
  tMat = tR * tT;
  tMat.translate( _focalPoint );

  _up    = tR * _up;  
  G3DPoint3d p = tMat * _eye;
//TODO rotate _up = tMat * _up ???
  setPosition( p );  

  // (*) but than _up must be rotated, too
  //TODO really? up is set in setup() called by setPosition()
  //tMat.axisRotate( angle, G3DVector3d(0, 1, 0) );
  //_up    = tR * _up;  
  //_right = _up.cross( _direction );
  //invalidateTM();
}

void 
G3DCamera::dolly( double f )
{
  G3DVector3d d( _direction * f );
  setPosition( _eye + d );
}

void 
G3DCamera::elevation( double angle )
{
  G3DMatrix tMat, tR, tTi;
  
  tMat.translate( _focalPoint );
  tR.rotate( _right, angle );  //TODO use (1,0,0) instead??
  tTi.translate( _focalPoint * (-1) );
  tMat *= tR;
  tMat *= tTi;

  G3DPoint3d p = tMat * _eye;
  setPosition( p );
}
     
void 
G3DCamera::pitch(double angle)
{
  G3DMatrix tMat, tR, tTi;
  
  tMat.translate( _eye * (-1) );
  tR.rotate( _right, angle );
  tTi.translate( _eye );
  tMat *= tR;
  tMat *= tTi;

  G3DPoint3d p = tMat * _focalPoint;
  setLookAt( p );
}

void 
G3DCamera::roll( double angle )
{
  if( angle == 0.0 )
    return; 

  _rollAngle += angle;

  G3DMatrix tMat;
  tMat.rotate( _direction, angle );
  _up    = tMat * _up;
  _right = _up.cross( _direction );

  invalidateTM(); 
}

void 
G3DCamera::tuck( double x, double y, double z )
{
  G3DVector3d delta((_right * x) + (_up * y) + (_direction * z));

  _eye        += delta;
  _focalPoint += delta;
  setup();
}

void 
G3DCamera::yaw( double angle )
{
  G3DMatrix tMat, tR, tTi;

  tMat.translate( _eye * (-1) );
  // here: (0,1,0) is better than _up, if yaw and pitch
  // are used together(*)
  tR.rotate( G3DVector3d(0, 1, 0), angle );
  tTi.translate( _eye );
  tMat *= tR;
  tMat *= tTi;

  _up = tMat * _up;
  G3DPoint3d p = tMat * _focalPoint;
  setLookAt( p );  
}


////////////////////////////////////////


// build ray through pixel(x,y)
void 
G3DCamera::buildRay( int x, int y, G3DRay& ray )
     // actually (x,y) are ints(mouse position)
{
  G3DVector3d tDir, tRight;
  // (1) find the vector from eye to viewplane center
  tDir   = _direction * (-_distance); //TODO not distance
  tRight = _right * (_width/(double)_height); // scale _right by aspect ratio

  // (2) add  the vector to the lower-left viewplane corner 
  tDir += tRight * -0.5;
  tDir += _up * -0.5;
  
  // (3) now calc vector from the lower-left corner to the pixel and add it
  tDir += tRight * (x/(double)_width) + _up*(y/(double)_height);

  // (2) and (3) can be composed into one statement, in order to save
  // some calculations
    
  tDir.normalize();

  ray.startpoint_ = _eye;
  ray.direction_  = tDir;
}

//TODO make this const by making the view matrix mutable
void
G3DCamera::unproject( int x, int y, G3DPoint3d &pnt )
{
  double dx, dy;  //device coordinates
  dx = (2.0*double(x)/_width) - 1.0;
  dy = 1.0 - (2.0*double(y)/_height);

  G3DMatrix m;
  getViewMatrix( m );
  pnt = m * pnt;

  if( isParallelMode() )
    {
      G3DMatrix p;
      getProjectionMatrix( p );
      p.invert4();
      G3DPoint3d win( dx, dy, 0 );
      win = p * win;
      pnt[0] = win[0];
      pnt[1] = win[1];
    }
  else
    {
      double persp = -1 * pnt[2] * tan( deg2rad(_fov)/2 );
      pnt[0] = dx * persp * _width / _height;
      pnt[1] = dy * persp;
    }

  m.invert4();
  pnt = m * pnt;
}

void
G3DCamera::setWindowSize( int w, int h )
{
  if( w != _width || h != _height )
    {
      _width  = w;
      _height = h;
      invalidatePM();
    }
}


void
G3DCamera::computeProjectionMatrix()
{
  GLenum oldMatrixMode;
  glGetIntegerv(GL_MATRIX_MODE, (GLint *)&oldMatrixMode);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  double tAspect = _width/(double)_height;

  if( isParallelMode() )
    {
      float w2, h2;

      w2 = _parallelScale / 2;
      h2 = w2 / tAspect;
      glOrtho( -w2 - _windowCenter[0], 
	       w2 - _windowCenter[0], 
	       -h2 - _windowCenter[1], 
	       h2 - _windowCenter[1],
	       -_zfar, _zfar);         //TODO -_zfar may be to much
    }
  else
    gluPerspective(_fov, tAspect, _znear, _zfar);

  double buf[16];
  glGetDoublev( GL_MODELVIEW_MATRIX, buf );
  _pm.setTo( buf );

  glPopMatrix();
  glMatrixMode( oldMatrixMode );

  _pmValid = true;
}

void
G3DCamera::computeParallelTMatrix()
{
  //_tm.setToIdentity(); ??? TODO
  switch (viewMode_) {
  case Top:
    _tm.rotateX(deg2rad(90));
    break;
  case Bottom:
    _tm.rotateX(deg2rad(-90));
    break;
  case Front:
    _tm.setToIdentity();
    break;
  case Back:
    _tm.rotateY(deg2rad(180));
    break;
  case Left:
    _tm.rotateY(deg2rad(90));
    break;
  case Right:
    _tm.rotateY(deg2rad(-90));
    break;
  }
}

void 
G3DCamera::computePerspectiveTMatrix()
{
  G3DMatrix   tVM, tMat;
  G3DVector3d tUp; 

  _right.normalize();
  tUp    = _direction.cross( _right );  //TODO ???

  tVM.setRow( 0, _right );
  tVM.setRow( 1, tUp );
  tVM.setRow( 2, _direction );

  tMat.translate( _eye * (-1) );
  _tm = tVM * tMat;  // * tPRPTranslate;
}

void 
G3DCamera::computeTMatrix()
{
  if( !_tmValid )
    {
      if( isParallelMode() )
	computeParallelTMatrix();
      else 
	computePerspectiveTMatrix();
      _tmValid = true;
    }
}

void 
G3DCamera::getProjectionMatrix( G3DMatrix& pm )
{
  if (!_pmValid) 
    computeProjectionMatrix();
  pm = _pm;
}

void 
G3DCamera::getViewMatrix(G3DMatrix& tm)
{
  if (!_tmValid)
    computeTMatrix();
  tm = _tm;
}

void
G3DCamera::render( void )
{
  G3DMatrix tM;  

  glViewport( 0, 0, _width, _height );
  glMatrixMode( GL_PROJECTION );
  getProjectionMatrix( tM );
  glLoadMatrixd( (const double*)tM );
  
  glMatrixMode(GL_MODELVIEW);
  getViewMatrix( tM );
  glLoadMatrixd( (const double*)tM );
}


////////////////////


void 
G3DCamera::setViewMode( TViewMode mode )
{
  if( mode != viewMode_ )
    {
      viewMode_ = mode;
      if( isParallelMode() )
	{
	  _windowCenter[0] = 0;
	  _windowCenter[1] = 0;
	}
      invalidatePM();
      invalidateTM();
   }
}
 
void 
G3DCamera::setParallelScale(double pScale)
{
  float minScale = 0.01;
  
  if( pScale < minScale )
    pScale = minScale;

  _parallelScale = pScale;
  if( isParallelMode() )
    invalidatePM();  
}

double 
G3DCamera::getParallelScale() const
{
  return _parallelScale;
}
 
void 
G3DCamera::setWindowCenter( double x, double y )
{
  _windowCenter[0]=x;
  _windowCenter[1]=y;
  if( isParallelMode() )
    invalidatePM();
}

void 
G3DCamera::getWindowCenter( double& x, double& y )
{
  x = _windowCenter[0];
  y = _windowCenter[1];
}


//==================================END=============================
