
/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: camera.h,v 1.12 1998/06/22 20:12:21 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __CAMERA2_H_
#define __CAMERA2_H_


#include <string.h>
#include "g3d_math.h"


//TODO what are common functions of all camera types?

/** An instance of this class represents a simple perspective camera.
    The physical model is that of a pin-hole camera. The eye point
    is the location of the pin-hole and the zoom value gives the
    distant from the pin-hole to the projection plane (film).

    The point of interest is given by _eyePoint + _lookAt
    */
class G3DCamera
{
public:
    ///
  G3DCamera();
    ///
  G3DCamera( const G3DVector3d& pos, const G3DVector3d& lookAt,
	     const G3DVector3d& up );
  G3DCamera( const G3DCamera& );
  virtual ~G3DCamera();

    /// calc ray through pixel(x,y)
  void   buildRay( int x, int y, G3DRay& ray );

  // ## get  All vectors returnd have length==1.0

  double       getAspectRatio( void ) const { return(_width/(double)_height);}
    ///Returns a 0-vector when in parallel mode!
  G3DVector3d  getDirection( void ) const;
    ///
  double       getDistance() const { return _distance; };
    ///
  double       getFOV() const { return _fov; }
    ///
  const G3DPoint3d& getLookAt( void ) const { return _focalPoint; }
    ///Get camera name.    
  char*        getName( void ) { return(_name); }; //TODO const?
    ///
  const G3DPoint3d& getPosition( void ) const { return _eye; }
    ///Compute and return the projection matrix.
  void         getProjectionMatrix( G3DMatrix& );
    ///
  G3DVector3d  getRight( void ) const;
    ///Compute and return the view matrix.
  void         getViewMatrix( G3DMatrix& );
    ///
  G3DVector3d  getUp( void ) const;
    ///
  double       getZFar() const { return _zfar; };
    ///
  double       getZNear() const { return _znear; };
    /**Calculate a point on the plane parallel to the viewing plane
       at distance pnt.z such that the ray from the cameras position
       to pnt goes through the point (x,y) on the view plane.
       The point is returned in pnt.*/
  void         unproject( int x, int y, G3DPoint3d &pnt );

  G3DCamera&   operator=( const G3DCamera& );

    ///Set the field of view angle given in degree.
  void  setFOV( double );
    ///Set distance
  //void  setDistance( double );
    ///
  void  setLookAt( const TVector<double,3>& );
    ///Set the camera name.
  void  setName( char *name ) { _name = strdup(name); };
    ///
  void  setPosition( const TVector<double,3>& );
    ///Set camera parameter
  void  setUVN( const TVector<double,3>&, const TVector<double,3>&,
		const TVector<double,3>& );
    ///This one is called from the GLWidget when its size changes
  void  setWindowSize( int, int );
    ///Set far clipping value.
  void  setZFar( double );
    ///Set near clipping value.
  void  setZNear( double );


    /**Rotate around up(*) vector at focal point, ie. rotate eye around
       focalpoint. Angle is in radians.*/
  void azimuth( double angle );
    ///Move eye point along viewplane normal. Focal point is fixed.
  void dolly(double);
    /**Rotate around right vector at focalpoint, ie. rotate eye
      around focalpoint. Angle is in radians.*/
  void elevation( double );
    ///Rotate _focalpoint around _right vector.
  void pitch( double );
    ///Set up projection and modelview matrices.
  void render( void );
    ///Rotate camera around viewplane normal
  void roll( double );
    ///Move camera along its axes (UVN).
  void tuck( double, double, double );
    /**Rotate _focalpoint around _up vector, NO! we use (0,1,0) instead,
       this has the advantage that the xz-plane always stays aligned to
       the viewport.*/
  void yaw( double );

    /**Pass one of the following into setParallelView(..).
       y-axis is up.*/
  enum TViewMode { Perspective = 0, Top, Bottom, Front, Back, Left, Right };

    /// get parallel zoom factor
  double    getParallelScale() const;
    ///
  TViewMode getViewMode() const { return viewMode_; }
    ///
  void      getWindowCenter(double&,double&);
    /// is camera in parallel mode?
  bool      isParallelMode() const { return viewMode_ != Perspective; }
  void      setViewMode( TViewMode );
    /// set parallel zoom factor
  void      setParallelScale(double);
    ///
  void setWindowCenter(double,double);


protected:
    ///
  void computeProjectionMatrix();
    ///
  void computeParallelTMatrix();
    ///
  void computePerspectiveTMatrix();
    /// first matrix on MODEL_VIEW stack, camera transformation
  void computeTMatrix();
    ///
  const G3DMatrix& getPM();
    ///
  const G3DMatrix& getTM();
    ///Initialize camera with default values.
  void init();
    ///Invalidate the projection matrix.
  void invalidatePM() { _pmValid = false; };
    ///Invalidate the transformation matrix.
  void invalidateTM() { _tmValid = false; };
    ///Compute internal variables.
  void setup();


protected:
    ///Name of the camera.
  char        *_name;
    ///The eye point, you guessed it!
  G3DPoint3d _eye;
    ///Center of viewplane.  _focalPoint = _eye + _direction * _distance
  G3DPoint3d _focalPoint;
    /// uvn system used to build _tm
  G3DVector3d _up, _right, _direction;
    ///Distance between _eyePoint <-> _focalPoint
  double _distance;
    ///field of view in x direction???
  double _fov;
    /// near/far clipping planes
  double _znear, _zfar;

    ///Projection matrix, camera transformation matrix.
  G3DMatrix _pm, _tm;
    ///If _pm is requested and _pmValid==false, then _pm is recalculated.
  bool      _pmValid, _tmValid;
    ///
  double    _rollAngle;
    ///Extensions of the viewport window.
  int       _width, _height;

  TViewMode  viewMode_;
    /// parallel mode zoom factor
  float _parallelScale;
    /// center of 2d window
  float _windowCenter[2];
    /// type of parallel view, see above enum{}
};


// other possible cameras are:
/*
    orthograpic projection
    camera with lenses simulation
    */

#endif

