/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: iterator.h,v 1.13 1998/07/23 20:15:47 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __ITERATOR_H_
#define __ITERATOR_H_


#include <GL/gl.h>
#include "g3d_math.h"
#include "actions.h"

/**Derived types of this class may used to iterate over elements in
   aggregates. They also may have a state. However, be carefully with
   the usage of state-based iterators.
   */
template<class T>
class GIterator
{
public:
  /**Abstract iterator function. Must return true to stop iteration for
     remaining elements.*/
  virtual bool exec(T&) = 0;
};


/**An iterator that calls the shading methods of primitives or light sources.
   To be used in shader or shadable (sub)class(es).
*/
template<class T>
class GCompileListIterator: public GIterator<T>
{
public:
  GCompileListIterator( int lnum ) : num_(lnum) { ; };

  virtual bool exec( T& t ) {
    t.compileList( num_ );
    return(false);
  }
protected:
  int num_;
};


/**An iterator that calls the shading methods of primitives or light sources.
   To be used in shader or shadable (sub)class(es).
*/
template<class T>
class GLightingIterator: public GIterator<T>
{
public:
  GLightingIterator( int num ) : num_(num) { ; } ;
  virtual bool exec( T& t ) {
    t.enable( num_ );
    return(false);
  }
  int getLNum( void ) { return num_; };

protected:
  int num_;
};


/**An iterator that calls the shading methods of primitives or light sources.
   To be used in shader or shadable (sub)class(es).
*/
template<class T>
class GSelectionIterator: public GIterator<T>
{
public:
  GSelectionIterator( void ) : num_(0) { ; };

  virtual bool exec( T& t ) {
    glLoadName( num_++ );
    t.select();
    return(false);
  }
protected:
  int num_;
};


/**An iterator that calls the shading methods of primitives or light sources.
   To be used in shader or shadable (sub)class(es).
*/
template<class T>
class GShadingIterator: public GIterator<T>
{
public:
  GShadingIterator( int lnum ): lnum_(lnum) { ; };

  virtual bool exec( T& t ) {
    t.display( lnum_ ); 
    return(false);
  }
protected:
  int lnum_;
};


/**Send the given action to the object.
 */
template<class T>
class SendActionIterator : public GIterator<T>
{
public:
  SendActionIterator( ActionData &action ) : action_(action) {; }

  virtual bool exec( T& t ) {
    t.doAction( action_ );
    return(false);
  }
protected:
  ActionData &action_;
};


/**
An iterator that calls the intersect methods of primitives or light sources.
*/
/*
template<class T>
class GIntersectIterator: public GIterator<T>
{
public:
  GIntersectIterator( G3DRay *ray ) {
    _ray      = ray;
    _object   = NULL;
    _distance = G3D_INFINITY;  //TODO better initialisation?
  };

  virtual bool exec( T& t ) {
    T *obj;
    double dist = t.intersect( _ray, &obj );
    if( dist < _distance )
      {
	_distance = dist;
	_object   = obj;
      }
    return(false);
  }

  double getDistance( void ) { return( _distance ); };
  T*     getObject( void ) { return( _object ); };

protected:
  T       *_object;
  G3DRay  *_ray;
  double  _distance;
};
*/


#endif
