/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: light.cpp,v 1.10 1998/07/23 20:15:48 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * Implementation of light sources
 *
 */


#include <math.h>
#include <GL/gl.h>

#include "light.h"
#include "w_light.h"

const float cmax = 1.0;


//This function is not inline to force the compiler to output
//the virtual table for this class.
G3DLight::G3DLight( void ) :
  color_(cmax,cmax,cmax)
{
  isOn_   = true;
  wflag_  = false;
  widget_ = NULL;
}

G3DLight::G3DLight( const G3DLight &orig ) :
  G3DObject(orig), color_(orig.color_)
{
  isOn_   = orig.isOn_;
  wflag_  = false; //TODO orig.wflag_;
  widget_ = NULL;
}

uint
G3DLight::getParameter( Param *pList )
{
  if( pList )
    { //TODO isOn_ ??
      pList->color = &color_;
      pList++;
      pList->flt   = intensity_;
      pList++;
      pList->logic = castShadows_;
    }
  return( 3 );
}

void
G3DLight::select( void )
{
}

void
G3DLight::setParameter( const Param *pList )
{
  if( pList )
    {
      pList++;
      intensity_ = pList->flt;
      pList++;
      castShadows_ = pList->logic;
    }
}


//-------------------------------------------------------------------

G3DAmbientLight::G3DAmbientLight()
{
  const float  cval = 0.15;
  type_  = T_AMBIENT_LIGHT;
  //G3DColor col(cval, cval, cval);
  color_ = G3DColor(cval, cval, cval); //col;
}

G3DAmbientLight::G3DAmbientLight( const G3DColor &col )
{
  type_ = T_AMBIENT_LIGHT;
  setColor(col);
}

void
G3DAmbientLight::enable( int& )
{
  if( isOn_ )
    {
      GLfloat tmpcol[4] = {0.0, 0.0, 0.0, 1.0};

      glLightfv( GL_LIGHT0, GL_DIFFUSE,  tmpcol );
      glLightfv( GL_LIGHT0, GL_SPECULAR, tmpcol );
      glLightfv( GL_LIGHT0, GL_AMBIENT,  color_ );
      glEnable( GL_LIGHT0 );
    }
}

const char*
G3DAmbientLight::GetTypeName( void )
{
  return( "Ambient_Light" );
}

const char*
G3DAmbientLight::getTypeName( void ) const
{
  return( G3DAmbientLight::GetTypeName() );
}


//=======================================================================


G3DDirectionalLight::G3DDirectionalLight( const G3DDirectionalLight &orig ) :
  G3DLight(orig), direction_(orig.direction_)
{
}

void
G3DDirectionalLight::enable( int &num )
{
//TODO if( num > lightMax_ ) report error return true??
  //if( widget_ )
   //TODO widet->shade();
//TODO maybe call a special function like shadeDirectionalLight( direction_);
  if( isOn_ )
    {
      GLfloat tmpF4[4];
      GLenum  tEnum = (GLenum)( GL_LIGHT0 + num );

      tmpF4[0] = direction_[0];
      tmpF4[1] = direction_[1];
      tmpF4[2] = direction_[2];
      tmpF4[3] = 0;   //This makes it a directional light

      glLightfv( tEnum, GL_DIFFUSE, (const float*)color_ );
      glLightfv( tEnum, GL_POSITION, tmpF4 );
      glEnable( tEnum );
      ++num;
    }
}

uint
G3DDirectionalLight::getParameter( Param *pList )
{
  uint num = G3DLight::getParameter( pList );
  if( pList )
    {
      pList += num;
      pList->vector3 = &direction_;
    }
  return( num + 1 );
}

const char*
G3DDirectionalLight::GetTypeName( void )
{
  return( "Directional_Light" );
}

const char*
G3DDirectionalLight::getTypeName( void ) const
{
  return( G3DDirectionalLight::GetTypeName() );
}


//=======================================================================


G3DPointLight::G3DPointLight( void )
{
  type_ = T_POINT_LIGHT;
  attenuation_[0] = 1;
  attenuation_[1] = 0;
  attenuation_[2] = 0;
  setOrigin( G3DPoint3d(0,5,5) );
}


G3DPointLight::G3DPointLight( const G3DPointLight &orig ) :
  G3DLight(orig)
{
  attenuation_[0] = orig.attenuation_[0];
  attenuation_[1] = orig.attenuation_[1];
  attenuation_[2] = orig.attenuation_[2];
  //TODO if( orig.widget_ ) widget_ = new G3DPointLightWidget( orig.widget_ );
}

G3DObject*
G3DPointLight::copy( void ) const
{
  return( new G3DPointLight(*this) );
}

void
G3DPointLight::doAction( ActionData &action )
{
  switch( action.type )
    {
    case A_ADD_SHAPE:
      wflag_ = true;
    case A_SELECT:
      if( widget_ == NULL )
	widget_ = new G3DPointLightWidget( this );
      break;
    case A_DESELECT:
      if( wflag_ )
	break;
    case A_REM_SHAPE:
      delete widget_;
      widget_ = NULL;
      wflag_  = false;
      break;
    case A_FIND_SELECTION:
      if( widget_ )
	widget_->doAction( action );
      break;
      //TODOcase A_RESET_WIDGETS:
    }
  G3DObject::doAction( action );
}


void
G3DPointLight::enable( int &num )
{
  glPushMatrix();
  glMultMatrixd( (const double*)trafo_ );

  if( widget_ )
    ((G3DPointLightWidget*)widget_)->display();
  if( isOn_ )
    {
      GLfloat tmpF4[4];
      GLenum  tEnum = (GLenum)(GL_LIGHT0 + num);

      tmpF4[0] = 0;
      tmpF4[1] = 0;
      tmpF4[2] = 0;
      tmpF4[3] = 1.0;   //This makes it a point light

      glLightfv( tEnum, GL_DIFFUSE,  (const float*)color_ );
      glLightfv( tEnum, GL_SPECULAR, (const float*)color_ );
      glLightfv( tEnum, GL_POSITION, tmpF4 );
      glEnable( tEnum );

      glLightf( tEnum, GL_CONSTANT_ATTENUATION,  attenuation_[0] ); 
      glLightf( tEnum, GL_LINEAR_ATTENUATION,    attenuation_[1] ); 
      glLightf( tEnum, GL_QUADRATIC_ATTENUATION, attenuation_[2] );

      glEnable( tEnum );
      ++num;
    }
  glPopMatrix();
}

Field G3DPointLight::FieldList_[NUM_LIGHT_PARAMS+3] =
{ {"color", s_color}, {"intensity", s_flt}, {"shadows", s_logic},
 {"att_1", s_flt}, {"att_2", s_flt}, {"att_3", s_flt}
};

const Field*
G3DPointLight::getFieldList( void ) const
{
  return( FieldList_ );
}

uint
G3DPointLight::getParameter( Param *pList )
{
  uint num = G3DLight::getParameter( pList );
  if( pList )
    {
      pList += num;
      pList->flt   = attenuation_[0];
      pList++;
      pList->flt   = attenuation_[1];
      pList++;
      pList->flt   = attenuation_[2];
    }
  return( num + 3 );
}

const char*
G3DPointLight::GetTypeName( void )
{
  return( "Point_Light" );
}

const char*
G3DPointLight::getTypeName( void ) const
{
  return( G3DPointLight::GetTypeName() );
}

void
G3DPointLight::select( void )
{
  if( widget_ )
    {
      glPushMatrix();
      glMultMatrixd( (const double*)trafo_ );
      ((G3DPointLightWidget*)widget_)->select();
      glPopMatrix();
    }
}

void
G3DPointLight::setParameter( const Param *pList )
{
  if( pList )
    {
      G3DLight::setParameter( pList );
      pList += NUM_LIGHT_PARAMS;
      attenuation_[0] = pList->flt;
      pList++;
      attenuation_[1] = pList->flt;
      pList++;
      attenuation_[2] = pList->flt;
    }
}


//=======================================================================


G3DSpotLight::G3DSpotLight( void ) :
  position_( 0, 0, 1), direction_( 0, 0, -1)
{
  type_           = T_SPOT_LIGHT;
  angle_          = 45;
  attenuation_[0] = 1;
  attenuation_[1] = 0;
  attenuation_[2] = 0;
}


G3DSpotLight::G3DSpotLight( const G3DSpotLight &orig ) :
  G3DLight(orig), position_(orig.position_),direction_(orig.direction_)
{
  angle_          = orig.angle_;
  attenuation_[0] = orig.attenuation_[0];
  attenuation_[1] = orig.attenuation_[1];
  attenuation_[2] = orig.attenuation_[2];
}

void
G3DSpotLight::enable( int &num )
{
  if( isOn_ )
    {
      GLfloat tmpF4[4];
      GLenum  tEnum = (GLenum)(GL_LIGHT0 + num);

      tmpF4[0] = position_[0];
      tmpF4[1] = position_[1];
      tmpF4[2] = position_[2];
      tmpF4[3] = 1.0;   //This makes it a point light

      //GL_AMBIENT is 0 for tEnum > 0
      glLightfv( tEnum, GL_DIFFUSE,  (const float*)color_ );
      glLightfv( tEnum, GL_SPECULAR, (const float*)color_ );

      glLightfv( tEnum, GL_POSITION, tmpF4 );
      tmpF4[0] = direction_[0];
      tmpF4[1] = direction_[1];
      tmpF4[2] = direction_[2];
      glLightfv( tEnum, GL_SPOT_DIRECTION, tmpF4 );

      glLightf( tEnum, GL_SPOT_EXPONENT, 0 );
      glLightf( tEnum, GL_SPOT_CUTOFF, angle_ );
      glLightf( tEnum, GL_CONSTANT_ATTENUATION,  attenuation_[0] );
      glLightf( tEnum, GL_LINEAR_ATTENUATION,    attenuation_[1] );
      glLightf( tEnum, GL_QUADRATIC_ATTENUATION, attenuation_[2] );

      glEnable( tEnum );
      ++num;
    }
}

uint
G3DSpotLight::getParameter( Param *pList )
{
  uint num = G3DLight::getParameter( pList );
  if( pList )
    {
      pList += num;
      pList->vector3 = &direction_;
      pList++;
      pList->point3  = &position_;
      pList++;
      pList->flt     = attenuation_[0];
      pList++;
      pList->flt     = attenuation_[1];
      pList++;
      pList->flt     = attenuation_[2];
      pList++;
      pList->flt     = angle_;
    }
  return( num + 6 );
}

const char*
G3DSpotLight::GetTypeName( void )
{
  return( "Spot_Light" );
}

const char*
G3DSpotLight::getTypeName( void ) const
{
  return( G3DSpotLight::GetTypeName() );
}


//==================================END================================
