/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997, 1998. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: light.h,v 1.10 1998/07/23 20:15:48 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __LIGHT_H_
#define __LIGHT_H_


#include "object.h"
#include "g3d_math.h"
#include "color.h"

#define LIGHT_MAX  8  //TODO get this number from OpenGL at run time.
#define NUM_LIGHT_PARAMS 3


/**Rendering-independent(?) base class of light sources.
   The intensity of the light is measured in SI units: W m^-2 sr^-1.
   It is a factor to be multiplied with the color.
   Lights can be turned on and off.
   */
class G3DLight : public G3DObject
{
protected:
  ///Default is a white light source. TODO what intensity?
  G3DLight( void );
  G3DLight( const G3DLight& );

public:
    /**Setup the parameter for shading calculations.*/
  virtual void enable( int& ) = 0;

    ///
  const G3DColor& getColor( void ) const { return( color_ ); }
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint getParameter( Param* );
    ///
  bool       isOn( void ) const { return isOn_; };
    ///
  virtual void select( void );
    ///
  void       setColor( const G3DColor &col ) { color_ = col; };
    ///
  void       setOff( void )  { isOn_ = false; };
    ///
  void       setOn( void )   { isOn_ = true; };
  virtual void  setParameter( const Param* );

//TODO static initClass();

protected:
  G3DColor color_;
  G3DObject *widget_;

    ///In W m^-2 sr^-1. Default is 1.0.
  float    intensity_;
  bool     isOn_, wflag_;
  bool     castShadows_;
};


/**Ambient light specifies the base amount of light that is added to the
   illumination of all surfaces in a scene. Ambient light has no apparent
   source or location; its intensity is constant, and it does not cast
   shadows. The default ambient intensity = (0.15, 0.15, 0.15, 1.0).

   The OpenGL shader always uses LIGHT0 to set the ambient light intensity.
   All ambient light values are absolute.
   */
class G3DAmbientLight : public G3DLight
{
public:
    ///
  G3DAmbientLight();
    ///
  G3DAmbientLight( const G3DColor &col );
    ///
  G3DAmbientLight( const G3DAmbientLight &orig ) : G3DLight(orig) {};

  virtual void enable( int& );
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
};


/** A directional light is a light that emits parallel rays in a specific
    direction. A directional light may be set to cast shadows.

    The visualisation of the light is an arrow rooted at the center of
    the locale coordinate system and pointing in the direction of
    direction_.
    */
class G3DDirectionalLight : public G3DLight
{
public:
    ///
  G3DDirectionalLight( void ) : direction_(0.0, 0.0, -1.0)
  { type_ = T_DIRECTIONAL_LIGHT; };
    ///
  G3DDirectionalLight( const G3DDirectionalLight& );

  virtual void enable( int& );
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint getParameter( Param* );
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
  void setDirection( const TVector<double,3> &tv ) { direction_ = tv; };

protected:
  G3DVector3d direction_;
};


/** A point light is a light that emits rays in all directions from a
    specific point source. A point light may suffer attenuation over
    distance, and may cast shadows (default is to cast).
    */
class G3DPointLight : public G3DLight
{
public:
    ///
  G3DPointLight( void );
    ///
  G3DPointLight( const G3DPointLight& );

    ///
  virtual G3DObject* copy( void ) const;
    ///
  virtual void doAction( ActionData& );
    ///
  virtual void enable( int& );
  virtual const Field* getFieldList( void ) const;
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint getParameter( Param* );
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
    ///
  virtual void select( void );
  virtual void setParameter( const Param* );
    ///
  //virtual void setPosition( const G3DPoint3d &pos ) { position_ = pos; };

protected:
  float        attenuation_[3];
  static Field FieldList_[NUM_LIGHT_PARAMS+3];
};


/**
  A Spot light
 */
class G3DSpotLight : public G3DLight
{
public:
  G3DSpotLight( void );
  G3DSpotLight( const G3DSpotLight& );

  virtual void  enable( int& );
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint  getParameter( Param* );
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );

protected:
  G3DPoint3d  position_;
  G3DVector3d direction_;
  float       angle_;
  float       attenuation_[3];

};

//TODO G3DAreaLight


#endif
