/*
 * Project : "MindsEye, math lib"
 * Copyright (c) 1996, 1997, 1998. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: g3d_math.cpp,v 1.9 1998/07/23 20:15:57 source Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * OVERVIEW
 * ========
 * 3 dimensional Vector based on  the "temporary base class idiom".
 * See the article in the german magazine iX issue 11/96 p166.
 *
 */

#include "include/config.h"
#include <stdlib.h>
#include "tvector.h"
#include "g3dvector.h"
#include "3dlib/g3d_math.h"

double meround( double d );

double meround( double d )
{
  double fl, ce, t1, t2;

  fl = floor( d );
  ce = ceil( d );
  t1 = d - fl;
  t2 = ce - d;
  if( t1 <= t2 )
    return( fl );
  else
    return( ce );
}



template <class Tp, int N>
TVector<Tp,N>& TVector<Tp,N>::cwDivide( const TVector<Tp,N> &tv )
{
  const Tp *src = (const Tp*)tv;
        Tp *dst = _data;

  for( int i = N; i > 0; i-- )
    *(dst++) /= *(src++);
  return( *this );
}


template <class Tp, int N>
TVector<Tp,N>& TVector<Tp,N>::cwMultiply( const TVector<Tp,N> &tv )
{
  const Tp *src = (const Tp*)tv;
        Tp *dst = _data;

  for( int i = N; i > 0; i-- )
    *(dst++) *= *(src++);
  return( *this );
}

template <class T, int N>
T TVector<T,N>::dot( const TVector<T,N> &tv ) const
{
  const T *src1  = _data;
  const T *src2  = tv._data;
        T result = 0;

  for( int i = N; i > 0; i-- )
    result += *(src1++) * *(src2++);

  return( result );
}


template <class T, int N>
bool TVector<T,N>::isZero( void ) const
{
  const T *src = _data;

  for( int i = N; i > 0; i-- )
    if( fabs(*src) > EPSILON )
      return( false );
    else
      src++;
  return( true );
}

template <class Tp, int N>
Tp TVector<Tp,N>::length( void ) const
{
  const Tp* src = _data;
        Tp  res = 0;

  for( int i = N; i > 0; i-- )
    {
      res += *src * *src;
      src++;
    }
  if( res > FG_EPSILON )
    return( sqrt(res) );
  else
    return( 0 );
}

template <class T, int N>
void
TVector<T,N>::max( const TVector<T,N> &tv )
{
  const T* src = (const T*)tv;
        T* dst = _data;

  for( int i = N; i > 0; i-- )
    {
      if( *dst < *src )
        *dst = *src;
      src++;  dst++;
    }
}

template <class T, int N>
void
TVector<T,N>::min( const TVector<T,N> &tv )
{
  const T* src = (const T*)tv;
        T* dst = _data;

  for( int i = N; i > 0; i-- )
    {
      if( *dst > *src )
        *dst = *src;
      src++;  dst++;
    }
}

template <class T, int N>
TVector<T,N>&
TVector<T,N>::normalize()
{
  T* src = _data;
  T  norm = 0;

  for( int i = N; i > 0; i-- )
    {
      norm += *src * *src;
      src++;
    }

  if( norm > FG_EPSILON )
    {
      norm = sqrt(norm);
      T* src2 = _data;
      for( int j = N; j > 0; j-- )
	*(src2++) /= norm;
    }
  return( *this );
}

template <class Tp, int N>
TVector<Tp,N>
TVector<Tp,N>::random( void )
{
  TVector<Tp,3> tv;
  Tp* dst = tv.data();
  
  for( int i = N; i > 0; i-- )
    {
      *(dst++) = ((double)rand()) / RAND_MAX;
    }
  return( tv );
}

template <class Tp, int N>
TVector<Tp,N>&
TVector<Tp,N>::round( void )
{
  Tp *dst = _data;

  for( int i = N; i > 0; i-- )
    {
      *dst = meround( *dst );
      dst++;
    }

  return( *this );
}

template <class Tp, int N>
TVector<Tp,N>&
TVector<Tp,N>::round( const TVector<Tp,N> &spacing )
{
  const Tp *src = (const Tp*)spacing;
        Tp *dst = _data, tmp;

  for( int i = N; i > 0; i-- )
    {
      if( *src != 0 )
	{
	  tmp = *dst / *src;
	  tmp = meround( tmp );
	  *(dst++) = tmp * *(src++);
	}
    }

  return( *this );
}

template <class T, int N>
ostream& operator<<( ostream& os, const TVector<T,N> &tv )
{
  const T *src = tv._data;

  for( int i = 0; i < N; i++ )
    {
      os << *(src++);
      if( i != (N-1) )
	os << IOSeparator;
    }
  return( os );
}

template <class T, int N>
istream& operator>>( istream& os, TVector<T,N> &tv )
{
  T *dst = tv._data;

  for( int i = 0; i < N; i++ )
    os >> *(dst++);
  return( os );
}


//====================================================================


// TODO template <class T>
void minmax( G3DPoint3<double> &vmin, G3DPoint3<double> &vmax,
	     const G3DVector3<double> &e1, const G3DVector3<double> &e2,
	     const G3DVector3<double> &e3 )
{
  G3DPoint3d orig(vmin), pt;

  pt   = orig + e1;
  vmin.min( pt );
  vmax.max( pt );
  pt  += e2;
  vmin.min( pt );
  vmax.max( pt );
  pt  += e3;
  vmin.min( pt );
  vmax.max( pt );
  pt   = orig + e2;
  vmin.min( pt );
  vmax.max( pt );
  pt  += e3;
  vmin.min( pt );
  vmax.max( pt );
  pt   = orig + e3;
  vmin.min( pt );
  vmax.max( pt );
  pt  += e1;
  vmin.min( pt );
  vmax.max( pt );
}


//====================================================================


template <class T>
TVector<T,3> G3DVector3<T>::cross( const G3DVector3 &v ) const
{
  TVector<T,3> tv;
        T* dest = tv.data();
  const T* dat1 = _data;
  const T* dat2 = (const T*)v;

  *(dest++) = dat1[1]*dat2[2] - dat1[2]*dat2[1];
  *(dest++) = dat1[2]*dat2[0] - dat1[0]*dat2[2];
  *dest     = dat1[0]*dat2[1] - dat1[1]*dat2[0];
  return( tv );  //TODO 
}


template <class T>
TVector<T,3>  G3DVector3<T>::getOrtho( void ) const
{
  TVector<T,3> tv;
  T *dest = tv.data();

  if( (_data[0] == 0) && (_data[1] == 0) )
    {
      *(dest++) = 1;
      *(dest++) = 0;
      *(dest++) = 0;
    }
  else
    {
      *(dest++) = _data[1];
      *(dest++) = -_data[0];
      *(dest++) = 0.0;
    }

  return( tv );
}


template<class T>
void
G3DVector3<T>::rotateY( double alpha )
{
  double cos_val, sin_val;
  T      t0, t2;

  cos_val = cos( alpha );
  sin_val = sin( alpha );

  t0 = cos_val * _data[0] + sin_val * _data[2];
  t2 = cos_val * _data[2] - sin_val * _data[0];
  _data[0] = t0;
  _data[2] = t2;
}


template<class T>
void
G3DVector3<T>::rotateX( double alpha )
{
  double cos_val, sin_val;
  T      t1, t2;

  cos_val = cos( alpha );
  sin_val = sin( alpha );

  t1 = cos_val * _data[1] - sin_val * _data[2];
  t2 =  sin_val * _data[1] + cos_val * _data[2];
  _data[1] = t1;
  _data[2] = t2;
}


//--------------------------------------------------------------

// explicitly instantiate the template classes with the
// types we want to use.

#ifdef EXPLICIT_TEMPLATES
#ifdef __GNUG__

template class TVector<double,3>;
template class G3DVector3<double>;
template class G3DPoint3<double>;
template istream& operator>>(istream&, TVector<double,3>&);
template ostream& operator<<(ostream&, const TVector<double,3>&);
//template void minmax( G3DVector3<double>&, G3DVector3<double>& );

#else

void  Vector_dumbSPARC()
{
  TVector<double> tv;
  G3DVector3<double> v;
  G3DPoint3<double> p;

  cout << tv;
}

#endif
#endif  //EXPLICIT_TEMPLATES


//===================================END================================
