/*
 * Project : "MindsEye, math lib"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: g3dvector.h,v 1.7 1998/07/23 20:15:58 source Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * 3 dimensional Vector based on  the "temporary base class idiom".
 * See the article in the german magazine iX issue 11/96 p166.
 *
 */


#ifndef __G3D_VECTOR_H_
#define __G3D_VECTOR_H_


#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <math.h>

#include "tvector.h"



/**A 3 dimensional vector. Although derived from class TVector, instances
   of this class store their data in a member variable. This allows for
   better memory efficiency of arrays of vectors.
 */
template <class T>
class G3DVector3 : public TVector<T, 3>
{
public:
    ///
  G3DVector3( void ) : TVector<T,3>() { setTo(0.0, 0.0, 0.0); }
    ///
  G3DVector3( T a ) : TVector<T,3>() { setTo(a,a,a); }
    ///
  G3DVector3( T x, T y, T z );
    ///
  G3DVector3( const T *vec ) : TVector<T,3>(vec) {};
    ///TODO is this a copy constructor?
  G3DVector3( const TVector<T,3> &tv ) : TVector<T,3>(tv) {};

    ///return the cross product of this with v
  TVector<T,3>  cross( const G3DVector3 &v) const;
    ///Return a vector orthogonal to this one. It will be in the xy plane.
  TVector<T,3>  getOrtho( void ) const;

  void          rotateX( double alpha );
  void          rotateY( double alpha );
  void          rotateZ( double alpha );

  inline void   setTo( T, T, T );
  T     x( void ) const { return _data[0]; };
  T     y( void ) const { return _data[1]; };
  T     z( void ) const { return _data[2]; };

  inline G3DVector3<T>&  operator=( const TVector<T,3> &tv );

    ///create new TVector for result
  inline TVector<T,3>    operator+( const G3DVector3<T> &v ) const;
    ///return result in tv
  inline TVector<T,3>&   operator+( TVector<T,3> &tv ) const;
    ///create new TVector for result
  inline TVector<T,3>    operator-( const G3DVector3<T> &v ) const;
    ///return result of subtractiion in tv
  inline TVector<T,3>&   operator-( TVector<T,3> &tv ) const;
  inline G3DVector3<T>&  operator+=( const TVector<T,3> &tv );

    //return a new TVector with the result of the scaling
  inline TVector<T,3>    operator*(T) const;
    ///return a pointer to the data
  inline G3DVector3<T>&  operator*=( T );

};



//===============Methods for class G3DVector===========================


template<class T>
inline G3DVector3<T>::G3DVector3( T x, T y, T z ) :
  TVector<T,3>( _data )
{
  _data[0] = x;
  _data[1] = y;
  _data[2] = z;
}


template <class T>
inline void G3DVector3<T>::setTo( T x, T y, T z )
{
  T* dest   = _data;

  *(dest++) = x;
  *(dest++) = y;
  *dest     = z;
}


template <class T>
inline G3DVector3<T>& G3DVector3<T>::operator=( const TVector<T,3> &tv )
{
        T* dest = _data;
  const T* src  = (const T*)tv;

  *(dest++) = *(src++);
  *(dest++) = *(src++);
  *dest     = *src;

  return( *this );
}


template <class T>
inline TVector<T,3> G3DVector3<T>::operator+( const G3DVector3<T> &v ) const
{
  TVector<T,3> tv;

  const T* sum1 = _data;
  const T* sum2 = (const T*)v;
        T* dest = tv.data();

  *(dest++) = *(sum1++) + *(sum2++);
  *(dest++) = *(sum1++) + *(sum2++);
  *dest     = *sum1 + *sum2;
  return( tv );
}


template <class T>
inline TVector<T,3>& G3DVector3<T>::operator+( TVector<T,3> &tv ) const
{
  const T* sum1 = _data;
        T* dest = tv.data();

  *(dest++) += *(sum1++);
  *(dest++) += *(sum1++);
  *dest     += *sum1;

  return( tv );
}


template <class T>
inline TVector<T,3> G3DVector3<T>::operator-( const G3DVector3<T> &v ) const
{
  TVector<T,3> tv;

  const T* sum1 = _data;
  const T* sum2 = v._data;
        T* dest = tv.data();

  *(dest++) = *(sum1++) - *(sum2++);
  *(dest++) = *(sum1++) - *(sum2++);
  *dest     = *sum1 - *sum2;
  return( tv );
}


template <class T>
inline TVector<T,3>& G3DVector3<T>::operator-( TVector<T,3> &tv ) const
{
  const T* sum1 = _data;
        T* dest = tv.data();

  *dest = *(sum1++) - *dest;  ++dest;
  *dest = *(sum1++) - *dest;  ++dest;
  *dest = *sum1 - *dest;

  return( tv );
}

template <class T>
inline G3DVector3<T>& G3DVector3<T>::operator+=( const TVector<T,3> &tv )
{
        T* sum1 = _data;
  const T* sum2 = (const T*)tv;

  *(sum1++) += *(sum2++);
  *(sum1++) += *(sum2++);
  *sum1     += *sum2;

  return( *this );
}


template <class T>
inline TVector<T,3> G3DVector3<T>::operator*( T mult ) const
{
  TVector<T,3> tv;
        T* dest = tv.data();
  const T* src  = _data;

  *(dest++) = *(src++) * mult;
  *(dest++) = *(src++) * mult;
  *dest     = *src * mult;

  return( tv );
}


template <class T>
inline G3DVector3<T>& G3DVector3<T>::operator*=( T mult )
{
  T* dest = _data;

  *(dest++) *= mult;
  *(dest++) *= mult;
  *dest     *= mult;

  return( *this );
}

#endif
