/*
 * Project : "MindsEye, math lib"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: tvector.h,v 1.13 1998/07/23 20:15:59 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * Classes for Vectors based on  the "temporary base class idiom".
 * See the article in the german magazine iX issue 11/96 p166.
 *
 */

#ifndef __TVECTOR_H_
#define __TVECTOR_H_


#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <math.h>

//TODO
#define EPSILON  1.0E-15


template <class T> class G3DVector3;
template <class T> class G3DPoint3;
class G3DMatrix;


/**This is a general vector class especially designed for temporary values.
   If the result of an arithmetic operation is a vector, the result is
   stored in the 1st operand.
   */
template <class T, int N>
class TVector
{
  friend class G3DVector3<T>;
  friend class G3DPoint3<T>;
  friend class G3DMatrix;

    /**Write the components to the ostream, separated by a space.
       The last caracter is a space. */
  friend ostream&  operator<<( ostream& os, const TVector<T,N> &tv );
  friend istream&  operator>>( istream& os, TVector<T,N> &tv );

public:
    ///The data area is not initialised!
  TVector( void ) {};
    ///
  TVector( const T* vec );
    ///
  TVector( const TVector<T, N>& tv );

    ///Componentwise divide this vector by tv.
  TVector<T,N>&  cwDivide( const TVector<T,N> &tv );
    ///Componentwise multiply this vector with tv.
  TVector<T,N>&  cwMultiply( const TVector<T,N> &tv );
    ///Calcualte the dot product of this vector with vector tv
  T      dot( const TVector<T,N> &tv ) const;
    ///Test if all components of this vector are zero ( |x_i| < epsilon).
  bool   isZero( void ) const;
    ///Calculate the lenght of this vector.
  T      length( void ) const;
    ///Calculate the squared lenght of this vector.
  double lengthSquared( void ) const;

    ///Set this vector to the component wise maxiumum of this and tv
  void   max( const TVector<T,N> &tv );
    ///Set this vector to the component wise mainimum of this and tv
  void   min( const TVector<T,N> &vt );

    ///Scale this vector to unit length. Do nothing if length < FG_EPSILON.
  TVector<T,N>&  normalize( void );
    /**TODO generate a random vector. The length of the generated
      vector will be evenly distributed in the open intervall (0,1). */
  TVector<T,N>   random( void );
    ///Round each component of this vector.
  TVector<T,N>&  round( void );
    /**Round each commponent to the nearest grid value. If a component in
       spacing is 0 the corresponding component in the vector is not changed.*/
  TVector<T,N>&  round( const TVector<T,N> &spacing );

    ///Compare each pair of components. If all pairs are equal return 1, else 0
  inline int           operator==( const TVector<T,N> &tv ) const;
    ///Compare each pair of components.
  inline int           operator<=( const TVector<T,N> &tv ) const;
    ///
  inline int           operator!=( const TVector<T,N> &tv ) const;
    ///
  inline TVector<T,N>& operator+( const TVector<T,N> &tv );
    ///
  inline TVector<T,N>& operator-( const TVector<T,N> &tv );
    ///
  inline TVector<T,N>& operator*(T);
    ///
  inline T&            operator[]( int );
    ///
  inline T             operator[]( int i ) const { return _data[i]; };
    ///Return a const pointer to the data of the vector.
  inline operator const T*() const;
    // Return a pointer to the data of the vector
  inline T* data() ; 

protected:
  T _data[N];

};


//===============Methods for class TVector===========================


template<class T, int N>
inline TVector<T,N>::TVector( const T *vec )
{
  T* dest = _data;

  for( int i = N; i > 0; i-- )
    *(dest++) = *(vec++);
}


template<class T, int N>
inline TVector<T,N>::TVector( const TVector<T, N>& tv )
{
        T* dest = _data;
  const T* src  = tv._data;

  for( int i = N; i > 0; i-- )
    *(dest++) = *(src++);
}

template <class T, int N>
inline int TVector<T,N>::operator==( const TVector<T,N> &tv ) const
{
  const T *src1 = _data;
  const T *src2 = tv._data;

  for( int i = N; i > 0; --i )
    if( *(src1++) != *(src2++) )
      return 0;
  return 1;
}

template <class T, int N>
inline int TVector<T,N>::operator<=( const TVector<T,N> &tv ) const
{
  const T *src1 = _data;
  const T *src2 = tv._data;

  for( int i = N; i > 0; --i )
    if( *(src1++) > *(src2++) )
      return 0;
  return 1;
}

template <class T, int N>
inline int TVector<T,N>::operator!=( const TVector<T,N> &tv ) const
{
  const T *src1 = _data;
  const T *src2 = tv._data;

  for( int i = N; i > 0; --i )
    if( *(src1++) != *(src2++) )
      return 1;
  return 0;
}

template <class T, int N>
inline TVector<T, N>& TVector<T,N>::operator+( const TVector<T,N> &tv )
{
        T* sum1 = _data;
  const T* sum2 = tv._data;

  for( int i = N; i > 0; i-- )
    *(sum1++) += *(sum2++);

  return( *this );
}


template <class T, int N>
inline TVector<T, N>& TVector<T,N>::operator-( const TVector<T,N> &tv )
{
        T* sum1 = _data;
  const T* sum2 = tv._data;

  for( int i = N; i > 0; i-- )
    *(sum1++) -= *(sum2++);

  return( *this );
}


template <class T, int N>
inline TVector<T,N>& TVector<T,N>::operator*( T mult )
{
  T* dest = _data;

  for( int i = N; i > 0; i-- )
    *(dest++) *= mult;
  return( *this );
}


template <class T, int N>
inline T& TVector<T,N>::operator[]( int i )
{
#ifdef DEBUG
  if( i >= 0 && i < N )
#endif
    return( _data[i] );
#ifdef DEBUG
  else
    return( _data[0] );  //TODO throw exception??
#endif
}


template <class T, int N>
TVector<T,N>::operator const T*( ) const
{
  return (const T*) _data ;
}

template <class T, int N>
T* TVector<T,N>::data()
{
  return( &(_data[0]) );
}


#endif
