/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: object.h,v 1.11 1998/07/23 20:15:48 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __OBJECT_H_
#define __OBJECT_H_


#include <stdio.h>
#include "include/config.h"
#include "g3d_math.h"
#include "t_actions.h"
#include "t_fields.h"
#include "t_objects.h"


class G3DColor;


enum G3DObjectState { visible = 1, selected = 2 };


/**Base class for all 3d Objects in a scene. An object may be
   a group (including the scene itself), a geometric primitive,
   a widget, a light source or a sound source.

   The class hirarchy based on G3DObject has two purposes:

     - it serves as an organisation and management tool for
       the objects in a scene.

     - it has an interface to a real time shading system. This
       is currently openGL.

   There are interfaces which allow for higher order functions
   to be implemented outside of the class hirarchy. This allows
   for a lot of flexibility.

   The attribute system.
   Each object is described by its geometry (in case of a group
   the geometry is described by the set of its children) and
   the attributes.

   Internaly the attributes are classified into the follwoing
   categorys:

     - modeling matrix
     - shading attributes
     - deffered shading attributes
     - constraints
     - behaviour
     - misc

   There is no problem to invent new attributes and assign them.
   A possible use may be the storage of physical propertys for
   kinematic calculations.


   implicit attributes:
   name, type, trafo, texture(?), bbox
 */
class G3DObject
{
protected:
    ///
  G3DObject( void );
    ///Create an Object with p as parent.
  G3DObject( G3DObject *p );
    ///
  G3DObject( const G3DObject& );

public:
  virtual ~G3DObject();

    ///Do nothing here.
  virtual void  adjust( void );
    ///TODO Apply the modeling matrix to the object and set the matrix to 1.
  virtual void  applyTransform( void );
    ///Make a deep copy of the object.
  virtual G3DObject* copy( void ) const;
    /**This is a general method to send messages to a single object
      or to each object in the whole scene. If the class of an object
      does not understand the message it is passed to the parent class.*/
  virtual void  doAction( ActionData& );
    ///The roation is specified via the 3 Euler angles.
  virtual const G3DVector3d& getAngles( void ) const;
    ///Return a list with the names and types of the objects parameter.
  virtual const Field* getFieldList( void ) const;
    ///TODO
  virtual void  getFunctionList( void ) const;
    ///
  virtual void  getLocalTrafo( G3DMatrix &m ) const;
    ///
  const char*   getName( void ) const { return(name_); };
    ///Get Origin of local coordinate system.
  virtual const G3DPoint3d& getOrigin( void ) const;
    /**Get a list of params describing the geometry of the object.
       If plist is Null you get just the number of elements.*/
  virtual uint  getParameter( Param* );
    ///
  G3DObject*    getParent( void ) { return(parent_); };
    ///Get the position in the local coordinate system.
  virtual void  getPosition( G3DPoint3d& ) const;
  virtual const G3DVector3d& getScale( void ) const;
    ///
  G3DObjectState getState( void ) const { return( state_ ); }
    ///Get the type identifier. Note that there is no method to set this value.
  t_type        getType( void ) const { return( type_ ); };
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
    ///Return a
  static  const char* GetTypeName( void );
    /**Multiply all parent transformation matrices up to the root of
       the scene. If local is true, the last matrix is the one of the
       object itself. */
  virtual void   getWorldTrafo( G3DMatrix &m, bool local=true ) const;
    ///Select subobjects of an object.
  virtual void   selectSubObjects( void );
    ///Set the roation.
  virtual void   setAngles( const TVector<double,3>& );
    ///
  virtual void   setLocalTrafo( const G3DMatrix &m );
    ///
  void           setName( const char* );
    ///Set origin of local coordinate system.
  virtual void   setOrigin( const TVector<double,3>& );
    ///The opposite of getParameter.
  virtual void   setParameter( const Param* );
    ///
  virtual void   setParent( G3DObject *pt );
    ///TODO more general? Do we need this?
  virtual void   setPosition( const G3DPoint3d& );
  virtual void   setScale( const TVector<double,3> &scale );
    ///TODO delete _selectColor?
  static void    setSelectColor( const G3DColor *c ) { selectColor_ = c; };

protected:
  void  rebuildTrafo( void );

protected:
  G3DObject  *parent_;
    ///runtime type information
  t_type     type_;
  char       *name_;  //TODO a general info attribute?

  G3DMatrix   trafo_;
  G3DPoint3d  origin_;
  G3DVector3d scale_;
  //G3DVector3d axis_;
  G3DVector3d angles_;

    ///reference counter
  //int        count_;
  G3DObjectState state_;

  enum { selName_shape, selName_cpt, selName_normal, selName_widget,
         selName_light };

  static const G3DColor *selectColor_;

};


#endif
