/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_group.cpp,v 1.18 1998/06/22 20:12:23 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <GL/gl.h>
#include "s_group.h"
#include "Attributes/texture.h"


class FindBBoxIterator : public GIterator<G3DShape>
{
public:
  FindBBoxIterator( G3D_BBAligned &box ) : box_(box) {}
  virtual bool exec( G3DShape &t )
    {
      t.getLocalTrafo( mat_ );
      box_.addBBox( t.getBoundingBox(), mat_ );
      return( false );
    };

private:
  G3D_BBAligned &box_;
  G3DMatrix     mat_;
};

class LightActionIterator : public GIterator<G3DLight>
{
public:
  LightActionIterator( ActionData &action ) : action_(action) {};
  virtual bool exec( G3DLight &t )
  {
    t.doAction( action_ );
    return( false );
  }
private:
  ActionData &action_;
};

class ShapeActionIterator : public GIterator<G3DShape>
{
public:
  ShapeActionIterator( ActionData &action ) : action_(action) {};
  virtual bool exec( G3DShape &t )
  {
    t.doAction( action_ );
    return( false );
  }
private:
  ActionData &action_;
};


//---------------------------------------------------------------------


G3DGroup::G3DGroup( void )
{
  type_           = T_S_GROUP;
  selectChildren_ = false;
  resetBoundingBox();
}

G3DGroup::G3DGroup( const G3DGroup &orig ) :
  G3DShape( orig )
{
  selectChildren_ = orig.selectChildren_;

  shapeList_ = orig.shapeList_;
  lightList_ = orig.lightList_;
  resetBoundingBox(); //TODO
}


void
G3DGroup::addObject( G3DObject *obj )
{
  if( obj->getType() & T_SHAPE )
    addShape( (G3DShape*)obj );
  else if( obj->getType() & T_LIGHT )
    addLight( (G3DLight*)obj );
}

void
G3DGroup::addShape( G3DShape *obj )
{
  shapeList_.append( obj );
  obj->setParent( this );
  if( shapeList_.size() > 1 )
    {
      G3DMatrix mat;
      obj->getLocalTrafo( mat );
      bbox_.addBBox( obj->getBoundingBox(), mat );
    }
  else
    resetBoundingBox();  //fresh start for the first element
}

void
G3DGroup::compileList( int lnum )
{
  glPushMatrix();
  glMultMatrixd( (const double*)trafo_ );

  G3DAListIterator it = attributes_.get( this, TA_TEXTURE, true );
  if( *it )
    (*it)->activate();
  GLightingIterator<G3DLight> lonit( lnum );
  lightList_.doWithElements( lonit );
  int lnum2 = lonit.getLNum();

  GCompileListIterator<G3DShape> cit( lnum2 );
  shapeList_.doWithElements( cit );

  for( int i = lnum; i < LIGHT_MAX; i++ )
    glDisable( (GLenum)(i + GL_LIGHT0 ) );

  glPopMatrix();
}

G3DObject*
G3DGroup::copy( void ) const
{
  return( new G3DGroup(*this) );
}

void
G3DGroup::doAction( ActionData &action )
{
  switch( action.type )
    {
    case A_TOGGLE_CHILD_ACCESS:
      if( selectChildren_ )
	selectChildren_ = false;
      else
	selectChildren_ = true;
      break;
    case A_GENERIC_ALL:
      //TODO lights, sounds, ...
    case A_GENERIC_SHAPES:
      {
	((GenericAction&)action).evaluate( this );
	ShapeActionIterator siter( action );
	doWithShapes( siter );
	if( action.type == A_GENERIC_SHAPES )
	  {	
	    ((GenericAction&)action).postAction( this );  //TODO ???
	    break;
	  }
      }
    case A_GENERIC_LIGHTS:
      {
	//TODO ? ((GenericAction&)action).evaluate( this );
	LightActionIterator liter( action );
	doWithLights( liter );
	((GenericAction&)action).postAction( this );	
	break;
      }
    case A_FIND_SELECTION:
      {
	GLuint n, otype;
	((SelectData&)action).peekNextNum( otype );
	if( otype == selName_shape )
	  {
	    if( selectChildren_ )
	      {
		((SelectData&)action).nextNum();
		((SelectData&)action).getNextNum( n );
		G3DShape *tmp = getShape( n );
		if( tmp )
		  tmp->doAction( action );
	      }
	    else
	      ((SelectData&)action).setSelection( this );
	    break;
	  }
	else if( otype == selName_light )
	  {
	    if( selectChildren_ )
	      {
		((SelectData&)action).nextNum();
		((SelectData&)action).getNextNum( n );
		G3DLight *tmp = getLight( n );
		if( tmp )
		  tmp->doAction( action );
	      }
	    else
	      ((SelectData&)action).setSelection( this );
	    break;
	  }   
	//else Fall through
      }
    default:
      G3DShape::doAction( action );
    }
}


void
G3DGroup::doWithLights( GIterator<G3DLight> &it )
{
  lightList_.doWithElements( it );
}


void
G3DGroup::doWithShapes( GIterator<G3DShape> &it )
{
  shapeList_.doWithElements( it );
}

const char*
G3DGroup::GetTypeName( void )
{
  return( "Group" );
}

const char*
G3DGroup::getTypeName( void ) const
{
  return( G3DGroup::GetTypeName() );
}

double
G3DGroup::intersect( G3DRay *ray, G3DObject **obj )
{
  //BBox test
  //TODO if level == 0
  //return after BBox test
  /*TODO
  GIntersectIterator<G3DShape> sit( ray );
  shapeList_.doWithElements( sit );

  double dist = sit.getDistance();
  //cout << "node dist: " << dist << endl;
  if( (dist > 0) && (dist < G3D_INFINITY) )
      *obj = sit.getObject();
  return( dist );
  */
  return( 0 );
}


int
G3DGroup::numberOfShapes( void ) const
{
  return( shapeList_.size() );
}

bool
G3DGroup::removeObject( G3DObject *obj )
{
  if( obj->getType() & T_SHAPE )
    return( removeShape((G3DShape*)obj) );
  else if( obj->getType() & T_LIGHT )
    return( removeLight((G3DLight*)obj) );
}

/* TODO
void
G3DGroup::reset( void )
{
  shapeList_.reInitialize();
  resetBoundingBox();
}
*/

void
G3DGroup::resetBoundingBox( void )
{
  bbox_.reset();
  FindBBoxIterator bit( bbox_ );
  doWithShapes( bit );

  if( parent_ )  //TODO
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DGroup::select( void )
{
  if( state_ & visible )
    {
      glPushMatrix();
      glMultMatrixd( (const double*)trafo_ );

      glPushName( selName_light );
      glPushName( 0 );
      GSelectionIterator<G3DLight> lit;
      lightList_.doWithElements( lit );
      glPopName(); glPopName();

      glPushName( selName_shape );
      glPushName( 0 );
      GSelectionIterator<G3DShape> sit;
      shapeList_.doWithElements( sit );
      glPopName(); glPopName();

      glPopMatrix();
      //resetGLState();
    }
};


void
G3DGroup::shade( int lnum ) const
{
  glPushAttrib( GL_CURRENT_BIT | GL_LIGHTING_BIT );  //TODO

  GLightingIterator<G3DLight> lit( lnum );
  lightList_.doWithElements( lit );
  int lnum2 = lit.getLNum();

  GShadingIterator<G3DShape> sit( lnum2 );
  shapeList_.doWithElements( sit );

  for( int i = lnum; i < LIGHT_MAX; i++ )  //TODO don't use LIGHT_MAX
    glDisable( (GLenum)(i + GL_LIGHT0 ) );

  glPopAttrib();
}


//================================END======================================
