/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_group.h,v 1.17 1998/06/22 20:12:24 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __GROUP_H_
#define __GROUP_H_


#include "shape.h"
//#include "bbox.h"
#include "aggregate.h"
#include "iterator.h"
#include "light.h"



/**A group node for organizing the scene in a hirachical way.
   A group encapsulates its children with respect to the shading
   process. This means, the current shading state is saved when
   the group is entered and restored when the group is left.
   A group manages a list of shapes and a list of lights.

   The hirarchiy set up with groups is strctly contents based. Thinks
   like kinematic relationships are treated apart from this.
 */
class G3DGroup : public G3DShape
{
public:
  G3DGroup( void );
  G3DGroup( const G3DGroup& );

    ///Add a light source.
  inline  void   addLight( G3DLight* );
    ///
  void           addObject( G3DObject* );
    ///Add a shape to the group.
  void           addShape( G3DShape* );
    /**Build an OpenGL display List. The function is called for every
       shape in the group. There is no display list associated with
       the group itself.*/
  virtual void compileList( int lnum );
    ///Make a deep copy of the object.
  virtual G3DObject* copy( void ) const;
    ///
  virtual void doAction( ActionData& );
    ///Iterate over the lights.
  void   doWithLights( GIterator<G3DLight>& );
    ///Iterate over the shapes.
  void   doWithShapes( GIterator<G3DShape>& );
    ///Return the shape at position i in the list of shapes.
  G3DLight* getLight( int i ) { return lightList_.getElement(i); }
    ///Return the shape at position i in the list of shapes.
  G3DShape* getShape( int i ) { return shapeList_.getElement(i); }
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
    ///Not used.
  virtual double intersect( G3DRay*, G3DObject** );
    ///
  virtual int    numberOfShapes( void ) const;
    ///
  int    numberOfLights( void ) const { return( lightList_.size() ); }
    ///Remove a light from the group.
  inline  bool   removeLight( G3DLight* );
  bool           removeObject( G3DObject* );
    ///Remove a shape from the group.
  inline  bool   removeShape( G3DShape* );
    ///Remove all shapes and lights.
  //virtual void   reset( void );  //TODO virtual??
    ///
  virtual void   resetBoundingBox( void );
    /**Call select for all shapes in the group. This is used for OpenGL
       selection mode.*/
  virtual void   select( void );
    ///Calls display() for all shapes in the group.
  virtual void   shade( int ) const;

protected:
  G3DDLList<G3DShape> shapeList_;
  G3DDLList<G3DLight> lightList_;

  bool                selectChildren_;
};



inline
void G3DGroup::addLight( G3DLight *light )
{
  lightList_.append( light );
  light->setParent( this );
}

inline
bool G3DGroup::removeLight( G3DLight *ltg )
{
  lightList_.remove( ltg );
  return( true );
}

inline
bool G3DGroup::removeShape( G3DShape *obj )
{
  shapeList_.remove( obj );
  return( true );
}


#endif
