/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_simple.h,v 1.14 1998/07/23 20:15:52 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __S_PRIMS_H__
#define __S_PRIMS_H__

#include "shape.h"
#include "s_glshapes.h"


///a box object using OpenGL to draw itself on the screen. <s_shapes.h>
/**Instances of this class represent boxes and generalisations,
   where the opposing faces are parallel to each other.
 */
class G3DS_Box : public G3DShape
{
public:
  G3DS_Box( void );
  G3DS_Box( const G3DVector3d &orig, const G3DVector3d &e1,
	    const G3DVector3d &e2, const G3DVector3d &e3 );
  G3DS_Box( const G3DS_Box& );
  virtual ~G3DS_Box();

    ///
  virtual G3DObject* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Copy pointers to the geometric data to the list. The order is:
       origin, e1, e2, e3*/
  virtual uint   getParameter( Param* );
    ///
  virtual void   getPosition( G3DPoint3d& ) const;
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
    ///
  virtual double intersect( G3DRay *ray, G3DObject **obj );
    ///
  virtual void   resetBoundingBox( void );
    ///
  virtual void   setPosition( const G3DPoint3d& );
    ///
  virtual void   setParameter( const Param* );
    ///
  virtual void   shade( int ) const;

protected:
  void  tesselate( void );

protected:
  G3DPoint3d  _origin;
  G3DVector3d _e1, _e2, _e3;  //Position + geometry of the box

  G3DVector3d _points[8]; //TODO points
  G3DVector3d _norm[6];

  static int   Faces_[6][4];
  static Field FieldList_[4];

};


/**A cone is defined by an axes and two radii.
 */
class G3DS_Cone : public G3DShape
{
public:
  G3DS_Cone( void );
  G3DS_Cone( const G3DS_Cone& );
  virtual ~G3DS_Cone();

    ///
  virtual G3DObject* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Copy pointers to the geometric data to the list. The order is:
       r_bottom, r_top, origin, orientation*/
  virtual uint  getParameter( Param* );
    ///
  virtual void  getPosition( G3DPoint3d& ) const;
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
    ///
  virtual void  resetBoundingBox( void );
    ///
  virtual void  setParameter( const Param* );
   ///
  virtual void  setPosition( const G3DPoint3d& );
     ///
  virtual void  shade( int ) const;

protected:
  void  tesselate( void );

protected:
  //GL primitives used to render the thing
  G3DGLQuadStrip _mantle;
  G3DS_Polygon   _top, _bottom;

  //these values define the cylinder
  G3DPoint3d     _origin;
  G3DVector3d    _orientation; //, _major, _minor;
  double  _botr, _topr;

  int            _cupFlag;
  static Field FieldList_[4];
};


/**.A cylinder is defined by an origin (at the center of the bottom cap)
   and three mutually perpendicular vectors that define the orientation and
   the major and minor radii of the cylinder. A cylinder may include a top
   cap, a bottom cap, or both.
   The specification of a major radii is useful to manipulate the texture
   coordinates.
   */
class G3DS_Cylinder : public G3DShape
{
public:
  ///
  G3DS_Cylinder( double r, double h );
  G3DS_Cylinder( G3DVector3d &orient, G3DVector3d &mj,
		 G3DVector3d &mi, G3DVector3d &orig );
  G3DS_Cylinder( const G3DS_Cylinder& );
  virtual ~G3DS_Cylinder();

    ///
  virtual G3DObject* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Copy pointers to the geometric data to the list. The order is:
       origin, orientation, major, minor radius vector*/
   virtual uint  getParameter( Param* );
    ///
  virtual void   getPosition( G3DPoint3d& ) const;
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
    ///
  virtual double intersect( G3DRay*, G3DObject** );
    ///
  virtual void   resetBoundingBox( void );
    ///
  virtual void   setParameter( const Param* );
    ///
  virtual void   setPosition( const G3DPoint3d& );
    ///
  virtual void   shade( int ) const;

protected:
  /**The mantle is made outof a quadstripe, the caps are polygons.*/
  void  tesselate( void );

protected:
  //GL primitives used to render the thing
  G3DGLQuadStrip _mantle;
  G3DS_Polygon   _top, _bottom;

  //these values define the cylinder
  G3DPoint3d     _origin;
  G3DVector3d    _orientation, _major, _minor;
  int            _cupFlag;
  static Field FieldList_[4];

};


/**.A sphere is defined by an origin and 3 axes */
class G3DS_Sphere : public G3DShape
{
public:
  G3DS_Sphere( double r );
  G3DS_Sphere( double r, const G3DPoint3d &center );
  G3DS_Sphere( const G3DS_Sphere& );
  virtual ~G3DS_Sphere();

    ///
  virtual G3DObject* copy( void ) const;
    ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList( void ) const;
    /**Copy pointers to the geometric data to the list. The order is:
       center, radius*/
  virtual uint   getParameter( Param* );
    ///
  virtual void   getPosition( G3DPoint3d& ) const;
    ///Return a Descriptive name of the shape.
  virtual const char* getTypeName( void ) const;
  static  const char* GetTypeName( void );
    ///
  virtual double intersect( G3DRay*, G3DObject** );
    ///
  virtual void   resetBoundingBox( void );
    ///
  virtual void   setParameter( const Param* );
    ///
  virtual void   setPosition( const G3DPoint3d& );
    ///
  virtual void   shade( int ) const;

protected:
  void  tesselate( void );

protected:
  G3DGLQuadStrip *_mantle;

  G3DPoint3d _center;
  double     _radius;
  int        _stacks;
  static Field FieldList_[2];

};


#endif
