/*
 * Project : "MindsEye, 3D Library"
 * Copyright (c) 1997. All Rights Reserved.
 * Author(s):   Martin Kuball
 *
 * $Id: s_torus.cpp,v 1.5 1998/07/23 20:15:53 source Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * implementation of classes for geometric primitives
 */


#include <stdlib.h>
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "s_torus.h"


G3DTorus::G3DTorus( double rad1, double rad2 ) :
  _center(0,0,0)
{
  type_   = T_TORUS;
  _mantle = NULL;
  _rad1   = rad1;
  _rad2   = rad2;
  tesselate();
};

G3DTorus::G3DTorus( double rad1, double rad2, const G3DPoint3d &center ) :
  _center(center)
{
  type_   = T_TORUS;
  _mantle = NULL;
  _rad1   = rad1;
  _rad2   = rad2;
  tesselate();
};

G3DTorus::G3DTorus( const G3DTorus &orig ) :
  G3DShape(orig)
{
  _center = orig._center;
  _rad1   = orig._rad1;
  _rad2   = orig._rad2;
  _mantle = NULL;
  tesselate();
}

G3DTorus::~G3DTorus()
{
  delete[] _mantle;
}


G3DObject*
G3DTorus::copy( void ) const
{
  return( new G3DTorus(*this) );
}

Field G3DTorus::FieldList_[3] =
{ {"center", s_point3}, {"rad1", s_dbl}, {"rad2", s_dbl} };

const Field*
G3DTorus::getFieldList( void ) const
{
  return( FieldList_ );
}


uint
G3DTorus::getParameter( Param *pList )
{
  if( pList )
    {
      pList->point3 = &_center;
      pList++;
      pList->dbl    = _rad1;
      pList++;
      pList->dbl    = _rad2;
    }
  return( 3 );
}

void
G3DTorus::getPosition( G3DPoint3d &pos ) const
{
  pos = _center;
}

const char*
G3DTorus::GetTypeName( void )
{
  return( "Torus" );
}

const char*
G3DTorus::getTypeName( void ) const
{
  return( G3DTorus::GetTypeName() );
}

double
G3DTorus::intersect( G3DRay*, G3DObject** )
{
}

void
G3DTorus::resetBoundingBox()
{
  G3DVector3d max( _rad1+_rad2, _rad2, _rad1+_rad2 );

  bbox_.setTo( _center - max, _center + max );
  if( parent_ )
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DTorus::setParameter( const Param *pList )
{
  if( pList )
    {
      _center  = *(pList->point3);
      pList++;
      _rad1    = pList->dbl;
      pList++;
      _rad2    = pList->dbl;
      tesselate();
    }
}

void
G3DTorus::setPosition( const G3DPoint3d &pos )
{
  _center = pos;
  tesselate();
}

void
G3DTorus::shade( int ) const
{
  for( int i = 0; i < _stacks; i++ )
    _mantle[i].shade();
}


void
G3DTorus::tesselate( void )
{
  double x, y, z, nsign = 1.0;  //TODO based on inside/outside Attribute
  double n1, n2, rad;
  double rho, drho, theta, dtheta;
  G3DGLQuadStrip *strip;
  int k, slices = 6;
  G3DVector3d normal;
  G3DPoint3d  point;

  _stacks = 10;
  delete[] _mantle;
  _mantle = new G3DGLQuadStrip[_stacks];
  //if (qobj->Orientation==GLU_INSIDE)
  //nsign = -1.0;

  drho   = 2.0 * M_PI / (double)_stacks;
  dtheta = 2.0 * M_PI / (double)slices;

  for( int i = 0; i < _stacks; i++ )
     {
       strip = &(_mantle[i]);
       strip->setStripeNumber( slices );
       rho = i * drho;
       k = 0;
       for( int j = 0; j <= slices; j++ )
	 {
	   theta = ( (j == slices) ? 0.0 : j * dtheta );

	   n1 = cos(rho) * sin(theta);
	   n2 = sin(rho) * sin(theta) * nsign;
	   y  = -cos(theta);
	   normal.setTo( n1, y, n2 );
	   strip->setNormal( k, normal * nsign );

	   rad = _rad1 + (_rad2 * sin(theta));
	   x   = rad * cos( rho );
	   y  *= _rad2;
	   z   = rad * sin( rho ) * nsign;
	   point.setTo( x, y, z );
	   point += _center;
	   strip->setVertex( k++, point );

	   n1 = cos(rho+drho) * sin(theta);
	   n2 = sin(rho+drho) * sin(theta) * nsign;
	   normal.setTo( n1, y, n2 );
	   strip->setNormal( k, normal * nsign );
	   x = rad * cos( rho+drho );
	   z = rad * sin( rho+ drho ) * nsign;
	   point.setTo( x, y, z );
	   point += _center;
	   strip->setVertex( k++, point );
	 }
     }

  resetBoundingBox();
}


//==================================END==================================
