/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_trigrid.cpp,v 1.8 1998/07/23 20:15:53 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "s_trigrid.h"
#include "s_point.h"


G3DTriGrid::G3DTriGrid( uint rows, uint cols )
{
  type_ = T_TRIGRID;

  _rows = rows;
  _cols = cols;
  _vertices = new G3DPoint3d[rows * cols];
  _normals  = new G3DVector3d[rows * cols];

  G3DPoint3d *point = &(_vertices[0]);
  G3DPoint3d vertex, xstep(1.0/(cols-1), 0, 0), zstep(0,0,-1.0/(rows-1));
  for( uint i = 0; i < rows; i++ )
    {
      vertex = zstep * i;
      for( uint j = 0; j < cols; j++ )
	{
	  *point = vertex;
	  vertex += xstep;
	  ++point;
	}
    }
  calcSmoothNormals(); //TODO no implicit call??
  //resetBoundingBox();  //TODO
}

G3DTriGrid::G3DTriGrid( const G3DTriGrid &orig ):
  G3DShape( orig )
{
  _rows = orig._rows;
  _cols = orig._cols;

  G3DVector3d *dnorm, *snorm;
  G3DPoint3d  *dvert, *svert;

  _vertices = new G3DPoint3d[_rows * _cols];
  _normals  = new G3DVector3d[_rows * _cols];

  dnorm = _normals;
  snorm = orig._normals;
  dvert = _vertices;
  svert = orig._vertices;
  for( uint i = _cols*_rows; i > 0; i-- )
    {
      *(dvert++) = *(svert++);
      *(dnorm++) = *(snorm++);
    }
  //resetBoundingBox();
}

G3DTriGrid::~G3DTriGrid( )
{
  delete[] _vertices;
  delete[] _normals;
}

void
G3DTriGrid::adjust( void )
{
  calcSmoothNormals();  //TODO only region arund selected point?
  G3DShape::adjust();
}

G3DObject*
G3DTriGrid::copy( void ) const
{
  return( new G3DTriGrid(*this) );
}

//TODO I'm still looking for a cleaner and shorter(?) implementation
void
G3DTriGrid::calcSmoothNormals( void )
{
  G3DVector3d v1, v2, v3, n, *fn, *fnormals;
  G3DPoint3d  *vert1, *vert2;
  uint        nfaces = (_rows-1)*(_cols-1)*2;

  fnormals = new G3DVector3d[nfaces];
  fn       = fnormals;
  vert1 = _vertices;
  vert2 = &(_vertices[_cols]);
  for( uint i = 0; i < _rows - 1; i++ )
    {
      for( uint j = 0; j < _cols-1; j++ )
	{
	  v1 = *vert2 - *vert1;
	  v3 = *(vert1++);
	  v2 = *vert2 - *vert1;
	  v3 = v3 - *vert2;
	  n = v3.cross(v1) + v3.cross(v2) + v1.cross(v2);
	  *(fn++) = n.normalize();

	  v1 = *(vert2++);
	  v1 = v1 - *vert2;
	  v3 = *vert1 - *vert2;
	  n = v2.cross(v3) + v2.cross(v1) + v1.cross(v3);
	  *(fn++) = n.normalize();
	}
    }

  G3DVector3d zero(0,0,0);
  G3DVector3d *n1 = _normals;
  for( uint k = _rows*_cols; k > 0; k--)
    *(n1++) = zero;

  fn = fnormals;
  G3DVector3d *n2;
  n1 = _normals;
  n2 = &(_normals[_cols]);
  for( uint l = nfaces; l > 0; l-=2 )
    {
      *(n1++) += *fn;
      *n2     += *fn;
      *n1     += *(fn++);
      *(n2++) += *fn;
      *n1     += *fn;
      *n2     += *(fn++);
    }
  n1 = _normals;
  for( uint m = _rows*_cols; m > 0; m-- )
    {
      n1->normalize();
      n1++;
    }
}

void
G3DTriGrid::doAction( ActionData &action )
{
  if( action.type == A_FIND_SELECTION )
    {
      GLuint n, m;
      if( ((SelectData&)action).peekNextNum(n) )
	{
	  if( n == selName_cpt )
	    {
	      ((SelectData&)action).nextNum();
	      ((SelectData&)action).getNextNum( m );
	      ((SelectData&)action).setSelection( new G3DVertex(_vertices[m],
								this) );
	      return;
	    }
	  else if( n == selName_normal )
	    {
	      ((SelectData&)action).nextNum();
	      ((SelectData&)action).getNextNum( m );
	      ((SelectData&)action).setSelection( new G3DNormal(_vertices[m],
								_normals[m],
								this) );
	      return;
	    }
	}
    }
  G3DShape::doAction( action );
}

Field G3DTriGrid::FieldList_[6] =
{ {"rows", s_number}, {"columns", s_number},
  {"vertices", m_point3}, {"num_vert", s_number},
  {"normals", m_vector3}, {"num_norm", s_number}
};

const Field*
G3DTriGrid::getFieldList( void ) const
{
  return( FieldList_ );
}

uint
G3DTriGrid::getParameter( Param *pList )
{
  pList->number  = _rows;
  pList++;
  pList->number  = _cols;
  pList++;
  pList->point3  = _vertices;
  pList++;
  pList->number  = _rows * _cols;
  pList++;
  pList->vector3 = _normals;
  pList++;
  pList->number  = _rows * _cols;

  return( 6 );
}

void
G3DTriGrid::getPosition( G3DPoint3d &pos ) const
{
  pos = _vertices[0];
}

const char*
G3DTriGrid::GetTypeName( void )
{
  return( "Trigrid" );
}

const char*
G3DTriGrid::getTypeName( void ) const
{
  return( G3DTriGrid::GetTypeName() );
}

double
G3DTriGrid::intersect( G3DRay *ray, G3DObject **obj )
{
  /*
  G3DPoint3d *vert;
  double     t, dist, min = G3D_INFINITY;
  uint       num;

  vert = _vertices;
  for( uint k = _rows*_cols; k > 0; k-- )
    {
      t = (*vert - ray->_startpoint).dot( ray->_direction );
      dist = ((ray->_direction * t) + ray->_startpoint - *vert).length();
      if( dist < min )
	{
	  num = k;
	  min = dist;
	}
    }
  if( min < G3D_INFINITY )
    *obj = new G3DCpt( _vertices[num-1] );
  return( min );
  */
}

void
G3DTriGrid::resetBoundingBox( void )
{
  G3DPoint3d *vert = _vertices;
  G3DPoint3d pmin, pmax;

  if( vert )
    {
      pmin = *vert;
      pmax = *(vert++);
      for( uint i = (_cols*_rows)-1; i > 0; i-- )
	{
	  pmin.min( *vert );
	  pmax.max( *(vert++) );
	}
      bbox_.setTo( pmin, pmax );
    }
  else
    bbox_.reset();

  if( parent_ )
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DTriGrid::setParameter( const Param *plist )
{
  if( plist )
    {
      _rows  = plist->number;  plist++;
      _cols  = plist->number;  plist++;
      if( _vertices != plist->point3 )
	{
	  delete[] _vertices;
	  _vertices = plist->point3;
	}
      plist += 2;
      if( _normals != plist->vector3 )
	{
	  delete[] _normals;
	  if( plist->vector3 == NULL )
	    {
	      _normals = new G3DVector3d[_rows * _cols];
	      calcSmoothNormals();
	    }
	  else
	    _normals  = plist->vector3;
	}
      resetBoundingBox();
    }
}

void
G3DTriGrid::setVertex( uint r, uint c, TVector<G3Dfloat,3> )
{
  //TODO
}

void
G3DTriGrid::shade( int ) const
{
  G3DVector3d *n1, *n2;
  G3DPoint3d  *vert1, *vert2;
  uint j;

  vert1 = _vertices;
  vert2 = &(_vertices[_cols]);
  n1    = _normals;
  n2    = &(_normals[_cols]);
  glDisable(GL_CULL_FACE);  //TODO
  for( uint i = 0; i < _rows - 1; i++ )
    {
      //serp = i & 0x1;
      glBegin( GL_TRIANGLE_STRIP );
      for( j = 0; j < _cols; j++ )
	{
	  // TODO TXTR_COORD(tcx, tcy);
	  glNormal3dv( (const double*)(*n1++) );
	  glVertex3dv( (const double*)(*vert1++) );
	  glNormal3dv( (const double*)(*n2++) );
	  glVertex3dv( (const double*)(*vert2++) );
	}
      glEnd();
    }
  glEnable(GL_CULL_FACE);
}

void
G3DTriGrid::shadeSubObjects( void ) const
{
  //G3DVector3d *n1, *n2;
  G3DPoint3d  *vert1;
  uint j;

  glPushName( selName_cpt );
  glPushName( 0 );
  vert1 = _vertices;
  glPointSize( 6 );
  glColor3f(0.2, 0.2, 0.9); //TODO get contrasting color from Texture
  for( uint k = 0; k < _rows*_cols; k++ )
    {
      glLoadName( k );
      glBegin( GL_POINTS );
      glVertex3dv( (const double*)(*(vert1++)) );
      glEnd();
    }
  glPopName(); glPopName();
}


//===================================END================================
