/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. van Velsen, M. Kuball
 *
 * $Id: io_dialogs.cpp,v 1.11 1998/07/23 20:16:02 source Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * MIODialog implements a dialog widget for loading and
 * saving 3d data in the various supported formats.
 */



#include <stdio.h>

#include <qwindefs.h>
#include <qfiledlg.h>
#include <qcombo.h>
#include <qtooltip.h>
#include <qlistbox.h>
#include <qlabel.h>
#include <qpushbt.h>

#include "io_dialogs.h"
#include "io_base.h"
#include "io_module.h"
#include "3dlib/scene.h"
#include "kernel/m_core.h"
#include "kernel/world_obj.h"
#include "kernel/world_list.h"

#define red stl_red
#define black stl_black
#include <map.h>
#undef red
#undef black


struct ltint
{
  bool operator()( int i1, int i2 ) const
  { return( i1 < i2 ); }
};

static map<int, IoFactory*,ltint>* IoFilterList_ = NULL;


//TODO this must be dynamic datastructures

const char *FormatList[] = 
{ "automatic",
  "MindsEye format (.me3d)",
  "3d-studio ascii (.asc)",
  "3d-studio binary (.3ds)",
  "3ds MAX (.max)",
  "Alias WaveFront (.obj)",
  "Autodesk DXF format (.dxf)",
  "Caligari TrueSpace (.cob)",
  "LightWave (.iff / .lwob)",
  "Neutral File Format (.nff / .enff )",
  "Persistence of Vision (.pov)",
  // "Silicon Graphics Inventor (.iv)",
  //"SoftImage (.?)",
  "Quickdraw Metafile (Apple) (.3dmf)",
  "Renderman (.rib)",
  "VRML (.wrl)",
  // "Stereo CAD-3D 2.0 (.3d / .3d2 / .3d4 )",
  // "ART (.art)",
  // "AVS ( .geom / .prop / .scr )",
  // "Cyberware ( .cy / .tmesh )",
  // "LightScape ( .ls / .lsa / .lsb )",
  // "NURBS ( .txt / .nurbs )",
  0
};

const char *FilterList[] =
{ "*", "*.me3d",
  "*.asc", "*.3ds", "*.max", "*.obj",
  "*.dxf", "*.cob", "*.iff", "*.nff",
  "*.pov", "*.3dmf", "*.rib", "*.wrl"
};

const int FormatTypes[] =
{ IO_AUTOMATIC, IO_FORMAT_ME3D,
  IO_FORMAT_3DSASC, IO_FORMAT_3DSBIN, IO_FORMAT_MAX,
  IO_FORMAT_ALIAS,  IO_FORMAT_DXF,    IO_FORMAT_TRUESP,
  IO_FORMAT_LW,     IO_FORMAT_NFF,    IO_FORMAT_POV,
  IO_FORMAT_3DMF,   IO_FORMAT_RIB,    IO_FORMAT_VRML
};

 
/*---------------------------------------------------------------------------*/

MEIoModule::MEIoModule( int module_id, int load_t ) :
  MEModule(module_id, load_t, "MEIoModule" )
{
  _selNum = 0;
  strcpy (m_info->module_name,"Dialog module");

  int offset = 5;

  setCaption ("Choose a file format");

  QLabel *label = new QLabel( "File format:", this );
    label->setGeometry( 7, offset, FORMAT_BWIDTH, label->height() );
  offset += label->height();

  QComboBox *combo = new QComboBox( FALSE, this );
    combo->setGeometry( 5, offset, FORMAT_BWIDTH, BUTTON_HEIGHT );
    combo->insertStrList( FormatList );
    connect( combo, SIGNAL(activated(int)), SLOT(activated(int)) );
  offset += BUTTON_HEIGHT + 10;
  _current_type = 0;

  label = new QLabel( "List of loaded scenes", this );
    label->setGeometry( 7, offset, FORMAT_BWIDTH, label->height() );
  offset += label->height();

  _worldList = new QListBox( this );
    _worldList->setGeometry( 5, offset, FORMAT_BWIDTH, 70 );
    connect( _worldList, SIGNAL(selected(int)), SLOT(useLoaded(int)) );
    connect( _worldList, SIGNAL(highlighted(int)), SLOT(useLoaded(int)) );
  offset += 70 + 10;

  //>------------- 

  QPushButton *quitButton = new QPushButton (this,"Cancel");
  quitButton->setText ("Cancel");
  quitButton->setGeometry (5,offset,80,25);
  connect (quitButton, SIGNAL(clicked()), this, SLOT(closeIt()) );

  _doitButton = new QPushButton (this,"doit");
  _doitButton->setText ("load");
  _doitButton->setGeometry (90,offset,80,25);
  connect( _doitButton, SIGNAL(clicked()), this, SLOT(doIO()) );

  //>------------- 

  QPushButton *helpButton = new QPushButton (this,"Help");
  helpButton->setText ("Help");
  helpButton->setGeometry (175,offset,80,25);

  adjustSize();
  setFixedSize( width (),height() ); // make sure it stays the same size

  setWorldList();
  connect( Kernel, SIGNAL(worldListChanged()), SLOT(setWorldList()) );
}

MEIoModule::~MEIoModule ()
{
}

/*--------------------------------------------------------------------------*/

int
MEIoModule::addFilter( IoFactory* fnt, const char* ename, const char *ftype )
{
  int type = IoFilterList_->size();
  (*IoFilterList_)[type] = fnt;
  
}

void
MEIoModule::activated( int nr )
{
  _current_type = nr;
}

void
MEIoModule::closeIt()
{
  close( );
}

void
MEIoModule::doIO()
{
  QString str;

  str = QFileDialog::getOpenFileName( NULL, FilterList[_current_type] );
  if( !str.isEmpty() )
    {
      IO_BaseClass *ioObj = getIoFilter( FormatTypes[_current_type], str );
      if( ioObj == NULL )
	return;

      MWorldList &worlds = Kernel->getWorldList();
      if( _action == ACTION_LOAD )
	{
	  if( _root == NULL )
	    {
	      G3DScene     *scene = new G3DScene();
	      MWorldObject *world = new MWorldObject( scene, 0 );
	      ioObj->file_load( world, (const char*)str );
	      const char *tmp = strrchr( (const char*)str, '/' );
	      if( tmp == NULL )
		tmp = (const char*)str;
	      else
		tmp++;
	      scene->setName( tmp );
	      Kernel->addWorld( (void*) world );
	    }
	  else
	    {
	      ioObj->import( _root, (const char*)str );
	      emit objectsImported();  //TODO
	    }
	}
      else if( (_action == ACTION_SAVE) && (worlds.size() >= _selNum) )
	{
	  ioObj->file_save( worlds[_selNum], (const char*)str, 0 );
	  const char *tmp = strrchr( (const char*)str, '/' );
	  if( tmp == NULL )
	    tmp = (const char*)str;
	  else
	    tmp++;
	  ((worlds[_selNum])->getScene())->setName( tmp );
	  Kernel->addWorld( NULL );
	}
      delete ioObj;
    }
}

IO_BaseClass*
MEIoModule::getIoFilter( int type, const char *str )
{
  //return createMe3dFilter();

  if( type == IO_AUTOMATIC && str )
    {
      const char *ctype = strrchr( str, '.' );
      if( ctype == NULL )
	return( NULL );
      ctype++;
      for( int i = 1; i < 13; i++ )
	{
	  if( strcmp( ctype, FilterList[i]+2 ) == 0 )
	    {
	      type = i;
	      break;
	    }
	}
    }
  map<int, IoFactory*, ltint>::iterator itr = IoFilterList_->find( type );
  if( itr != IoFilterList_->end() )
    return( ((*itr).second)() );
  else
    return( NULL );
}

void
MEIoModule::initStandardFilter( void )
{
  if( IoFilterList_ == NULL )
    {
      IoFilterList_ = new map<int,IoFactory*,ltint>;
      (*IoFilterList_)[IO_FORMAT_ME3D] = createMe3dFilter;
      (*IoFilterList_)[IO_FORMAT_NFF]  = createNffFilter;
      (*IoFilterList_)[IO_FORMAT_RIB]  = createRibFilter;
      (*IoFilterList_)[IO_FORMAT_POV]  = createPovFilter;
    }
}

void
MEIoModule::setAction( int action )
{
  _action = action;
  if( _action == ACTION_LOAD )
    _doitButton->setText( "load" );
  else
    _doitButton->setText( "save" );

  _root = NULL;
}

void
MEIoModule::setWorldList()
{
  MWorldList &worlds = Kernel->getWorldList();
  _worldList->clear();
  _worldList->setAutoUpdate( FALSE );

  MWorldList::iterator wld = worlds.begin();
  while( wld != worlds.end() )
    _worldList->insertItem( (*(wld++))->getName() );

  _worldList->setAutoUpdate( TRUE );
  _worldList->repaint();
  _selNum = 0;
}

void
MEIoModule::useLoaded( int num )
{
  _selNum = num;
}


//===================================END===============================
