/*
 * Project : "MindsEye, IO library"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: me3d.cpp,v 1.5 1998/07/23 20:16:14 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * MindsEye's own IO Format.
 *
 */

#include <stdlib.h>

#include "3dlib/light.h"
#include "3dlib/scene.h"
#include "3dlib/camera.h"
#include "3dlib/actions.h"
#include "3dlib/Attributes/attribute.h"
#include "vector.hh"
#include "me3d.h"
#include "kernel/world_list.h"
#include "kernel/world_obj.h"
#include "kernel/m_core.h"
#include "io/io_module.h"

#include <matrix.hh>

#include <iomanip.h>

#define ME3D_HEADER "MindsEye scene description V0.2"


void readParameter( ifstream &inf, Param*, const Field*, int );
void writeParameter( ofstream &outf, const Param*, const Field*, int );


class Me3dSaveAction : public GenericAction
{
public:
  Me3dSaveAction( ofstream &out ) :
    GenericAction(A_GENERIC_ALL), outf_(out) {};
  virtual EvalState evaluate( G3DObject* );
  virtual void      postAction( G3DObject* );

private:
  Param    params_[20];
  ofstream &outf_;
};


IO_BaseClass*
createMe3dFilter( void )
{
  return( new MEIoMe3d );
}


//----------------------------------------------------------------


int
MEIoMe3d::check_header( char* )
{
  return( 0 );  //TODO undefined
}


int
MEIoMe3d::exitLoad( int )
{
  _outpf.close();
  return( IO_BaseClass::exitLoad(1) );
}


int
MEIoMe3d::exitSave( int )
{
  cout << "loaded " << _objCount << " Objects" << endl;
  _inpf.close();
  return( 1 );
}


int
MEIoMe3d::getFormatInfo( int type )
{
  return( 0 ); //TODO return a FormatInfo object
}


int
MEIoMe3d::initLoad( const char *fname )
{
  int result;

  _inpf.open( fname );
  if( !_inpf.good() )
    {
      printf("# Open Input failed. Error %ld.\n", (long)result);
      return( READ_ERROR );
    }
  else
    {
      char buffer[80];
      _inpf.setf( ios::skipws );
      _inpf.get( buffer, 80, '\n' );
      if( strstr(buffer, ME3D_HEADER) != 0 )
	return( READ_SUCCESS );
      else
	{
	  printf("# Reading the header failed. Error %ld.\n",(long)result);
	  return( READ_ERROR );
	}
    }
}

int
MEIoMe3d::initSave( const char *fname, int )
{
  _outpf.open( fname );
  if( !_outpf.good() )
    {
      puts( "# Open Output failed" );
      return( READ_ERROR );
    }
  else
    {
      return( WRITE_SUCCESS );
    }
}

int
MEIoMe3d::loadObjects( G3DGroup *group )
{
  char name[100];

  _inpf >> setw(100) >> name;
  _inpf.ignore( 100000, '{' );
  if( strcmp(name, "Cameras") == 0 )
    {
      _inpf.ignore( 100000, '}' );
      //if( _world )
      //G3DCameraList *clist = _world->getCameras();
      _inpf >> setw(100) >> name;
      _inpf.ignore( 100000, '{' );
    }
  if( strcmp(name, "Scene") == 0 )
    {
      readObject( group );
    }
  else
    {  //TODO wrong format?
      G3DObject *obj = Kernel->createObject( name );
      if( obj )
	readObject( obj );
      else
	return( 0 );  //TODO set error
      group->addObject( obj );
    }
  return( 1 );
}

void
MEIoMe3d::readAttributes( G3DObject *obj )
{
  const Field *flist;
  Param       params[20];
  char        name[100];

  do {
    _inpf >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    G3DAttribute *attr = Kernel->createAttribute( name );
    if( attr )
      {
	_inpf.ignore( 100000, '{' );
	int num   = attr->getParameter( params );
	flist     = attr->getFieldList();
	readParameter( _inpf, params, flist, num );
	//attr->setParameter( params );
	if( obj->getType() & T_SHAPE )
	  ((G3DShape*)obj)->setAttribute( attr );
      }
    _inpf.ignore( 100000, '}' );
  } while( 1 );
}

void
MEIoMe3d::readObject( G3DObject *obj )
{
  const Field *flist;
  Param       params[20];
  char        name[100];  //TODO to big for recursion?

  do {
    _inpf >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    _inpf.ignore( 100000, '{' );

    if( strcmp(name, "description") == 0 )
      {
	int num   = obj->getParameter( params );
	flist     = obj->getFieldList();
	readParameter( _inpf, params, flist, num );
	obj->setParameter( params );
	_inpf.ignore( 100000, '}' );
      }
    else if( strcmp(name, "transformation") == 0 )
      {
	G3DVector3d v;
	_inpf >> v;
	obj->setScale( v );
	_inpf >> v;
	obj->setOrigin( v );
	_inpf >> v;
	obj->setAngles( v );
	_inpf.ignore( 100000, '}' );
      }
    else if( strcmp(name, "attributes") == 0 )
      {
	readAttributes( obj );
      }
    else if( strcmp(name, "subobjects") == 0 )
      {
	if( obj->getType() & T_GROUP )
	  readSubobjects( (G3DGroup*)obj );
      }
    else
      cout << "wrong format: " << name << endl;  //TODO error
  } while( 1 );
}

void
MEIoMe3d::readSubobjects( G3DGroup *grp )
{
  char  name[100];  //TODO to big for recursion?

  do {
    _inpf >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    _inpf.ignore( 100000, '{' );

    G3DObject *obj = Kernel->createObject( name );
    if( obj )
      {
	readObject( obj );
	grp->addObject( obj );
      }
    else
      cout << "unknown object: " << name << endl;  //TODO error
  } while( 1 );
}

EvalState
Me3dSaveAction::evaluate( G3DObject *obj )
{
  const Field *flist;
  int   num;

  outf_ << obj->getTypeName() << " {" << endl;
  {
    outf_ << " transformation { " << endl
          << "   " << obj->getScale() << endl
          << "   " << obj->getOrigin() << endl
          << "   " << obj->getAngles() << endl
	  << " }" << endl;
  }
  if( obj->getType() & T_SHAPE )
    {
      outf_ << " attributes {" << endl;
      G3DAListIterator iterator =
	((G3DShape*)obj)->getAttribute(TA_LOCAL_ATTRIBUTES);
      while( *iterator )
        {
	  if( !(*iterator)->isReference() )
	    {
	      outf_ << "  " << (*iterator)->getTypeName() << " {" << endl;
	      num   = (*iterator)->getParameter( params_ );
	      flist = (*iterator)->getFieldList();
	      writeParameter( outf_, params_, flist, num );
	      outf_ << "  }" << endl;
	    }
	  else
	    cout << "mio: ref" << endl;
	  ++iterator;
	}
      outf_ << " }" << endl;
    }

  outf_ << " description {" << endl;
  num   = obj->getParameter( params_ );
  flist = obj->getFieldList();
  writeParameter( outf_, params_, flist, num );
  outf_ << "  }" << endl;

  if( obj->getType() & T_GROUP )
    {
      outf_ << " subobjects {" << endl;
    }
  else
    outf_ << "  }" << endl;

  return( continueEval );
}

void
Me3dSaveAction::postAction( G3DObject* )
{
  outf_ << " }" << endl;
}

int
MEIoMe3d::saveObjects(  G3DScene *scene )
{
  _outpf << ME3D_HEADER << endl << endl;

  _outpf << "Cameras {" << endl;
  //TODO G3DCameraList *clist = _world->getCameras();
  //writeCameras( clist );
  _outpf << "}" << endl;

  Me3dSaveAction action( _outpf );
  scene->doAction( action );
  _outpf << "}" << endl;

  return( 0 );
}

void
MEIoMe3d::writeCamera( G3DCamera *cam )
{
  _outpf << " camera {" << endl;
  _outpf << "  " << cam->getName() << endl;
  _outpf << "  " << cam->getPosition() << endl;
  _outpf << "  " << cam->getLookAt() << endl;
  _outpf << "  " << cam->getFOV() << endl;
    //aspect
  _outpf << "  }" << endl;
}

void
MEIoMe3d::writeCameras( void )
{
  MCameraList *clist = _world->getCameras();
  MCameraList::iterator cam = clist->begin();
  while( cam != clist->end() )
    writeCamera( *(cam++) );
}


void
readParameter( ifstream &inf, Param *params, const Field *flist, int nfields )
{
  if( flist != NULL )
    {
      for( int i = 0; i < nfields; ++i )
	{
	  switch( flist->type )
	    {
	    case s_point3:
	      inf >> *(params->point3);  break;
	    case s_vector3:
	      inf >> *(params->vector3);  break;
	    case s_ctrlPt:
	      inf >> *(params->ctrlPt);  break;
	    case s_color:
	      inf >> *(params->color); break;
	    case s_dbl:
	      inf >> params->dbl; break;
	    case s_flt:
	      inf >> params->flt; break;
	    case s_number:
	      inf >> params->number; break;
	    case s_logic:
	      inf >> params->logic; break;
	      //TODO case n_string:
	      // inf >> params->string; break;

	    case m_point3:
	      {
		uint num;
		inf >> num;
		G3DPoint3d *pts;
		if( num > 0 )
		  pts = new G3DPoint3d[num];
		else
		  pts = NULL;
		params->point3 = pts; ++params;
		params->number = num;
		++flist; ++i;
		for( ; num > 0; --num )
		  {
		    inf >> *(pts);
		    ++pts;
                  }
		break;
	      }
	    case m_vector3:
	      {
		uint num;
		inf >> num;
		G3DVector3d *vec;
		if( num > 0 )
		  vec = new G3DPoint3d[num];
		else
		  vec = NULL;
		params->vector3 = vec;  ++params;
		params->number  = num;
		++flist;  ++i;
		for( ; num > 0; --num )
		  {
		    inf >> *(vec);
		    ++vec;
                  }
		break;
	      }
	    case m_vctrlPt:
	      {
		uint num;
		inf >> num;
		Vector_CtrlPnt4D *cpt = new Vector_CtrlPnt4D(num);
		inf >> *cpt;
		params->vCtrlPt = cpt;
		break;
	      }
	    case m_flt:
	      {
		uint num;
		inf >> num;
		float *vec;
		if( num > 0 )
		  vec = new float[num];
		else
		  vec = NULL;
		params->fltList = vec;  ++params;
		params->number  = num;
		++flist;  ++i;
		for( ; num > 0; --num )
		  inf >> *(vec++);
		break;
	      }
	    case m_number:
	      {
		uint num;
		inf >> num;
		uint *vec;
		if( num > 0 )
		  vec = new uint[num];
		else
		  vec = NULL;
		params->numList = vec;  ++params;
		params->number  = num;
		++flist;  ++i;
		for( ; num > 0; --num )
		  inf >> *(vec++);
		break;
	      }
	    case m_mctrlPt:
	      {
		uint rows, cols;
		inf >> rows >> cols;
		Matrix_CtrlPnt4D *cpt = new Matrix_CtrlPnt4D( rows, cols );
		inf >> *cpt;
		params->mCtrlPt = cpt;
		break;
	      }
	    }
	  ++flist;
	  ++params;
	}
    }
}

void
writeParameter( ofstream &outf, const Param *params, const Field *flist,
		int num )
{
  if( flist != NULL )
    {
      for( int i = 0; i < num; ++i )
	{
	  outf << "   "; //TODO ?? << flist->name << " ";
	  switch( flist->type )
	    {
	    case s_point3:
	      outf << *(params->point3) << endl;  break;
	    case s_vector3:
	      outf << *(params->vector3) << endl;  break;
	    case s_ctrlPt:
	      outf << *(params->ctrlPt) << endl;  break;
	    case s_color:
	      outf << *(params->color) << endl; break;
	    case s_dbl:
	      outf << params->dbl << endl; break;
	    case s_flt:
	      outf << params->flt << endl; break;
	    case s_number:
	      outf << params->number << endl; break;
	    case s_logic:
	      outf << params->logic << endl; break;
	    case s_string:
              outf << params->string << endl; break;  //TODO

	    case m_point3:
	      {
		G3DPoint3d *pt = params->point3;
		params++; flist++; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *pt << ' '; ++pt;
		  }
		outf << endl;
		break;
	      }
	    case m_vector3:
	      {
		G3DVector3d *vt = params->vector3;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *vt << ' '; ++vt;
		  }
		outf << endl;
		break;
	      }
	    case m_vctrlPt:
	      {
		Vector_CtrlPnt4D *cpt = params->vCtrlPt;
		outf << cpt->n << ' ';
		outf << *cpt << endl;
		break;
	      }
	    case m_mctrlPt:
	      {
		Matrix_CtrlPnt4D *cpt = params->mCtrlPt;
		outf << cpt->rows << ' ' << cpt->cols << ' ';
		outf << *cpt << endl;
		break;
	      }
	    case m_color:
	      // TODO  outf << *(params->color) << endl;
	      break;
	    case m_dbl:
	      {
		double *dbl = params->dblList;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *dbl << ' '; ++dbl;
		  }
		outf << endl;
		break;
	      }
	    case m_flt:
	      {
		float *flt = params->fltList;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *flt << ' '; ++flt;
		  }
		outf << endl;
		break;
	      }
	    case m_number:
	      {
		uint *num = params->numList;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *num << ' '; ++num;
		  }
		outf << endl;
		break;
	      }
	    }
	  ++flist;
	  ++params;
	}
    }
}

//====================================END================================
