/*
 * Project : "MindsEye, IO library"
 * Copyright (c)1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: nffobjects.h,v 1.11 1998/07/23 20:16:15 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERWIE
 * =======
 * Declaration of classes representing nff objects
 *
 */

#ifndef __NFFOBJECTS_H__
#define __NFFOBJECTS_H__

#include "fstream.h"
#include "3dlib/t_fields.h"

extern void prerror( char* );

class G3DShape;
class G3DCamera;
class G3DNurbsCurve;
class G3DNurbsSurface;
class G3DATexture;


/**Reads and writes an nff description of the surface parameter.
 */
class NFFMaterial
{
public:
  NFFMaterial( int );
  ~NFFMaterial( );

  G3DAttribute* buildG3DColor( void );
  void read( ifstream &inpf );
  void write( ofstream &opf, G3DAttribute* );

  enum{ inp, outp };

protected:
  G3DAColor *_mat, *_last;
  short     _orig;
};


/**Dummy class to reference every NFF Object via a common pointer
 */
class NFFObject
{
public:
  NFFObject( void ) { _ok = 1; };
  virtual ~NFFObject( ) {};

  virtual G3DShape* buildG3DShape( void );
  int  isValid( void ) { return( _ok ); };

protected:
  int _ok;
};


/**Read and write an nff Box object.
 */
class NFFBox : public NFFObject
{
public:
  virtual G3DShape *buildG3DShape( void );
  void read( ifstream &inpf );
  static void write( ofstream &outf, Param *list, G3DMatrix &m );

protected:
  double x, y, z, x1, y1, z1;
};


class NFFCamera : public NFFObject
{
public:
  NFFCamera( ifstream &inpf, char* );

  G3DCamera* buildG3DCamera( void );
  static void write( ofstream &opf, const G3DCamera *cam );

protected:
  double fromx, fromy, fromz;
  double atx, aty, atz;
  double upx, upy, upz;
  double zoom;
  double near;

};

/**Load and write cone and cylinder objects. A cylinder is a cone with
   identical top and bottom radius.
   */
class NFFCone : public NFFObject
{
public:
  virtual G3DShape* buildG3DShape( void );
  void   read( ifstream &inpf );
  static void write( ofstream &outf, Param *list, G3DMatrix &m, int );

protected:
  float bX, bY, bZ, bR, aX, aY, aZ, aR;
};

/**A NURBS curve. This is not strictly part of the NFF format, but
  an enhancement. The format is:
  pc
  degree, num_knots, num_points
  */
class NFFCurve : public NFFObject
{
public:
  NFFCurve( void ) { _curve = NULL; _used = 0; }
  virtual ~NFFCurve( );
  virtual G3DShape* buildG3DShape( void );
  void    read( ifstream &inpf );
  static  void write( ofstream &outf, Param *list, G3DMatrix &m );

protected:
  G3DNurbsCurve *_curve;
  short         _used;
};

/** */
class NFFLight : public NFFObject
{
public:
  NFFLight( ifstream &inpf, char* );
  G3DLight* buildG3DLight( void );

protected:
  G3DPoint3d pos;
  float cx, cy, cz;

};


class NFFNurbs : public NFFObject
{
public:
  NFFNurbs( void ) { nurbs_ = NULL; used_ = false; }
  virtual ~NFFNurbs( );
  virtual G3DShape* buildG3DShape( void );
  void    read( ifstream &inpf );
  static  void write( ofstream &outf, Param *list, G3DMatrix &m );

protected:
  G3DNurbsSurface *nurbs_;
  bool            used_;
};


class NFFPPatch : public NFFObject
{
public:
  NFFPPatch( void );
  ~NFFPPatch();

  virtual G3DShape* buildG3DShape( void );
  void read( ifstream &inpf );

protected:
  G3DPoint3d  *vertices_;
  G3DVector3d *normals_;
  int          tNum_, tMax_;

};


class NFFPolygon : public NFFObject
{
public:
  NFFPolygon( void );
  ~NFFPolygon( );

  virtual G3DShape* buildG3DShape( void );
  void   read( ifstream &inpf );
  static void write( ofstream &outf, Param *list, G3DMatrix &m );

protected:
  G3DPoint3d  *vertices_;
  int         tNum_;

};


class NFFSphere : public NFFObject
{
public:
  virtual G3DShape* buildG3DShape( void );
  void    read( ifstream &inpf );
  static  void write( ofstream &outf, Param *list, G3DMatrix &m );

protected:
  float x, y, z, r;

};


#endif
