/*
 * Project : "MindsEye, IO Library"
 * Copyright (c) 1997. All Rights Reserved.
 * Author(s): Nicholas Yue (nicholas_yue@acm.org)
 *
 * $Id: io_rib.cpp,v 1.10 1998/07/23 20:16:15 source Exp $
 *
 * COPYRIGHT
 * ========= 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 
 * OVERVIEW
 * ========
 * I/O lib for writing Renderman Interface Bit Stream format files.
 */
#include <time.h>
#include <assert.h>

#include "3dlib/scene.h"
#include "3dlib/camera.h"
#include "io_rib.h"
#include "kernel/world_obj.h"


IO_BaseClass*
createRibFilter( void )
{
  return( new IO_RIB );
}


IO_RIB::IO_RIB()
{
}

IO_RIB::~IO_RIB()
{
}

int IO_RIB::check_header(char *fname)
{
  return IO_BaseClass::check_header(fname);
}

int IO_RIB::getFormatInfo(int type)
{
  return 0;
}

int IO_RIB::exitSave(int result)
{
  _outpf.close();
  return 0;
}

int IO_RIB::initLoad(const char *fname)
{
  _outpf.close();
  return 0;
}

int  IO_RIB::exitLoad(int result)
{
  _outpf.close();
  return IO_BaseClass::exitLoad(result);
}


int IO_RIB::initSave(const char *fname, int flag)
{
  _outpf.open(fname);
  strncpy(filename,fname,255);
  if (!_outpf.good()) {
    puts( "# Open Output failed" );
    return( READ_ERROR );
  } else
    return( WRITE_SUCCESS );
}

int IO_RIB::loadObjects( G3DGroup* )
{
  return 0;
}

int IO_RIB::saveObjects(G3DScene *group)
{
  register unsigned int i,j,k,n;
  G3DShape* objectRef;
  G3DCamera* cameraRef;
  IO_RIB_Lights itl(_outpf);
  itl.resetLightCounter();
  IO_RIB_Geometry itg(_outpf);

  float radius,zmin,zmax,thetamax,alpha,beta,gamma,mag,umin,umax,vmin,vmax;
  time_t t;
  int nu,nv,nrow,ncol,uorder,vorder;
  time(&t);
  
  // RIB file Header
  _outpf << "##RenderMan RIB-Structure 1.0" << endl;
  _outpf << "##Scene " << endl;
  _outpf << "##CreationDate " << ctime(&t);
  _outpf << "##Creator MindsEye 0.5.20 April, 1998" << endl;
  _outpf << "##Frames 1" << endl;
  _outpf << "##Shaders plastic" << endl;
  _outpf << "##CapabilitiesNeeded Programmable Shaders, Trim Curve" << endl;
  _outpf << "##RIB I/O By Nicholas Yue nicholas_yue@acm.org" << endl;
  _outpf << "version 3.03" << endl;
  // Camera Header
  //maku  cameraRef = (_world->getCameras())->getElement( 0 );
  cameraRef = _world->getActiveCamera();
  _outpf << "Display \"ri.tif\" \"file\" \"rgb\"" << endl;
  writeCamera(cameraRef);


  // RIB Frame Count
  _outpf << "FrameBegin 1" << endl;

  // RIB World Definition

  _outpf << "WorldBegin" << endl;
  // Main contents

  // RIB Lights
  // itl.incrementLightCounter();
  group->doWithLights(itl);

  // RIB Lights
  group->doWithShapes(itg);

  // Trailing Stuff
  _outpf << "WorldEnd" << endl;
  _outpf << "FrameEnd" << endl;

  return 0;
}


void IO_RIB_Geometry::minmax(const float* input,unsigned int size,float& min,float& max)
{
  min = MAXFLOAT;
  max = -MAXFLOAT;
  for (register int i=0;i<size;i++) {
    if (input[i]>max)
      max = input[i];
    if (input[i]<min)
      min = input[i];
  }
}

void IO_RIB::writeCamera(G3DCamera* cameraRef)
{
  // Refer to "The RenderMan Companion : A Programmer's Guide to Realistic
  // Computer Graphics", Chapter 8, Steve Upstill, Addison Wesley, 1992, ISBN
  // 0-201-50868-0 for more details
  G3DPoint3d cameraFrom, cameraTo;
  G3DVector3d cameraDirection;
  float xzlen,yzlen,yrot,xrot;

  // Do the calculation
  cameraFrom = cameraRef->getPosition();
  cameraTo = cameraRef->getLookAt();
  cameraDirection = cameraRef->getDirection();
  xzlen = sqrt(cameraDirection.x()*cameraDirection.x()+
	       cameraDirection.z()*cameraDirection.z());
  if (fabs(xzlen)<MINFLOAT)
    yrot = (cameraDirection.y()<0)?180:0;
  else
    yrot = 180.0*acos(cameraDirection.z()/xzlen)/M_PI;

  yzlen = sqrt(cameraDirection.y()*cameraDirection.y()+xzlen*xzlen);
  assert(fabs(yzlen)>MINFLOAT);
  xrot = 180.0*acos(xzlen/yzlen)/M_PI;

  // Write out the information
  _outpf << "Projection \"perspective\"  \"fov\" "
	 << cameraRef->getFOV() << endl;
  if (cameraDirection.y()>0)
    _outpf << "Rotate " << xrot << " 1 0 0" << endl;
  else
    _outpf << "Rotate " << -xrot << " 1 0 0" << endl;
  if (cameraDirection.x()>0)
    _outpf << "Rotate " << -yrot << " 0 1 0" << endl;
  else
    _outpf << "Rotate " << yrot << " 0 1 0" << endl;
    
  _outpf << "Translate " << cameraFrom << endl;
  
}

bool IO_RIB_Lights::exec(G3DLight& t)
{
  G3DPoint3d location;
  _lightCounter++;
  if( t.getType() & T_LIGHT ) {
    _outf << "# CONTAINS LIGHT " << endl;
    switch (t.getType()) {
    case T_AMBIENT_LIGHT :
      _outf << "# T_AMBIENT_LIGHT encountered" << endl;
      _outf << "LightSource \"ambientlight\" " << _lightCounter << " \"intensity\" [1] " << endl;
      break;
    case T_DIRECTIONAL_LIGHT :
      _outf << "# T_DIRECTIONAL_LIGHT encountered" << endl;
      break;
    case T_POINT_LIGHT :
      _outf << "# T_POINT_LIGHT encountered" << endl;
      t.getPosition(location);
      _outf << "LightSource \"pointlight\" " << _lightCounter << " \"intensity\" [1] \"lightcolor\" [1 1 1] \"from\" [ "
	    << location.x() << " "
	    << location.y() << " "
	    << location.z() << " ]" << endl;
      break;
    case T_SPOT_LIGHT :
      _outf << "# T_SPOT_LIGHT encountered" << endl;
      break;
    case T_AREA_LIGHT :
      _outf << "# T_AREA_LIGHT encountered" << endl;
      break;
    }
  }
  return( false );
}

bool IO_RIB_Geometry::exec(G3DShape& t)
{
  if( !(t.getType() & T_GROUP) ) {
    G3DMatrix    m;
    G3DAListIterator atr = t.getAttribute( TA_TEXTURE );
    // _mat.write( _outf, atr );
    t.getParameter( _params );
    t.getWorldTrafo( m );
    switch( t.getType() ) {
    case T_BLOB:
      writeBlob( m );
      break;
    case T_BOX:
      writeBox( m );
      break;
    case T_CONE:
      writeCone( m );
      break;
    case T_CYLINDER:
      writeCylinder( m);
      break;
    case T_DIFFERENCE:
      writeDifference( m);
      break;
    case T_DISK:
      writeDisk( m);
      break;
    case T_INTERSECTION:
      writeIntersection( m);
      break;
    case T_NURBS_CURVE:
      writeNurbsCurve( m );
      break;
    case T_NURBS_SURFACE:
      writeNurbsSurface( m );
      break;
    case T_POINT:
      writePoint( m );
      break;
    case T_POLYGON:
      writePolygon( m );
      break;
    case T_SPHERE:
      writeSphere( m );
      break;
    case T_TORUS:
      writeTorus( m );
      break;
    case T_UNION:
      writeUnion( m );
      break;
    default :
      break;
    }
  }
  return( false );
}

void IO_RIB_Geometry::writeBox(G3DMatrix& m)
{
  // G3DS_Box bref = objectRef;
  _outf << "AttributeBegin # Box\n";
  writeTransformation(m);
  writeShader();
  _outf << "Translate " <<
    _params[0].vector3->x() << "\t" <<
    _params[0].vector3->y() << "\t" <<
    _params[0].vector3->z() << endl;
  _outf << "Patch # Top\n"
	 <<	"\"bilinear\" \"P\" [ "
	 << -_params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << "]" << endl;
  _outf << "Patch # Bottom\n"
	 <<	"\"bilinear\" \"P\" [ "
	 << -_params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << "]" << endl;
  _outf << "Patch # Left\n"
	 <<	"\"bilinear\" \"P\" [ "
	 << -_params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << "]" << endl;
  _outf << "Patch # Right\n"
	 <<	"\"bilinear\" \"P\" [ "
	 << _params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << "]" << endl;
  _outf << "Patch # Front\n"
	 <<	"\"bilinear\" \"P\" [ "
	 << -_params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << _params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << "]" << endl;
  _outf << "Patch # Back\n"
	 <<	"\"bilinear\" \"P\" [ "
	 << -_params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << -_params[2].vector3->z() << " "
    //===================================
	 << -_params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << _params[3].vector3->x() << " "
	 << -_params[1].vector3->y() << " "
	 << _params[2].vector3->z() << " "
    //===================================
	 << "]" << endl;
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writeCone(G3DMatrix& m)
{
  register int i;
  float mag,radius,height,thetamax;

  _outf << "AttributeBegin # Cone\n";
  writeTransformation(m);
  writeShader();
  /*
  _outf << "Translate " << _params[0].vector3->x()
	 << "\t" << _params[0].vector3->y() << "\t"
	 << _params[0].vector3->z() << endl;
  mag = sqrt(_params[1].vector3->x() * _params[1].vector3->x() +
	     _params[1].vector3->y() * _params[1].vector3->y() +
	     _params[1].vector3->z() * _params[1].vector3->z());
  radius = sqrt(_params[3].vector3->x() * _params[3].vector3->x() +
		_params[3].vector3->y() * _params[3].vector3->y() +
		_params[3].vector3->z() * _params[3].vector3->z());
		*/
  
  _outf << "Translate " <<
    _params[0].point3->x() << " " <<
    _params[0].point3->y() << " " <<
    _params[0].point3->z() << endl;

  G3DVector3d rotationAxis = _params[1].vector3->cross(G3DVector3d(0,1,0));
  height = sqrt (
	      (_params[1].vector3->x()*_params[1].vector3->x()) +
	      (_params[1].vector3->y()*_params[1].vector3->y()) +
	      (_params[1].vector3->z()*_params[1].vector3->z()) );
  float angle = acos(_params[1].vector3->y()/height);
  _outf << "Rotate " <<
    angle << " " <<
    rotationAxis.x() << " " <<
    rotationAxis.y() << " " <<
    rotationAxis.z() << endl;

  radius = _params[2].dbl;

  thetamax = 360.0;
  _outf << "Cone " << height << " "
	 << radius << " " << thetamax  << endl;
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writeCylinder(G3DMatrix& m)
{
  register int i;
  float mag,radius,zmin,zmax,thetamax;

  _outf << "AttributeBegin # Cylinder\n";
  writeTransformation(m);
  writeShader();
  _outf << "Translate " << _params[0].vector3->x()
	 << "\t" << _params[0].vector3->y() << "\t"
	 << _params[0].vector3->z() << endl;
  mag = sqrt(_params[1].vector3->x() * _params[1].vector3->x() +
	     _params[1].vector3->y() * _params[1].vector3->y() +
	     _params[1].vector3->z() * _params[1].vector3->z());
  radius = sqrt(_params[3].vector3->x() * _params[3].vector3->x() +
		_params[3].vector3->y() * _params[3].vector3->y() +
		_params[3].vector3->z() * _params[3].vector3->z());
  zmin = 0.0;
  zmax = mag;
  thetamax = 360.0;
  _outf << "Cylinder " << radius << " "
	 << zmin << " " << zmax << " " << thetamax  << endl;
  _outf << "Disk 0 " << radius << " " << thetamax  << endl;
  _outf << "Disk " << mag << " " << radius << " " << thetamax  << endl;
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writePolygon(G3DMatrix& m)
{
  _outf << "AttributeBegin # Polygon\n";
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writeSphere(G3DMatrix& m)
{
  register int i;
  float mag,radius,zmin,zmax,thetamax;

  _outf << "AttributeBegin # Cylinder\n";
  writeTransformation(m);
  writeShader();
  _outf << "Translate " << _params[0].vector3->x()
	 << "\t" << _params[0].vector3->y() << "\t"
	 << _params[0].vector3->z() << endl;
  radius = _params[1].dbl;
  zmin = -radius;
  zmax = radius;
  thetamax = 360.0;
  _outf << "Sphere " << radius << " "
	 << zmin << " " << zmax << " " << thetamax  << endl;
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writeNurbsCurve(G3DMatrix& m)
{
  _outf << "AttributeBegin # NurbsCurve\n";
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writeNurbsSurface(G3DMatrix& m)
{
  int nu,nv,nrow,ncol,uorder,vorder;
  float umin,umax,vmin,vmax;
  register int i,j,k,n;

  _outf << "AttributeBegin # NurbsSurface\n";
  writeTransformation(m);
  writeShader();
  nu=(int)(_params[2].number);
  nv=(int)(_params[4].number);
  uorder = (int)(_params[0].flt)+1;
  vorder = (int)(_params[1].flt)+1;
  minmax(_params[3].fltList,
	 nu,
	 umin,umax);
  minmax(_params[5].fltList,
	 nv,
	 vmin,vmax);
  nrow=(int)(_params[6].number);
  ncol=(int)(_params[7].number);
  _outf << "NuPatch\n";
  
  _outf << nu - uorder << endl;
  
  _outf << uorder << endl;
  _outf << "[ " ;
  for (i=0;i<nu;i++)
    _outf << _params[3].fltList[i] << endl;
  _outf << "]" << endl;
  _outf << umin << " " << umax <<endl;
  
  _outf << nv - vorder << endl;
  _outf << vorder << endl;
  _outf << "[ " ;
  for (i=0;i<nv;i++)
    _outf << _params[5].fltList[i] << endl;
  _outf << "]" << endl;
  _outf << vmin << " " << vmax <<endl;
  n=0;
  _outf << "\"Pw\" [ " ;
  for (i=0;i<nrow;i++) {
    for (j=0;j<ncol;j++) {
      for (k=0;k<4;k++)
	_outf << _params[8].fltList[n++] << " ";
      _outf << endl;
    }
    _outf << endl;
  }
  _outf << "]" << endl;
  _outf << "AttributeEnd # NurbsSurface\n";
}

void IO_RIB_Geometry::writeBlob(G3DMatrix& m)
{
  _outf << "AttributeBegin # Blob\n";
  writeTransformation(m);
  writeShader();
  _outf << "AttributeEnd # Blob\n";
}

void IO_RIB_Geometry::writeDifference(G3DMatrix& m)
{
  _outf << "SolidBegin \"difference\" # Difference\n";
  _outf << "SolidEnd # Difference\n";
}

void IO_RIB_Geometry::writeDisk(G3DMatrix& m)
{
  float height,radius,thetamax;

  height = 1.0;
  radius = 1.0;
  thetamax = 360.0;

  _outf << "AttributeBegin # Blob\n";
  writeTransformation(m);
  writeShader();
  _outf << "Disk "
	<< height << " "
	<< radius << " "
	<< thetamax
	<< endl;
  _outf << "AttributeEnd # Blob\n";
}

void IO_RIB_Geometry::writeIntersection(G3DMatrix& m)
{
  _outf << "SolidBegin \"intersection\" # Intersection\n";
  _outf << "SolidEnd # Intersection\n";
}

void IO_RIB_Geometry::writePoint(G3DMatrix& m)
{
  _outf << "AttributeBegin\n";
  writeTransformation(m);
  writeShader();
  _outf << "Points \"P\" [ 1 1 1 ] \"width\" [1] " << endl;
  _outf << "AttributeEnd\n";
}

void IO_RIB_Geometry::writeTorus(G3DMatrix& m)
{
  float majorrad,minorrad,phimin,phimax,thetamax;

  phimin = -180.0;
  phimax = 180.0;
  thetamax = 360.0;
  _outf << "AttributeBegin # Torus\n";
  writeTransformation(m);
  writeShader();
  _outf << "Translate " <<
    _params[0].point3->x() << "\t" <<
    _params[0].point3->y() << "\t" <<
    _params[0].point3->z() << endl;

  majorrad = _params[1].dbl;
  minorrad = _params[2].dbl;

  _outf << "Torus "
	<< majorrad << " "
	<< minorrad << " "
	<< phimin << " "
	<< phimax << " "
	<< thetamax << " "
	<< endl;
  _outf << "AttributeEnd # Torus\n";
}

void IO_RIB_Geometry::writeUnion(G3DMatrix& m)
{
  _outf << "SolidBegin \"union\" # Union\n";
  _outf << "SolidEnd # Union\n";
}

void IO_RIB_Geometry::writeTransformation( G3DMatrix& m )
{
  register int i;
  const double *p = m;
  _outf << "ConcatTransform \n";
  _outf << "[ ";
  for (i=0;i<16;i++)
    _outf << p[i] << " ";
  _outf << "]" << endl;
}

void IO_RIB_Geometry::writeShader()
{
  _outf << "Surface \"plastic\"" << endl;
  _outf << "Color [ .2 1 .1 ]" << endl;
}
