/*
 * vi:ts=4:sw=4:
 *
 * Project : "Config Manager"
 * Copyright 1997. All Rights Reserved.
 * 
 * $RCSfile: m_config.h,v $
 *
 * $Revision: 1.1.1.1 $
 * 
 * $Author: source $ 
 * 
 * $Locker:  $
 * 
 * $State: Exp $
 * 
 * 
 * COPYRIGHT
 * =========
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *
 * OVERVIEW
 * ========
 *
 * See the readme file for detailed information on how these sources work
 *
 */
 
#ifndef __GENERAL_CONFIG_H__
#define __GENERAL_CONFIG_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
   
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <time.h>

/// show debug info when parsing
//#define __CREATE_DEBUG__    1

/// define version number
#define VERSION_MAJOR       0
/// define version revision
#define VERSION_MINOR      84

#ifndef PATH_MAX
#define PATH_MAX         1024
#endif

/// maximum number of entries to one variable  
#define MAX_VARS           50

/// no config file found, created new one ( not fatal )
#define NO_CONFIG          -2
/// user's home dir not found ( fatal )
#define NO_HOMEDIR         -1                          
/// everything checks out 
#define CONFIG_OK           0
/// there was an error while creating the config file ( fatal )
#define ERROR_WRITING       1
/// there was an error while creating the config directory ( fatal )
#define ERROR_CREATEDIR     2 

#define DEFAULT_INT         0
#define DEFAULT_FLOAT    0.00
#define DEFAULT_STRING "none"

#define DEFAULT_CONFIG ".config"

//>------- define some behavioural #defines for future use

#define DEF_DESTRUCTWRITE 1 // write file when the destructor is called
#define DEF_NOTEXISTWRITE 2 // write a file when there is no file found when constructing

/// conf_entry is the actual data structure used for the variable entries
struct conf_entry
{
 int   nr_entries;

 char *var_name;
 char *var_types;
 void *var_array [MAX_VARS];
 
 struct conf_entry *next;
};

/// group_entry is the data structure which holds several conf_entries
struct group_entry
{
 char  *group_name;

 struct conf_entry *var_top;
 struct conf_entry *var_current;

 struct group_entry *next;
};

/// This is the main Config Manager class
class ConfigClass
{

  public:

    //>-------- public vars
    
    /// can be used to determen what's going on inside the constructor 
    char config_status;

    /// pointer to next ConfigClass structure
    ConfigClass *next;

    //>-------- public member functions 

    /// Default constructor, called internally when creating child   
    ConfigClass ();
    /// Takes the name of a config file as an argument
    ConfigClass (char *config_file);
    /// Use this one when you want the use the name of the program as config file
    ConfigClass (int argc,char **argv);
    /// Takes the name of a config file as an argument and also determens package name
    ConfigClass (char *config_file,char *progr_name);
    /// This config class is used by plugins to create a child instance
    ConfigClass (ConfigClass *parent);
    
    /// Writes the config file and destroys the data used.
    ~ConfigClass ();
    
    /**
        Adds a new group to the data-tree. It is only used internally at parse 
        time.
    */
    void insert_group (char *);

    /**
        Adds a new entry to the data-tree. It is only used internally at parse-   
        time.
    */
    void insert_entry (char *);

    /**
        This is function wich will be called when the user wants to add an entry  
        to the current config. The function will first see if it can match the    
        request to an existing variable. If it can't find a suitable variable     
        then a new one will be created.                                           
    */
    void insert_entry (char *name,char *var_string,...);

    /**
        You can use this function if you feel confident you know where and how
        the data is located inside the config data structure.
    */
    void insert_entry_to_group (char *name,char *group,char *var_string,...);

    /**
        This function is the config wrapper for the flex parser function. it will
        only be called by the constructor. Other calls in the program will cause it
        to clear the entire data-structure and start all over.
    */    
    int  read_config (void);

    /**
        Force the writing of the entire config file. Used by the destructor to 
        create the config file. Can also be used in other parts. 
    */
    int  write_config (void);
    
    /**
        Determens what the contents of the config primitives will be if an entry
        is added that was not already in the config file.
    */
    void set_defaults (int,float,char *);
    
    /// Creates a new group entry for this instance of the config manager
    void create_group (char *token);
    /// Creates a new variable in floating group in the current config instance
    void create_var (char *token);

    /**
       Creates a new integer entry in variable pointed to by: 
       "floating_entry->var_array []"
    */ 
    void add_integer (int token);

    /**
       Creates a new float entry in variable pointed to by: 
       "floating_entry->var_array []"
    */   
    void add_float (float token);

    /**
       Creates a new string entry in variable pointed to by: 
       "floating_entry->var_array []"
    */
    void add_string (char *token);
   
  private:
  
    //>-------- private vars
     
    ConfigClass *parent;
    
    /// if the there are plugins to this instance, then this pointer != NULL
    ConfigClass *plugin_top;     
    /// if the there are plugins to this instance, then this pointer is the end of the list    
    ConfigClass *plugin_current;

    /// default value for integers    
    int   default_int;
    /// default value for floats
    float default_float;
    /// default string, will be copied to target
    char *default_string;
  
    /// contains absolute path to user's home directory
    char *home_string;  
    /// this contains the name of the package/program, is optional however
    char *program_name;
    /// name of the config file used by this instance
    char *file_name;
    /// pointer to "file_name"
    FILE *file_pointer;
    
    /// top of groups for this instance
    struct group_entry *group_top;
    /// pointer to end of groups list
    struct group_entry *group_current;
    
    /// pointer used by the matching system
    struct group_entry *floating_group; 
    /// pointer used by the matching system
    struct conf_entry  *floating_entry;
        
    //>-------- private member functions 

    /**
        Steps through all variable entries in the group. It starts at the variable
        pointed to by "temp_entry" and writes to "temp_file"   
    */         
    int    write_entries (FILE *,struct conf_entry *); 

    /**
       This is an internal function that is called when the config file is 
       parsed at construct time. It will assume the pointer "group_top"
       is valid and try to use it to match an existing group
    */
    group_entry *find_group (char *);

    /**
       This is an internal function that is called when the config file is 
       parsed at construct time. It will assume the pointer "floating_group"
       is valid and try to use it to match an existing entry.
    */
    conf_entry *find_entry (char *);    

    /**
        When the ConfigClass constructor is called with a "parent" as an argument
       this function will create the appropriate links in the parent instance.
    */    
    void add_child  (ConfigClass *);

    /**
       Construct a valid string pointer to the user's home directory
    */    
    int create_homedir  (void);    
    /**
       Create valid string containing the resulting configuration directory. This
       If this directory does not exists, it will be created and a README file 
       will be written stating the nature of the config directory.       
    */     
    int create_confdir  (char *);

    /**
       Create a valid string pointer containing the absolute path to the config
       file used by the config instance. The string will stay valid until the
       destructor is called.
    */
    char *create_conffile (char *);

    /**
       Remove entire data-tree including all child config instances    
    */    
    void   clear_data      (void);
};

#endif // __GENERAL_CONFIG_H__
  