/*
 * Project : "MindsEye"
 * Copyright 1996 - 1998. All Rights Reserved.
 * Author(s): M. van Velsen, M. Kuball
 *
 * $Id: m_core.cpp,v 1.17 1998/07/23 20:16:17 source Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * OVERVIEW
 * ========
 *
 */

#ifdef HAVE_CONFIG_H
#include "include/config.h"
#endif	 // HAVE_CONFIG_H

#include <qpixmap.h>

#include "3dlib/shape.h"
#include "3dlib/camera.h"
#include "3dlib/s_line.h"
#include "3dlib/s_nurbsurf.h"
#include "dialogs/about/m_aboutdlg.h"
#include "io/io_dialogs.h"
#include "material/m_mat.h"
#include "modeler/modeler.h"
#include "network/m_network.h"
#include "nurbsedit/shaper.h"
#include "treeedit/scene_dlg.h"
#include "viewer/m_viewer.h"
#include "world_obj.h"
#include "m_core.h"
#include "m_console.h"
#include "m_kmap.h"
#include "world_list.h"
#include "include/m_messg.h"
#include "include/version.h"
#include "plugin/dll.h"
#include "init_queues.h"
#include "init_widgets.h"


//The global visible instance of the kernel
MCore *Kernel = NULL;


/*---------------------------------------------------------------------------*/
MCore::MCore( int argc, char **argv )  //TODO no arguments
{ 
 // init kernel vars
  the_console    = NULL;
  worldList_.push_back( new MWorldObject );

  //TODO move this to setup setup_modules() ?
  for (int i = 0; i < NR_MODULES; i++)
    {
      modules [i].load_type      = LOAD_ON_CALL; // cool if the user can change this
      modules [i].the_module     = NULL;
      modules [i].the_module_inf = NULL;
      modules [i].toggled        = FALSE;
    }
  //TODO it should be possible to start more than one instance of a module

  kernel_blocked = FALSE; // now we're ready :)
}

MCore::~MCore (void)
{
  puts( "delete mcore" ); //TODO the destructor is never called??
  // shutdown all running modules ......
  for( int i = 1; i < NR_MODULES; i++ ) // skip console
    {
      if( modules[i].toggled == TRUE )
	{
	  printf( "Killing module: [%s]\n",
		  modules[i].the_module_inf->module_name );
	  delete modules[i].the_module;
	  delete modules[i].the_module_inf;
	}
    }
  delete pluginMgr_;
  delete message_system;
}
/*--------------------------------------------------------------------------*/
void
MCore::addAttributeFactory( const char *name, AcreateFunc *func )
{
  attributeFactories_[name] = (CreateFunc*)func;
}

void
MCore::addAttributeWidget( TMatWidget *widget )
{
  attributeWidgets_.push_back( widget );
}

void
MCore::addObjectClass( const char* name )
{
  objClassList_[name] = new ObjList;
}

void
MCore::addObjectFactory( const char *name, OcreateFunc *func )
{
  objectFactories_[name] = (CreateFunc*)func;
}

void
MCore::addObjectName( const char *cname, const char *name, const char *type )
{
  map<const char*, ObjList*, ltstr>::iterator itr
    = objClassList_.find( cname );
  if( itr != objClassList_.end() )
    {
      (*((*itr).second))[name] = type;
    }
}

void
MCore::addRenderQueue( MERenderQueue *queue )
{
  renderQueues_.push_back( queue );
}

void
MCore::addWorld( void* data )
{
  //the_console->m_puts( "A world was loaded" );
  if( data )
    worldList_.push_back( (MWorldObject*)data );
  emit worldListChanged();
}

/*--------------------------------------------------------------------------*/

void
MCore::close_module( int module_id )
{
  if( modules[module_id].load_type == LOAD_ON_CALL )
    {
      the_console->m_printf( "Module [%s] is shutting down",
			     modules [module_id].the_module_inf->module_name );
      delete modules[module_id].the_module;
      modules[module_id].the_module     = NULL;
      modules[module_id].the_module_inf = NULL;
      modules[module_id].toggled        = FALSE;
    }
  else
    {
      the_console->m_printf( "Module [%s] is closing",
			     modules [module_id].the_module_inf->module_name);
      modules[module_id].the_module->hide ();
      modules[module_id].toggled = FALSE;
    }
}

G3DAttribute*
MCore::createAttribute( const char *classname )
{
  map<const char*, CreateFunc*, ltstr>::iterator itr
    = attributeFactories_.find( classname );
  if( itr != attributeFactories_.end() )
    return( (G3DAttribute*)(((*itr).second)()) );
  else
    return( NULL );
}

G3DObject*
MCore::createObject( const char *classname )
{
  map<const char*, CreateFunc*, ltstr>::iterator itr
    = objectFactories_.find( classname );
  if( itr != objectFactories_.end() )
    return( (G3DObject*)(((*itr).second)()) );
  else
    return( NULL );
}

vector<void*>&
MCore::getAttributeWidgets( void )
{
  return( attributeWidgets_ );
}

MEModule*
MCore::getModule( int id )
{
  if( (id >= 0) && (id < NR_MODULES) );
  return( modules[id].the_module );
}

const char**
MCore::getObjClassList( void )
{
  const char **ret = new (const char*)[objClassList_.size()+1];
  const char **tmp = ret;
  map<const char*, ObjList*, ltstr>::iterator it = objClassList_.begin();
  while( it != objClassList_.end() )
    *(tmp++) = (*(it++)).first;
  *tmp = NULL;
  return( ret );
}

const char**
MCore::getObjList( const char *cname )
{
  map<const char*, ObjList*, ltstr>::iterator itr = objClassList_.find(cname);
  if( itr != objClassList_.end() )
    {
      ObjList *lst = (*itr).second;
      const char **ret = new (const char*)[lst->size()+1];
      const char **tmp = ret;
      ObjList::iterator it = lst->begin();
      while( it != lst->end() )
	*(tmp++) = (*(it++)).first;
      *tmp = NULL;
      return( ret );
    }
  else
    {
      const char **ret = new (const char*)[1];
      *ret = NULL;
      return( ret );
    }
}

const char*
MCore::getObjTypeName( const char *cname, const char *desc )
{
  map<const char*, ObjList*, ltstr>::iterator itr = objClassList_.find(cname);
  if( itr != objClassList_.end() )
    {
      ObjList *lst = (*itr).second;
      ObjList::iterator it = lst->find( desc );
      if( it != lst->end() )
	return( (*it).second );
    }
  return( NULL );
}
/*
void
MCore::getObjectWidget( G3DShape *obj )
{
  PropertyDialog *tmp;

  tmp = NULL; //_propertyDialogs[obj->getType()];
  //if( tmp )
  //tmp->show();
  //TODO else show some generic dialog
}
*/

vector<void*>&
MCore::getRenderQueues( void )
{
  return( renderQueues_ );
}

MCore*
MCore::loadKernel( int argc, char **argv )
{
  if( Kernel == NULL )
    Kernel = new MCore( argc, argv );

  // go through the various setup stages
  Kernel->setup_app( argc, argv );
  Kernel->setup_chooser ();  //TODO setup_modules() ?
  Kernel->setup_console ();  //  "
  Kernel->setup_sns ();
  Kernel->setup (); 

  return Kernel;
}

void
MCore::load_module( int module_id )
{
  //TODO QApplication::setOverrideCursor( waitCursor );

  if( kernel_blocked == FALSE )  //TODO is this thread safe?
    {
      kernel_blocked = TRUE;

      switch( module_id )
	{
	case MODULE_CONFIG:
	  break; 
  
	case MODULE_SCRIPT:
	  break; 
 
	case MODULE_TREEDLG:
	  {
	    MSceneGraphDlg *dlg =
	      new MSceneGraphDlg( module_id,
				  modules [module_id].load_type );
	    modules[module_id].the_module = dlg;
	    break;
	  }

	case MODULE_MODEL:
	  {
	    MModeler *modeler = 
	      new MModeler( module_id,
			    modules [module_id].load_type );
	    modules[module_id].the_module = modeler;
	    break;
	  }

	case MODULE_MATERIAL:
	  modules[module_id].the_module
	    = new MEMaterialEditor( module_id,
				    modules [module_id].load_type );
	  break; 

	case MODULE_VIEWER:
	  modules[module_id].the_module
	    = new MViewer (module_id,
			   modules [module_id].load_type);
	  break; 
	  
	case MODULE_NET:
	  modules[module_id].the_module
	    = new MNetwork (module_id,
			    modules [module_id].load_type);
	  break; 

	case MODULE_LOAD:
	case MODULE_SAVE:
	  modules[module_id].the_module
	    = new MEIoModule (module_id,
			      modules [module_id].load_type);
	  if( module_id == MODULE_LOAD )
	    ((MEIoModule*)(modules [module_id].the_module))
	      ->setAction(ACTION_LOAD);
	  else
	    {
	      ((MEIoModule*)(modules [module_id].the_module))
		->setAction(ACTION_SAVE);
	    }
	  break;

	case MODULE_HELP:
	  break;

	case MODULE_SHAPER:
	  modules[module_id].the_module =
	    new ME2DShaper( module_id, modules[module_id].load_type );
	  break;
	}
      
      kernel_blocked=FALSE;
    }

  //TODO QApplication::restoreOverrideCursor();
}

void
MCore::process_message (struct message *m)
{
  switch (m->message_id)
    {
    case K_MESS_DATAINFO :
      break;

    case K_MESS_MODULES  :
      break;

    case K_MESS_STATUS   :
      the_console->m_printf ("I'm feeling fine Dave.");
      break; 

    case K_MESS_WHO      :
      the_console->m_printf ("Multiuser mode not implemented yet, sorry.");
      break; 
      
    case K_MESS_QUIT     :
      shutdown ();
      break;

    case K_MESS_LOADSCENE:
      toggle_module( MODULE_LOAD );
      break;
    case K_MESS_SAVESCENE:
      toggle_module( MODULE_SAVE );
      break;
    case K_MESS_IMPORTOBJ:
      toggle_module( MODULE_LOAD );
      ((MEIoModule*)(modules[MODULE_LOAD].the_module))
	->setRoot( (G3DGroup*)(m->contents) );
      break;
    case K_MESS_NEWSCENE:
      {
	MWorldObject *tmp = new MWorldObject;
	addWorld( tmp );
	break;
      }

    default:
      the_console->m_printf ("[Kernel]: Unrecoqnized message id: [%d]",
			     m->message_id);
      break;
    }
}

int MCore::run (void)
{
  //TODO test if setup was successful?

#ifdef WITH_KERNEL_WIDGET
  kernel_widget->show();
#endif
  modules [MODULE_CONSOLE].toggled = TRUE;
  //TODO show only if user wants to
  //the_console->show ();
  the_console->m_printf( "Entering message loop........" );

  int result = mindseye_app->exec();

  return (result);
}

/*---------------------------------------------------------------------------*/

int
MCore::setup (void)
{
  the_console->m_printf( "Booting MindsEye v%d.%d.%d ...", MAJOR,MINOR,PATCH );

 // booting message subsystem
  message_system = new MMessageServer( DEFAULT_QSIZE );
  connect (message_system,
	   SIGNAL (console_output (const char *)),
	   the_console,
	   SLOT (m_puts (const char *)));
  the_console->m_printf ("Message subsystem ready,setting queuesize to [%d]",
			 DEFAULT_QSIZE);
 // message system ready

  the_console->m_printf( "Setting up 3d class library and plugins" );
  init3dlib( this );
  MEIoModule::initStandardFilter();
  initRenderQueues( this );
  initAttributeWidgets( this );

  pluginMgr_ = new MEPluginManager;
  //TODO pluginMgr_->loadDllsFromDir( "system dir", "" );
  pluginMgr_->loadDllsFromDir( "./plugins", "" );  //user specified dir
  pluginMgr_->print();

  the_console->m_puts( CopyrightMessage );
  return (0);
}

int
MCore::setup_app( int argc,char **argv )
{
  mindseye_app = new QApplication( argc, argv );
  mindseye_app->setStyle( MotifStyle );
  mindseye_app->setColorMode (QApplication::CustomColors);

  // >------------ determine font style

  app_font = new QFont( "Helvetica", 10 );
  app_font->setBold( TRUE );

  mindseye_app->setFont( *app_font, TRUE );
  mindseye_app->setStyle( WindowsStyle );

  background = new QPixmap;
  background->loadFromData (backgr_bmp_data,backgr_bmp_len);

  // >------------ QApplication for MindsEye should be setup now

  return (0);
}

int
MCore::setup_chooser (void)
{
#ifdef WITH_KERNEL_WIDGET
  kernel_widget = new MKernelWidget( 0, 0 );
#ifdef WITH_KMAP_ICONS
  kernel_widget->setBackgroundPixmap(*background);
#endif
  mindseye_app->setMainWidget( kernel_widget );
#else
  kernel_widget = NULL;
#endif
  return (0);
}

int
MCore::setup_console (void)
{
  //modules[0].the_module = TODO
  the_console = new MKernelConsole( 0, 0 );
#ifndef WITH_KERNEL_WIDGET
  mindseye_app->setMainWidget( the_console );
#endif
  return (0);
}

int
MCore::setup_sns (void)
{
#ifdef WITH_KERNEL_WIDGET
  connect( kernel_widget,
	   SIGNAL (toggle_module (int)),
	   this,
	   SLOT (toggle_module (int)) );
#endif
}

/*---------------------------------------------------------------------------*/
void
MCore::shutdown (void)
{
  qApp->quit();
}

void
MCore::toggle_module( int module_id )
{
  if( (module_id == MODULE_CONSOLE) && (the_console != NULL) )
    {
      // The console is the only main part which does not use the module
      // services. This is because it should always be available
      if( modules[0].toggled == TRUE )
	{
	  the_console->hide (); 
	  modules[0].toggled = FALSE;
	}
      else
	{
	  the_console->show ();
	  modules[0].toggled = TRUE;
	}
    }
  else if( module_id == MODULE_ABOUT )
    {
      MAboutDialog *tmp = new MAboutDialog();
      tmp->show();
    }
  else
    {
      // this is the place where the actual behaviour of the modules
      // is determined.
      if( modules[module_id].toggled == TRUE )
	{
	  close_module( module_id );
	}
      else
	{
	  if( modules[module_id].load_type == LOAD_ON_CALL )
	    {
	      load_module( module_id );
	      if( modules[module_id].the_module == NULL )
		return;

	      modules[module_id].the_module_inf
		= modules[module_id].the_module->get_module_info();
	      //TODO move this into MModule?
	      the_console->m_printf( "Module [%s] v%d.%d.%d is running",
			     modules[module_id].the_module_inf->module_name,
			     modules[module_id].the_module_inf->major_version,
			     modules[module_id].the_module_inf->minor_version,
			     modules[module_id].the_module_inf->patch_level);
	      connect( modules[module_id].the_module,
		       SIGNAL(close_me(int)),
		       this, SLOT(close_module(int)) );
	    }
	  else
	    {
	      the_console->m_printf( "Module [%s] v%d.%d.%d is shown",
			     modules[module_id].the_module_inf->module_name,
			     modules[module_id].the_module_inf->major_version,
		     	     modules[module_id].the_module_inf->minor_version,
		      	     modules[module_id].the_module_inf->patch_level);
	    }
	  modules[module_id].toggled = TRUE; 
	  modules[module_id].the_module->show();
	} 
    } 
}


//====================================END=============================
