/*
 * Project : "MindsEye"
 * Copyright 1996, 1998. All Rights Reserved.
 * Author(s) M. Kuball, F. Meier, M. van Velsen
 *
 * $Id: m_core.h,v 1.6 1998/03/19 21:22:31 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * OVERVIEW
 * ========
 *
 */

#ifdef HAVE_CONFIG_H
#include "include/config.h"
#endif	 // HAVE_CONFIG_H

#include "dll.h"
#include "classinfo.h"

#include <sys/types.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>

class MCore;
extern MCore *Kernel;

/////////////////////////////////////////////////////////////////////

inline
unsigned getEhnoVersion()   // application version
{
  return EHNO_VERSION;
}

//--------------------------------------------------------------------
#define HAVE_DLFCN_H

#ifdef HAVE_DLFCN_H
#include <dlfcn.h>

inline void*
pl_open( const char *szDlName )
{
  return dlopen(szDlName, RTLD_LAZY );
}

inline void
pl_close( void *handle)
{
  dlclose( handle );
}

inline const char*
pl_error( void )
{
  return dlerror();
}

inline void *
pl_symbol( void *handle, const char *symbol )
{
  return dlsym( handle, symbol );
}
//-----------------------------------------------------------------------
#else
#ifdef HAVE_DL_H

#include <dl.h>

/* 
 * HP-UX section, thanks to Alf Clement <alf@hpbidrd.bbn.hp.com>
 * for providing me the details
 * /

inline void*
pl_open( const char *szDlName )
{
  return shl_load( szDlName, BIND_IMMEDIATE|BIND_VERBOSE ,0 );
}

inline void
pl_close( void *handle)
{
  shl_close( handle );
}

inline const char*
pl_error( void )
{
  return "undefined yet";
}

inline void *
pl_symbol( void *handle, const char *symbol )
{
  void *addr;
  return shl_findsym( &handle, symbol, TYPE_UNDEFINED, &addr );
}

/*---------------------------------------------------------------------------*/

#endif /* HAVE_DL_H    */
#endif /* HAVE_DLFCN_H */


////////// DllInfo //////////////////////////////////////////////////


DllInfo::DllInfo(void *handle,  const char *dllName, const char *desc ) :
  mHandle(handle), mDllFilename(dllName), desc_(desc)
{
}

void DllInfo::free()
{
  if( mHandle )
    pl_close( mHandle );
}


/////////////////////////////////////////////////////////////////////


MEPluginManager::~MEPluginManager()
{
  unloadAllDlls();
}

void
MEPluginManager::loadDll( const char *pDllName )
{
  void *handle = pl_open( pDllName );
  if( !handle )
    {
      cerr << "[MEPluginManager::loadDll] dlopen " << pl_error() << endl;
      return;
    }
  
  Pl_init    *plInit;
  Pl_close   *plClose;
  const char *plDescription, *plCopyright;
  const char *error;
  unsigned   plVersion;

  plVersion = *((unsigned*)pl_symbol( handle, "pluginVersion" ));
  if( (error = pl_error()) != NULL )
    {
      cerr << "[dlsym] pluginVersion: " << error << endl;
      pl_close( handle );
      return;
    }
  if( plVersion != getEhnoVersion() )
    {
      cerr << "[MEPluginManager::loadDll] plug-in version mismatch for: \""
	   << pDllName << "\" " << plVersion
	   << " " << getEhnoVersion() << endl;
      return;
    }

  plInit = (Pl_init*)pl_symbol( handle, "pluginInit" );
  if( (error = pl_error()) != NULL )
    {
      cerr << "[dlsym] pluginInit: " << error << endl;
      pl_close( handle );
      return;
    }
  plDescription = *((const char**)pl_symbol( handle, "pluginDescription" ));
  cout << plDescription << endl;
  if( (error = pl_error()) != NULL )
    {
      cerr << "[dlsym] pluginDescription: " << error << endl;
      pl_close( handle );
      return;
    }

  ////////////////////////////////////////

  if( plInit( Kernel ) )
    {
      DllInfo *dllInfo = new DllInfo( handle, pDllName, plDescription );
      //TODO ? plugins_.reserve( plugins_.size()+1 );
      plugins_.push_back( dllInfo );      
    }
   else 
    pl_close( handle );
}

//TODO use wildcard (2nd argument)
bool
MEPluginManager::loadDllsFromDir(const char *pDirName, const char*)
{
  struct dirent *de;
  DIR    *dir;
  char   dllFile[1024];  //TODO
  
  if( (dir = opendir(pDirName)) == NULL ) 
    {
      perror( "[MEPluginManager::loadDllsFromDir] opendir" );  
      return false;
    }

  while( (de = readdir(dir)) )
    {
      if( (de->d_name)[0] == '.'   //don't open hidden files
	  || (de->d_name)[0] != '_' )
	continue;
      strcpy( dllFile, pDirName );
      strcat( dllFile, "/" );
      strcat( dllFile, de->d_name );
      loadDll( dllFile );    // HIIIIIIER "pDirName + de.d_name"
    }
 
  closedir( dir );

  return true;
}

int
MEPluginManager::numDllsLoaded() const
{
  return plugins_.size();
}

DllInfo&
MEPluginManager::operator[] (int i)
{
  return *(plugins_[i]);
}

void
MEPluginManager::unloadAllDlls()
{
  for( vector<DllInfo*>::iterator dll = plugins_.begin();
       dll != plugins_.end(); dll++)
    {
      (*dll)->free();
      delete (*dll);   
    }
}

void MEPluginManager::print()
{
  cout << "-- MEPluginManager --------------------------------------" << endl;
  for( unsigned i=0; i<plugins_.size(); i++ ) 
  {
    DllInfo* dllInfo = plugins_[i];

    cout << " DLL " 
	 << " desc: \"" << dllInfo->getDescription() << "\"" << endl;	 
  }
  cout << "-- " << plugins_.size() << " DLL(s) ----------------------" << endl;
}


//===================================END=====================================

