/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: world_obj.cpp,v 1.15 1998/07/23 20:16:18 source Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"

#include "3dlib/camera.h"
#include "3dlib/scene.h"
#include "3dlib/light.h"
#include "3dlib/aggregate.h"
#include "3dlib/selectGroup.h"
#include "3dlib/iterator.h"
#include "world_obj.h"
#include "world_list.h"


MWorldObject::MWorldObject( void )
{
  flag_  = 0;  //TODO not used, yet
  scene_ = new G3DScene;
  scene_->setName( "Scratch" );

  G3DPointLight *light = new G3DPointLight;
  light->setPosition( G3DPoint3d(0, 10, 10) );
  scene_->addLight( light );

  activeGroup_   = scene_;
  selection_     = NULL;
  G3DCamera *cam = new G3DCamera;
  cam->setName( "Camera 1" );

  cameraList_ = new MCameraList;
  cameraList_->push_back( cam );
}


MWorldObject::MWorldObject( G3DScene *scene, MCameraList *cameras )
{
  scene_ = scene;
  if( cameras )
    cameraList_ = cameras;
  else
    {
      cameraList_    = new MCameraList;
      G3DCamera *cam = new G3DCamera;
      cam->setName( "Camera 1" );
      cameraList_->push_back( cam );
    }
  selection_   = NULL;
  activeGroup_ = scene_;
  flag_        = 0;
}

MWorldObject::~MWorldObject()
{
  delete scene_;

  MCameraList::iterator itr = cameraList_->begin();
  while( itr != cameraList_->end() )
    delete *(itr++);
  delete cameraList_;
}

void
MWorldObject::addCamera( G3DCamera* camera )
{
  cameraList_->push_back( camera );
}

void
MWorldObject::addLight( G3DLight* light )
{
  activeGroup_->addLight( light );
}

void
MWorldObject::addObject( G3DObject* obj )
{
  if( obj != NULL )
    {
      if( obj->getType() & T_SHAPE )
	addShape( (G3DShape*)obj );
      else if( obj->getType() & T_LIGHT )
	addLight( (G3DLight*)obj );
    }
}

void
MWorldObject::addShape( G3DShape* shape )
{
  activeGroup_->addShape( shape );
}

G3DObject*
MWorldObject::addToSelection( G3DObject *sel )
{
  if( selection_ == NULL )
    {
      selectObject( sel );
    }
  else
    {
      if( selection_->getType() != T_SEL_GROUP )
	{
	  G3DObject *tmp = selection_;
	  selection_ = new G3DSelectionGroup;
	  ((G3DSelectionGroup*)selection_)->appendObject( tmp );
	  //TODO emit objectSelected( selection_ );
	}
      if( ((G3DSelectionGroup*)selection_)->findObject(sel) >= 0 )
	{
	  ((G3DSelectionGroup*)selection_)->removeObject( sel );
          ActionData action(A_DESELECT);
	  sel->doAction( action );
	}
      else
	{
          ActionData action(A_SELECT);
	  sel->doAction(action );
	  ((G3DSelectionGroup*)selection_)->appendObject( sel );
	  emit decorateObject( sel );
	}
    }

  return( selection_ );
}

void
MWorldObject::deleteLight( G3DLight *light )
{
  G3DGroup *parent = (G3DGroup*)(light->getParent());
  if( parent->removeLight(light) )
    {
      if( light == selection_ )
	{
	  emit objectSelected( NULL );
	  selection_ = NULL;
	}
      delete light;
      emit sceneChanged( parent );
    }
}

void
MWorldObject::deleteObject( G3DObject *obj )
{
  if( obj == NULL )
    obj = selection_;
  if( obj == NULL )
    return;

  if( obj->getType() & T_SHAPE )
    deleteShape( (G3DShape*)obj );
  else if( obj->getType() & T_LIGHT )
    deleteLight( (G3DLight*)obj );
  else if( obj->getType() & T_SEL_GROUP )
    {
      ((G3DSelectionGroup*)obj)->eraseObjects();
      delete obj;
      emit objectSelected( NULL );
    }
}

void
MWorldObject::deleteShape( G3DShape *obj )
{
  G3DGroup *parent = (G3DGroup*)(obj->getParent());
  if( parent && parent->removeShape(obj) )
    {
      if( obj == selection_ )
	{
	  emit objectSelected( NULL );
	  selection_ = NULL;
	}
      delete obj;
      emit sceneChanged( parent );
    }
}

void
MWorldObject::doWithScene( ActionData &action )
{
  scene_->doAction( action );
  emit sceneChanged( scene_ );
}

void
MWorldObject::doWithSelection( ActionData &action )
{
  if( selection_ )
    {
      if( selection_->getType() & T_SHAPE )
	selection_->doAction( action );
      //TODO else ( T_SEL_GROUP )
    }
}

const char*
MWorldObject::getName() const
{ 
  return( scene_->getName() );
}


class RegroupObjectsIterator : public GIterator<G3DObject>
{
public:
  RegroupObjectsIterator( G3DGroup *grp ) : deselect_(A_DESELECT)
    { grp_ = grp; }

  virtual bool exec( G3DObject &obj )
  {
    obj.doAction( deselect_ );
    ((G3DGroup*)(obj.getParent()))->removeObject( &obj ); //scene inconsistent
    grp_->addObject( &obj );    //scene consistent again
    return( false );
  }

private:
  G3DGroup   *grp_;
  ActionData deselect_;
};

void
MWorldObject::groupSelectedObjects( void )
{
  if( selection_ && (selection_->getType() == T_SEL_GROUP) )
    {
      G3DObject *parent =
	((G3DSelectionGroup*)selection_)->getObject(0)->getParent();
      if( parent )
	{
	  G3DGroup *grp = new G3DGroup;
	  RegroupObjectsIterator iter( grp );
	  ((G3DSelectionGroup*)selection_)->doWithObjects( iter );
	  ((G3DGroup*)parent)->addShape( grp );
	  delete selection_;
	  selection_ = grp;
	  ActionData action(A_SELECT);
	  grp->doAction( action );
	  emit decorateObject( selection_ );
	  emit sceneChanged( scene_ );
	  emit objectSelected( selection_ );
	}
    }
}

uint
MWorldObject::numberOfCameras( void )
{
  return( cameraList_->size() );
}

void
MWorldObject::selectObject( G3DObject *sel )
{
  if( selection_ ) //TODO special actions??
    {
      ActionData action(A_DESELECT);
      selection_->doAction( action );
      if( selection_->getType() == T_SEL_GROUP )
	delete selection_;
    }

  selection_ = sel;
  if( sel )
    {
      ActionData action(A_SELECT);
      selection_->doAction( action );
      emit decorateObject( selection_ );
    }

  emit objectSelected( selection_ );
}

void
MWorldObject::setActiveCamera( const G3DCamera *cam )
{
  activeCamera_ = *cam;
}

void
MWorldObject::setActiveGroup( G3DGroup *grp )
{
  if( grp && (grp->getType() & T_GROUP) )
    activeGroup_ = grp;
  else
    activeGroup_ = scene_;
}

void
MWorldObject::update( void )
{
  emit objectChanged( );
}


void
MWorldObject::update( G3DGroup *group )
{
  emit sceneChanged( group );
}


//==================================END==================================
