/****************************************************************************
** $Id: qgl.cpp,v 1.3 1997/07/08 13:25:16 aavit Exp $
**
** Implementation of OpenGL classes for Qt
**
** Created : 970112
**
** Copyright (C) 1997 by Troll Tech AS.  All rights reserved.
**
*****************************************************************************/

#include "qglpix.h"

#if defined(Q_GLX)
#define INT8  dummy_INT8
#define INT32 dummy_INT32
#include <GL/glx.h>
#undef  INT8
#undef  INT32
#endif

/*
makeCurrent: else if( PDT_PIXMAP )
chooseContext: make a non-direct context
*/


/*!
  \class QGLWidget qgl.h
  \brief The QGLWidget class is a widget for rendering OpenGL graphics.

  It is easy to render OpenGL graphics in Qt applications. You can create
  a subclass of QGLWidget and reimplement two functions: resizeGL() and
  paintGL(). resizeGL() is called whenever the widget has been resized.
  paintGL() is called when the widget needs to be updated.

  \code
    class DrawGL : public QGLWidget
    {
        QOBJECT;   // Include this if you want to use Qt signals & slots etc.

    public:
        DrawGL( QWidget *parent, const char *name )
	    : QGLWidget(parent,name) {}

    protected:
        void paintGL()
	{
	  // draw the scene
	}

	void resizeGL( int w, int h )
	{
	  // setup viewport, display lists etc.
	}
    };
  \endcode

  If you need to repaint from other places than paintGL() (a typical
  example is when using \link QTimer timers\endlink to animate scenes)
  you must call makeCurrent() before you draw and probably swapBuffers()
  when you are finished.

  Like QGLContext, QGLWidget has advanced functions for requesting
  a new contex format and you can even set a new context.
*/


/*!
  Constructs an OpenGL widget with a \a parent widget and a \a name.

  The \link QGLFormat::defaultFormat() default format\endlink is
  used.

  \sa QGLFormat::defaultFormat()
*/

QGLPixmap::QGLPixmap( int w, int h )
    : QPixmap(w, h, -1)
{
  //w, h > 0!!
  //setBackgroundColor( black );
  glcx = 0;
  setContext( new QGLContext(QGLFormat::defaultFormat(), this) );
}


/*!
  \fn const QGLFormat &QGLPixmap::format() const
  Returns the widget's format.
  \sa setFormat()
*/

/*!
  \fn bool QGLPixmap::doubleBuffer() const
  Returns TRUE if double buffering is set for this widget.
  \sa QGLFormat::doubleBuffer()
*/


/*!
  \fn bool QGLPixmap::isValid() const
  Returns TRUE if the widget was able to satisfy the specified constraints 
*/

/*!
  \fn void QGLPixmap::makeCurrent()
  Makes this pixmap the current drawable for OpenGL operations.
*/

/*!
  Sets a new format for this widget. The widget becomes invalid if the
  requested format cannot be satisfied.
  \sa format(), setContext(), isValid()
*/

void QGLPixmap::setFormat( const QGLFormat &format )
{
    glcx->setFormat( format );
    glcx->create();
}


/*!
  \fn const QGLContext *QGLPixmap::context() const
  Returns the current context.
  \sa setContext()
*/

/*!
  Sets a new context for this widget. The old context is deleted.
  \sa context(), setFormat()
*/

void QGLPixmap::setContext( QGLContext *context )
{
    if ( context == 0 ) {
#if defined(CHECK_NULL)
	warning( "QGLPixmap::setContext: Cannot set null context" );
#endif
    }
    if ( context->device() != this ) {
#if defined(CHECK_STATE)
	warning( "QGLPixmap::setContext: Context must refer this widget" );
#endif
    }
    delete glcx;
    glcx = context;
    if ( !glcx->isValid() ) {
	if ( !glcx->create() )
	    return;
    }
#if defined(Q_GLX)
    XVisualInfo *vi = (XVisualInfo*)glcx->vi;
    if ( XVisualIDFromVisual(vi->visual) ==
	 XVisualIDFromVisual((Visual*)QPaintDevice::x11Visual()) )
	return;

    /*
      Here we can optimize:
	1) Create one colormap for each visual.
	2) Recreate window only if necessary.
    */
    Colormap cmap;
    cmap = XCreateColormap( dpy, RootWindow(dpy,vi->screen),
			    vi->visual, AllocNone );
    XSetWindowAttributes a;
    a.colormap = cmap;
    a.background_pixel = backgroundColor().pixel();
    a.border_pixel = black.pixel();
    Window p = RootWindow( dpy, DefaultScreen(dpy) );
    glXCreateGLXPixmap( dpy, vi, handle() );
    glcx->setPixmap( );
#endif
}


/*!
  \fn void QGLPixmap::updateGL()

  Updates the widget by calling paintGL().
*/


/*!
  Handles paint events. Calls the virtual function paintGL().
*/

void QGLPixmap::paintEvent( QPaintEvent * )
{
    glcx->makeCurrent();
    paintGL();
    if ( glcx->format().doubleBuffer() )
	glcx->swapBuffers();
    else
	glFlush();
}


/*****************************************************************************
  QGL classes overview documentation.
 *****************************************************************************/

/*!
\page qgl.html

<title>Qt OpenGL Classes</title>
</head><body bgcolor="#ffffff">

<h1 align=center>Qt OpenGL Classes</h1>
<hr>


<h2>Introduction</h2>

OpenGL is a standard API for rendering 3D graphics.

OpenGL only deals with 3D rendering and does not worry about GUI
programming issues. You must create the user interface with another
toolkit, such as Motif on the X platform, Microsoft Foundation Classes
(MFC) under Windows -- or Qt on both platforms.

<h2>The QGL Classes</h2>

The OpenGL support classes in Qt are:
<ul>
<li> <strong>\link QGLPixmap QGLPixmap\endlink:</strong> An easy-to-use Qt
  widget for rendering OpenGL scenes.
<li> <strong>\link QGLContext QGLContext\endlink:</strong> An OpenGL context.
<li> <strong>\link QGLFormat QGLFormat\endlink:</strong> Spesifies the
device settings of a context.
</ul>

Many applications need only the high-level QGLPixmap class. The other QGL
classes provide advanced features.

*/
