/*
 * Project : "MindsEye"
 * Copyright (c)1996, 1997, 1998. All Rights Reserved.
 * Author(s) M. van Velsen, M. Kuball
 *
 * $Id: m_viewp.h,v 1.19 1998/07/23 20:16:24 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * OVERVIEW
 * ========
 * Modeling is done in a viewport. A viewport can operate in 3
 * basic modes, which are:
 *
 *   Flat mode: 2 Dimensional view of scene ( FRONT,BACK,LEFT,RIGHT,UP,DOWN )
 *
 *   Camera mode: This mode uses user defined viewing types (can also be spot)
 *
 *   Scene: A mode based on LightWave which uses a free movable camera
 *
 * The basic operations have been logically split up into:
 *
 * Modeler Operations
 *   |
 *   |
 *   +-- Movement 
 *   |   |
 *   |   +-- Panning  modes: ( Scene , Flat )
 *   |   |
 *   |   +-- Rotating modes: ( Scene , Camera )
 *   |   |
 *   |   +-- Zooming  modes: ( Scene , Camera , Flat )
 *   | 
 *   +-- Selecting
 *       |
 *       +-- Object
 *       |
 *       +-- Polygon
 *       |
 *       +-- Vertex 
 *
 */
 
#ifndef __MVIEWPORT_H__
#define __MVIEWPORT_H__

#include <qgl.h>
#include <qwidget.h>
#include <qframe.h>
#include <qcursor.h>

#include "3dlib/g3d_math.h"
#include "3dlib/camera.h"


enum MEViewportMode { ViewMode, ObjectMode };
enum { OPERATION_OBJECT, OPERATION_CAMERA };
enum { DirectMode, DisplayListMode };


class MWorldObject;
class G3DCamera;
class G3DObject;
class G3DGroup;
class G3DScene;
class G3DColor;
class QPopupMenu;
class MObjectPopup;
class MModeler;
class MViewPortLayout;
class TransformControler;
class MECameraControler;


/**This widget uses OpenGL to display a scene. It handles the basic user
   interaction based on the policy set by the owner.
   */
class MGLPort : public QGLWidget
{
    Q_OBJECT

public:
    ///
  MGLPort( QCursor *new_hand, QPopupMenu *lpopup,
	   QWidget *parent, const char *name=0 );
  ~MGLPort( );

    ///Set focal point to (0, 0, 0).
  void centerCamera( void );
    ///
  G3DCamera* getCamera( void );
    ///Perspective oder parallel camera.
  void setCameraType( G3DCamera::TViewMode );
    ///Set
  void setDisplayListMode( int );
    ///Set interaction mode: camera movement or object selection
  void setOperationMode( int );
    ///
  void setModel( MModeler *model ) { _model = model; };
    ///
  void setWorld( MWorldObject* );
    ///
  void show_axes    (bool);
    ///
  void show_grid    (bool);
    ///

signals:
  void viewChanged();

public slots:
    /// WIRE/FLAT/SMOOTH
  void change_display(int);
    // CAMERA/FRONT/BACK .... 
  //void change_view(int);
    ///
  void objectSelected(G3DObject*);

protected:
    ///Draw axes of global coordinate system.
  void       drawAxes( void );
    ///Draw a grid //TODO local grid object?
  void       drawGrid (void);
    /// get the selected objects from the elem list
  G3DObject* examineHitBuffer( int hits, GLuint *ptr, G3DObject *obj );
   /**.Build a ray from the camera into the scene and find the object
     with the nearest intersection point in front of the camera.
     start may be 0 or a group of objects.
     mode specifys if the test should stop at a group node or descend
     down to the bottom objects. */
  G3DObject* findElement( int x, int y, int mode, G3DGroup *start );
  void       focusInEvent( QFocusEvent* );
  void       focusOutEvent( QFocusEvent* );
   ///
  void       initPicking( QMouseEvent*, GLsizei, GLuint[] );
    /// Initialize the shader (OpenGL).
  void       initShader( void );
    ///
  void       keyPressEvent( QKeyEvent* e );
    ///
  void       mouseMoveEvent( QMouseEvent* );
    ///
  void       mousePressEvent( QMouseEvent* );
    ///
  void       mouseReleaseEvent( QMouseEvent* );
  /** Draw the scene.
      1. initialize OpenGl
      2. disable all light sources
      3. display each element of the scene.*/
  void       paintGL();
    /// setup/initiate picking + draw the objects in elem
  G3DObject* pickObject( QMouseEvent*, G3DGroup *elem );
    ///
  G3DObject* pickSubObject( QMouseEvent*, G3DObject *obj );
    ///
  void       resetLights( void );
    ///Set the new size of the viewport.
  void       resizeGL( int, int );
  void       setControler( QMouseEvent* );
    ///
  void       updateDisplayList( void );

protected:
    //
  MWorldObject *world_;
  G3DScene     *scene_;
  G3DCamera    *camera_;
  G3DObject    *_selected;
  G3DObject    *_active;
    //
  G3DColor  *_selectColor;
    //
  MModeler  *_model;
    //
  TransformControler *tControler_;
  MECameraControler  *camControler_;
    //
  QPopupMenu *_defaultPopup;
  QCursor    *hand_cursor;

    ///
  bool see_grid;
    ///
  bool see_axes;
    ///Camera or objects
  int  opMode_;
    ///Use some caching to speed up displaying the scene.
  int  dispListMode_;
    ///Display the scen as a wire frame, flat shaded or smooth shaded.
  char  display_type;

  int   _mouse_button;
};



/**This widget is a wrapper for the MGLPort widget to draw a nice
   frame around it.
   */
class MViewPort : public QFrame
{
    Q_OBJECT

    friend class MViewPortLayout;

public:
    MViewPort( QCursor *new_hand, QPopupMenu *lpopup,
	       QWidget *parent, const char *name=0 );

signals:
    void activateView( MViewPort* );

protected:
    virtual void enterEvent ( QEvent * );
 
protected:
    MGLPort glPort_;

};


#endif // __MVIEWPORT_H__
