/*
 * Project : "MindsEye"
 * Copyright (c) 1996, 1997 by Martin Kuball. All Rights Reserved.
 * Author(s): M. Kuball, M. van Velsen, F. Meier
 * 
 * $Id: tcontroler.h,v 1.7 1998/07/23 20:16:25 source Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 *
 */

#ifndef _T_CONTROLER_H_
#define _T_CONTROLER_H_

#include <qobject.h>
#include <qevent.h>

#include "3dlib/g3d_math.h"


class G3DObject;
class G3DCamera;
class Grid;
class MWorldObject;


/**Abstract base class for Controler which are used to Transform
   3d objects. They translate mouse movements into suitable 3d
   coordinates.
   For special input devices you have to reimplement the transform
   function;
  */
class TransformControler : public QObject
{
    Q_OBJECT

public:
  TransformControler( void ) : QObject(NULL, NULL)
    {
      _selected = NULL;
      grid_     = NULL;
    };
  virtual ~TransformControler( );

    ///
  void    setCoordSys( G3DCamera* );
    ///
  void    setGrid( Grid *grid ) { grid_ = grid; };
    ///
  void    setSelectedObject( G3DObject *sel ) { _selected = sel; };
    ///
  virtual void start( QMouseEvent* ) = 0;
    ///
  virtual void stop( MWorldObject* ) = 0;
    ///The interface may need to be extended for other input devices.
  virtual void transform( QMouseEvent*, int ) = 0;
    ///Set transformation values directly, without using the mouse.
  virtual void transform( float, float, float );

signals:
  void setInfoText(const char*);

protected:
  G3DObject   *_selected;
  G3DVector3d direction_;
  Grid        *grid_;
  G3DCamera   *camera_;
};


/**Objects of this class serve as a Translation controler.
   They perform a translation on the selected object based on
   the given mouse coordinates and the internal state
 */
class MoveControler : public TransformControler
{
    Q_OBJECT

public:
  MoveControler( void ) {};

    ///
  virtual void start( QMouseEvent* );
    ///
  virtual void stop( MWorldObject* );
    ///Translate the selected object.
  virtual void transform( QMouseEvent*, int );

protected:
    ///
  void doMove( int, int, int );

protected:
  G3DMatrix   map2local_;
  G3DVector3d orig_;
  G3DPoint3d  position_, mousepos_, mstart_;
  int         oldy_, oldz_;
};


/**Objects of this class serve as a rotation controler.
   They perform a rotation on the selected object based on
   the given mouse coordinates and the internal state.
 */
class RotateControler : public TransformControler
{
    Q_OBJECT

public:
  RotateControler( void ) {};

    ///
  virtual void start( QMouseEvent* );
    ///
  virtual void stop( MWorldObject* );
    ///Translate the selected object.
  virtual void transform( QMouseEvent*, int buttonstate );

protected:
    ///Find the angle for the given mouse position.
  double calcAngle( G3DVector3d &mouse );
    ///
  void   doRotate( int, int );
    ///Initialize the translation from mouse pos to angle.
  void   initAngle( G3DVector3d&, G3DVector3d &e1, G3DVector3d &e2 );

protected:
  G3DMatrix   mrot_;
  G3DVector3d axis_, gaxis_, lcenter_, radius_;
  G3DPoint3d  gcenter_, mouse_, origin_;

  bool        colinear_;
  double      angle0_, length_;
};


/**Objects of this class serve as a controler for scaling objects.
   A scaling controler must be asigned a widget object.
 */
class ScaleControler : public TransformControler
{
    Q_OBJECT

public:
  ScaleControler( void ) {};

    ///
  virtual void start( QMouseEvent* );
    ///
  virtual void stop( MWorldObject* );
    ///Translate the selected object.
  virtual void transform( QMouseEvent*, int );
  virtual void transform( float, float, float );

protected:
    ///
  void doScale( int, int, int );

protected:
  G3DMatrix   mlocal_, map2local_;
  G3DVector3d svector_, scale_, transl_;
  G3DPoint3d  mousepos_, gfixed_, lfixed_;
  int        _oldx, _oldy, _oldz;
};



/*
class SixDControler : public MoveControler, virtual RotateControler
{
    Q_OBJECT

  virtual void start( QMouseEvent* );
  virtual void transform( QMouseEvent*, int );

protected:
    
};
*/


#endif
